/*****************************************************************************
   vortexcontrol.c - Vortex Control Panel
   Copyright (C) 2004 by Claudia
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
******************************************************************************/
/*

Requirement:-

   Hardware:           Aureal Vortex Sound Cards au8810, au8820 or au8830.

   Software Library:   alsa-lib-1.0.5 or later
                       gtk-2.2.0 or above
                       openal(optional)

Functions of Vortex Control Panel

1) AC97 Mixer
2) Hardware Equalizer with EQ Peak Meter
3) Midi (Download soundfont to hardware wavetable synthesier) (Only ARL parser implemented)
4) Speaker Test/Hardware Mixing
5) SDPIF OUT for au8830/Diamond MX300,SQ1500 and SPDIF IN/OUT for TBS Montego Quadzilla
6) 3D Sound (require OpenAL)
7) Joystick Calibration (N/A)
8) Modem (N/A)
9) Speaker Calibration (N/A)
10)HRTF (N/A)

Change Log
   0.0.1 
         Alpha version

   0.0.2 
         10 band Graphic EQ Peak

   0.0.3
         Hardware Equalizer with 5 pre-defined set of EQ.
         A dynamic maximum value of EQ Peak.
         GUI of AC97 Mixer.

   0.0.4
	 AC97 Mixer (Including specific feature of STAC9708)
	 Speaker Test 
	 GUI of 3D Sound. 
	 GUI of SPDIF. 

   0.0.5 
         Speaker test 
            -use sine wave when outls.wav, outrs.wav, out1lq.wav, out1rq.wav, out2lq.wav and 
             out2rq.wav not found in the same directory of vortexcontrol.
         Hardware Mixing
            -playback of multiple sound concurrently (up to 16 mono/8 stereo/4 quad sound).
            -add scope to monitor playback
         3D Sound 
            -16 sound sources (fixed location, circular and linear movement).
            -occlusion, obstruction, exclusion.
            -reflection (open area, closed wall). 
            -sound sources and listener can be relocated by mouse click.
         SPDIF 
            - SPDIF/SR merge into mixer

   0.0.6 
         Wavetable
            - ARL parser. 
	        - no chunk size of INFO,sdta,pdta 
	        - chunk size of LIST is zero for INFO list 
	        - chunk size of IMOD and PMOD are multiples of 12 bytes
	        - chunk size of shdr is multiples of 48 bytes
	        - 16 bits samples
	        - sample is bad when startsample = endsample = startloop = endloop
                - unknown generators with oper > rootkey.
                - preset has modulator.
            - DLS parser.
            - midi parser.
            - play the individual sample in ARL/DLS using adb engine instead of wt engine.
            - sample with rate > 48000 Hz cannot be play through adb engine.
         Scope 
            - fix bug in scope for a quad channels wav file.
            - capture scope still not working. 
	 3D Sound
            - openal (using software renderer - attuentation by distance, doppler effect)

   0.1.0
         Wavetable
            - parsing au10gm.arl, aurealgm, au30gml.arl or gm.dls
              arl is similar to sf2 except:-
              1) no chunk size for INFO,sdta,pdta chunks 
              2) chunk size of LIST is zero for INFO list 
              3) chunk size of IMOD and PMOD are multiples of 12 bytes
              4) chunk size of shdr is multiples of 48 bytes
              5) 8/16 bits samples
              6) bad samples if startsample = endsample = startloop = endloop 
         SPDIF 
            - change IEC958_AES3_CON_FS ( 32000, 44100, 48000 )
            - IEC958 not function until au88x0 support ICE958
         Package
            - add acinclude.m4, configure.in, Makefile.am, README
		aclocal
		autoconf
		autoheader
		automake -a
		./configure --enable-openal
		make
		make install
	    use the above procedure instead of
		gcc -Wall vortexcontrol.c -o vortexcontrol `pkg-config --cflags gtk+-2.0` -lasound -lopenal `pkg-config --libs gtk+-2.0`

Remarks:-
a) Please provide the following output and your aureal cards when you report bugs.
   1) "lspci -vvn"
   2) "cat /proc/asound/au88x0/codec97#0/ac97#0-0"
   3) "amixer contents"
b) Please use "alsactl restore" after boot,"alsactl save" before shutdown to retain setting.
c) The left or right channel of different source can be selected for capture.
d) Need to obtain the subsystem vendor/device id to detect specific aureal sound cards 
   e.g. OPTICAL IN of Turtle Beach Quadzilla, Modem, SPDIF OUT of SQ1500.
e) AC97 mixer highly depends on the ordering of the playback switch and playback volume.
   It may fail to work if there is any change in the ordering of the hctl or ctl elements.
f) The front and rear channels are swapped after the first play. You have to recompile the au88x0
   driver by using stopfifo() instead of pausefifo() when cmd is SNDRV_PCM_TRIGGER_STOP in
   the subroutine snd_vortex_pcm_trigger() of au88x0_pcm.c   

*/
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <assert.h>
#include <math.h>
#include <pthread.h>
#include <sys/time.h>
#include <gtk/gtk.h>
#include <alsa/asoundlib.h>


#define VC_VERSION          	"0.1.0"
#define MAX_CARD_NUMBERS 	8                     
#define MAX_NO_OF_PLAYBACK	16                    // Volume Control & Mute 
#define MAX_NO_OF_AC97_3D       2                     // AC97 3D controls
#define MAX_NO_OF_CAPTURE       8                     // Recording Source controls
#define MAX_NO_OF_ENABLE        5                     // Boolean controls
#define MAX_NO_OF_ENUM          5                     // Enumeration elem
#define MAX_NO_OF_ENUM_RADIO    16                    // Enumeration radio
#define NO_OF_BAND       	10                    // Equalizer
#define NO_OF_EQSET             6                     // Predefined Settings of Equalizer
#define MAX_NO_OF_MOVE_TYPE     4
#define MAX_NO_OF_3D_SOURCE     16
#define MAX_NO_OF_PRESET_SAMPLES 100               
#define MAX_NO_OF_INST_SAMPLES  50

#ifdef  HAVE_LIBOPENAL
#include <AL/al.h>
#include <AL/alc.h>
#include <AL/alu.h>
#include <AL/alut.h>
#define NUM_BUFFERS 16	// Number of buffers to be Generated

ALvoid 		*Context;
ALCdevice 	*Device;
ALuint		source[MAX_NO_OF_3D_SOURCE];
ALuint		g_Buffers[NUM_BUFFERS];		// Array of Buffers
ALvoid DisplayALError( const char *text, ALint errorcode)
{
	g_print("%s%s\n", text, alGetString(errorcode));
}
#endif
/******************************************************************************************/

#define card_is_au8810    ( strcmp(driver,"au8810") == 0 )
#define card_is_au8820    ( strcmp(driver,"au8820") == 0 )
#define card_is_au8830    ( strcmp(driver,"au8830") == 0 )

#define ac97_has_sdac     ( strstr(snd_ctl_card_info_get_mixername(hw_info),"STAC9708") > 0 )

// some au8830 may use other quad codecs such as Tritech 29028 or Wolfsun 9704

/******************************************************************************************/

snd_ctl_t *ctl;
snd_ctl_card_info_t *hw_info;

snd_hctl_t *hctl;

snd_hctl_elem_t *elem;
snd_hctl_elem_t *prev_elem;
snd_ctl_elem_info_t *info;
snd_ctl_elem_id_t *id;
snd_ctl_elem_value_t *val;

snd_hctl_elem_t *eqpower_elem;
snd_hctl_elem_t *eqpeak_elem;
snd_hctl_elem_t *eq_adj_elem[NO_OF_BAND];
snd_hctl_elem_t *mix_adj_elem[MAX_NO_OF_PLAYBACK];
snd_hctl_elem_t *mix_mute_elem[MAX_NO_OF_PLAYBACK];
snd_hctl_elem_t *mix_capsrc_elem;
snd_hctl_elem_t *mix_3d_elem[MAX_NO_OF_AC97_3D];
snd_hctl_elem_t *mix_enum_elem[MAX_NO_OF_ENUM];
snd_hctl_elem_t *mix_enable_elem[MAX_NO_OF_ENABLE];
snd_ctl_elem_value_t *spdif_out;
snd_ctl_elem_value_t *spdif_in;
snd_aes_iec958_t iec958;

struct {
	snd_hctl_elem_t *hrtf_elem;
	snd_hctl_elem_t *ild_elem;
	snd_hctl_elem_t *itd_elem;
	snd_hctl_elem_t *filter_elem;
} alsa_3d_snd[16];

int alsa_card_number;
int spdif_id;
const char *spdif_str = SND_CTL_NAME_IEC958("", PLAYBACK, DEFAULT);

snd_pcm_format_t   format      = SND_PCM_FORMAT_S16; /* sample format */
unsigned int       rate        = 48000;	             /* stream rate */
unsigned int       channels    = 1;	             /* count of channels */
unsigned int       buffer_time = 500000;	     /* ring buffer length in us */
unsigned int       period_time = 100000;	     /* period time in us */
double             freq        = 440;                /* sinusoidal wave frequency in Hz */
snd_pcm_uframes_t  buffer_size;
snd_pcm_uframes_t  period_size;

#define	WAVE_FORMAT_PCM			(0x0001)

typedef struct
{
	FILE *file;
	short format_tag, channels, block_align, bits_per_sample, eof, going;
	long samples_per_sec, avg_bytes_per_sec;
	unsigned long position, length;
	int seek_to, data_offset;
} WaveFile;

/******************************************************************************************/

gint      eqpeak_timer_tag;
gint      v3d_timer_tag;

GtkWidget *window;

GtkObject *mix_adj[MAX_NO_OF_PLAYBACK][2];
GtkWidget *mix_vscale[MAX_NO_OF_PLAYBACK][2];

GtkWidget *mix_lock_button[MAX_NO_OF_PLAYBACK];
GtkWidget *mix_mute_button[MAX_NO_OF_PLAYBACK];

GtkObject *ac97_3d_adj[MAX_NO_OF_AC97_3D];
GtkWidget *ac97_3d_hscale[MAX_NO_OF_AC97_3D];

GtkObject *ac97_spsa_adj;
GtkWidget *ac97_spsa_hscale;

GtkWidget *mix_rec_radio[MAX_NO_OF_CAPTURE][2];
GtkWidget *mix_rec_lock;
GtkWidget *mix_enum_radio[MAX_NO_OF_ENUM_RADIO];         
GtkWidget *mix_enable_button[MAX_NO_OF_ENABLE];

GtkObject *eq_adj[NO_OF_BAND][2];
GtkWidget *eq_vscale[NO_OF_BAND][2];
GtkWidget *eqpeak_drawing[NO_OF_BAND];
GtkWidget *eqset_button[NO_OF_EQSET];
GtkWidget *eq_speaker_button;
GtkWidget *eqlock_button;
GtkWidget *eqgang_button;
GtkWidget *eqpower_button;

GtkWidget *spdif_rate_radio[3];                  // 32000, 44100, 48000

GtkWidget *speaker_button[3][2];
GtkWidget *speaker_set_radio[3];                 // no of speaker set
GtkObject *speaker_freq_adj;
GtkWidget *speaker_freq_hscale;

GtkWidget *scope_drawing[16];
GtkWidget *capture_scope_drawing;

GtkWidget *hwmix_label[16];
GtkWidget *hwmix_rate_label[16];
GtkWidget *hwmix_chn_label[16];
GtkWidget *hwmix_vbox[16];
GtkWidget *hwmix_play_button[16];
GtkWidget *hwmix_pause_button[16];
GtkWidget *hwmix_stop_button[16];
GtkWidget *hwmix_loop_button[16];
GtkWidget *hwmix_del_button[16];
GtkWidget *hwmix_total_chn_label;
GtkWidget *hwmix_add_button;
GtkWidget *hwmix_table;
GtkWidget *hwmix_start_capture_button;
GtkWidget *hwmix_stop_capture_button;

struct {
    GtkObject *vc_adj[4];
    GtkWidget *vc_hscale[4];
    int created;
} hwmix_vc[16];
int wt_sample[600];
int wt_preset_sample[MAX_NO_OF_PRESET_SAMPLES];
int wt_inst_sample[MAX_NO_OF_INST_SAMPLES];
int wt_dls_sample[512];
int wt_dls_inst[256];
int wt_dls_preset[256];

GtkListStore *hwmix_list_store;
GtkWidget    *hwmix_tree;

GtkWidget *wt_preset_table;
GtkWidget *wt_preset_button[140];
GtkWidget *wt_inst_table;
GtkWidget *wt_inst_button[200];
GtkWidget *wt_sample_table;
GtkWidget *wt_sample_button[600];
GtkWidget *wt_preset_sample_button[MAX_NO_OF_PRESET_SAMPLES];
GtkWidget *wt_inst_sample_button[MAX_NO_OF_INST_SAMPLES];
GtkWidget *wt_info_table;
//GtkWidget *wt_info_label[20];
GtkWidget *wt_scope_drawing;
GtkWidget *wt_stop_button;
GtkWidget *wt_midi_button;
GtkWidget *wt_play_midi_button;
GtkWidget *wt_stop_midi_button;

GtkWidget *wt_dls_preset_button[256];
GtkWidget *wt_dls_inst_button[256];
GtkWidget *wt_dls_sample_button[512];
GtkWidget *wt_dls_inst_sample_button[128];

GtkWidget *v3d_drawing;
GtkObject *v3d_scaleup_adj;
GtkWidget *v3d_scaleup_spin;
GtkWidget *v3d_showpath_button;

GtkObject *v3d_x_adj[MAX_NO_OF_3D_SOURCE];
GtkObject *v3d_y_adj[MAX_NO_OF_3D_SOURCE];
GtkObject *v3d_z_adj[MAX_NO_OF_3D_SOURCE];

GtkObject *v3d_w_adj[MAX_NO_OF_3D_SOURCE];
GtkObject *v3d_h_adj[MAX_NO_OF_3D_SOURCE];
GtkObject *v3d_speed_adj[MAX_NO_OF_3D_SOURCE];
GtkObject *v3d_r_adj[MAX_NO_OF_3D_SOURCE];
GtkObject *v3d_o_adj[MAX_NO_OF_3D_SOURCE];
GtkObject *v3d_e_adj[MAX_NO_OF_3D_SOURCE];
GtkObject *v3d_v_adj[MAX_NO_OF_3D_SOURCE];
GtkObject *v3d_oh_o_adj[MAX_NO_OF_3D_SOURCE];
GtkObject *v3d_oh_speed_adj[MAX_NO_OF_3D_SOURCE];
GtkObject *v3d_cone_inner_adj[MAX_NO_OF_3D_SOURCE];
GtkObject *v3d_cone_outer_adj[MAX_NO_OF_3D_SOURCE];

GtkWidget *v3d_w_hscale[MAX_NO_OF_3D_SOURCE];
GtkWidget *v3d_h_hscale[MAX_NO_OF_3D_SOURCE];
GtkWidget *v3d_speed_spin[MAX_NO_OF_3D_SOURCE];
GtkWidget *v3d_r_hscale[MAX_NO_OF_3D_SOURCE];
GtkWidget *v3d_o_spin[MAX_NO_OF_3D_SOURCE];
GtkWidget *v3d_e_spin[MAX_NO_OF_3D_SOURCE];
GtkWidget *v3d_v_hscale[MAX_NO_OF_3D_SOURCE];
GtkWidget *v3d_oh_o_spin[MAX_NO_OF_3D_SOURCE];
GtkWidget *v3d_oh_speed_spin[MAX_NO_OF_3D_SOURCE];
GtkWidget *v3d_cone_inner_spin[MAX_NO_OF_3D_SOURCE];
GtkWidget *v3d_cone_outer_spin[MAX_NO_OF_3D_SOURCE];

GtkWidget *v3d_nb;
GtkWidget *v3d_x_pos_label[MAX_NO_OF_3D_SOURCE];
GtkWidget *v3d_y_pos_label[MAX_NO_OF_3D_SOURCE];
GtkWidget *v3d_z_pos_label[MAX_NO_OF_3D_SOURCE];

GtkWidget *v3d_x_spin[MAX_NO_OF_3D_SOURCE];
GtkWidget *v3d_y_spin[MAX_NO_OF_3D_SOURCE];
GtkWidget *v3d_z_spin[MAX_NO_OF_3D_SOURCE];

GtkWidget *v3d_z_hscale[MAX_NO_OF_3D_SOURCE];
GtkWidget *v3d_move_button[MAX_NO_OF_3D_SOURCE];
GtkWidget *v3d_play_button[MAX_NO_OF_3D_SOURCE];
GtkWidget *v3d_stop_button[MAX_NO_OF_3D_SOURCE];
GtkWidget *v3d_loop_button[MAX_NO_OF_3D_SOURCE];
GtkWidget *v3d_wave_button[MAX_NO_OF_3D_SOURCE];
GtkWidget *v3d_move_nb[MAX_NO_OF_3D_SOURCE];
GtkWidget *v3d_move_type_radio[MAX_NO_OF_3D_SOURCE][MAX_NO_OF_MOVE_TYPE];

GtkObject *v3d_listener_x_adj;
GtkObject *v3d_listener_y_adj;
GtkObject *v3d_listener_z_adj;
GtkWidget *v3d_listener_x_spin;
GtkWidget *v3d_listener_y_spin;
GtkWidget *v3d_listener_z_spin;

GtkObject *v3d_listener_direction_adj;
GtkWidget *v3d_listener_direction_hscale;

GtkWidget *v3d_wavetracing_button;
GtkObject *v3d_wavelength_adj;
GtkWidget *v3d_wavelength_hscale;
/*
GtkObject *v3d_listener_elevation_adj;
GtkObject *v3d_listener_speed_adj;
GtkWidget *v3d_listener_speed_spin;
GtkWidget *v3d_listener_elevation_hscale;
*/
GtkWidget *v3d_occlusion_button[4];
GtkWidget *v3d_exclusion_button[4];
GtkWidget *v3d_obstruction_button[4];
GtkWidget *v3d_reflection_radio[3];

GtkObject *v3d_obstruction_radius_adj;
GtkWidget *v3d_obstruction_radius_hscale;

GtkObject *v3d_obstruction_size_adj;
GtkWidget *v3d_obstruction_size_hscale;

GtkObject *v3d_reflection_adj;
GtkWidget *v3d_reflection_hscale;

GtkObject *v3d_reverb_adj;
GtkWidget *v3d_reverb_hscale;

struct {
	GtkObject *hrtf_adj[6];
	GtkObject *ild_adj[2];
	GtkObject *itd_adj[2];
	GtkObject *filter_adj[4];
	GtkWidget *hrtf_hscale[6];
	GtkWidget *ild_hscale[2];
	GtkWidget *itd_hscale[2];
	GtkWidget *filter_hscale[4];
} alsa_3d_gtk[16];

static GdkGC *penGreenShadow = { NULL, };
static GdkGC *penGreenLight = { NULL, };
static GdkGC *penOrangeShadow = { NULL, };
static GdkGC *penOrangeLight = { NULL, };
static GdkGC *penRedShadow = { NULL, };
static GdkGC *penRedLight = { NULL, };
static GdkGC *penBlack = { NULL, };
static GdkGC *penGrey = { NULL, };
static GdkGC *penWhite = { NULL, };
static GdkGC *pen3DSource[MAX_NO_OF_3D_SOURCE] = { NULL, };
static GdkGC *penWT[16] = { NULL, };

static GdkPixmap *pixmap[NO_OF_BAND] = { NULL, };
static GdkPixmap *v3d_pixmap = { NULL, };
static GdkPixmap *scope_pixmap[16] = { NULL, };
static GdkPixmap *wt_scope_pixmap = { NULL, };
static GdkPixmap *capture_scope_pixmap = { NULL, };

/******************************************************************************************/

const float PI = 3.1415926535f;
char driver[64];
char card[16];

int vc[MAX_NO_OF_PLAYBACK];
int cap[MAX_NO_OF_CAPTURE];

struct {
     int elem_idx;
     int idx;
     int value;
} mix_enum[MAX_NO_OF_ENUM_RADIO];

int stac9708_pcmoutpath_idx;
int stac9708_4speakerstereo_idx;
int sdac_master_idx,sdac_mixer_idx;
int ac97_3d_enable;

int no_of_playback;
int no_of_capture_source;
int no_of_enum_elem;
int no_of_enum_radio;
int no_of_enable;
int no_of_ac97_3d;
int iec958_present;

int no_of_speaker_set;

int spdif_rate[3]={32000,44100,48000};

int eq_lock;
int eq_power;
int eq_ungang;

struct {
   int left[NO_OF_BAND];
   int right[NO_OF_BAND];
} eqpeak_level;

int eqpeak_min,eqpeak_max;

int no_of_scope=0;
int scope_type;
//int scope_format=SND_PCM_FORMAT_S16_LE;
//int scope_channels=2;
//int scope_bits_per_sample;
gint16 scope_data[16][4096];

char *midi_key[256]={"C0","C#0","D0","D#0","E0","F0","F#0","G0","G#0","A0","A#0","B0",
"C1","C#1","D1","D#1","E1","F1","F#1","G1","G#1","A1","A#1","B1",
"C2","C#2","D2","D#2","E2","F2","F#2","G2","G#2","A2","A#2","B2",
"C3","C#3","D3","D#3","E3","F3","F#3","G3","G#3","A3","A#3","B3",
"C4","C#4","D4","D#4","E4","F4","F#4","G4","G#4","A4","A#4","B4",
"C5","C#5","D4","D#4","E5","F5","F#5","G5","G#5","A5","A#5","B5",
"C6","C#6","D6","D#6","E6","F6","F#6","G6","G#6","A6","A#6","B6",
"C7","C#7","D7","D#7","E7","F7","F#7","G7","G#7","A7","A#7","B7",
"C8","C#8","D8","D#8","E8","F8","F#8","G8","G#8","A8","A#8","B8",
"C9","C#9","D9","D#9","E9","F9","F#9","G9","G#9","A9","A#9","B9",
"C10","C#10","D10","D#10","E10","F10","F#10","G10",
"128","129","130","131","132","133","134","135","136","137","138","139","140","141","142","143",
"144","145","146","147","148","149","150","151","152","153","154","155","156","157","158","159",
"160","161","162","163","164","165","166","167","168","169","170","171","172","173","174","175",
"176","177","178","179","180","181","182","183","184","185","186","187","188","189","190","191",
"192","193","194","195","196","197","198","199","200","201","202","203","204","205","206","207",
"208","209","210","211","212","213","214","215","216","217","218","219","220","221","222","223",
"224","225","226","227","228","229","230","231","232","233","234","235","236","237","238","239",
"240","241","242","243","244","245","246","247","248","249","250","251","252","253","254","255"};

char *controllers[128]={"Bank Select",
 			"Modulation wheel",
			"Breath control",
			"",
			"Foot controller",
			"Portamento time",
			"Data Entry",
			"Channel Volume",
			"Balance",
			"",
			"Pan",   
			"Expression Controller",
			"Effect control 1",
			"Effect control 2",
			"",
			"",
			"General Purpose Controller #1",
			"General Purpose Controller #2",
			"General Purpose Controller #3",
			"General Purpose Controller #4",
			"",
			"",
			"",
			"",
			"",
			"",
			"",
			"",
			"",
			"",
			"",
			"",
			"Bank Select",
 			"Modulation wheel",
			"Breath control",
			"",
			"Foot controller",
			"Portamento time",
			"Data Entry",
			"Channel Volume",
			"Balance",
			"",
			"Pan",   
			"Expression Controller",
			"Effect control 1",
			"Effect control 2",
			"",
			"",
			"General Purpose Controller #1",
			"General Purpose Controller #2",
			"General Purpose Controller #3",
			"General Purpose Controller #4",
			"",
			"",
			"",
			"",
			"",
			"",
			"",
			"",
			"",
			"",
			"",
			"",
			"Damper pedal on/off (Sustain)",
			"Portamento on/off",
			"Sutenuto on/off",
                        "Soft pedal on/off",
			"Legato Footswitch",
			"Hold 2",
	 		"Sound Controller 1 (Sound Variation)",
			"Sound Controller 2 (Timbre)",
			"Sound Controller 3 (Release Time)",
			"Sound Controller 4 (Attack Time)",
			"Sound Controller 5 (Brightness)",
			"Sound Controller 6",
			"Sound Controller 7",
			"Sound Controller 8",
			"Sound Controller 9",
			"Sound Controller 10",
			"General Purpose Controller #5",
			"General Purpose Controller #6",
			"General Purpose Controller #7",
			"General Purpose Controller #8",
			"Portamento Control",
			"",
			"",   
			"",   
			"",   
			"",   
			"Effects 1 Depth",
			"Effects 2 Depth",
			"Effects 3 Depth",
			"Effects 4 Depth",
			"Effects 5 Depth",
			"Data entry +1",
			"Data entry -1",
			"Non-Registered Parameter Number LSB",
			"Non-Registered Parameter Number MSB",
			"Registered Parameter Number LSB",
			"Registered Parameter Number MSB",
			"",
			"",
			"",
			"",
			"",
			"",
			"",
			"",
			"",
			"",
			"",
			"",
			"",
			"",
			"",
			"",
			"",
			"",
			"All Sound Off",
			"Reset All Controllers",
			"Local control on/off",
			"All notes off",
			"Omni mode off (+ all notes off)",
			"Omni mode on (+ all notes off)",
			"Poly mode on/off (+ all notes off)",
			"Poly mode on (incl mono=off +all notes off)"
			};                        

typedef unsigned char byte;
typedef unsigned short uint16;
typedef short int16;
typedef unsigned int uint32;
typedef int int32;

typedef struct
	{
	byte byLo;
	byte byHi;
	} rangesType;

typedef union
	{
	rangesType ranges;
	int16 shAmount;
	uint16 wAmount;
	} genAmountType;

typedef enum
	{
	monoSample = 1,
	rightSample = 2,
	leftSample = 4,
	linkedSample = 8
	} SFSampleLink;

enum {
	/* level 0; chunk */
	UNKN_ID, RIFF_ID, LIST_ID, SFBK_ID,
	/* level 1; id only */
	INFO_ID, SDTA_ID, PDTA_ID,
	/* info stuff; chunk */
	IFIL_ID, ISNG_ID, IROM_ID, INAM_ID, IVER_ID, IPRD_ID, ICOP_ID, 
        ISFT_ID, ICMT_ID, IENG_ID, ICRD_ID,
	/* sample data stuff; chunk */
	SNAM_ID, SMPL_ID,
	/* preset stuff; chunk */
	PHDR_ID, PBAG_ID, PMOD_ID, PGEN_ID,
	/* inst stuff; chunk */
	INST_ID, IBAG_ID, IMOD_ID, IGEN_ID,
	/* sample header; chunk */
	SHDR_ID,
};

	static struct idstring {
		char *str;
		int id;
	} idlist[] = {
		{"RIFF", RIFF_ID},		/* "RIFF" chunk */
		{"LIST", LIST_ID},		/* "LIST" chunk */
		{"sfbk", SFBK_ID},		/* "sfbk" SoundFont chunk */
		{"INFO", INFO_ID},		/* "INFO" chunk */
		{"sdta", SDTA_ID},		/* "sdta" sample DATA chunk */
		{"snam", SNAM_ID},		/* "snam" SoundFont Bank name */
		{"smpl", SMPL_ID},		/* "smpl" sample data subchunk in SDTA */
		{"pdta", PDTA_ID},		/* "pdta" Preset/Instrument/Sample params */
		{"phdr", PHDR_ID},		/* "phdr" preset header */
		{"pbag", PBAG_ID},		/* "pbag" preset bag (zone) */
		{"pmod", PMOD_ID},		/* "pmod" preset modulator */
		{"pgen", PGEN_ID},		/* "pgen" preset generator */
		{"inst", INST_ID},		/* "inst" instrument header */
		{"ibag", IBAG_ID},		/* "ibag" instrument bag (zone) */
		{"imod", IMOD_ID},              /* "imod" instrument modulator */
		{"igen", IGEN_ID},              /* "igen" instrument generator */
		{"shdr", SHDR_ID},              /* "shdr" sample header */
		{"ifil", IFIL_ID},              /* "ifil" SoundFont version */
		{"isng", ISNG_ID},              /* "isng" target sound engine */
		{"irom", IROM_ID},              /* "irom" ROM name */
		{"iver", IVER_ID},              /* "iver" ROM version */
		{"INAM", INAM_ID},              /* "INAM" SoundFont name */
		{"IPRD", IPRD_ID},              /* "IPRD" product SoundFont is intended for */
		{"ICOP", ICOP_ID},              /* "ICOP" copyright message */
		{"ISFT", ISFT_ID},		/* "ISFT" tools used to create and alter */
		{"ICMT", ICMT_ID},              /* "ICMT" comments */
		{"IENG", IENG_ID},              /* "IENG" sound designers and engineers */
                {"ICRD", ICRD_ID},              /* "ICRD" creation date */
	};


enum {
	SF_startAddrs,         
        SF_endAddrs,
        SF_startloopAddrs,     
        SF_endloopAddrs,       
        SF_startAddrsHi,       
        SF_lfo1ToPitch,        
        SF_lfo2ToPitch,        
        SF_env1ToPitch,        
        SF_initialFilterFc,    
        SF_initialFilterQ,     
        SF_lfo1ToFilterFc,     
        SF_env1ToFilterFc,     
        SF_endAddrsHi,        
        SF_lfo1ToVolume,       
        SF_env2ToVolume,       
        SF_chorusEffectsSend,  
        SF_reverbEffectsSend,  
        SF_panEffectsSend,    
        SF_auxEffectsSend,    
        SF_sampleVolume,       
        SF_unused3,
        SF_delayLfo1,          
        SF_freqLfo1,          
        SF_delayLfo2,         
        SF_freqLfo2,           
        SF_delayEnv1,          
        SF_attackEnv1,         
        SF_holdEnv1,             
        SF_decayEnv1,            
        SF_sustainEnv1,         
        SF_releaseEnv1,         
        SF_autoHoldEnv1,
        SF_autoDecayEnv1,
        SF_delayEnv2,            
        SF_attackEnv2,           
        SF_holdEnv2,            
        SF_decayEnv2,           
        SF_sustainEnv2,          
        SF_releaseEnv2,          
        SF_autoHoldEnv2,
        SF_autoDecayEnv2,
        SF_instrument,          
        SF_nop,
        SF_keyRange,             
        SF_velRange,             
        SF_startloopAddrsHi,     
        SF_keynum,             
        SF_velocity,             
        SF_initAtten,              
        SF_keyTuning,
        SF_endloopAddrsHi,       
        SF_coarseTune,
        SF_fineTune,
        SF_sampleId,
        SF_sampleFlags,
        SF_samplePitch,         
        SF_scaleTuning,
        SF_keyExclusiveClass,
        SF_rootKey,
	SF_EOF,
};
char *sf_gen_text[SF_EOF] = {
	"startAddrs",         
        "endAddrs",
        "startloopAddrs",     
        "endloopAddrs",       
        "startAddrsHi",       
        "lfo1ToPitch",        
        "lfo2ToPitch",        
        "env1ToPitch",        
        "initialFilterFc",   
        "initialFilterQ",     
        "lfo1ToFilterFc",     
        "env1ToFilterFc",     
        "endAddrsHi",         
        "lfo1ToVolume",       
        "env2ToVolume",       
        "chorusEffectsSend",  
        "reverbEffectsSend",  
        "panEffectsSend",     
        "auxEffectsSend",     
        "sampleVolume",       
        "unused3",
        "delayLfo1",          
        "freqLfo1",           
        "delayLfo2",          
        "freqLfo2",          
        "delayEnv1",          
        "attackEnv1",         
        "holdEnv1",             
        "decayEnv1",            
        "sustainEnv1",        
        "releaseEnv1",         
        "autoHoldEnv1",
        "autoDecayEnv1",
        "delayEnv2",          
        "attackEnv2",           
        "holdEnv2",             
        "decayEnv2",        
        "sustainEnv2",         
        "releaseEnv2",          
        "autoHoldEnv2",
        "autoDecayEnv2",
        "instrument",          
        "nop",
        "keyRange",            
        "velRange",            
        "startloopAddrsHi",     
        "keynum",              
        "velocity",             
        "initAtten",              
        "keyTuning",
        "endloopAddrsHi",       
        "coarseTune",
        "fineTune",
        "sampleId",
        "sampleFlags",
        "samplePitch",          
        "scaleTuning",
        "keyExclusiveClass",
        "rootKey",
};


typedef struct {
	char id[4];
	int32 size;
} SFChunk;

typedef struct {
	char id[4];
	char *desc;
} SFDesc;

typedef struct {
	int16 oper;
	int16 amount;
} SFGen;

typedef struct {
	int16 srcoper;
        int16 destoper;
        int16 amount;
        int16 amtsrcoper;
        int16 transoper;
	uint16 unknown;              // look like 32 bits alignment
} SFMod;

typedef struct {
	int16 GenNdx;
	int16 ModNdx;
} SFBag;

typedef struct {
	char name[20];
	uint16 bagNdx;
	uint16 preset, bank;
	int32 lib, genre, morphology;
} SFPreset;

typedef struct {
	char name[20];
	uint16 bagNdx;
} SFInst;

typedef struct  {
	char name[20];
	int32 startsample, endsample;
	int32 startloop, endloop;

	int32 samplerate;
	byte originalpitch;
	byte pitchcorrection;
	uint16 samplelink;
	uint16 sampletype;  /*1=16 bit mono , 17=8 bit mono */
} SFSampleInfo;

typedef struct {
	int nbags;
	SFBag *bag;
} SFBagList;

typedef struct {
	int ngens;
	SFGen *gen;
} SFGenList;

typedef struct {
	int nmods;
	SFMod *mod;
} SFModList;

typedef struct {
	int  loaded;
	char filename[FILENAME_MAX+1];
	char *sf_name;

	struct  {
		uint16 major;
		uint16 minor;
        } version;

	SFDesc info[8];

	long samplepos;
	int32 samplesize;

	long infopos, infosize;

	int npresets;
	SFPreset *preset;
	
	int nsamples;
	SFSampleInfo *sample;

	int ninsts;
	SFInst *inst;

	SFBagList prbags, inbags;
	SFGenList prgens, ingens;
	SFModList prmods, inmods;
	
	struct {
		struct {
	  	    int sample;
		    int wt;
		} note[128];
		int  preset;
		int  inst;
	} midi[128],drum[10];
	unsigned char *data;
} SFInfo;


static SFInfo sfinfo;

/**************************************************************************************/

enum {  FOURCC_LIST,FOURCC_RIFF,FOURCC_DLS, FOURCC_COLH,FOURCC_MSYN,
        FOURCC_WVPL,FOURCC_PTBL,FOURCC_PATH,FOURCC_WAVE,FOURCC_LINS,
        FOURCC_INS, FOURCC_INSH,FOURCC_LRGN,FOURCC_RGN, FOURCC_RGNH,FOURCC_LART,FOURCC_ART1,
	FOURCC_WLNK,FOURCC_WSMP,FOURCC_FMT, FOURCC_DATA,FOURCC_VERS,
        FOURCC_INFO,FOURCC_ICOP,FOURCC_ICMT,FOURCC_IENG,FOURCC_INAM,FOURCC_ISBJ,FOURCC_ISFT,
        FOURCC_UNKN};
 
	static struct dlsidstring {
		int id;
		char *str;
	} dlsidlist[] = {
                { FOURCC_LIST,"LIST"},
		{ FOURCC_RIFF,"RIFF"},
		{ FOURCC_DLS ,"DLS "},
		{ FOURCC_COLH,"colh"},
	        { FOURCC_MSYN,"msyn"},
		{ FOURCC_WVPL,"wvpl"},
		{ FOURCC_PTBL,"ptbl"},
		{ FOURCC_PATH,"path"},
		{ FOURCC_WAVE,"wave"},
 		{ FOURCC_LINS,"lins"},
		{ FOURCC_INS, "ins "},
		{ FOURCC_INSH,"insh"},
		{ FOURCC_LRGN,"lrgn"},
		{ FOURCC_RGN, "rgn "},
		{ FOURCC_RGNH,"rgnh"},
		{ FOURCC_LART,"lart"},
		{ FOURCC_ART1,"art1"},
		{ FOURCC_WLNK,"wlnk"},
		{ FOURCC_WSMP,"wsmp"},
		{ FOURCC_FMT ,"fmt "},
		{ FOURCC_DATA,"data"},
		{ FOURCC_VERS,"vers"},
		{ FOURCC_INFO,"INFO"},
		{ FOURCC_ICOP,"ICOP"},
		{ FOURCC_ICMT,"ICMT"},
		{ FOURCC_IENG,"IENG"},
		{ FOURCC_INAM,"INAM"},
		{ FOURCC_ISBJ,"ISBJ"},
		{ FOURCC_ISFT,"ISFT"}
	};
 /*
    Articulation connection graph definitions 
 */
 
 /* Generic Sources */
 #define CONN_SRC_NONE              0x0000
 #define CONN_SRC_LFO               0x0001
 #define CONN_SRC_KEYONVELOCITY     0x0002
 #define CONN_SRC_KEYNUMBER         0x0003
 #define CONN_SRC_EG1               0x0004
 #define CONN_SRC_EG2               0x0005
 #define CONN_SRC_PITCHWHEEL        0x0006
 
 /* Midi Controllers 0-127 */
 #define CONN_SRC_CC1               0x0081
 #define CONN_SRC_CC7               0x0087
 #define CONN_SRC_CC10              0x008a
 #define CONN_SRC_CC11              0x008b
 
 /* Registered Parameter Numbers */
 #define CONN_SRC_RPN0             0x0100
 #define CONN_SRC_RPN1             0x0101
 #define CONN_SRC_RPN2             0x0102
 
 /* Generic Destinations */
 #define CONN_DST_NONE              0x0000
 #define CONN_DST_ATTENUATION       0x0001
 #define CONN_DST_RESERVED          0x0002
 #define CONN_DST_PITCH             0x0003
 #define CONN_DST_PAN               0x0004
 
 /* LFO Destinations */
 #define CONN_DST_LFO_FREQUENCY     0x0104
 #define CONN_DST_LFO_STARTDELAY    0x0105
 
 /* EG1 Destinations */
 #define CONN_DST_EG1_ATTACKTIME    0x0206
 #define CONN_DST_EG1_DECAYTIME     0x0207
 #define CONN_DST_EG1_RESERVED      0x0208
 #define CONN_DST_EG1_RELEASETIME   0x0209
 #define CONN_DST_EG1_SUSTAINLEVEL  0x020a
 
 /* EG2 Destinations */
 #define CONN_DST_EG2_ATTACKTIME    0x030a
 #define CONN_DST_EG2_DECAYTIME     0x030b
 #define CONN_DST_EG2_RESERVED      0x030c
 #define CONN_DST_EG2_RELEASETIME   0x030d
 #define CONN_DST_EG2_SUSTAINLEVEL  0x030e
 
 #define CONN_TRN_NONE              0x0000
 #define CONN_TRN_CONCAVE           0x0001
 
 typedef struct {
   uint32    dwVersionMS;
   uint32    dwVersionLS;
 }DLSVERSION;
 				   
 typedef struct {
   uint16   usSource;
   uint16   usControl;
   uint16   usDestination;
   uint16   usTransform;
   int32     lScale;
   }CONNECTION;
  
 typedef struct  {
   uint32    cbSize;            /* size of the connection list structure */
   uint32    cConnections;      /* count of connections in the list */
   } CONNECTIONLIST;
  
 typedef struct {
   uint16 usLow;
   uint16 usHigh;
 }RGNRANGE;
 
 #define F_INSTRUMENT_DRUMS      0x80000000
 
 typedef struct {
   uint32 ulBank;
   uint32 ulInstrument;
 }MIDILOCALE;
  
 #define F_RGN_OPTION_SELFNONEXCLUSIVE  0x0001
 
 typedef struct {
   RGNRANGE RangeKey;            /* Key range */
   RGNRANGE RangeVelocity;       /* Velocity Range */
   uint16   fusOptions;          /* Synthesis options for this range */
   uint16   usKeyGroup;          /* Key grouping for non simultaneous play
                                    0 = no group, 1 up is group 
                                    for Level 1 only groups 1-15 are allowed */
 }RGNHEADER;
 
 typedef struct {
   uint32      	cRegions;          /* Count of regions in this instrument */
   MIDILOCALE 	Locale;            /* Intended MIDI locale of this instrument */
 }INSTHEADER;
 
 typedef struct {
   uint32      	cInstruments;      /* Count of instruments in the collection */
 }DLSHEADER;
 
 /*****  For level 1 only WAVELINK_CHANNEL_MONO is valid  ****
   ulChannel allows for up to 32 channels of audio with each bit position
   specifiying a channel of playback */
 
 #define WAVELINK_CHANNEL_LEFT    0x0001l
 #define WAVELINK_CHANNEL_RIGHT   0x0002l
 
 #define F_WAVELINK_PHASE_MASTER  0x0001
 
 typedef struct { /* any paths or links are stored right after struct */
   uint16   fusOptions;     /* options flags for this wave */
   uint16   usPhaseGroup;   /* Phase grouping for locking channels */
   uint32    ulChannel;      /* channel placement */
   uint32    ulTableIndex;   /* index into the wave pool table, 0 based */
 }WAVELINK;
 
 #define POOL_CUE_NULL  0xffffffffl
 
 typedef struct { 
   uint32    ulOffset;       /* Offset to the entry in the list */
 }POOLCUE;
 
 typedef struct {
   uint32    cbSize;            /* size of the pool table structure */
   uint32    cCues;             /* count of cues in the list */
   } POOLTABLE;
 
 /*
    Structures for the "wsmp" chunk
 */
 
 #define F_WSMP_NO_TRUNCATION     0x0001l
 #define F_WSMP_NO_COMPRESSION    0x0002l
 
 
 typedef struct {
   uint32   cbSize;
   uint16  usUnityNote;         /* MIDI Unity Playback Note */
   int16   sFineTune;           /* Fine Tune in log tuning */
   int32    lAttenuation;        /* Overall Attenuation to be applied to data */
   uint32   fulOptions;          /* Flag options */
   uint32   cSampleLoops;        /* Count of Sample loops, 0 loops is one shot */
   } WSMPL;
  
 #define WLOOP_TYPE_FORWARD   0
 
 typedef struct {
   uint32 cbSize;
   uint32 ulType;              /* Loop Type */
   uint32 ulStart;             /* Start of loop in samples */
   uint32 ulLength;            /* Length of loop in samples */
 } WLOOP;
 
typedef struct  {
	char id[4];
	int32 size;
} DLSChunk;

typedef struct {
    CONNECTIONLIST  connection_header;
    CONNECTION   *connection_list;
} LART;

typedef struct {
    WSMPL   wsmpl;
    WLOOP   *wloop_list;
} SAMPLE;

typedef struct {
    RGNHEADER  range;
    WAVELINK   wavelink;
    SAMPLE     *sample;
    LART       art;
} REGION;

typedef struct {
    INSTHEADER header;
    REGION     *region_list;
    LART       art;
    char       *name;
} INSTInfo;

typedef struct {
    uint16     formattag;
    uint16     channels;
    uint32     samplespersec;
    uint32     avgbytespersec;
    uint16     blockalign;
    uint16     bitspersample;
 } FMTInfo;

typedef struct {
    POOLCUE    poolcue;
    uint32     size;
    uint16     *data;
    FMTInfo    fmt;
    SAMPLE     *sample;
    char       *name;
} CUE;

typedef struct {
   int 		loaded;
   DLSHEADER 	header;
   DLSVERSION 	version;	
   char         *name;

   INSTInfo     *inst_list;

   POOLTABLE    pooltable;
   CUE          *cue_list;
   struct {
       int      inst;
       struct {
	   int  cue;
           int  wt;
       } note[128];
   } midi[128],drum[10];
} DLSInfo;

DLSInfo dlsinfo;

/******************************************************************************************/

typedef struct {
	char id[4];
	int32 size;
} MIDIChunk;

typedef struct {
        uint32   size;
	unsigned char *data;
	unsigned long pos;
	long     currenttime;
        long     deltatime;
        short    prev;
	short    next;
	unsigned char run_stat;
} TRACKInfo;

typedef struct {
        short format;		/* File format */
	short ntrks;		/* Number of tracks in the file */
	short division;		/* Quarter-note division */
	TRACKInfo *track_list;
	long  currenttime;
	long  nexttime;
	int   loaded;
	int   channel[16];
        short playing_tracks;
} MIDIInfo ;

MIDIInfo midiinfo;


#define MFE_OK		  0	/* No error */
#define MFE_EOF		 -1	/* Premature end of file */
#define MFE_HKWEXP	 -2	/* Header keyword expected */
#define MFE_ILLHLEN	 -3	/* Illegal header length */
#define MFE_ILLFORM	 -4	/* Illegal file format */
#define MFE_TKWEXP	 -5	/* Track keyword expected */
#define MFE_ILLSETYPE	 -6	/* Illegal SysEx message type */
#define MFE_ILLSTAT	 -7	/* Illegal status byte */
#define MFE_ILLMLEN	 -8	/* Illegal meta-event length */
#define MFE_ILLMVAL	 -9	/* Illegal value in meta-event */
#define MFE_UKMETA	-10	/* Unknown meta-event */
#define MFE_NOEOT	-11	/* No end-of-track meta-event */


char *err_msgs[] = {
	"No error",
	"End of file",
	"'MThd' expected",
	"Illegal header length",
	"Illegal file format",
	"'MTrk' expected",
	"Illegal system-exclusive message type",
	"Illegal status byte",
	"Illegal meta-event length",
	"Illegal value in meta-event",
	"Unknown meta-event",
	"Missing end-of-track meta-event",
	NULL
};


/* META-EVENT MESSAGE TYPES */
#define META_SEQNUM		0x00	/* Sequence number */
#define META_TEXT		0x01	/* Text event */
#define META_COPYRIGHT		0x02	/* Copyright notice */
#define META_SEQNAME		0x03	/* Sequence/track name */
#define META_INSTNAME		0x04	/* Instrument name */
#define META_LYRIC		0x05	/* Lyric */
#define META_MARKER		0x06	/* Marker */
#define META_CUEPT		0x07	/* Cue point */
#define META_PROGRAM            0x08    /* Program */
#define META_DEVICE             0x09    /* Device */
#define META_EOT		0x2f	/* End of track */
#define META_TEMPO		0x51	/* Set tempo */
#define META_SMPTE		0x54	/* SMPTE offset */
#define META_TIMESIG		0x58	/* Time signature */
#define META_KEYSIG		0x59	/* Key signature */
#define META_SEQSPEC		0x7f	/* Sequencer-specific event */
/******************************************************************************************/
#define NEW(type,nums)	(type*)malloc(sizeof(type) * (nums))

#define READID(var,fd)	fread(var, 4, 1, fd)
#define READSTR(var,fd)	fread(var, 20, 1, fd)

#define READCHUNK(var,fd)	fread(&var, 8, 1, fd)
#define READDW(var,fd)	fread(&var, 4, 1, fd)
#define READW(var,fd)	fread(&var, 2, 1, fd)
#define READB(var,fd)	fread(&var, 1, 1, fd)
#define SKIPB(fd)	{char dummy; fread(&dummy, 1, 1, fd);}
#define SKIPW(fd)	{short dummy; fread(&dummy, 2, 1, fd);}
#define SKIPDW(fd)	{int dummy; fread(&dummy, 4, 1, fd);}

#define FSKIP(size,fd)	fseek(fd, size, SEEK_CUR)

#define XCHG_SHORT(x) ((((x)&0xFF)<<8) | (((x)>>8)&0xFF))
#define XCHG_LONG(x) ((((x)&0xFF)<<24) | \
		      (((x)&0xFF00)<<8) | \
		      (((x)&0xFF0000)>>8) | \
		      (((x)>>24)&0xFF))

/******************************************************************************************/

int wt_preset_sample_idx[MAX_NO_OF_PRESET_SAMPLES];

int wt_inst_sample_idx[MAX_NO_OF_INST_SAMPLES];

int wt_select_dls_inst;

int wt_scope_type=0;

typedef struct {
	pthread_t memory_play_thread;

	snd_pcm_t          *pcm;
	snd_pcm_format_t   format;
	snd_pcm_uframes_t  period_size;
	snd_pcm_uframes_t  buffer_size;
	int idx;
	int button_type;
	int force_start,paused,going,loop;
	unsigned char *data8;
	int16         *data16;
	long current_sample,total_written;
	long start_sample,end_sample,start_loop,end_loop;
	int channels,bits_per_sample,samples_per_sec;
} WT_WAVE;

WT_WAVE wt_wave[16];

pthread_t 		capture_thread;
snd_pcm_t               *capture_pcm;
snd_pcm_format_t  	capture_format;
snd_pcm_uframes_t       capture_buffer_size;
snd_pcm_uframes_t       capture_period_size;
long                    capture_position;
int			capture_channels;
unsigned int		capture_rate;
int			capture_scope_type=0;
int			capture_scale=1;
gboolean                capturing;

unsigned char capture_data[8192];

int no_of_hwmixer;

typedef struct {
   pthread_t play_thread;
   int       button_type;
   int       idx;
   gboolean  v3d;
   gboolean  loop;

   snd_pcm_t 		*alsa_pcm;
   snd_pcm_hw_params_t 	*hwparams;
   snd_pcm_sw_params_t 	*swparams;

   unsigned int rate;
   unsigned int channels;
   snd_pcm_format_t format;
   int alsa_can_pause;
   gboolean audio_error;

   WaveFile *wf;
   char filename[FILENAME_MAX+1];
     
   snd_pcm_channel_area_t   *areas;
   int 			 alsa_bps;
   guint64		alsa_total_written;
   gboolean mmap;
   gboolean force_start, going, paused;
   gpointer buffer;

} HWMIX_WAVE;

HWMIX_WAVE hwmix[16],speaker_wave[4],v3d_wave[16];

int hwmix_total_channels;
int hwmix_fs_idx;

int v3d_page;
int v3d_init;
int no_of_3d_move;
int no_of_3d_play;

int v3d_reverb;
int v3d_raytracing;
int v3d_wavetracing;
int v3d_1st_order_reflection;
int v3d_2nd_order_reflection;
int v3d_fs_idx;

int v3d_ox,v3d_oy,v3d_m2;
int v3d_source_select;
int v3d_source_select_idx;
int v3d_listener_select;
int v3d_listener_change_direction;

int v3d_showpath;
int v3d_obstruction_circular;
int v3d_reflect_from_obj;

double v3d_wavelength;

typedef struct {
    double left;
    double right;
    double bottom;  
    double top;
} V3D_RECT;

struct {
   double x;
   double y;
   double z;

   double x_range;
   double y_range;
   double angle; 
   double speed;

   double p;
   double o;

   double p_range;
   double increment;

   double oh_angle;
   double oh_speed;

   double direction;
   double cone_inner_angle;
   double cone_outer_angle;

   int  dir;

   int  idx;
   int  move;
   int  move_type;

   int  play;

   int no_of_1st_reflect;
   struct {
	double x,y,z;
        double ix,iz,iy;
   } reflection[5];

   int no_of_2nd_reflect;
   struct {
	double x,y,z;
	double ix1,iy1,iz1;
	double ix2,iy2,iz2;
   } reflection2[25];

   char filename[FILENAME_MAX+1];
   
} v3d_source[MAX_NO_OF_3D_SOURCE];


struct {
   double x;
   double y;
   double z;

   double orientation;
   double elevation;
   double direction;
   double speed;
} v3d_listener;

struct {
	V3D_RECT rc;
        int flag;
} v3d_obstruction[4],v3d_reflection,v3d_occlusion[4];

struct {
	V3D_RECT rc[2];
        int flag;
} v3d_exclusion[4];

V3D_RECT *reflect_rc1,*reflect_rc2;


V3D_RECT v3d_obj[20];
int v3d_no_of_obj;


double v3d_obstruction_radius,v3d_obstruction_min_radius;
double v3d_obstruction_size;
double v3d_scaleup;

/******************************************************************************************/

int initial_o[MAX_NO_OF_3D_SOURCE]={ 8, 0, 4,12, 1, 2, 3, 5, 6, 7, 9,10,11,13,14,15};

int mix_idx[20]={0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19};
int hwmix_idx[16]={0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15};
int hwmix_vc_idx[64]={0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,
                    16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,
                    32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,
                    48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63};

int alsa_3d_idx[96]={0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,
                    16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,
                    32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,
                    48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,
                    64,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,
                    80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95};
struct {
   int left[NO_OF_BAND];
   int right[NO_OF_BAND];
   char setname[16];
} eqset[NO_OF_EQSET] = {  
                { {-5,-2, 1, 3, 8, 5, 0,-5,-7,-4},{-5,-2, 1, 3, 8, 5, 0,-5,-7,-4},"Jazz"},
                { {18,13, 9, 4,-1,-6,-3, 0, 6, 1},{18,13, 9, 4,-1,-6,-3, 0, 6, 1},"Rock"},
                { { 6, 5, 4, 3, 7,-5,-1, 4, 2,-2},{ 6, 5, 4, 3, 7,-5,-1, 4, 2,-2},"Classical"},
                { { 0, 0, 0, 2, 3, 6, 4,-2,-4, 0},{ 0, 0, 0, 2, 3, 6, 4,-2,-4, 0},"Pop"},
                { {20,14,11, 5, 0,-5,-3, 0, 8, 0},{20,14,11, 5, 0,-5,-3, 0, 8, 0},"Dance"},
                { { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},"Reset"} };

char scope_idx[16]={0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15};
char *scope_type_name[4]={"Dot","Line","Filled","Filled2"};

// speaker compensation

struct {
//    unsigned short left[NO_OF_BAND];
//    unsigned short right[NO_OF_BAND];
    char name[32];
} eq_speaker[5];
   
char *speaker_set_name[3] =   {"Headphone", "Stereo speakers", "Quad speakers (STAC9708)" };
char *speaker_stereo_name[2]= {"Left","Right"};
char *speaker_quad_name[4]=   {"Front Left","Front Right","Rear Left","Rear Right"};
char *speaker_stereo_wave[2]= {"outls.wav","outrs.wav"};
char *speaker_quad_wave[4]=   {"out1lq.wav","out1rq.wav","out2lq.wav","out2rq.wav"};
char *v3d_oeo[3]=             {"Occlusion","Exclusion","Obstruction"};
char *v3d_oeo_label[4]=       {"Left","Right","Front","Rear"};
char *v3d_movement_type[MAX_NO_OF_MOVE_TYPE]=   {"Fixed","Circular","Linear","Overhead"};
char *v3d_reflection_type[3]= {"Open area","Closed wall(Square)","Closed wall(Circle)"};

struct category_str {
	int val;
	const char *name;
};

static struct category_str con_category[] = {
	{ IEC958_AES1_CON_DAT, "DAT" },
	{ IEC958_AES1_CON_VCR, "VCR" },
	{ IEC958_AES1_CON_MICROPHONE, "microphone" },
	{ IEC958_AES1_CON_SYNTHESIZER, "synthesizer" },
	{ IEC958_AES1_CON_RATE_CONVERTER, "rate converter" },
	{ IEC958_AES1_CON_MIXER, "mixer" },
	{ IEC958_AES1_CON_SAMPLER, "sampler" },
	{ IEC958_AES1_CON_PCM_CODER, "PCM coder" },
	{ IEC958_AES1_CON_IEC908_CD, "CD" },
	{ IEC958_AES1_CON_NON_IEC908_CD, "non-IEC908 CD" },
	{ IEC958_AES1_CON_GENERAL, "general" },
};

#define ARRAY_SIZE(x) (int)(sizeof(x)/sizeof(x[0]))

/******************************************************************************************/

static GdkGC *get_pen(int idx,int nRed, int nGreen, int nBlue)
{
	GdkColor *c;
	GdkGC *gc;
	c = (GdkColor *)g_malloc(sizeof(GdkColor));
	c->red = nRed;
	c->green = nGreen;
	c->blue = nBlue;
	gdk_color_alloc(gdk_colormap_get_system(), c);
	gc = gdk_gc_new(pixmap[idx]);
	gdk_gc_set_foreground(gc, c);
	return gc;
}

static int get_band(GtkWidget *widget)
{
	int idx;
	for (idx=0; idx<NO_OF_BAND; idx++) {
		if ( widget == eqpeak_drawing[idx] ) {
			return idx;
		};
	};	
	return 0;
}

static void redraw_meters(int idx, int width, int height)
{
	int segment_width,segments,green_segments,red_segments,orange_segments;
	int seg,segs_on1,segs_on2;
        int seg_left2;
        
	segment_width = eq_ungang ?  (width >> 1) - 8 : width - 12 ;
	segments = (height - 6) / 4;
	green_segments = (segments / 4) * 3;
	red_segments = 2;
	orange_segments = segments - green_segments - red_segments;
	
	eqpeak_max = MAX(MAX( eqpeak_max, eqpeak_level.left[idx]),eqpeak_level.right[idx]);

	if ( eq_ungang ) {
	    segs_on1=segments*eqpeak_level.left[idx]/eqpeak_max;
            segs_on2=segments*eqpeak_level.right[idx]/eqpeak_max;
	    seg_left2=2 + (width >> 1);
	}
	else {
	    segs_on1 = segments*((eqpeak_level.left[idx]+eqpeak_level.right[idx])>>1)/eqpeak_max;
        };

	for (seg = 0; seg < green_segments; seg++) {
		gdk_draw_rectangle(pixmap[idx],
				   segs_on1 > 0 ? penGreenLight : penGreenShadow,
				   TRUE,
				   6, 3 + ((segments - seg - 1) << 2),
				   segment_width,
				   3);
		segs_on1--;
		if ( eq_ungang) {
			gdk_draw_rectangle(pixmap[idx],
					   segs_on2 > 0 ? penGreenLight : penGreenShadow,
					   TRUE,
					   seg_left2,
					   3 + ((segments - seg - 1) << 2),
					   segment_width,
					   3);
 		    	segs_on2--;
                };
	};
	for (seg = green_segments; seg < green_segments + orange_segments; seg++) {
		gdk_draw_rectangle(pixmap[idx],
				   segs_on1 > 0 ? penOrangeLight : penOrangeShadow,
				   TRUE,
				   6, 3 + ((segments - seg - 1) << 2),
				   segment_width,
				   3);
		segs_on1--;
		if ( eq_ungang ) {
			gdk_draw_rectangle(pixmap[idx],
				segs_on2 > 0 ? penOrangeLight : penOrangeShadow,
					   TRUE,
					   seg_left2,
					   3 + ((segments - seg - 1) << 2),
					   segment_width,
					   3);
		  	segs_on2--;
		};
	};
	for (seg = green_segments + orange_segments; seg < segments; seg++) {
		gdk_draw_rectangle(pixmap[idx],
				   segs_on1 > 0 ? penRedLight : penRedShadow,
				   TRUE,
				   6, 3 + ((segments - seg - 1) << 2),
				   segment_width,
				   3);
		segs_on1--;
		if ( eq_ungang) {
			gdk_draw_rectangle(pixmap[idx],
					   segs_on2 > 0 ? penRedLight : penRedShadow,
					   TRUE,
					   seg_left2,
					   3 + ((segments - seg - 1) << 2),
					   segment_width,
					   3);
			segs_on2--;
		};
	};
}

void redraw_10_meters(void)
{
	int idx;
	GtkWidget *widget;
        for (idx=0; idx<NO_OF_BAND; idx++) {
	    widget=eqpeak_drawing[idx];		
	    if ( GTK_WIDGET_VISIBLE(widget) ) {
		    redraw_meters(idx, widget->allocation.width, 
			   widget->allocation.height);
		    gdk_draw_rectangle(pixmap[idx],
			   widget->style->black_gc,
			   TRUE,
			   0, 0,
			   widget->allocation.width,
			   widget->allocation.height);
		};
	};
}

gint eqmeters_configure_event(GtkWidget *widget, GdkEventConfigure *event)
{
	int idx = get_band(widget);
	if (pixmap[idx] != NULL)
		gdk_pixmap_unref(pixmap[idx]);
	pixmap[idx] = gdk_pixmap_new(widget->window,
				     widget->allocation.width,
				     widget->allocation.height,
				     -1);
	penGreenShadow = get_pen(idx, 0, 0, 0);
	penGreenLight = get_pen(idx, 0, 0xffff, 0);
	penOrangeShadow = get_pen(idx, 0, 0, 0);
	penOrangeLight = get_pen(idx, 0xffff, 0x99ff, 0);
	penRedShadow = get_pen(idx, 0, 0, 0);
	penRedLight = get_pen(idx, 0xffff, 0, 0);
	penBlack = get_pen(idx, 0, 0, 0);
	penGrey = get_pen(idx, 0x8000, 0x8000, 0x8000);
	penWhite = get_pen(idx, 0xffff, 0xffff, 0xffff);

	pen3DSource[0] = get_pen(idx, 0, 0, 0xffff);
	pen3DSource[1] = get_pen(idx, 0xffff, 0, 0);
	pen3DSource[2] = get_pen(idx, 0, 0xffff, 0);
	pen3DSource[3] = get_pen(idx, 0xffff, 0xffff, 0xffff);
	     
	pen3DSource[5] = get_pen(idx, 0xffff, 0xffff, 0);
	pen3DSource[4] = get_pen(idx, 0, 0xffff, 0xffff);
	pen3DSource[6] = get_pen(idx, 0xffff, 0, 0xffff);

	pen3DSource[7] = get_pen(idx, 0, 0, 0x8000);
	pen3DSource[8] = get_pen(idx, 0x8000, 0, 0);
	pen3DSource[9] = get_pen(idx, 0, 0x8000, 0);

	pen3DSource[10] = get_pen(idx, 0, 0x8000, 0xffff);
	pen3DSource[11] = get_pen(idx, 0xffff, 0, 0x8000);
	pen3DSource[12] = get_pen(idx, 0x8000, 0xffff, 0);

	pen3DSource[13] = get_pen(idx, 0x8000, 0, 0xffff);
	pen3DSource[14] = get_pen(idx, 0xffff, 0x8000, 0);
	pen3DSource[15] = get_pen(idx, 0, 0xffff, 0x8000);

	gdk_draw_rectangle(pixmap[idx],
			   widget->style->black_gc,
			   TRUE,
			   0, 0,
			   widget->allocation.width,
			   widget->allocation.height);
	redraw_meters(idx, widget->allocation.width, widget->allocation.height);
	return TRUE;
}

gint eqmeters_expose_event(GtkWidget *widget, GdkEventExpose *event)
{
	int idx = get_band(widget);
	redraw_meters(idx, widget->allocation.width, widget->allocation.height); 
	gdk_draw_pixmap(widget->window,
			widget->style->black_gc,
			pixmap[idx],
			event->area.x, event->area.y,
			event->area.x, event->area.y,
			event->area.width, event->area.height);
	return FALSE;
}

gint equalizer_update_eqpeak_callback(gpointer data)
{
	GtkWidget *widget;
	int err;
	int idx;
	snd_ctl_elem_info_alloca(&info);
	snd_ctl_elem_value_alloca(&val);
	
        if ( (err= snd_hctl_elem_info(eqpeak_elem, info)) < 0) {
		g_print("snd_hctl_elem_info : %s\n",snd_strerror(err));
	}
	else {
	    if ( snd_ctl_elem_info_get_count(info) == 20 ) {
		if ((err = snd_hctl_elem_read(eqpeak_elem,val)) < 0) {
	 	    g_print("snd_hctl_elem_read : %s\n",snd_strerror(err));
		}
		else {
		    for (idx=0; idx<NO_OF_BAND; idx++) {
			eqpeak_level.left[idx]=snd_ctl_elem_value_get_integer(val,idx);
			eqpeak_level.right[idx]=snd_ctl_elem_value_get_integer(val,idx+NO_OF_BAND);

//			eqpeak_min = MIN(MIN( eqpeak_min, eqpeak_level.left[idx]),eqpeak_level.right[idx]);
			eqpeak_max = MAX(MAX( eqpeak_max, eqpeak_level.left[idx]),eqpeak_level.right[idx]);

		    }; 
		    for (idx=0; idx<NO_OF_BAND; idx++) {
 		        widget = eqpeak_drawing[idx];
		        if ( GTK_WIDGET_VISIBLE(widget)) {
		            redraw_meters(idx, widget->allocation.width, widget->allocation.height);
			    gdk_draw_pixmap(widget->window,
					widget->style->black_gc,
					pixmap[idx],
					0, 0,
					0, 0,
					widget->allocation.width, widget->allocation.height);	
                        };
		    };
		    return eq_power;
		};
	    };
        };
    	return eq_power;
}

static void eq_adj_left_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	snd_ctl_elem_value_t *val;
	int idx;
	int err;
	snd_ctl_elem_info_alloca(&info);
	snd_ctl_elem_value_alloca(&val);
	for (idx=0; idx<NO_OF_BAND; idx++) {
 	    if ( adj == GTK_ADJUSTMENT(eq_adj[idx][0]) ) {
		elem=eq_adj_elem[idx];
         	if ( (err= snd_hctl_elem_info(elem, info)) < 0) {
			g_print("snd_hctl_elem_info : %s\n",snd_strerror(err));
		}
		else {
		    if ((err = snd_hctl_elem_read(elem,val)) < 0) {
			g_print("snd_hctl_elem_read : %s\n",snd_strerror(err));
		    }
		    else {
			if ( snd_ctl_elem_info_is_writable(info) ) {
			    snd_ctl_elem_value_set_integer(val,0,adj->value);
                            if ( eq_lock ) {
 			        snd_ctl_elem_value_set_integer(val,1,adj->value);
            		        gtk_adjustment_set_value(GTK_ADJUSTMENT(eq_adj[idx][1]),adj->value);
                	    };
			    if ( (err = snd_hctl_elem_write(elem,val)) < 0 ) {
			        g_print("snd_hctl_elem_write : %s\n",snd_strerror(err));
			    };
			    return;
			};
		    };
		};
	    };
	};
}

static void eq_adj_right_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	snd_ctl_elem_value_t *val;
	int idx;
	int err;

	snd_ctl_elem_info_alloca(&info);
	snd_ctl_elem_value_alloca(&val);
	for (idx=0; idx<NO_OF_BAND; idx++) {
 	    if ( adj == GTK_ADJUSTMENT(eq_adj[idx][1]) ) {
		elem=eq_adj_elem[idx];
         	if ( (err= snd_hctl_elem_info(elem, info)) < 0) {
			g_print("snd_hctl_elem_info : %s\n",snd_strerror(err));
		}
		else {
		    if ((err = snd_hctl_elem_read(elem,val)) < 0) {
			g_print("snd_hctl_elem_read : %s\n",snd_strerror(err));
		    }
		    else {
		        if ( snd_ctl_elem_info_is_writable(info) ) {
			    snd_ctl_elem_value_set_integer(val,1,adj->value);
                            if ( eq_lock ) {
 			        snd_ctl_elem_value_set_integer(val,0,adj->value);
	           	        gtk_adjustment_set_value(GTK_ADJUSTMENT(eq_adj[idx][0]),adj->value);
			    };
			    if ( (err = snd_hctl_elem_write(elem,val)) < 0 ) {
				g_print("snd_hctl_elem_write : %s\n",snd_strerror(err));
			    };
			    return;
			};
		    };
	        };
	    };
	};
}

void eqset_toggle_button_cb (GtkWidget *widget, gpointer data)
{
    int idx,k;
    if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
// equalizer setting override speaker compensation
//        gtk_widget_set_sensitive(eq_speaker_button,FALSE);
	for (k=0; k<NO_OF_EQSET; k++) {
	    if ( widget == eqset_button[k] ) {
	        for (idx=0; idx<NO_OF_BAND; idx++) {
	            gtk_adjustment_set_value(GTK_ADJUSTMENT(eq_adj[idx][0]),0x3e96+eqset[k].left[idx]*800);
	            gtk_adjustment_set_value(GTK_ADJUSTMENT(eq_adj[idx][1]),0x3e96+eqset[k].right[idx]*800);
                };
	    }
	    else {
		if ( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(eqset_button[k]))) {
		     gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(eqset_button[k]),FALSE);
		};
	    };
	};    
     } 
     else {
//        gtk_widget_set_sensitive(eq_speaker_button,TRUE);
     };
}
/*
void eq_speaker_toggle_button_cb (GtkWidget *widget, gpointer data)
{
}
*/
void eqpower_toggle_button_cb (GtkWidget *widget, gpointer data)
{
    int idx,err;

    snd_ctl_elem_info_alloca(&info);
    snd_ctl_elem_value_alloca(&val);
    elem=eqpower_elem;
    if ( (err= snd_hctl_elem_info(elem, info)) < 0) {
	g_print("snd_hctl_elem_info : %s\n",snd_strerror(err));
    }
    else {
	if ((err = snd_hctl_elem_read(elem,val)) < 0) {
	    g_print("snd_hctl_elem_read : %s\n",snd_strerror(err));
	}
	else {
	    if ( snd_ctl_elem_info_is_writable(info) ) {
	        snd_ctl_elem_value_set_boolean(val,0,gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)));
	        if ( (err = snd_hctl_elem_write(elem,val)) < 0 ) {
	  	    g_print("snd_hctl_elem_write : %s\n",snd_strerror(err));
		};

     	        if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
		    eq_power=1;
	            eqpeak_timer_tag=gtk_timeout_add(40, equalizer_update_eqpeak_callback, NULL);
	        } 
                else {    
		    eq_power=0;
		    gtk_timeout_remove(eqpeak_timer_tag);
    	        };
		for (idx=0; idx<NO_OF_BAND; idx++) {
    		    eqpeak_level.left[idx]=0;
		    eqpeak_level.right[idx]=0;
		};
                eqpeak_min=0;
		eqpeak_max=eqpeak_min+64;
		redraw_10_meters();
	    };
	};
    };
}

void eqlock_toggle_button_cb (GtkWidget *widget, gpointer data)
{
    if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) 
	eq_lock=1;
    else 
	eq_lock=0;
}

void eqgang_toggle_button_cb (GtkWidget *widget, gpointer data)
{
    if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) 
	eq_ungang=0;
    else
    	eq_ungang=1;
    redraw_10_meters();
}

static void create_equalizer(GtkWidget *main, GtkWidget *notebook, int page)
{
	GtkWidget *label;
	GtkWidget *vbox;
        GtkWidget *hbox;
	GtkWidget *table;
	char idname[128];
	int i,k;
	int count;
	int err;

	snd_ctl_elem_id_alloca(&id);
	snd_ctl_elem_info_alloca(&info);
	snd_ctl_elem_value_alloca(&val);


	vbox = gtk_vbox_new(FALSE, 0);
	gtk_widget_show(vbox);

	gtk_container_add(GTK_CONTAINER(notebook), vbox);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 6);


        label = gtk_label_new("Hardware Equalizer");
        gtk_widget_show(label);
	gtk_notebook_set_tab_label(GTK_NOTEBOOK(notebook), 
				   gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), page), 
				   label);

	for (i=0; i<NO_OF_BAND; i++) {
		eqpeak_level.left[i]=0;
		eqpeak_level.right[i]=0;
	};		
	eqpeak_min=0;
	eqpeak_max=64;
	
	hbox=gtk_hbox_new(FALSE,0);
	gtk_widget_show(hbox);
	gtk_container_add(GTK_CONTAINER(vbox), hbox);


	for (i=0; i<NO_OF_EQSET; i++) {
 	    eqset_button[i]=gtk_toggle_button_new_with_label(eqset[i].setname);
	    gtk_widget_show(eqset_button[i]);		
	    gtk_container_add(GTK_CONTAINER(hbox), eqset_button[i]);
            gtk_signal_connect (GTK_OBJECT (eqset_button[i]), "toggled",
                       GTK_SIGNAL_FUNC (eqset_toggle_button_cb), NULL);
	};
/*
	label=gtk_label_new(" ");
	gtk_widget_show(label);
	gtk_container_add(GTK_CONTAINER(hbox), label);

	eq_speaker_button=gtk_toggle_button_new_with_label("Speaker");
        gtk_widget_show(eq_speaker_button);
        gtk_container_add(GTK_CONTAINER(hbox), eq_speaker_button);
        gtk_signal_connect (GTK_OBJECT (eq_speaker_button), "toggled",
                       GTK_SIGNAL_FUNC (eq_speaker_toggle_button_cb), NULL);

*/
	label=gtk_label_new(" ");
	gtk_widget_show(label);
	gtk_container_add(GTK_CONTAINER(hbox), label);


	eqlock_button=gtk_toggle_button_new_with_label("Lock");
	gtk_widget_show(eqlock_button);		
	gtk_container_add(GTK_CONTAINER(hbox), eqlock_button);
        gtk_signal_connect (GTK_OBJECT (eqlock_button), "toggled",
                       GTK_SIGNAL_FUNC (eqlock_toggle_button_cb), NULL);

	eqgang_button=gtk_toggle_button_new_with_label("Gang");
	gtk_widget_show(eqgang_button);		
	gtk_container_add(GTK_CONTAINER(hbox), eqgang_button);
        gtk_signal_connect (GTK_OBJECT (eqgang_button), "toggled",
                       GTK_SIGNAL_FUNC (eqgang_toggle_button_cb), NULL);


	eqpower_button=gtk_toggle_button_new_with_label("Power");
	gtk_widget_show(eqpower_button);		
	gtk_container_add(GTK_CONTAINER(hbox), eqpower_button);
        gtk_signal_connect (GTK_OBJECT (eqpower_button), "toggled",
                       GTK_SIGNAL_FUNC (eqpower_toggle_button_cb), NULL);


	table = gtk_table_new(3,NO_OF_BAND,FALSE);
        gtk_widget_show(table);
	gtk_container_add (GTK_CONTAINER (vbox), table);

	i=0;
	for (elem = snd_hctl_first_elem(hctl); elem; elem = snd_hctl_elem_next(elem)) {
            if ( (err= snd_hctl_elem_info(elem, info)) < 0) {
		g_print("snd_hctl_elem_info : %s\n",snd_strerror(err));
	    }
	    else {
 	        snd_hctl_elem_get_id(elem, id);		
		strcpy(idname,snd_ctl_elem_id_get_name(id));
		if ( memcmp(&idname[0],"EQ",2) == 0 ) {
		    count=snd_ctl_elem_info_get_count(info);
		    switch (snd_ctl_elem_info_get_type(info)) {
    		    case SND_CTL_ELEM_TYPE_BOOLEAN:
			eqpower_elem=elem;
		        if ((err = snd_hctl_elem_read(elem, val)) < 0) {
			    g_print("snd_hctl_elem_read : %s\n",snd_strerror(err));
			}
			else {
//  turn on/off display of EQ peak.
                            eq_power= snd_ctl_elem_value_get_boolean(val,0);
                        };
			break;
		    case SND_CTL_ELEM_TYPE_INTEGER:
			switch (count) {
			case 20:
// set up EQ peak min,max 
			    eqpeak_elem=elem;
			    if ((err = snd_hctl_elem_read(elem, val)) < 0) {
			        g_print("snd_hctl_elem_read : %s\n",snd_strerror(err));
			    }
			    else {
			        eqpeak_min=snd_ctl_elem_info_get_min(info);
			        eqpeak_max=snd_ctl_elem_info_get_max(info);

				for (i=0; i<NO_OF_BAND; i++) {
				    eqpeak_level.left[i]=snd_ctl_elem_value_get_integer(val,i);
				    eqpeak_level.right[i]=snd_ctl_elem_value_get_integer(val,i+NO_OF_BAND);
			            eqpeak_drawing[i] = gtk_drawing_area_new();
		 	            gtk_widget_set_name(eqpeak_drawing[i], "EQ Band");
	                            gtk_widget_set_usize(eqpeak_drawing[i], 45, 202);
	                            gtk_widget_show(eqpeak_drawing[i]);

				    gtk_table_attach_defaults(GTK_TABLE(table),eqpeak_drawing[i],i,i+1,2,3);
 	                            gtk_signal_connect(GTK_OBJECT(eqpeak_drawing[i]), 
					  	"expose_event",
			                       	GTK_SIGNAL_FUNC(eqmeters_expose_event), NULL);
	                            gtk_signal_connect(GTK_OBJECT(eqpeak_drawing[i]), 							"configure_event",
			                       	GTK_SIGNAL_FUNC(eqmeters_configure_event), NULL);
	                            gtk_widget_set_events(eqpeak_drawing[i], GDK_EXPOSURE_MASK);
			        };
			    };	
			    i=0;				
			    break;
		        case 2:
// EQ level adjustment
			    idname[strlen(idname)-2]=0;
			    label= gtk_label_new(idname+3);
			    gtk_table_attach_defaults(GTK_TABLE(table),label,i,i+1,1,2);	
			    gtk_widget_show(label);

			    if ((err = snd_hctl_elem_read(elem, val)) < 0) {
			        g_print("snd_hctl_elem_read : %s\n",snd_strerror(err));
			    }
			    else {
			        hbox=gtk_hbox_new(FALSE,0);
			        gtk_widget_show(hbox);

			        label= gtk_label_new("   ");
			        gtk_container_add(GTK_CONTAINER(hbox), label);
			        gtk_widget_show(label);

			        eq_adj_elem[i]=elem;
				
			        for (k=0; k<2; k++) {
			   	    eq_adj[i][k] = gtk_adjustment_new (
					(gdouble)(snd_ctl_elem_value_get_integer(val,k)*1.0), 
					snd_ctl_elem_info_get_min(info), 
					(snd_ctl_elem_info_get_max(info)+1.0), 
            				1.0,1.0,1.0);
				    eq_vscale[i][k]=gtk_vscale_new(GTK_ADJUSTMENT(eq_adj[i][k]));
//				    gtk_scale_set_value_pos(GTK_SCALE(eq_vscale[i][k]),2);
				    gtk_scale_set_draw_value(GTK_SCALE(eq_vscale[i][k]),0);
				    gtk_scale_set_digits(GTK_SCALE(eq_vscale[i][k]),0);
				    gtk_range_set_inverted(GTK_RANGE(eq_vscale[i][k]),TRUE);
	                            gtk_widget_set_usize(eq_vscale[i][k], 22, 102);
				    gtk_widget_show(eq_vscale[i][k]);
				    gtk_container_add(GTK_CONTAINER(hbox), eq_vscale[i][k]);
				    if ( k == 0 ) {
	       				gtk_signal_connect (GTK_OBJECT (eq_adj[i][k]), 
						"value_changed",
					GTK_SIGNAL_FUNC (eq_adj_left_value_changed_cb), 
						NULL);
				    }
				    else {
	       				gtk_signal_connect (GTK_OBJECT (eq_adj[i][k]), 
						"value_changed",
					GTK_SIGNAL_FUNC (eq_adj_right_value_changed_cb), 
						NULL);					
				    };
			        };

			        label= gtk_label_new("   ");
			        gtk_container_add(GTK_CONTAINER(hbox), label);
			        gtk_widget_show(label);

			        gtk_table_attach_defaults(GTK_TABLE(table),hbox,i,i+1,0,1);
                                i++;
			    };
			    break;
			};
			break;
		    default:
			break;
		    };
		};
	    };
	};
}

/******************************************************************************************/

static void mixer_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
    int err;
    snd_hctl_elem_t *elem;

    snd_ctl_elem_info_alloca(&info);
    snd_ctl_elem_value_alloca(&val);

    elem=data;
    if ( (err= snd_hctl_elem_info(elem, info)) < 0) {
	g_print("snd_hctl_elem_info : %s\n",snd_strerror(err));
    }
    else {
	if ((err = snd_hctl_elem_read(elem,val)) < 0) {
	    g_print("snd_hctl_elem_read : %s\n",snd_strerror(err));
	}
	else {
	    if ( snd_ctl_elem_info_is_writable(info) ) {
	        snd_ctl_elem_value_set_integer(val,0,adj->value);
	        if ( (err = snd_hctl_elem_write(elem,val)) < 0 ) {
		    g_print("snd_hctl_elem_write : %s\n",snd_strerror(err));
		};
	    };
	};
    };
}

static void mix_adj_left_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
    int idx,err;
    int *p;
    snd_hctl_elem_t *elem;
    snd_ctl_elem_info_alloca(&info);
    snd_ctl_elem_value_alloca(&val);

    p=data;
    idx=*p;
    elem=mix_adj_elem[idx];	    
    if ( (err= snd_hctl_elem_info(elem, info)) < 0) {
	g_print("snd_hctl_elem_info : %s\n",snd_strerror(err));
    }
    else {
	if ((err = snd_hctl_elem_read(elem,val)) < 0) {
	    g_print("snd_hctl_elem_read : %s\n",snd_strerror(err));
	}
	else {
	    if ( snd_ctl_elem_info_is_writable(info) ) {
	        snd_ctl_elem_value_set_integer(val,0,adj->value);
                if ( gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(mix_lock_button[idx]))) {
                    snd_ctl_elem_value_set_integer(val,1,adj->value);
	            gtk_adjustment_set_value(GTK_ADJUSTMENT(mix_adj[idx][1]),adj->value);
       		};
	        if ( (err = snd_hctl_elem_write(elem,val)) < 0 ) {
		    g_print("snd_hctl_elem_write : %s\n",snd_strerror(err));
		};
	    };
	};
    };
}

static void mix_adj_right_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
    int idx,err;
    int *p;
    snd_hctl_elem_t *elem;
    snd_ctl_elem_info_alloca(&info);
    snd_ctl_elem_value_alloca(&val);

    p=data;
    idx=*p;
    elem=mix_adj_elem[idx];
    if ( (err= snd_hctl_elem_info(elem, info)) < 0) {
	g_print("snd_hctl_elem_info : %s\n",snd_strerror(err));
    }
    else {
	if ((err = snd_hctl_elem_read(elem,val)) < 0) {
	    g_print("snd_hctl_elem_read : %s\n",snd_strerror(err));
	}
	else {
	    if ( snd_ctl_elem_info_is_writable(info) ) {
	        snd_ctl_elem_value_set_integer(val,1,adj->value);
                if ( gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(mix_lock_button[idx]))) {
                    snd_ctl_elem_value_set_integer(val,0,adj->value);
     	            gtk_adjustment_set_value(GTK_ADJUSTMENT(mix_adj[idx][0]),adj->value);
       		};
	        if ( (err = snd_hctl_elem_write(elem,val)) < 0 ) {
		    g_print("snd_hctl_elem_write : %s\n",snd_strerror(err));
		};
	    };
	};
    };
}

void mixer_mute_toggle_button_cb (GtkWidget *widget, gpointer data)
{
    int err,count,i;
    snd_hctl_elem_t *elem;
    snd_ctl_elem_info_alloca(&info);
    snd_ctl_elem_value_alloca(&val);

    elem=data;
    if ( (err= snd_hctl_elem_info(elem, info)) < 0) {
	g_print("snd_hctl_elem_info : %s\n",snd_strerror(err));
    }
    else {
	count=snd_ctl_elem_info_get_count(info);
	if ((err = snd_hctl_elem_read(elem,val)) < 0) {
	    g_print("snd_hctl_elem_read : %s\n",snd_strerror(err));
	}
	else {
	    if ( snd_ctl_elem_info_is_writable(info) ) {
	        if ( gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
                    for (i=0; i<count; i++) {
                        snd_ctl_elem_value_set_boolean(val,i,FALSE);
                    }
		}
	        else {
		    for (i=0; i<count; i++) {
                        snd_ctl_elem_value_set_boolean(val,i,TRUE);
		    };
                };			
	        if ( (err = snd_hctl_elem_write(elem,val)) < 0 ) {
		    g_print("snd_hctl_elem_write : %s\n",snd_strerror(err));
		};
	    };
	};
    };
}

void mixer_rec_left_toggle_button_cb (GtkWidget *widget, gpointer data)
{
    int err;
    int left_src,right_src;
    int *p;
    snd_hctl_elem_t *elem;
    snd_ctl_elem_info_alloca(&info);
    snd_ctl_elem_value_alloca(&val);
 
    elem=mix_capsrc_elem;	
    p=data;
    left_src=*p;
    if ( (err= snd_hctl_elem_info(elem, info)) < 0) {
	g_print("snd_hctl_elem_info : %s\n",snd_strerror(err));
    }
    else {
	if ((err = snd_hctl_elem_read(elem, val)) < 0) {             
	    g_print("snd_hctl_elem_read : %s\n",snd_strerror(err));
	}
	else {
            if (  gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(widget)) ) {
		right_src=snd_ctl_elem_value_get_integer(val,1);
                snd_ctl_elem_value_set_integer(val,0,left_src);
	        if ( gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(mix_rec_lock)) ) {
	 	    right_src=left_src;
	            gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mix_rec_radio[right_src][1]),TRUE);
		};
                snd_ctl_elem_value_set_integer(val,1,right_src);
		if ( (err = snd_hctl_elem_write(elem,val)) < 0 ) {
		    g_print("snd_hctl_elem_write : %s\n",snd_strerror(err));
		};
		return;
	    };
        };
    };			
}

void mixer_rec_right_toggle_button_cb (GtkWidget *widget, gpointer data)
{
    int err;
    int *p;
    int left_src,right_src;
    snd_hctl_elem_t *elem;
    snd_ctl_elem_info_alloca(&info);
    snd_ctl_elem_value_alloca(&val);

    elem=mix_capsrc_elem;	
    p=data;
    right_src=*p;	
    if ( (err= snd_hctl_elem_info(elem, info)) < 0) {
	g_print("snd_hctl_elem_info : %s\n",snd_strerror(err));
    }
    else {
	if ((err = snd_hctl_elem_read(elem, val)) < 0) {             
	    g_print("snd_hctl_elem_read : %s\n",snd_strerror(err));
	}
	else {
            if (  gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(widget)) ) {
		left_src=snd_ctl_elem_value_get_integer(val,0);
                snd_ctl_elem_value_set_integer(val,1,right_src);
	        if ( gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(mix_rec_lock)) ) {
		    left_src=right_src;
                    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mix_rec_radio[left_src][0]),TRUE);
		};
                snd_ctl_elem_value_set_integer(val,0,left_src);
	        if ( (err = snd_hctl_elem_write(elem,val)) < 0 ) {
		    g_print("snd_hctl_elem_write : %s\n",snd_strerror(err));
		};
		return;
	    };
        };
    };			
}

void mixer_enable_button_cb (GtkWidget *widget, gpointer data)
{
    int err;
    snd_hctl_elem_t *elem;
    snd_ctl_elem_info_alloca(&info);
    snd_ctl_elem_value_alloca(&val);

    elem=data;
    if ( (err= snd_hctl_elem_info(elem, info)) < 0) {
	g_print("snd_hctl_elem_info : %s\n",snd_strerror(err));
    }
    else {
	if ((err = snd_hctl_elem_read(elem,val)) < 0) {
	    g_print("snd_hctl_elem_read : %s\n",snd_strerror(err));
	}
	else {
	    if ( snd_ctl_elem_info_is_writable(info) ) {
		if ( gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)) ) {
 		    snd_ctl_elem_value_set_boolean(val,0,TRUE);
		}
		else {
		    snd_ctl_elem_value_set_boolean(val,0,FALSE);
		};
	        if ( (err = snd_hctl_elem_write(elem,val)) < 0 ) {
		    g_print("snd_hctl_elem_write : %s\n",snd_strerror(err));
		};
	    };
	};
    };
}

void mixer_enum_button_cb (GtkWidget *widget, gpointer data)
{
    int idx,err;
    int *p;
    snd_hctl_elem_t *elem;
    snd_ctl_elem_info_alloca(&info);
    snd_ctl_elem_value_alloca(&val);
    p=data;
    idx=*p;
    elem=mix_enum_elem[mix_enum[idx].elem_idx];	
    if ( widget == mix_enum_radio[idx] ) {
       if ( (err= snd_hctl_elem_info(elem, info)) < 0) {
            g_print("snd_hctl_elem_info : %s\n",snd_strerror(err));
       }
       else {
            if ((err = snd_hctl_elem_read(elem, val)) < 0) {             
                g_print("snd_hctl_elem_read : %s\n",snd_strerror(err));
            }
            else {
                if (  gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(widget)) ) {
                    snd_ctl_elem_value_set_integer(val,0,mix_enum[idx].value);
                    if ( (err = snd_hctl_elem_write(elem,val)) < 0 ) {
      	                g_print("snd_hctl_elem_write : %s\n",snd_strerror(err));
		    };
  	        };
	    };
        };
    };			
}

static void create_mixer(GtkWidget *main, GtkWidget *notebook, int page)
{
	GtkWidget *label;
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkWidget *capture_vbox;
	GtkWidget *left_right_table;
	GtkWidget *ac97_3d_hbox;
	GtkWidget *misc_hbox;
	GtkWidget *enable_hbox;
	GtkWidget *mixer_table;
	GtkWidget *frame;

	int i,m;
	int err;
        int count,type;
	int prev_boolean,current_boolean;
	int prev_count;
	unsigned int item,items;
	int cap_left_right[2];
	int enum_value;
        int  adj_min,adj_max;
        char idname[128];
	char newname[128];

	iec958_present=0;

	vbox = gtk_vbox_new(FALSE, 0);
	gtk_widget_show(vbox);
	gtk_container_add(GTK_CONTAINER(notebook), vbox);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 6);

        label = gtk_label_new(snd_ctl_card_info_get_mixername(hw_info));
        gtk_widget_show(label);
	gtk_notebook_set_tab_label(GTK_NOTEBOOK(notebook), 
				   gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), page), 
				   label);


	mixer_table=gtk_table_new(5,MAX_NO_OF_PLAYBACK,FALSE);
        gtk_widget_show(mixer_table);
	gtk_container_add (GTK_CONTAINER (vbox), mixer_table);
//	gtk_container_set_border_width(GTK_CONTAINER(mixer_table), 6);

	capture_vbox = gtk_vbox_new(FALSE, 0);
	gtk_widget_show(capture_vbox);
	gtk_container_add(GTK_CONTAINER(vbox), capture_vbox);
//	gtk_container_set_border_width(GTK_CONTAINER(capture_vbox), 6);

	misc_hbox = gtk_hbox_new(FALSE, 0);
	gtk_widget_show(misc_hbox);
	gtk_container_add(GTK_CONTAINER(vbox), misc_hbox);
//	gtk_container_set_border_width(GTK_CONTAINER(misc_hbox), 6);

	ac97_3d_hbox=gtk_hbox_new(FALSE,0);
	gtk_widget_show(ac97_3d_hbox);
        gtk_container_add(GTK_CONTAINER(vbox), ac97_3d_hbox);		
//	gtk_container_set_border_width(GTK_CONTAINER(ac97_3d_hbox), 6);

	enable_hbox = gtk_hbox_new(FALSE, 0);
	gtk_widget_show(enable_hbox);
	gtk_container_add(GTK_CONTAINER(vbox), enable_hbox);
//	gtk_container_set_border_width(GTK_CONTAINER(enable_hbox), 6);


	no_of_playback=0;
	no_of_enum_elem=0;
	no_of_enum_radio=0;
	no_of_capture_source=0;
	no_of_ac97_3d=0;
	sdac_mixer_idx=-1;

 	snd_ctl_elem_id_alloca(&id);
	snd_ctl_elem_info_alloca(&info);
	snd_ctl_elem_value_alloca(&val);

	for (elem = snd_hctl_first_elem(hctl); elem; elem = snd_hctl_elem_next(elem)) {
            if ( (err= snd_hctl_elem_info(elem, info)) < 0) {
		g_print("snd_hctl_elem_info : %s\n",snd_strerror(err));
	    }
	    else {
                count =snd_ctl_elem_info_get_count(info);
		type = snd_ctl_elem_info_get_type(info);
		if ((err = snd_hctl_elem_read(elem, val)) < 0) {             
		    g_print("snd_hctl_elem_read : %s\n",snd_strerror(err));
		}
		else {
		    snd_hctl_elem_get_id(elem, id);
		    strcpy(idname,snd_ctl_elem_id_get_name(id));
//		    g_print("%s\n",idname);
		    switch (type) {
		    case SND_CTL_ELEM_TYPE_INTEGER:
			adj_min=snd_ctl_elem_info_get_min(info);
			adj_max=snd_ctl_elem_info_get_max(info);

                        if ( strstr(idname,"Volume") > 0 ) {			    
			    m=no_of_playback;
			    if ( strstr(idname,"Playback Volume") > 0 )
			        idname[strlen(idname)-16]=0;

			    if ( strstr(idname,"Surround") > 0 ) {
			        if ( ac97_has_sdac ) {
				    if ( prev_count == 2 ) {
					sdac_master_idx=m;
				    }
				    else {
					sdac_mixer_idx=m;
				    };
				};
			    };

                            for (i=0; i<strlen(idname); i++) 
				if ( idname[i] == ' ' )
				    idname[i]='\n';

			    label=gtk_label_new(idname);
			    gtk_label_set_line_wrap (GTK_LABEL(label),TRUE);
			    gtk_widget_show(label);
			    gtk_table_attach_defaults(GTK_TABLE(mixer_table),label,m,m+1,0,1);	

                            mix_adj_elem[m]=elem;
			    hbox = gtk_hbox_new(FALSE, 0);
			    gtk_widget_show(hbox);
			    gtk_table_attach_defaults(GTK_TABLE(mixer_table),hbox,m,m+1,1,2);	
			    vc[m]=m;
			    for (i=0; i<count; i++) {
			        mix_adj[m][i]=gtk_adjustment_new(
					(gdouble)(snd_ctl_elem_value_get_integer(val,i)*1.0), 	
					(gdouble)(adj_min), 
					(gdouble)(adj_max+1.0), 
        	    			(gdouble)1.0,(gdouble)1.0,(gdouble)1.0);

			        mix_vscale[m][i]=gtk_vscale_new(GTK_ADJUSTMENT(mix_adj[m][i]));
			        gtk_widget_show(mix_vscale[m][i]);
			        gtk_container_add(GTK_CONTAINER(hbox), mix_vscale[m][i]);

				gtk_scale_set_value_pos(GTK_SCALE(mix_vscale[m][i]),3);
			        gtk_scale_set_draw_value(GTK_SCALE(mix_vscale[m][i]),0);
				gtk_scale_set_digits(GTK_SCALE(mix_vscale[m][i]),0);

			        gtk_range_set_inverted(GTK_RANGE(mix_vscale[m][i]),TRUE);

        	                gtk_widget_set_usize(mix_vscale[m][i], 22, 102);
				if ( count == 1 ) {
	 			        gtk_signal_connect (GTK_OBJECT (mix_adj[m][i]), 
						"value_changed",
						GTK_SIGNAL_FUNC (mixer_adj_value_changed_cb), 
						mix_adj_elem[m]);
				}
				else {
				    if ( i == 0 ) {
	 			        gtk_signal_connect (GTK_OBJECT (mix_adj[m][i]), 
						"value_changed",
					GTK_SIGNAL_FUNC (mix_adj_left_value_changed_cb), 
						&vc[m]);
				    }
				    else {
	 			        gtk_signal_connect (GTK_OBJECT (mix_adj[m][i]), 
						"value_changed",
					GTK_SIGNAL_FUNC (mix_adj_right_value_changed_cb), 
						&vc[m]);
				    };
				};
                            };

			    if ( count == 2 ) {
                                mix_lock_button[m]=gtk_toggle_button_new_with_label("Lock");
			        gtk_widget_show(mix_lock_button[m]);
			        gtk_table_attach_defaults(GTK_TABLE(mixer_table),mix_lock_button[m],m,m+1,2,3);	
				gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mix_lock_button[m]),(snd_ctl_elem_value_get_integer(val,0) == snd_ctl_elem_value_get_integer(val,1)));
                            };
			    mix_mute_elem[m]=prev_elem;
                            mix_mute_button[m]=gtk_toggle_button_new_with_label("Mute");
			    gtk_widget_show(mix_mute_button[m]);
			    gtk_table_attach_defaults(GTK_TABLE(mixer_table),mix_mute_button[m],m,m+1,3,4);	
		            gtk_signal_connect (GTK_OBJECT (mix_mute_button[m]), 							"toggled",
			                        GTK_SIGNAL_FUNC (mixer_mute_toggle_button_cb),
						mix_mute_elem[m]);
	                    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mix_mute_button[m]),( ! prev_boolean) );
	
			    no_of_playback++;
			}
			else {
			    if ( strstr(idname,"EQ") == 0 ) {
				if ( strstr(idname,"3D") > 0 ) {
				    m=no_of_ac97_3d;
				    mix_3d_elem[m]=elem;
			            frame= gtk_frame_new(idname);
			            gtk_widget_show(frame);
			            gtk_container_add(GTK_CONTAINER(ac97_3d_hbox), frame);

			            ac97_3d_adj[m]=gtk_adjustment_new(
				        (gdouble)(snd_ctl_elem_value_get_integer(val,0)*1.0), 	
					(gdouble)(adj_min), 
				        (gdouble)(adj_max+1.0), 
        	    			(gdouble)1.0,(gdouble)1.0,(gdouble)1.0);
			            ac97_3d_hscale[m]=gtk_hscale_new(GTK_ADJUSTMENT(ac97_3d_adj[m]));
			            gtk_widget_show(ac97_3d_hscale[m]);				

		                    gtk_scale_set_value_pos(GTK_SCALE(ac97_3d_hscale[m]),0);
			            gtk_scale_set_draw_value(GTK_SCALE(ac97_3d_hscale[m]),1);
				    gtk_scale_set_digits(GTK_SCALE(ac97_3d_hscale[m]),0);

        	                    gtk_widget_set_usize(ac97_3d_hscale[m], 52, 22);

			            gtk_container_add(GTK_CONTAINER(frame), ac97_3d_hscale[m]);

                                    gtk_signal_connect (GTK_OBJECT (ac97_3d_adj[m]), 
                                                        "value_changed",
                       			GTK_SIGNAL_FUNC (mixer_adj_value_changed_cb),
							mix_3d_elem[m]);

				    no_of_ac97_3d++;
				}
				else {
                                    if ( strstr(idname,"SPSA") > 0 ) {

// This control is used by ac97 2.2 ( No use for au88x0 )

			                frame= gtk_frame_new(idname);
			                gtk_widget_show(frame);
			                gtk_container_add(GTK_CONTAINER(ac97_3d_hbox), frame);

			                ac97_spsa_adj=gtk_adjustment_new(
				             (gdouble)(snd_ctl_elem_value_get_integer(val,0)*1.0), 
					     (gdouble)(adj_min), 
				             (gdouble)(adj_max+1.0), 
        	    			     (gdouble)1.0,(gdouble)1.0,(gdouble)1.0);
			                ac97_spsa_hscale=gtk_hscale_new(GTK_ADJUSTMENT(ac97_spsa_adj));
			                gtk_widget_show(ac97_spsa_hscale);
		                        gtk_scale_set_value_pos(GTK_SCALE(ac97_spsa_hscale),0);
			                gtk_scale_set_draw_value(GTK_SCALE(ac97_spsa_hscale),1);
				        gtk_scale_set_digits(GTK_SCALE(ac97_spsa_hscale),0);
        	                        gtk_widget_set_usize(ac97_spsa_hscale, 52, 22);
			                gtk_container_add(GTK_CONTAINER(frame), ac97_spsa_hscale);
 
				    }
				    else {
 			                g_print("%s not implemented (vc)\n",idname);
                                    };
				};
			    };
			};
			break;
		    case SND_CTL_ELEM_TYPE_ENUMERATED:
		        enum_value=snd_ctl_elem_value_get_integer(val,0);

			if ( strstr(idname,"XTALK") == 0 ) {
			    items = snd_ctl_elem_info_get_items(info);
			    frame= gtk_frame_new(idname);
			    gtk_widget_show(frame);
		  
			    if ( strstr(idname,"Capture Source") > 0 ) {
                                gtk_container_add(GTK_CONTAINER(capture_vbox), frame);			         
                                mix_capsrc_elem=elem;
			        no_of_capture_source=items;

				left_right_table=gtk_table_new(2,no_of_capture_source,FALSE);
			        gtk_widget_show(left_right_table);
				gtk_container_add (GTK_CONTAINER (frame), left_right_table);

				mix_rec_lock=gtk_toggle_button_new_with_label("Lock");
				gtk_widget_show(mix_rec_lock);
	                        gtk_table_attach_defaults(GTK_TABLE(left_right_table),mix_rec_lock,0,1,0,2);	
				for (i=0; i<no_of_capture_source; i++)
				    cap[i]=i;
				for (i=0; i<2; i++) {
					cap_left_right[i]=snd_ctl_elem_value_get_integer(val,i);
				};
                                for (i=0; i<2; i++) {
				    label=gtk_label_new(speaker_stereo_name[i]); 
				    gtk_widget_show(label);
		                    gtk_table_attach_defaults(GTK_TABLE(left_right_table),label,1,2,i,i+1);	
 			            for (item = 0; item < no_of_capture_source; item++) {
			                snd_ctl_elem_info_set_item(info, item);
			                if ((err = snd_hctl_elem_info(elem, info)) < 0) {
				            g_print("snd_hctl_elem_info error: %s\n", snd_strerror(err));
		  	                }
			                else {
                                           if ( item == 0 ) {
                                               mix_rec_radio[item][i] = gtk_radio_button_new_with_label (NULL, snd_ctl_elem_info_get_item_name(info));
				           }
                                           else {
                                               mix_rec_radio[item][i] = gtk_radio_button_new_with_label(gtk_radio_button_group (GTK_RADIO_BUTTON (mix_rec_radio[0][i])), snd_ctl_elem_info_get_item_name(info));
				           };
				           gtk_widget_show(mix_rec_radio[item][i]);
			                   gtk_table_attach_defaults(GTK_TABLE(left_right_table),mix_rec_radio[item][i],item+2,item+3,i,i+1);	

					   if ( i == 0 ) {
                                               gtk_signal_connect (GTK_OBJECT (mix_rec_radio[item][i]), 								
                                                        "toggled",
                       		GTK_SIGNAL_FUNC (mixer_rec_left_toggle_button_cb), &cap[item]);
					   }
					   else { 
                                               gtk_signal_connect (GTK_OBJECT (mix_rec_radio[item][i]), 								
                                                        "toggled",
                       		GTK_SIGNAL_FUNC (mixer_rec_right_toggle_button_cb), &cap[item]);
                                           };
				       };
				   };
			       };
	                       gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(mix_rec_lock),(cap_left_right[0] == cap_left_right[1]));

			       for (i=0; i<2; i++ ) {
		                   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(mix_rec_radio[cap_left_right[i]][i]),TRUE);
			       };

			   }
			   else {
                               gtk_container_add(GTK_CONTAINER(misc_hbox), frame);

		               hbox = gtk_hbox_new(FALSE, 0);
			       gtk_widget_show(hbox);
	                       gtk_container_add(GTK_CONTAINER(frame), hbox);

                               mix_enum_elem[no_of_enum_elem]=elem;
                               m=no_of_enum_radio;
			       if ( strstr(idname,"PCM Out & Mute") > 0 ) 
				   stac9708_pcmoutpath_idx=no_of_enum_elem;

			       for (item = 0; item < items; item++) {
			           snd_ctl_elem_info_set_item(info, item);
			           if ((err = snd_hctl_elem_info(elem, info)) < 0) {
				       g_print("snd_hctl_elem_info error: %s\n", snd_strerror(err));
			           }
			           else {
                                       mix_enum[no_of_enum_radio].elem_idx=no_of_enum_elem;
                                       mix_enum[no_of_enum_radio].idx=no_of_enum_radio;
                                       mix_enum[no_of_enum_radio].value=item;

                                       if ( item == 0 ) {
                                           mix_enum_radio[no_of_enum_radio] = gtk_radio_button_new_with_label (NULL, snd_ctl_elem_info_get_item_name(info));
				       }
                                       else {
                                           mix_enum_radio[no_of_enum_radio] = gtk_radio_button_new_with_label(gtk_radio_button_group (GTK_RADIO_BUTTON (mix_enum_radio[m])), snd_ctl_elem_info_get_item_name(info));
				       };
                                       gtk_box_pack_start(GTK_BOX(hbox), mix_enum_radio[no_of_enum_radio], TRUE, TRUE, 2);
				       gtk_widget_show(mix_enum_radio[no_of_enum_radio]);
                                       gtk_signal_connect (GTK_OBJECT (mix_enum_radio[no_of_enum_radio]),
                                                        "toggled",
                       			 GTK_SIGNAL_FUNC (mixer_enum_button_cb), &mix_enum[no_of_enum_radio].idx);
                                       no_of_enum_radio++;
		                   };
                                };
				gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(mix_enum_radio[m+enum_value]),TRUE);
				no_of_enum_elem++;
  			    };
                        };
			break;
		    case SND_CTL_ELEM_TYPE_BOOLEAN:
			current_boolean=snd_ctl_elem_value_get_boolean(val,0);
			if ((( strstr(idname,"Playback Switch") == 0 ) &&
           	            ( strstr(idname,"EQ") == 0 ) &&
                            ( strstr(idname,"Capture") == 0 )) ||
			    ( strstr(idname,"4-Speaker Stereo") > 0 )) {

			    if ( strstr(idname,"Playback Switch") > 0 )
			        idname[strlen(idname)-16]=0;
			    if ( strstr(idname,"Playback") > 0 )
				idname[strlen(idname)-9]=0;
			    if ( strstr(idname,"Sigmatel") > 0 ) {
				strcpy(newname,idname);
				strcpy(idname,&newname[9]);
			    };
			   
                            m=no_of_enable;
			    mix_enable_elem[m]=elem;
			    if  ( strstr(idname,"4-Speaker Stereo") > 0 ) {
			        stac9708_4speakerstereo_idx=m;
			    };

			    mix_enable_button[m]=gtk_toggle_button_new_with_label(idname);
			    gtk_widget_show(mix_enable_button[m]);
			    if ( strstr(idname,"3D") > 0 ) {
				ac97_3d_enable=m;
                                gtk_container_add(GTK_CONTAINER(ac97_3d_hbox), mix_enable_button[m]);
			    }
			    else {
                                gtk_container_add(GTK_CONTAINER(enable_hbox), mix_enable_button[m]);
                            };
                            gtk_signal_connect (GTK_OBJECT (mix_enable_button[m]),
                                                        "toggled",
                       		GTK_SIGNAL_FUNC (mixer_enable_button_cb), mix_enable_elem[m]);

			    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mix_enable_button[m]),
					current_boolean );
			    no_of_enable++;			    
			};
			prev_elem=elem;
			prev_boolean=current_boolean;
			prev_count=count;
			break;
		    case SND_CTL_ELEM_TYPE_IEC958:
			iec958_present=1;
			break;		    
		    default:
		        g_print("%s not implemented (type)\n",idname);
			break;
		    };
		};
	    };	
	};
	gtk_table_resize(GTK_TABLE(mixer_table),5,no_of_playback);
}

/******************************************************************************************/
static int read_le_long(FILE * file, long *ret)
{
	unsigned char buf[4];

	if (fread(buf, 1, 4, file) != 4)
		return 0;
	*ret = (buf[3] << 24) | (buf[2] << 16) | (buf[1] << 8) | buf[0];
	return TRUE;
}

static int read_le_short(FILE * file, short *ret)
{
	unsigned char buf[2];

	if (fread(buf, 1, 2, file) != 2)
		return 0;
	*ret = (buf[1] << 8) | buf[0];
	return TRUE;
}

/*******************************************************************************/

static int chunkid(char *id)
{
	char s[5];
	int i;
        memcpy(&s[0],id,4);
        s[4]=0;
	for (i = 0; i < sizeof(idlist)/sizeof(idlist[0]); i++) 
		if (strncmp(id, idlist[i].str, 4) == 0)
			return idlist[i].id;
	return UNKN_ID;
}

#define	printchunkid(s) g_print("%s %c%c%c%c\n",s,chunk.id[0],chunk.id[1],chunk.id[2],chunk.id[3])

void readstrandprint(SFChunk *chunk,FILE *fd)
{
	char buffer[256];
	fread(buffer, chunk->size,1, fd);
	buffer[chunk->size] = 0;
	g_print("%c%c%c%c : %s\n",chunk->id[0],chunk->id[1],chunk->id[2],chunk->id[3],buffer);
}

void read_dump_chunk(SFChunk *chunk, FILE *fd)
{
	unsigned char *data,*p;
	long k;
	int i,j;
	g_print("%c%c%c%c\n",chunk->id[0],chunk->id[1],chunk->id[2],chunk->id[3]);
	data=malloc(chunk->size);
	if ( data != NULL ) {
		fread(data,chunk->size,1,fd);
		p=data;
	        k=0;
	        while ( k < chunk->size ) {
		    i=16;
		    if ( (k+16) > chunk->size )
		        i=chunk->size-k;
		    g_print("%.8lx : ",k);
 	            for (j=0; j<i; j++) {
		        g_print("%.2x ",*p);
		        p++;
		        k++;
	            };
	            g_print("\n");
	        };
		free(data);
	};
}
			
static int parse_arl_info(int size, SFInfo *sf, FILE *fd)
{
	SFChunk chunk;
	char *buffer;
	int no_of_info;
	sf->infopos = ftell(fd);
	sf->infosize = size;

	no_of_info=0;
	while (TRUE) {
		READCHUNK(chunk, fd); size += 8;
		if (feof(fd))
			return -1;
		switch (chunkid(chunk.id)) {
		case IFIL_ID:
			READW(sf->version.major, fd);
			READW(sf->version.minor, fd);
			g_print("IFIL : %d.%d\n",sf->version.major,sf->version.minor);
			break;
		case INAM_ID:
			/* name of the font */
			sf->sf_name = (char*)malloc(chunk.size + 1);
			fread(sf->sf_name, chunk.size, 1, fd);
			sf->sf_name[chunk.size] = 0;
			g_print("INAM : %s\n",sf->sf_name);
			break;	
		case ICRD_ID:
		case IPRD_ID:
		case ICMT_ID:
		case ISFT_ID:
		case ISNG_ID:
		case IROM_ID:
		case IENG_ID:
		case ICOP_ID:
			buffer=malloc(chunk.size+1);
			fread(buffer,chunk.size,1,fd);
			buffer[chunk.size]=0;
			sf->info[no_of_info].desc=buffer;
			memcpy(&sf->info[no_of_info].id[0],&chunk.id[0],4);
			g_print("%c%c%c%c : %s\n",
				sf->info[no_of_info].id[0],
				sf->info[no_of_info].id[1],
				sf->info[no_of_info].id[2],
				sf->info[no_of_info].id[3],
				sf->info[no_of_info].desc);
			no_of_info++;
			break;
		case IVER_ID:
			FSKIP(chunk.size, fd);
			break;
		case LIST_ID:
			sf->infosize=size;
			return 0;
		default:
			printchunkid("info skip");
			read_dump_chunk(&chunk, fd);
			break;
		}
		size += chunk.size;
	};
	return 1;
}

static void load_preset(int size, SFInfo *sf, FILE *fd)
{
	int i;
	sf->npresets = size / 38;
	sf->preset = NEW(SFPreset, sf->npresets);
	g_print("no of presets=%d\n",sf->npresets);
	for (i = 0; i < sf->npresets; i++) {
		READSTR(sf->preset[i].name, fd);
		READW(sf->preset[i].preset, fd);
		READW(sf->preset[i].bank, fd);
		READW(sf->preset[i].bagNdx, fd);
		READDW(sf->preset[i].lib, fd); 
		READDW(sf->preset[i].genre, fd); 
		READDW(sf->preset[i].morphology, fd); 
	}
}

static void load_inst(int size, SFInfo *sf, FILE *fd)
{
	int i;
	sf->ninsts = size / 22;
	sf->inst = NEW(SFInst, sf->ninsts);
	g_print("no of insts=%d\n",sf->ninsts);
	for (i = 0; i < sf->ninsts; i++) {
		READSTR(sf->inst[i].name, fd);
		READW(sf->inst[i].bagNdx, fd);
	}
}

static void load_bag(int size, SFBagList *bagp, FILE *fd)
{
	int i;
	bagp->nbags = size / 4;
	bagp->bag = NEW(SFBag, bagp->nbags);
	g_print("no of bags = %d\n",bagp->nbags);
	for (i = 0; i < bagp->nbags; i++) {
		READW(bagp->bag[i].GenNdx, fd);
		READW(bagp->bag[i].ModNdx, fd);
	}
}

static void load_gen(int size, SFGenList *genp, FILE *fd)
{
	int i;
	genp->ngens = size / 4;
	genp->gen = NEW(SFGen, genp->ngens );
	g_print("no of gens = %d\n",genp->ngens);
	for (i = 0; i < genp->ngens; i++) {
		READW(genp->gen[i].oper, fd);
		READW(genp->gen[i].amount, fd);
	}
}

static void load_mod(int size ,SFModList *modp, FILE *fd)
{
	int i;
	modp->nmods = size / 12;
	modp->mod = NEW(SFMod, modp->nmods);
	g_print("no of mods = %d\n",modp->nmods);
	for (i=0 ; i< modp->nmods; i++) {
		READW(modp->mod[i].srcoper ,fd);
		READW(modp->mod[i].destoper ,fd);
		READW(modp->mod[i].amount ,fd);
		READW(modp->mod[i].amtsrcoper ,fd);
		READW(modp->mod[i].transoper ,fd);	
		READW(modp->mod[i].unknown, fd);
/*
		g_print("src=%d  dest=%d  amount=%d  amtsrc=%d trans=%d u=%.4x\n",
			modp->mod[i].srcoper,modp->mod[i].destoper,
			modp->mod[i].amount,modp->mod[i].amtsrcoper,
			modp->mod[i].transoper,modp->mod[i].unknown);
*/
	};	
}

static void load_sample_header(int size, SFInfo *sf, FILE *fd)
{
	int i;
        sf->nsamples = size / 48;
	sf->sample = NEW(SFSampleInfo, sf->nsamples);
        g_print("no of samples=%d\n",sf->nsamples);
	for (i = 0; i < sf->nsamples; i++) {
		READSTR(sf->sample[i].name,fd);
		READDW(sf->sample[i].startsample, fd);
		READDW(sf->sample[i].endsample, fd);
		READDW(sf->sample[i].startloop, fd);
		READDW(sf->sample[i].endloop, fd);

		READDW(sf->sample[i].samplerate, fd);
		READB(sf->sample[i].originalpitch, fd);
		READB(sf->sample[i].pitchcorrection, fd);
		READW(sf->sample[i].samplelink, fd);
		READW(sf->sample[i].sampletype, fd);
		SKIPW(fd);
/*
		g_print("Sample %d - start=%d end=%d sloop=%d eloop=%d rate=%dHz pitch=%.2x pc=%.2x link=%d type=%d  %s\n",
			i,sf->sample[i].startsample,sf->sample[i].endsample,
			sf->sample[i].startloop,sf->sample[i].endloop,
			sf->sample[i].samplerate,
			sf->sample[i].originalpitch,
			sf->sample[i].pitchcorrection,
			sf->sample[i].samplelink,
			sf->sample[i].sampletype,sf->sample[i].name);
                };
		if ( sf->sample[i].samplerate > 48000 )
		    sf->sample[i].samplerate=48000;
*/
	}
}

static int parse_arl_sdta(int size, SFInfo *sf, FILE *fd)
{
	SFChunk chunk;
	while (TRUE) {
		READCHUNK(chunk, fd); size += 8;
		if (feof(fd))
			return -1;
		switch (chunkid(chunk.id)) {
		case SMPL_ID:
			sf->samplepos = ftell(fd);
			sf->samplesize = chunk.size;
			g_print("Sample Pos %ld  size %d\n", sf->samplepos, sf->samplesize);
			sf->data=malloc(sf->samplesize);
			if ( sf->data != NULL ) 
				fread(sf->data,sf->samplesize,1,fd);
			break;
		case LIST_ID:
			return 0;
		default:
			printchunkid("sdta skip");
			read_dump_chunk(&chunk, fd);
			break;
		}
		size += chunk.size;
	};
	return 1;
}

static int parse_arl_pdta(int size, SFInfo *sf, FILE *fd)
{
	SFChunk chunk;
	while (TRUE) {
		READCHUNK(chunk, fd); size += 8;
		if (feof(fd))
			return -1;
		switch (chunkid(chunk.id)) {
		case PHDR_ID:
			load_preset(chunk.size, sf, fd);
			break;
		case PBAG_ID:
			load_bag(chunk.size, &sfinfo.prbags, fd);
			break;
		case PGEN_ID:
			load_gen(chunk.size, &sfinfo.prgens, fd);
			break;
		case INST_ID:
			load_inst(chunk.size, sf, fd);
			break;
		case IBAG_ID:
			load_bag(chunk.size, &sfinfo.inbags, fd);
			break;
		case IGEN_ID:
			load_gen(chunk.size, &sfinfo.ingens, fd);
			break;
		case SHDR_ID:
			load_sample_header(chunk.size, sf, fd);
			break;
		case LIST_ID:
//			g_print("pdta size=%d\n",size);
			return 0;
		case PMOD_ID: 
			load_mod(chunk.size, &sfinfo.prmods, fd);
			break;
		case IMOD_ID: 
			load_mod(chunk.size, &sfinfo.inmods, fd);
			break;
		default:
			printchunkid("pdta skip");
			read_dump_chunk(&chunk, fd);
			break;
		}
		size += chunk.size;
	};
	return 1;
}

static int parse_arl_list(int size, SFInfo *sf, FILE *fd)
{
	SFChunk chunk;
	READID(chunk.id, fd); 	size += 4;
	switch (chunkid(chunk.id)) {
	case INFO_ID:
		return parse_arl_info(size, sf, fd);
	case SDTA_ID:
		return parse_arl_sdta(size, sf, fd);
	case PDTA_ID:
		return parse_arl_pdta(size, sf, fd);
	default:
		printchunkid("*** illegal id in level 1 : ");
		FSKIP(size, fd);
		return 0;
	}
}

void generate_arl_midi(SFInfo *sf)
{
	SFPreset *preset,*next_preset;
	SFInst *inst;
	SFBag *bag;
	SFGen *gen,*inst_gen;
	genAmountType keyrange_genamount;
	int i,j,k,n,no_of_drum,drum_idx;
	int b0,b1,g0,g1,ig0,ig1;
        for (i=0; i<128; i++) {
	    sf->midi[i].preset=-1;
	    sf->midi[i].inst=-1;
	    for (j=0; j<128; j++) 
		sf->midi[i].note[j].sample=-1;
	};
	preset=&sf->preset[0];
	next_preset=&sf->preset[1];
	no_of_drum=0;
	for (i = 0; i < (sf->npresets-1); i++) {
	    if ( preset->bank == 0 ) 
		sf->midi[preset->preset].preset=i;
	    else {
		drum_idx=no_of_drum;
		sf->drum[drum_idx].preset=i;
		no_of_drum++;
	    };
	    b0=preset->bagNdx;
	    b1=next_preset->bagNdx;

            bag=&sfinfo.prbags.bag[b0];
	    g0=bag->GenNdx;
	
            bag=&sfinfo.prbags.bag[b1];
	    g1=bag->GenNdx;

	    gen=&sfinfo.prgens.gen[g0];
	    for (j=g0; j<g1; j++) {
	        if ( gen->oper == SF_instrument ) {
                    inst=&sfinfo.inst[gen->amount];

		    b0=inst->bagNdx;
 	            bag=&sfinfo.inbags.bag[b0];
		    ig0=bag->GenNdx;
		    inst++;
		    b1=inst->bagNdx;
	            bag=&sfinfo.inbags.bag[b1];
		    ig1=bag->GenNdx;

 	            inst_gen=&sfinfo.ingens.gen[ig0];
		    for (k=ig0; k<ig1; k++) {
	 	        if ( inst_gen->oper == SF_keyRange ) 
			    keyrange_genamount.shAmount=inst_gen->amount;
		        if ( inst_gen->oper == SF_sampleId ) {
                            if ( preset->bank == 0 ) {
			        for (n=keyrange_genamount.ranges.byLo; n<=keyrange_genamount.ranges.byHi; n++) 
			   	    sf->midi[preset->preset].note[n].sample=inst_gen->amount;
			    }
			    else {
			        for (n=keyrange_genamount.ranges.byLo; n<=keyrange_genamount.ranges.byHi; n++) 
				    sf->drum[drum_idx].note[n].sample=inst_gen->amount;
			    };				
	                };
	                inst_gen++;
		    };
	        };
		gen++;
	    };				
	    preset++;
	    next_preset++;
	};
}

/*
 *      read Aureal Sound Bank file (sfbk) for au88x0 
 *      similar to SF2 but different address scheme 
 *        - no chunk size for INFO,sdta,pdta chunks 
 *        - chunk size of LIST is zero for INFO list 
 *        - chunk size of IMOD and PMOD are multiples of 12 bytes
 *        - chunk size of shdr is multiples of 48 bytes
 *        - 8/16 bits samples
 *        - bad samples if startsample = endsample = startloop = endloop 
 */
int vortex_parse_arl(SFInfo *sf, FILE *fd)
{
	SFChunk chunk;
	int i;

	memset(sf,0,sizeof(SFInfo));

	sf->preset = NULL;
	sf->sample = NULL;
	sf->inst = NULL;
	sf->sf_name = NULL;
	sf->data =NULL;

	for (i=0; i<8; i++)
	    sf->info[i].desc=NULL;

	sf->prbags.bag = NULL;
	sf->inbags.bag = NULL;
	sf->prgens.gen = NULL;
	sf->ingens.gen = NULL;
	sf->prmods.mod = NULL;
	sf->inmods.mod = NULL;

	READCHUNK(chunk, fd);
	if (chunkid(chunk.id) != RIFF_ID) {
		g_print("*** not a RIFF file\n");
		return -1;
	}
	READID(chunk.id, fd);
	if (chunkid(chunk.id) != SFBK_ID) {
		g_print( "*** not a wavetable file\n");
		return -1;
	}
	READCHUNK(chunk, fd);
	for (;;) {
		if (feof(fd))
			break;
		else if (chunkid(chunk.id) == LIST_ID) {
			if (parse_arl_list(chunk.size, sf, fd))
				break;
		} else {
			printchunkid("*** Illegal id in level 0: ");
			FSKIP(chunk.size, fd);
		}
	}
	generate_arl_midi(sf);
	sf->loaded=TRUE;
	return 0;
}

void vortex_free_arl(SFInfo *sf)
{
	int i;
	if (sf->preset) 
	    free(sf->preset);
	if (sf->inst)   
	    free(sf->inst);

	if (sf->sample) 
	    free(sf->sample);
	if (sf->sf_name) 
	    free(sf->sf_name);

	for (i=0; i<8; i++) 
	    if ( sf->info[i].desc ) 
		free(sf->info[i].desc);

	if (sf->data) 
	    free(sf->data);
	if (sf->prbags.bag) 
	    free(sf->prbags.bag);
	if (sf->prgens.gen) 
	    free(sf->prgens.gen);
	if (sf->inbags.bag) 
	    free(sf->inbags.bag);
	if (sf->ingens.gen) 
	    free(sf->ingens.gen);
	if (sf->prmods.mod) 
	    free(sf->prmods.mod);
	if (sf->inmods.mod) 
	    free(sf->inmods.mod);
	sf->loaded=FALSE;
}

static int is_arl_file(char * filename)
{
	FILE *file;

	char magic[4];
	unsigned long len;
	char *ext;

    ext = strrchr(filename, '.');
    if (ext) {
	if (!strcasecmp(ext, ".arl")) {
	    if (( file = fopen(filename, "rb"))) {
		fread(magic, 1, 4, file);
		if (strncmp(magic, "RIFF", 4)==0) {
 		    read_le_long(file, &len);
		    fread(magic, 1, 4, file);
		    if (strncmp(magic, "sfbk", 4)==0) {
            	        fclose(file);
	                return TRUE;
                    };
                };
		fclose(file);
            };
	};
    };	
    return FALSE;
}

/************************************************************************************/

void vortex_free_dls(DLSInfo *dls)
{
	INSTInfo 	*inst;
	REGION 		*region;
	CUE    		*cue;
	int i,j;
	inst=&dls->inst_list[0];
	for (i=0; i<dls->header.cInstruments; i++) {
            if ( inst->region_list ) {
                region=&inst->region_list[0];
		for (j=0; j<inst->header.cRegions; j++) {
		    if ( region->art.connection_list ) 
			free(region->art.connection_list);
		    region++;
		};               
		free(inst->region_list);
	    };
	    if ( inst->art.connection_list )
		free(inst->art.connection_list);
	    if ( inst->name ) 
		free(inst->name);
	   
	    inst++;
	};
	if ( dls->inst_list ) free(dls->inst_list);

	if ( dls->cue_list) {
            cue=&dls->cue_list[0];
            for (i=0; i<dls->pooltable.cCues; i++) {
		if ( cue->data )
		    free(cue->data);
   	        if ( cue->name ) 
		    free(cue->name);
                cue++;
            };
	    free(dls->cue_list);
	};
	if ( dls->name ) free(dls->name);
	dls->loaded=FALSE;
}

static int dlschunkid(char *id)
{
	char s[5];
	int i;
        memcpy(&s[0],id,4);
        s[4]=0;
	for (i = 0; i < sizeof(dlsidlist)/sizeof(dlsidlist[0]); i++) {
		if (strncmp(id, dlsidlist[i].str, 4) == 0)
			return dlsidlist[i].id;
	}
	g_print("unknown id %s %.2x %.2x %.2x %.2x\n",s,s[0],s[1],s[2],s[3]);
	vortex_free_dls(&dlsinfo);
	exit(1);
	return FOURCC_UNKN;
}

void hexdump(unsigned char *data,long length)
{
	unsigned char *p;
	long k;	
	int i,j;
	p=data;
	k=0;
	while ( k < length ) {
	i=16;
	if ( (k+16) > length )
		i=length-k;
		g_print("%.8lx : ",k);
 	        for (j=0; j<i; j++) {
		        g_print("%.2x ",*p);
		        p++;
		        k++;
	        };
	        g_print("\n");
	};
}

void dump_dls_chunk(DLSChunk *chunk, FILE *fd,int flag)
{
	unsigned char *data;

	data=malloc(chunk->size);
	if ( data != NULL ) {
	    fread(data,chunk->size,1,fd);
	    if ( flag ) {
		hexdump(data,chunk->size);
	    };
	    free(data);
	};
}

void parse_lart_list(DLSInfo *dls,FILE *fd,int list_size,LART *art)
{
	DLSChunk chunk;
	CONNECTION *connection;
	int size;
	int k;
	size=list_size;
        READCHUNK(chunk,fd); size-=8;
        if  (dlschunkid(chunk.id) == FOURCC_ART1 ) {
	    fread(&art->connection_header,sizeof(CONNECTIONLIST),1,fd);
	    art->connection_list=NEW(CONNECTION,art->connection_header.cConnections);
	    if ( art->connection_list != NULL ) {
		memset(art->connection_list,0,sizeof(CONNECTION)*art->connection_header.cConnections);
		connection=&art->connection_list[0];
		for (k=0; k<art->connection_header.cConnections; k++) {
		    fread(connection,sizeof(CONNECTION),1,fd);
		    connection++;
	        };
	    };
	    size-=chunk.size;
	};
}

void parse_info_list(DLSInfo *dls,FILE *fd,int list_size,char **name)
{
	DLSChunk chunk;
	int size,str_size;
	char buffer [256];
	size=list_size;
        while ( size > 0 ) {
	    READCHUNK(chunk,fd);   size-=8;
	    str_size=chunk.size;
	    if ( chunk.size & 1 )
		str_size++;
            switch (dlschunkid(chunk.id)) {
	    case FOURCC_INAM:
		*(name)=malloc(str_size);
		fread(*(name),str_size,1,fd);
		size-=str_size;
		break;
	    case FOURCC_ICOP:
  	    case FOURCC_ICMT:
	    case FOURCC_IENG:
	    case FOURCC_ISBJ:
	    case FOURCC_ISFT:
	        fread(&buffer,str_size,1,fd);
		size-=str_size;
		break;
/*
	    case FOURCC_IARL:
	    case FOURCC_IART:
	    case FOURCC_ICMS:
	    case FOURCC_ICRD:
	    case FOURCC_IGNR:
	    case FOURCC_IKEY:
	    case FOURCC_IMED:
	    case FOURCC_IPRD:
	    case FOURCC_ISRC:
	    case FOURCC_ISRF:
	    case FOURCC_ITCH:
*/
	    default:
		exit(1);
	    };
	};
}

SAMPLE * parse_wsmp(FILE *fd)
{
	SAMPLE *sample;
	WLOOP  *wloop;
	int k;
    	sample=NEW(SAMPLE,1);	
        fread(&sample->wsmpl,sizeof(WSMPL),1,fd);
	if ( sample->wsmpl.cSampleLoops > 0 ) {
	    sample->wloop_list=NEW(WLOOP,sample->wsmpl.cSampleLoops);
            for (k=0; k<sample->wsmpl.cSampleLoops; k++) {
		wloop=&sample->wloop_list[k];
   	        fread(wloop,sizeof(WLOOP),1,fd);
            };
	};
	return(sample);
}

void parse_lrgn_list(DLSInfo *dls,FILE *fd,int list_size,INSTInfo *inst)
{
	DLSChunk chunk;
	REGION *region;
	int j,size,rgn_size;
	size=list_size;
    inst->region_list=NEW(REGION,inst->header.cRegions);
    if ( inst->region_list != NULL ) {
	memset(inst->region_list,0,sizeof(REGION)*inst->header.cRegions);
        for (j=0; j<inst->header.cRegions; j++) {
	    region=&inst->region_list[j];
            READCHUNK(chunk,fd);   size-=8;
	    if ( dlschunkid(chunk.id) == FOURCC_LIST ) {
	        READID(chunk.id,fd);     size-=4;
	        if ( dlschunkid(chunk.id) == FOURCC_RGN ) {
		    rgn_size=chunk.size-4;
	            READCHUNK(chunk,fd);   size-=8;    rgn_size-=8;
	            if ( dlschunkid(chunk.id) == FOURCC_RGNH ) {
		        fread(&region->range,sizeof(RGNHEADER),1,fd);
		        size-=chunk.size;   rgn_size-=chunk.size;
		        READCHUNK(chunk,fd);  size-=8;  rgn_size-=8;
		    }
		    else {
			g_print("*** Missing rngh chunk\n");
			return;
		    };
		    if ( dlschunkid(chunk.id) == FOURCC_WSMP ) {
			region->sample=parse_wsmp(fd);
			size-=chunk.size;   rgn_size-=chunk.size;
		        READCHUNK(chunk,fd);  size-=8;    rgn_size-=8;
		    };
		    if ( dlschunkid(chunk.id) == FOURCC_WLNK ) {
		        fread(&region->wavelink,sizeof(WAVELINK),1,fd);
		        size-=chunk.size;    rgn_size-=chunk.size;
                    }
		    else {
			g_print("*** Missing wlnk chunk\n");
			return;
		    };
		    if ( rgn_size > 0 ) {
		        READCHUNK(chunk,fd);  size-=8;    rgn_size-=8;		
    		        if ( dlschunkid(chunk.id) == FOURCC_LIST ) {
		            READID(chunk.id,fd);  size-=4; rgn_size-=4;
	                    if ( dlschunkid(chunk.id) == FOURCC_LART ) { 
                                parse_lart_list(dls,fd,chunk.size-4,&region->art);
				size-=(chunk.size-4); rgn_size-=(chunk.size-4);
	                    };
			};
		    };
                };
            };
        };
    };
}

void parse_lins(DLSInfo *dls,FILE *fd,int list_size)
{
	DLSChunk chunk;
	INSTInfo *inst;
	int size,inst_size;
	int i;
    	size=list_size;
    dls->inst_list=NEW(INSTInfo,dls->header.cInstruments);
    if ( dls->inst_list != NULL ) {
	memset(dls->inst_list,0,sizeof(INSTInfo)*dls->header.cInstruments);
    	for (i=0; i<dls->header.cInstruments; i++) {        
	    inst=&dls->inst_list[i];
    	    READCHUNK(chunk,fd);   size-=8;
    	    if ( dlschunkid(chunk.id) == FOURCC_LIST ) {
    	        READID(chunk.id,fd);  size-=4;
    	        if ( dlschunkid(chunk.id) == FOURCC_INS ) {
		    inst_size=chunk.size-4;
		    READCHUNK(chunk,fd);  size-=8; inst_size-=8;
    	            if ( dlschunkid(chunk.id) == FOURCC_INSH ) {
		        fread(&inst->header,sizeof(INSTHEADER),1,fd);
			size-=chunk.size;   inst_size-=chunk.size;
        	    	READCHUNK(chunk,fd);     size-=8;   inst_size-=8;
	            	if ( dlschunkid(chunk.id) == FOURCC_LIST ) {
            	            READID(chunk.id,fd);   size-=4;   inst_size-=4;
	                    if ( dlschunkid(chunk.id) == FOURCC_LRGN ) { 
                                inst_size-=(chunk.size-4);
                                parse_lrgn_list(dls,fd,chunk.size-4,inst);
			    };
		        };
		        while ( inst_size > 0 ) {
    		            READCHUNK(chunk,fd); size-=8; inst_size-=8;
    		            if ( dlschunkid(chunk.id) == FOURCC_LIST ) {
                                inst_size-=chunk.size;
		                READID(chunk.id,fd);  size-=4; 
	                        if ( dlschunkid(chunk.id) == FOURCC_LART ) { 
                                    parse_lart_list(dls,fd,chunk.size-4,&inst->art);
	    	                };
	                        if ( dlschunkid(chunk.id) == FOURCC_INFO ) {
                                    parse_info_list(dls,fd,chunk.size-4,&inst->name);
			        };
			    };
        	        };
		        inst++;
	            };   
                };
            };
        };
    };
}

void parse_ptbl(DLSInfo *dls,FILE *fd,int pool_size)
{
	CUE *cue;
	int size;
	int i;
	size=pool_size;
	fread(&dls->pooltable,sizeof(POOLTABLE),1,fd);
	size-=sizeof(POOLTABLE);

	dls->cue_list=NEW(CUE,dls->pooltable.cCues);
	if ( dls->cue_list != NULL ) {
	    for (i=0; i<dls->pooltable.cCues; i++) {
	        cue=&dls->cue_list[i];
                fread(&cue->poolcue,sizeof(POOLCUE),1,fd);
	        size-=sizeof(POOLCUE);
            };
	};
}

void parse_wvpl(DLSInfo *dls,FILE *fd,int list_size)
{
	CUE *cue;
	DLSChunk chunk;
	int size,wave_size;
	int i;
	size=list_size;
	for (i=0; i<dls->pooltable.cCues; i++) {
	    cue=&dls->cue_list[i];
	    READCHUNK(chunk,fd);    size-=8;
            if (dlschunkid(chunk.id) == FOURCC_LIST) {
	        READID(chunk.id, fd); 	size-=4;
	        if (dlschunkid(chunk.id) == FOURCC_WAVE ) {
		    wave_size=chunk.size-4;
		    READCHUNK(chunk,fd);    size-=8;  wave_size-=8;
		    if ( dlschunkid(chunk.id) == FOURCC_FMT ) {
			fread(&cue->fmt,chunk.size,1,fd);
		        size-=chunk.size;   wave_size-=chunk.size;
		    }
		    else {
			g_print("*** missing fmt chunk\n");
                    };
		    READCHUNK(chunk,fd);    size-=8;  wave_size-=8;
		    if ( dlschunkid(chunk.id) == FOURCC_WSMP ) {
			cue->sample=parse_wsmp(fd);
			size-=chunk.size;  wave_size-=chunk.size;
		        READCHUNK(chunk,fd);  size-=8;   wave_size-=8;
		    };
		    if ( dlschunkid(chunk.id) == FOURCC_DATA ) {
			cue->data=malloc(chunk.size);
			fread(cue->data,chunk.size,1,fd);
			cue->size=chunk.size;
//		        dump_dls_chunk(&chunk,fd,FALSE);
		        size-=chunk.size;  wave_size-=chunk.size;
                    };
		    if ( wave_size > 0 ) {
			READCHUNK(chunk,fd); size-=8;  wave_size-=8;
    		        if ( dlschunkid(chunk.id) == FOURCC_LIST ) {
		            READID(chunk.id,fd);  size-=4;  wave_size-=4;
	                    if ( dlschunkid(chunk.id) == FOURCC_INFO ) {
				size-=(chunk.size-4);  
				wave_size-=(chunk.size-4);
				parse_info_list(dls,fd,chunk.size-4,&cue->name);	
			    };
			};
		    };
		};
	    };
	};
}

void generate_dls_midi(DLSInfo *dls)
{
	INSTInfo *inst;
	REGION *region;
	int i,j,k,m,n;
	for (i=0; i<128; i++) {
            dls->midi[i].inst=-1;
	    for (j=0; j<128; j++) 
	        dls->midi[i].note[j].cue=-1;
        }
        m=0;
    	for (i=0; i<dls->header.cInstruments; i++) { 
	    inst=&dls->inst_list[i];   
	    if ( inst->header.Locale.ulBank == 0 ) {
		k=inst->header.Locale.ulInstrument;
		dls->midi[k].inst=i;
	       	for (j=0; j<inst->header.cRegions; j++) {
		    region=&inst->region_list[j];
		    for (n=region->range.RangeKey.usLow; n<=region->range.RangeKey.usHigh; n++) {
			dls->midi[k].note[n].cue=region->wavelink.ulTableIndex;
		    };           
		};    
	    };
	    if ( inst->header.Locale.ulInstrument == F_INSTRUMENT_DRUMS ) {
                dls->drum[m].inst=i;
	       	for (j=0; j<inst->header.cRegions; j++) {
		    region=&inst->region_list[j];
		    for (n=region->range.RangeKey.usLow; n<=region->range.RangeKey.usHigh; n++) {
			dls->drum[m].note[n].cue=region->wavelink.ulTableIndex;
		    };           
		};    
                m++;
	    };
    	};	
}

int vortex_parse_dls(DLSInfo *dls,FILE *fd)
{
	DLSChunk chunk;
	int size;

	memset(dls,0,sizeof(DLSInfo));

	READCHUNK(chunk, fd); size=chunk.size;
	if (dlschunkid(chunk.id) != FOURCC_RIFF) {
		g_print("*** not a RIFF file\n");
		return -1;
	}
	READID(chunk.id, fd);    size-=4;
	if (dlschunkid(chunk.id) != FOURCC_DLS) {
		g_print( "*** not a DLS file\n");
		return -1;
	}
	READCHUNK(chunk,fd);   size-=8;
	if ( dlschunkid(chunk.id) == FOURCC_COLH) {
		READDW(dls->header.cInstruments,fd);
		size-=chunk.size;	
	}
	else {
	    g_print("*** Missing colh chunk\n");
	    return -1;
	};
	READCHUNK(chunk,fd);    size-=8;
	if ( dlschunkid(chunk.id) == FOURCC_VERS) {
		READDW(dls->version.dwVersionMS,fd);
		READDW(dls->version.dwVersionLS,fd);
                size-=chunk.size;
	}
	else {
	    g_print("*** Missing vers chunk\n");
	    return -1;
	};
	READCHUNK(chunk,fd);  size-=8;
	if ( dlschunkid(chunk.id) == FOURCC_MSYN) {
		dump_dls_chunk(&chunk,fd,FALSE);
                size-=chunk.size;
	};
	READCHUNK(chunk,fd); size-=8;
        if (dlschunkid(chunk.id) == FOURCC_LIST) {
	    READID(chunk.id, fd); 	size-=4;
	    if  (dlschunkid(chunk.id) == FOURCC_LINS ) {
		parse_lins(dls,fd,chunk.size-4);   
	    }
	    else {
	        g_print("*** Missing lins list\n");
	        return -1;
	    };
	}
	else {
	    g_print("*** Missing lins list\n");
	    return -1;	
	};
	READCHUNK(chunk,fd);       size-=8;
	if  (dlschunkid(chunk.id) == FOURCC_PTBL ) {
	    parse_ptbl(dls,fd,chunk.size);
	}
	else {
	    g_print("*** Missing ptbl chunk\n");
	    return -1;
	};
	READCHUNK(chunk,fd);   size-=8;
        if (dlschunkid(chunk.id) == FOURCC_LIST) {
	    READID(chunk.id, fd); 	size-=4;
	    if ( dlschunkid(chunk.id) == FOURCC_WVPL ) {
	        parse_wvpl(dls,fd,chunk.size);
            }
	    else {
	        g_print("*** Missing wvpl list\n");
	        return -1;
	    };
	}
	else {
	    g_print("*** Missing wvpl list\n");
	    return -1;
	};
	READCHUNK(chunk,fd);
        if ( dlschunkid(chunk.id) == FOURCC_LIST ) {
	    READID(chunk.id,fd);  size-=4;  
	    if ( dlschunkid(chunk.id) == FOURCC_INFO ) {
		size-=(chunk.size-4); 
		parse_info_list(dls,fd,chunk.size-4,&dls->name);	
	    };
	};
	generate_dls_midi(dls);
	dls->loaded=TRUE;
	return 0;
}

static int is_dls_file(char * filename)
{
	FILE *file;

	char magic[4];
	unsigned long len;
	char *ext;

    ext = strrchr(filename, '.');
    if (ext) {
	if (!strcasecmp(ext, ".dls")) {
	    if (( file = fopen(filename, "rb"))) {
		fread(magic, 1, 4, file);
		if (strncmp(magic, "RIFF", 4)==0) {
 		    read_le_long(file, &len);
		    fread(magic, 1, 4, file);
		    if (strncmp(magic, "DLS ", 4)==0) {
            	        fclose(file);
	                return TRUE;
                    };
                };
		fclose(file);
            };
	};
    };	
    return FALSE;
}

void print_dls_art(LART *art)
{
	CONNECTION  *connection;
	int k;
    if ( art != NULL ) {
	g_print("Size=%d\n",art->connection_header.cbSize);
        g_print("Connections=%d\n",art->connection_header.cConnections);
	for (k=0; k<art->connection_header.cConnections; k++) {
	    connection=&art->connection_list[k];
	    g_print("Source =%d\n",connection->usSource);
	    g_print("Control =%d\n",connection->usControl);
	    g_print("Destination =%d\n",connection->usDestination);
	    g_print("Transform =%d\n",connection->usTransform);
	    g_print("Scale =%d\n",connection->lScale);
        };
    };
}

void print_dls_wsmp(SAMPLE *sample)
{
	WLOOP *wloop;
	int k;
    if ( sample != NULL ) {
	g_print("Size=%d\n",sample->wsmpl.cbSize);
	g_print("UnityNote=%d\n",sample->wsmpl.usUnityNote);
	g_print("FineTune=%d\n",sample->wsmpl.sFineTune);
	g_print("Attenuation=%d\n",sample->wsmpl.lAttenuation);
	g_print("fulOptions=%d\n",sample->wsmpl.fulOptions);
	g_print("SampleLoops=%d\n",sample->wsmpl.cSampleLoops);
	if ( sample->wsmpl.cSampleLoops > 0 ) {
		for (k=0; k<sample->wsmpl.cSampleLoops; k++) {
		        wloop=&sample->wloop_list[k];
	        	g_print("size=%d\n",wloop->cbSize);
	        	g_print("Type=%d\n",wloop->ulType);
	        	g_print("Start=%d\n",wloop->ulStart);
	        	g_print("Length=%d\n",wloop->ulLength);
		};
        };
    };
}


void print_dls_region(INSTInfo *inst,REGION *region)
{
	int idx;
    if (( inst != NULL ) && ( region != NULL )) {
	idx=region-&inst->region_list[0];
        g_print("Region %d\n",idx);
        g_print("RangeKey= %d - %d (%s -%s)\n",
		region->range.RangeKey.usLow,region->range.RangeKey.usHigh,
		midi_key[region->range.RangeKey.usLow],midi_key[region->range.RangeKey.usHigh]);
        g_print("RangeVelocity= %d - %d\n",
		region->range.RangeVelocity.usLow,region->range.RangeVelocity.usHigh);
        g_print("fusOption=%d\n",region->range.fusOptions);
	g_print("KeyGroup=%d\n",region->range.usKeyGroup);
	    
        if ( region->sample != NULL )
	        print_dls_wsmp(region->sample);
            
        g_print("fusOption=%d\n",region->wavelink.fusOptions);
	g_print("PhaseGroup=%d\n",region->wavelink.usPhaseGroup);
	g_print("channel=%d\n",region->wavelink.ulChannel);
	g_print("TableIndex=%d\n",region->wavelink.ulTableIndex);
     
        print_dls_art(&region->art);
	g_print("\n");
    };
}

void print_dls_inst(INSTInfo *inst)
{
	REGION	*region;
    	int idx;
    	int j;
    if ( inst != NULL ) {
	idx=inst-&dlsinfo.inst_list[0];
	g_print("Inst %d - %s cRegions : %d   ulBank : %d   ulInstrument : %d\n",
			idx,inst->name,
			inst->header.cRegions,
			inst->header.Locale.ulBank,
			inst->header.Locale.ulInstrument);
        for (j=0; j<inst->header.cRegions; j++) {
 	    region=&inst->region_list[j];
	    print_dls_region(inst,region);
            g_print("\n");
	};
        print_dls_art(&inst->art);
    };
}

void print_dls_cue(CUE *cue)
{
    	int idx;
    if ( cue != NULL ) {
    	idx=cue-&dlsinfo.cue_list[0];
	g_print("Cue %d %s - %d\n",idx,cue->name,cue->poolcue.ulOffset);
	
	g_print("format = %d\n",cue->fmt.formattag);
	g_print("channels = %d\n",cue->fmt.channels);
	g_print("samples per sec = %d\n",cue->fmt.samplespersec);
	g_print("avg bytes per sec = %d\n",cue->fmt.avgbytespersec);
	g_print("block align = %d\n",cue->fmt.blockalign);
	g_print("bits per sample = %d\n",cue->fmt.bitspersample);
        if ( cue->sample != NULL )
	    print_dls_wsmp(cue->sample);
    };
}

void print_dls(DLSInfo *dls)
{

    INSTInfo 	*inst;
    REGION 	*region;
    CUE    	*cue;    
    int i,j;

    g_print("name = %s\n",dls->name);
    g_print("cInstruments = %d\n",dls->header.cInstruments);	
    g_print("version = %.8x.%.8x\n",dls->version.dwVersionMS,dls->version.dwVersionLS);


    for (i=0; i<dls->header.cInstruments; i++) { 
	inst=&dls->inst_list[i];   
    	g_print("Inst %d - %s cRegions : %d   ulBank : %d   ulInstrument : %d\n",
			i,inst->name,
			inst->header.cRegions,
			inst->header.Locale.ulBank,
			inst->header.Locale.ulInstrument);
        for (j=0; j<inst->header.cRegions; j++) {
            region=&inst->region_list[j];
	    print_dls_region(inst,region);
            g_print("\n");
	};    
	print_dls_art(&inst->art);
	g_print("\n\n");
    };
    g_print("cbSize =%d  cCues=%d\n",dls->pooltable.cbSize,dls->pooltable.cCues);

    for (i=0; i<dls->pooltable.cCues; i++) {
	cue=&dls->cue_list[i];
	print_dls_cue(cue);
    };
}

/**********************************************************************************/

static void wt_scope_render_pcm (void) {
  	int i;
  	gint32 ry, ly, try = 0, tly = 0;
  	gint16 *pointer16;
	unsigned char  *pointer8;
	int    width,height,h2;
	int idx;
	WT_WAVE *wt;
	gint32    hh;


 	    GDK_THREADS_ENTER();

//            width=wt_scope_drawing->allocation.width;
            h2=wt_scope_drawing->allocation.height/2;
            height=h2+h2;


     	    for (idx=0; idx<16; idx++) {
		wt=&wt_wave[idx];
                if  ( wt->going ) {

		    if ( wt->bits_per_sample == 16 ) {
		        pointer16=wt->data16;
		        pointer16+=wt->current_sample;
		        hh=65536/height;
		        if ( hh * height < 65536 )
		            hh++;
	            }
	            else {
		        pointer8=wt->data8;
		        pointer8+=wt->current_sample;
		        hh=256/height;
		        if ( hh * height < 256 )
		            hh++;
                    };
		    width=MIN(wt_scope_drawing->allocation.width,(wt->end_sample - wt->current_sample));

  	            for ( i = 0; i < width; i++ ) {
  	                if ( wt->bits_per_sample == 16 ) {
		            ly = *pointer16; 
		            pointer16++;
   		            ly /= hh;
			    if ( wt->channels == 2 ) {
	      	                ry += *pointer16; 
		                pointer16++;
	                        ry /= hh;
	                    };
		        }
	                else {
		            ly = *pointer8;
			    pointer8++;		            
   		            ly /= hh;
			    if ( wt->channels == 2 ) {
	      	                ry = *pointer8; 
		                pointer8++;
	                        ry /= hh;
		            };		
	                };
	                if ( wt_scope_type == 3 ) {
		            ly /= 2;
		            gdk_draw_line(wt_scope_pixmap, penWT[idx], i, h2-1-ly, i, h2+ly);
		            if (wt->channels == 2) {
		                ry /= 2;
		                gdk_draw_line(wt_scope_pixmap, penRedLight, i, h2-1-ry, i, h2+ry);
		            };
			}
			else {
	  	            ly += h2;
    		            ry += h2;
                            switch (wt_scope_type) {
	                    case 0:			// dot
	                        gdk_draw_point(wt_scope_pixmap, penWT[idx], i, ly); 
		                if ( wt->channels == 2 ) 
		                    gdk_draw_point(wt_scope_pixmap, penRedLight, i, ry);
		                break;
	                    case 1:			// line
		                if (i == 0) { 
		                    gdk_draw_point(wt_scope_pixmap, penWT[idx], 0, ly);
		                    tly = ly;
		                    if  (wt->channels == 2) {
			                gdk_draw_point(wt_scope_pixmap, penRedLight, 0, ry);
			                try = ry;		
		                    };
		                }
		                else {
	                            gdk_draw_line(wt_scope_pixmap, penWT[idx], i - 1, tly, i,    ly); 
	  	                    tly = ly;
		                    if  (wt->channels == 2) {
		                        gdk_draw_line(wt_scope_pixmap, penRedLight, i - 1, try, i, ry);  
		                        try = ry;
                                    };
      		                };
		                break;
	                    case 2:			// filled
		                if (ly != h2) 
		                    gdk_draw_line(wt_scope_pixmap, penWT[idx], i, ((ly > 0) ? h2 : h2-1), i, ly);		  
		                if (wt->channels == 2 ) {
		                    if (ry != h2) 
		                        gdk_draw_line(wt_scope_pixmap, penRedLight, i, ((ry > 0) ? h2 : h2-1), i, ry);  
		                };
		                break;
                            };
	                };
	            };
		};
	    };
	    gdk_window_clear(wt_scope_drawing->window);

  	    GDK_THREADS_LEAVE();
}

void redraw_wt_scope(void)
{
       if ( GTK_WIDGET_VISIBLE(wt_scope_drawing) ) {
		    wt_scope_render_pcm();
		    gdk_draw_rectangle(wt_scope_pixmap,
			   wt_scope_drawing->style->black_gc,
			   TRUE,
			   0, 0,
			   wt_scope_drawing->allocation.width,
			   wt_scope_drawing->allocation.height);
	};	
}

static GdkGC *get_wt_pen(int nRed, int nGreen, int nBlue)
{
	GdkColor *c;
	GdkGC *gc;
	c = (GdkColor *)g_malloc(sizeof(GdkColor));
	c->red = nRed;
	c->green = nGreen;
	c->blue = nBlue;
	gdk_color_alloc(gdk_colormap_get_system(), c);
	gc = gdk_gc_new(wt_scope_pixmap);
	gdk_gc_set_foreground(gc, c);
	return gc;
}

gint wt_scope_drawing_configure_event(GtkWidget *widget, GdkEventConfigure *event)
{
	if (wt_scope_pixmap != NULL)
		gdk_pixmap_unref(wt_scope_pixmap);
 	wt_scope_pixmap = gdk_pixmap_new(window->window, 
                          widget->allocation.width, widget->allocation.height,
			  gdk_visual_get_best_depth());

	penWT[0] = get_wt_pen(0xffff, 0xffff, 0xffff);
	penWT[1] = get_wt_pen(0xc000, 0xc000, 0xffff);
	penWT[2] = get_wt_pen(0xffff, 0xc000, 0xc000);
	penWT[3] = get_wt_pen(0xc000, 0xffff, 0xc000);
	     
	penWT[5] = get_wt_pen(0xffff, 0xffff, 0xc000);
	penWT[4] = get_wt_pen(0xc000, 0xffff, 0xffff);
	penWT[6] = get_wt_pen(0xffff, 0xc000, 0xffff);

	penWT[7] = get_wt_pen(0, 0, 0xC000);
	penWT[8] = get_wt_pen(0xC000, 0, 0);
	penWT[9] = get_wt_pen(0, 0xC000, 0);

	penWT[10] = get_wt_pen(0, 0xC000, 0xffff);
	penWT[11] = get_wt_pen(0xffff, 0, 0xC000);
	penWT[12] = get_wt_pen(0xC000, 0xffff, 0);

	penWT[13] = get_wt_pen(0xC000, 0, 0xffff);
	penWT[14] = get_wt_pen(0xffff, 0xC000, 0);
	penWT[15] = get_wt_pen(0, 0xffff, 0xC000);

  	gdk_draw_rectangle(wt_scope_pixmap, penBlack, TRUE, 0, 0, 
			  widget->allocation.width, widget->allocation.height);
  	gdk_window_set_back_pixmap(wt_scope_drawing->window, wt_scope_pixmap, 0);
	redraw_wt_scope(); 
	return TRUE;
}

gint wt_scope_drawing_expose_event(GtkWidget *widget, GdkEventExpose *event)
{
	gdk_window_clear_area (widget->window,
                         event->area.x, event->area.y,
                         event->area.width, event->area.height);
  	gdk_gc_set_clip_rectangle (widget->style->fg_gc[widget->state],&event->area);

	redraw_wt_scope();
  	gdk_gc_set_clip_rectangle (widget->style->fg_gc[widget->state],NULL);
	return TRUE;
}

void wt_scope_type_changed_cb (GtkWidget *widget, gpointer data)
{
	char select_item[32];
	int idx;
	strcpy(select_item,gtk_entry_get_text (GTK_ENTRY(widget)));
	for (idx=0; idx<4; idx++) {
            if ( strcmp(scope_type_name[idx],select_item) == 0 ) {
		wt_scope_type=idx;
	        redraw_wt_scope();
		return;
            };
	};
}

void hwmix_usleep(gint usec)
{
	struct timeval tv;
	
	tv.tv_sec = usec / 1000000;
	usec -= tv.tv_sec * 1000000;
	tv.tv_usec = usec;
	select(0, NULL, NULL, NULL, &tv);
}

static void wt_xrun_recover(WT_WAVE *wt)
{
	int err;
	if (snd_pcm_state(wt->pcm) == SND_PCM_STATE_XRUN) {
		if ((err = snd_pcm_prepare(wt->pcm)) < 0)
			g_print("snd_pcm_prepare() failed.\n");
	}
}

void wt_alsa_write(WT_WAVE *wt,gpointer data, int length)
{
	snd_pcm_sframes_t written_frames;
	while (length > 0) {
		int frames = snd_pcm_bytes_to_frames(wt->pcm, length);
		written_frames = snd_pcm_writei(wt->pcm, data, frames);		
		if (written_frames > 0) {
			int written = snd_pcm_frames_to_bytes(wt->pcm,written_frames);
			wt->total_written += written;
			length -= written;
			data = (char*) data + written;
		}
		else {
		    if (written_frames == -EPIPE) {
			wt_xrun_recover(wt);
		    }
		    else {
			g_print("snd_pcm_writei() error : %s\n",snd_strerror(-written_frames));
			break;
		    };
	        };
	}
}

static snd_pcm_sframes_t wt_alsa_get_avail(WT_WAVE *wt)
{
	snd_pcm_sframes_t ret;
	if ((ret = snd_pcm_avail_update(wt->pcm)) == -EPIPE)
		wt_xrun_recover(wt);
	else {
	    if (ret < 0) {
		g_print("wt alsa get avail - snd_pcm_avail_update() failed: %s\n",snd_strerror(-ret));
		return 0;
	    }
	    else
		return ret;
        };
	if ((ret = snd_pcm_avail_update(wt->pcm)) < 0) {
		g_warning("wt alsa get avail - snd_pcm_avail_update() failed: %s\n",snd_strerror(-ret));
		return 0;
	}
	return ret;
}

int wt_alsa_free(WT_WAVE *wt)
{
        int err;
	if (wt->paused)
		return 0;
	else {
		if (wt->force_start &&
		    snd_pcm_state(wt->pcm) == SND_PCM_STATE_PREPARED) {
			if ((err = snd_pcm_start(wt->pcm)) < 0)
				g_print("wt alsa free - snd_pcm_start() failed: %s\n", snd_strerror(err));
		}
		wt->force_start = TRUE;
		return snd_pcm_frames_to_bytes(wt->pcm, wt_alsa_get_avail(wt));
	}
}

static void *play_sample_loop(void *arg)
{
	WT_WAVE *wt;
	int16 *ptr16;
	unsigned char *current_ptr;
	long end_sample;
	int bytes_per_sample;
	int bytes, blk_size;
        int err;
	wt=arg;

	bytes_per_sample=1;
	blk_size = 512 * wt->channels;
	if ( wt->bits_per_sample == 16 ) {
	    blk_size <<= 1;
	    bytes_per_sample=2;
	};

	wt->current_sample=wt->start_sample;
	if ( wt->start_loop == wt->end_loop ) {
	    end_sample=wt->end_sample;
	    wt->loop=FALSE;
	}
	else
	    end_sample=wt->end_loop;

	while (wt->going) {
	    if ( wt->current_sample < end_sample ) {
		bytes = blk_size;
		if ( ( end_sample - wt->current_sample )*bytes_per_sample < bytes ) 
		    bytes = ( end_sample - wt->current_sample ) * bytes_per_sample;
		while (( wt_alsa_free(wt) < bytes ) && wt->going )
		    hwmix_usleep(10000);
	        
                if ( wt->bits_per_sample == 16 ) {
		    ptr16=wt->data16;
		    ptr16+=wt->current_sample;
		    current_ptr=(unsigned char *)(ptr16);
		}
		else {
		    current_ptr=wt->data8 + wt->current_sample;	
		};
		redraw_wt_scope();
		wt_alsa_write(wt, current_ptr, bytes);
		wt->current_sample += ( bytes/bytes_per_sample);
	    }
	    else {
		wt_alsa_free(wt);
                if ( wt->loop ) {
		    if ( wt->start_loop == wt->end_loop ) {
			wt->current_sample=wt->start_sample;
			end_sample=wt->end_sample;
		    }
		    else {
                        wt->current_sample=wt->start_loop;
			end_sample=wt->end_sample;
		    };
                }
                else {
 	            hwmix_usleep(10000);
		    switch (wt->button_type) {
		    case 1:
			gtk_widget_set_sensitive(wt_preset_sample_button[wt->idx],TRUE);
			break;
		    case 2:
			gtk_widget_set_sensitive(wt_inst_sample_button[wt->idx],TRUE);
			break;
		    case 3:
			gtk_widget_set_sensitive(wt_sample_button[wt->idx],TRUE);
			break;
		    case 4:
			gtk_widget_set_sensitive(wt_dls_inst_sample_button[wt->idx],TRUE);
			break;
		    case 5:
			gtk_widget_set_sensitive(wt_dls_inst_sample_button[wt->idx],TRUE);
			break;
		    case 6:
			gtk_widget_set_sensitive(wt_dls_sample_button[wt->idx],TRUE);
			break;

		    };               	
		    if ( wt->pcm != NULL ) {
		        if ((err = snd_pcm_close(wt->pcm)) < 0)
			    g_warning("snd_pcm_close() failed: %s",snd_strerror(err));
		        wt->pcm = NULL;
			wt->going=FALSE;
                   };
                    pthread_detach(wt->memory_play_thread);
		};
	    };
	};
	pthread_exit(NULL);
}

static int wt_alsa_setup(WT_WAVE *wt)
{
	int err;
	snd_pcm_hw_params_t *hwparams;
	snd_pcm_sw_params_t *swparams;
        unsigned int rrate;
        int dir;
	char device[32];

        strcpy(device,card);

	err = snd_pcm_open(&wt->pcm, device,SND_PCM_STREAM_PLAYBACK,SND_PCM_NONBLOCK);
	
	if ( err < 0 ) {
		g_warning("Failed to open pcm device (%s): %s",device, snd_strerror(err));
		wt->pcm = NULL;
		return -1;
	};

	snd_pcm_hw_params_alloca(&hwparams);
 	err = snd_pcm_hw_params_any(wt->pcm, hwparams);
	if ( err < 0 ) {
		g_warning("No configuration available for playback: %s", snd_strerror(err));
		return -1;
	};
	err = snd_pcm_hw_params_set_access(wt->pcm, hwparams,SND_PCM_ACCESS_RW_INTERLEAVED);
	if ( err < 0 ) {
		g_warning("Cannot set direct write mode: %s",snd_strerror(err));
		return -1;
        };
 	err = snd_pcm_hw_params_set_format(wt->pcm, hwparams, wt->format);
        if ( err < 0 ) {
		g_warning("Sample format not available for playback: %s",snd_strerror(err));
		return -1;
	}

  	err=snd_pcm_hw_params_set_channels(wt->pcm, hwparams, wt->channels);
        if (err < 0) {
                g_print("Channels count (%i) not available for playbacks: %s\n", wt->channels, snd_strerror(err));
                return -1;
        }
 	dir=0;
	err=snd_pcm_hw_params_get_period_size_max(hwparams,&wt->period_size,&dir);
	if ( err < 0 ) {
		g_print("snd_pcm_hw_params_get_period_size_max failed : %s\n",snd_strerror(err));
		return -1;
	};

	wt->period_size=wt->period_size/4;
	if ( wt->start_loop > wt->end_loop ) {
	    if ( 2*(wt->end_loop-wt->start_loop ) < wt->period_size )
		wt->period_size=wt->period_size/2;
	}
	else {
	    if ( 2*(wt->end_sample-wt->start_sample ) < wt->period_size )
		wt->period_size=wt->period_size/2;
	};

        err=snd_pcm_hw_params_set_period_size(wt->pcm,hwparams,wt->period_size,dir);
	if ( err < 0 ) {
		g_print("snd_pcm_hw_params_set_period_size failed : %s\n",snd_strerror(err));
		return -1;
	};

        wt->buffer_size=wt->period_size*4;
        err=snd_pcm_hw_params_set_buffer_size(wt->pcm,hwparams,wt->buffer_size);
	if ( err < 0 ) {
		g_print("snd_pcm_hw_params_set_buffer_size failed : %s\n",snd_strerror(err));
		return -1;
	};

        rrate=wt->samples_per_sec;
	err=snd_pcm_hw_params_set_rate_near(wt->pcm, hwparams, &rrate, 0);
	if ( err < 0 ) {
		g_warning("No usable samplerate available.");
		return -1;
	}
        if ( rrate != wt->samples_per_sec ) {
		g_warning("Rate doesn't match (requested %iHz, get %iHz)\n", wt->samples_per_sec, rrate);
                return -1;
        };

	err=snd_pcm_hw_params(wt->pcm, hwparams);
	if ( err < 0 ) {
		g_warning("snd_pcm_hw_params() failed: %s",snd_strerror(err));
		return -1;
	}

 	snd_pcm_sw_params_alloca(&swparams);

	snd_pcm_sw_params_current(wt->pcm, swparams);
	err = snd_pcm_sw_params_set_start_threshold(wt->pcm, swparams, wt->period_size);
        if ( err < 0 ) {
		g_warning("snd_pcm_sw_params_set_start_threshold() failed: %s", snd_strerror(err));
        };
	err=snd_pcm_sw_params(wt->pcm, swparams);
	if ( err < 0 ) {
		g_warning("snd_pcm_sw_params() failed: %s", snd_strerror(err));
		return -1;
	}
	return 0;
}

void wt_alsa_close(WT_WAVE *wt)
{
	int err, started;

	started = wt->going;
	wt->going = FALSE;

	if (wt->pcm != NULL) {
		if (started) {
			if ((err = snd_pcm_drop(wt->pcm)) < 0)
				g_warning("snd_pcm_drop() failed: %s",snd_strerror(err));
                };
		if ((err = snd_pcm_close(wt->pcm)) < 0)
			g_warning("snd_pcm_close() failed: %s",snd_strerror(err));
		wt->pcm = NULL;
	}
}

void play_wave_sample(WT_WAVE *wt)
{	
		if (wt_alsa_setup(wt) < 0) {
			wt_alsa_close(wt);
			return;
		}
	
		wt->total_written = 0;
		wt->going = TRUE;
		wt->paused=FALSE;
		wt->force_start=FALSE;
	
		snd_pcm_prepare(wt->pcm);
	
	        pthread_create(&wt->memory_play_thread, NULL, play_sample_loop, wt);
}

static void stop_play_sample(WT_WAVE *wt)
{
	if ( wt->going) {
		wt->going = FALSE;
		pthread_join(wt->memory_play_thread, NULL);
		wt_alsa_close(wt);
	}
}

void print_arl_sample(int idx)
{
	SFSampleInfo *sample;
/*
	SFPreset *preset,*next_preset;
	SFInst *inst,*next_inst;
	SFBag *bag;
	SFGen *gen,*pr_gen;
	int i,j,k,m,g0,g1;
*/
	sample=&sfinfo.sample[idx];
	g_print("Sample %d  %s  %d   %d   %d   %d   %dHz  %.2x (%s) %.2x %d %d\n",
			idx,
			sample->name,
			sample->startsample,
			sample->endsample,
			sample->startloop,
			sample->endloop,
			sample->samplerate,
                        sample->originalpitch,midi_key[sample->originalpitch],
                        sample->pitchcorrection,
			sample->samplelink,
			sample->sampletype);
/*
	next_inst=&sfinfo.inst[0];
        for (i=0; i< (sfinfo.ninsts-1) ; i++) {
	    inst=next_inst;
	    next_inst++;           

            bag=&sfinfo.inbags.bag[inst->bagNdx];
  	    g0=bag->GenNdx;


            bag=&sfinfo.inbags.bag[next_inst->bagNdx];
	    g1=bag->GenNdx;
   
            gen=&sfinfo.ingens.gen[g0];
            for (j=g0; j<g1; j++) {    	   
	        if ( gen->oper == SF_sampleId ) {
		    if ( gen->amount == idx ) {
		  	g_print("Inst %s\n",inst->name);


			next_preset=&sfinfo.preset[0];
			for (k=0; k< (sfinfo.npresets-1); k++) {
	    		    preset=next_preset;
	    		    next_preset++;

            		    bag=&sfinfo.prbags.bag[preset->bagNdx];
	    		    g0=bag->GenNdx;
	
            		    bag=&sfinfo.prbags.bag[next_preset->bagNdx];
	    		    g1=bag->GenNdx;

            		    pr_gen=&sfinfo.prgens.gen[g0];
	    		    for (m=g0; m<g1; m++) {
	                        if ( pr_gen->oper == SF_instrument ) {
		                    if ( pr_gen->amount == i ) {
		  	                 g_print("Preset %s\n",preset->name);
		                    };
		                };
		                pr_gen++;
	                    };
	                };			
		    };
		};
		gen++;
	    };
	};
*/
}

int timecent_to_msec(int timecent)
{
	return (int)(1000 * pow(2.0, (double)timecent / 1200.0));
}

int msec_to_timecent(int msec)
{
	if (msec <= 0) msec = 1;
	return (int)(log((double)msec / 1000.0) / log(2.0) * 1200.0);
}

int abscent_to_mHz(int abscents)
{
	return (int)(8176.0 * pow(2.0, (double)abscents / 1200.0));
}

int mHz_to_abscent(int mHz)
{
	return (int)(log((double)mHz / 8176.0) / log(2.0) * 1200.0);
}

float abscent_to_Hz(int abscents)
{
	return (8.176 * pow(2.0, (double)abscents / 1200.0));
}

int Hz_to_abscent(int Hz)
{
	return (int)(log((double)Hz / 8.176) / log(2.0) * 1200.0);
}

int play_arl_sample(int sample_idx,int type,int idx,GtkWidget *widget)
{
	SFSampleInfo *sample;
	WT_WAVE *wt;
	int i;
	if ( sample_idx < sfinfo.nsamples ) {
		sample=&sfinfo.sample[sample_idx];

		if ( sample->endsample > sample->startsample ) {
		    i=0;
		    while (( i<16 ) && (wt_wave[i].going))
			i++;
		    if ( i < 16 ) {
			wt=&wt_wave[i];
			wt->button_type=type;
			wt->idx=idx;

			wt->channels=1;
			wt->samples_per_sec=sample->samplerate;
			if (( sample->sampletype & 0x10 ) == 0x10 ) {
		    	    wt->format=SND_PCM_FORMAT_U8;
		    	    wt->bits_per_sample=8;
			    wt->data8=sfinfo.data;
			}
			else {
		    	    wt->format=SND_PCM_FORMAT_S16_LE;
		    	    wt->bits_per_sample=16;
			    wt->data16=(int16 *)(sfinfo.data);
		        };
			
			wt->start_sample=sample->startsample;
			wt->end_sample=sample->endsample;
			wt->start_loop=sample->startloop;
			wt->end_loop=sample->endloop;

			wt->loop = ( wt->end_sample > wt->start_sample );
/*
			if ( wt->end_sample > wt->start_sample ) {
		    	    if ( wt->end_loop == wt->start_loop ) 
		        	wt->loop=2*wt->samples_per_sec/(wt->end_sample - wt->start_sample);
		    	    else 
				wt->loop=2*wt->samples_per_sec/(wt->end_loop - wt->start_loop);
			};
*/
		        play_wave_sample(wt);
			return i;
        	    };		
		}
		else {
			g_print("sample %s is NULL\n",sample->name);
		};
	}
	return -1;
}

void print_gen(SFGen *gen)
{
	genAmountType gen_amount;
        gen_amount.shAmount=gen->amount;
	printf(" %.5d - %5d ",gen->oper,gen->amount);
	switch (gen->oper) {
	case SF_startAddrs:         
        case SF_endAddrs:
        case SF_startloopAddrs:     
        case SF_endloopAddrs:      
	    g_print(" %s %d\n",sf_gen_text[gen->oper],gen->amount);
	    break; 
        case SF_startAddrsHi:       
        case SF_endAddrsHi:        
        case SF_startloopAddrsHi:     
        case SF_endloopAddrsHi:       
	    g_print(" %s %d * 32K\n",sf_gen_text[gen->oper],gen->amount);
	    break;
        case SF_lfo1ToPitch:        
        case SF_lfo2ToPitch:        
        case SF_env1ToPitch:        
	    g_print(" %s %d cents\n",sf_gen_text[gen->oper],gen->amount);
	    break; 
        case SF_initialFilterFc:    
	    g_print(" %s %d Hz\n",sf_gen_text[gen->oper],gen->amount);
	    break; 
        case SF_initialFilterQ:     
	    g_print(" %s %d\n",sf_gen_text[gen->oper],gen->amount);
	    break; 
        case SF_env1ToFilterFc:     
	    g_print(" %s %d\n",sf_gen_text[gen->oper],gen->amount);
	    break;
        case SF_chorusEffectsSend:  
        case SF_reverbEffectsSend:  
        case SF_panEffectsSend:    
	    g_print(" %s %.1f %%\n",sf_gen_text[gen->oper],0.1*gen->amount);
	    break; 
        case SF_delayLfo1:     
        case SF_delayLfo2:         
        case SF_delayEnv1:          
        case SF_attackEnv1:         
        case SF_holdEnv1:             
        case SF_decayEnv1:      
        case SF_releaseEnv1:    
        case SF_delayEnv2:                       
        case SF_attackEnv2:           
        case SF_holdEnv2:            
        case SF_decayEnv2:           
        case SF_releaseEnv2:          
	    g_print(" %s %d msec\n",sf_gen_text[gen->oper],timecent_to_msec(gen->amount));
	    break;
        case SF_freqLfo1:          
        case SF_freqLfo2:          
	    g_print(" %s %.3fHz\n",sf_gen_text[gen->oper],abscent_to_Hz(gen->amount)); 
	    break;
        case SF_sustainEnv1:   
	    g_print(" %s %.1f\n",sf_gen_text[gen->oper],0.1*gen->amount);
	    break;            
        case SF_sustainEnv2:    
        case SF_initAtten:              
	    g_print(" %s %.1f dB\n",sf_gen_text[gen->oper],0.1*gen->amount);
	    break;            
        case SF_instrument:          
	    g_print(" %s %d %s\n",sf_gen_text[gen->oper],gen->amount,sfinfo.inst[gen->amount].name);
	    break;
        case SF_keyRange:             
	    g_print(" %s %.2x - %.2x  %s - %s\n",sf_gen_text[gen->oper],gen_amount.ranges.byLo,gen_amount.ranges.byHi,midi_key[gen_amount.ranges.byLo],midi_key[gen_amount.ranges.byHi]);
	    break;       
        case SF_velRange:      
	    g_print(" %s %.2x - %.2x\n",sf_gen_text[gen->oper],gen_amount.ranges.byLo,gen_amount.ranges.byHi);
	    break;       
        case SF_keynum:             
        case SF_velocity:             
	    g_print(" %s %d\n",sf_gen_text[gen->oper],gen->amount);
	    break; 
        case SF_sampleId:
	    print_arl_sample(gen->amount);
	    break;
        case SF_sampleFlags:
	    g_print(" %s %d\n",sf_gen_text[gen->oper],gen->amount);
	    break; 
        case SF_fineTune:
	    g_print(" %s %d%%\n",sf_gen_text[gen->oper],gen->amount);
	    break;
        case SF_rootKey:
	    g_print("        %s %d %s\n",sf_gen_text[gen->oper],gen->amount,midi_key[gen->amount]);
	     break;
        case SF_lfo1ToFilterFc:     
        case SF_lfo1ToVolume:     
        case SF_autoHoldEnv1:
        case SF_autoDecayEnv1:
        case SF_autoHoldEnv2:
        case SF_autoDecayEnv2:
        case SF_nop:
        case SF_coarseTune:
        case SF_samplePitch:         
        case SF_scaleTuning:
        case SF_keyExclusiveClass:
 	    g_print("        oper %d amount %d\n",gen->oper,gen->amount);
	     break;
//      case SF_env2ToVolume:       
//      case SF_auxEffectsSend:    
//      case SF_sampleVolume:       
//      case SF_unused3:
//      case SF_keyTuning:
//	case SF_EOF:
	    break;
	default:
	    g_print("        oper %d amount %d\n",gen->oper,gen->amount);
	    break;
	};
}

void print_bags(int idx,int b0,int b1,int g0,int g1,int m0,int m1)
{
	SFBag *bag;
	SFGen *gen;
	SFMod *mod;
	int j;

	if ( idx == 0 )
	    bag=&sfinfo.prbags.bag[b0];
	else
	    bag=&sfinfo.inbags.bag[b0];

	for (j=b0; j<b1 ; j++) {
	    g_print("bag %.3d  GenNdx %.5d    ModNdx %.3d\n",j,bag->GenNdx,bag->ModNdx);
 	    bag++;
	};		

	if ( idx == 0 )
	    gen=&sfinfo.prgens.gen[g0];
	else
	    gen=&sfinfo.ingens.gen[g0];
	for (j=g0; j<g1; j++) {
	    g_print("Generator %.5d - ",j);
	    print_gen(gen);
	    gen++;
	};

	if ( idx == 0 )
	    mod=&sfinfo.prmods.mod[m0];
	else
	    mod=&sfinfo.inmods.mod[m0];
	for (j=m0; j<m1; j++) {
	    g_print("Modulator %.3d - src %d   dest %d   amount %d   amt src %d   trans %d  unknown=%.4x\n",
		j,mod->srcoper,mod->destoper,mod->amount,mod->amtsrcoper,mod->transoper,mod->unknown);
	    mod++;
	};
}

void wt_preset_sample_button_clicked_cb (GtkWidget *widget, gpointer data)
{
	int *p,idx,sample_idx;
	p=data;
	sample_idx=*p;
	idx=p-&wt_preset_sample_idx[0];
	print_arl_sample(sample_idx);
	play_arl_sample(sample_idx,1,idx,widget);
}

void wt_preset_sample_button_toggled_cb (GtkWidget *widget, gpointer data)
{
	int *p,idx,sample_idx;
	p=data;
	sample_idx=*p;
	idx=p-&wt_preset_sample_idx[0];
	if ( gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(widget))) {
	    print_arl_sample(sample_idx);
	    wt_preset_sample[sample_idx]=play_arl_sample(sample_idx,1,idx,widget);
        }
        else {
	    stop_play_sample(&wt_wave[wt_preset_sample[sample_idx]]);
	};
}

void wt_preset_button_clicked_cb (GtkWidget *widget, gpointer data)
{
	GtkWidget *dlg;
	GtkWidget *frame;
	GtkWidget *hbox;
	GtkWidget *table;

	SFPreset *preset,*next_preset;
	SFBag *bag;
	SFInst *inst;
	SFGen *gen,*inst_gen;
	SFSampleInfo *sample;
	genAmountType keyrange_genamount;

	int b0,b1,g0,g1,m0,m1;
	int ig0,ig1;
	int i,j,k,m,n;
	char buffer[128];
	char keyrange[16];

	preset=data;
	next_preset=preset;
	next_preset++;
	b0=preset->bagNdx;
	b1=next_preset->bagNdx;

        bag=&sfinfo.prbags.bag[b0];
	g0=bag->GenNdx;
	m0=bag->ModNdx;
	
        bag=&sfinfo.prbags.bag[b1];
	g1=bag->GenNdx;
	m1=bag->ModNdx;
	
	g_print("Bank %d  Preset %d %s bag %d-%d\n",
		preset->bank,preset->preset,preset->name,b0,b1-1);
	print_bags(0,b0,b1,g0,g1,m0,m1);

	sprintf(buffer,"%s - Bank %d  preset %d",preset->name,preset->bank,preset->preset);
        dlg=gtk_dialog_new_with_buttons(buffer,
				GTK_WINDOW(window),
				GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
				GTK_STOCK_OK,
				GTK_RESPONSE_NONE,
				NULL);

        g_signal_connect_swapped (GTK_OBJECT (dlg), 
                             	"response", 
                             	G_CALLBACK (gtk_widget_destroy),
                             	GTK_OBJECT (dlg));

	hbox=gtk_hbox_new(TRUE,0);
	gtk_widget_show(hbox);
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dlg)->vbox), hbox);

	i=0;
	for (j=g0; j<g1; j++) {
            gen=&sfinfo.prgens.gen[j];
	    if ( gen->oper == SF_instrument ) {
                inst=&sfinfo.inst[gen->amount];
		frame=gtk_frame_new(inst->name);
		gtk_widget_show(frame);
		gtk_container_add(GTK_CONTAINER(hbox), frame);

	        table = gtk_table_new(8, 1, TRUE);
	        gtk_widget_show(table);
		gtk_container_add(GTK_CONTAINER(frame), table);
	
		b0=inst->bagNdx;
	        bag=&sfinfo.inbags.bag[b0];
		ig0=bag->GenNdx;
		inst++;
		b1=inst->bagNdx;
	        bag=&sfinfo.inbags.bag[b1];
		ig1=bag->GenNdx;

		m=0;
                n=0;
	        strcpy(keyrange,"");
		for (k=ig0; k<ig1; k++) {
		    inst_gen=&sfinfo.ingens.gen[k];
	 	    if ( inst_gen->oper == SF_keyRange ) {
			keyrange_genamount.shAmount=inst_gen->amount;
			if ( keyrange_genamount.ranges.byLo == keyrange_genamount.ranges.byHi )
			    strcpy(keyrange,midi_key[keyrange_genamount.ranges.byLo]);
			else
			    sprintf(keyrange,"%s - %s",midi_key[keyrange_genamount.ranges.byLo],midi_key[keyrange_genamount.ranges.byHi]); 
		    };
		    if ( inst_gen->oper == SF_sampleId ) {
		        if ( i < MAX_NO_OF_PRESET_SAMPLES ) {
		            wt_preset_sample_idx[i]=inst_gen->amount;	
			    sample=&sfinfo.sample[inst_gen->amount];
			    if ( sample->endsample > sample->startsample ) {
                                if ( strcmp(keyrange,"") == 0 )
                                    strcpy(buffer,sample->name);
				else
			            sprintf(buffer,"%s : %s",keyrange,sample->name);
                                if ( sample->endloop > sample->startloop ) {
		                    wt_preset_sample_button[i]=gtk_toggle_button_new_with_label(buffer);
			            gtk_widget_show(wt_preset_sample_button[i]);
	                            gtk_signal_connect (GTK_OBJECT (wt_preset_sample_button[i]),
					    "toggled",
		                             GTK_SIGNAL_FUNC (wt_preset_sample_button_toggled_cb),
					     &wt_preset_sample_idx[i]);	
			        }
			        else {
		                    wt_preset_sample_button[i]=gtk_button_new_with_label(buffer);
			            gtk_widget_show(wt_preset_sample_button[i]);
	                            gtk_signal_connect (GTK_OBJECT (wt_preset_sample_button[i]),
				   	     "clicked",
		                             GTK_SIGNAL_FUNC (wt_preset_sample_button_clicked_cb),
					     &wt_preset_sample_idx[i]);	
                                };
	   	                gtk_table_attach_defaults(GTK_TABLE(table),wt_preset_sample_button[i],n,n+1,m,m+1);	
			        m++;
		                i++;
			        if ( m == 12 ) {
				    m=0;
				    n++;
			        };
		                if ( n == 0 )
	                            gtk_table_resize(GTK_TABLE(table),m,1);
			        else
			            if ( m == 0 )
	                                gtk_table_resize(GTK_TABLE(table),12,n);
                                    else
	                                gtk_table_resize(GTK_TABLE(table),12,n+1);
		            };
			};
	            };
	        };
	    };
	};
	gtk_widget_show_all(dlg);
}

void wt_inst_sample_button_clicked_cb (GtkWidget *widget, gpointer data)
{
	int *p,sample_idx,idx;
	p=data;
	sample_idx=*p;
	idx=p-&wt_inst_sample_idx[0];
	print_arl_sample(sample_idx);
	play_arl_sample(sample_idx,2,idx,widget);
}

void wt_inst_sample_button_toggled_cb (GtkWidget *widget, gpointer data)
{
	int *p,sample_idx,idx;
	p=data;
	sample_idx=*p;
	idx=p-&wt_inst_sample_idx[0];
	if ( gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget))) {
	    print_arl_sample(sample_idx);
	    wt_inst_sample[sample_idx]=play_arl_sample(sample_idx,2,idx,widget);
        }
	else {
	    stop_play_sample(&wt_wave[wt_inst_sample[sample_idx]]);
	};
}

void wt_inst_button_clicked_cb (GtkWidget *widget, gpointer data)
{
	GtkWidget *dlg;
	GtkWidget *table;
	SFGen *gen;
	SFInst *inst,*next_inst;
	SFBag *bag;
	SFSampleInfo *sample;
	genAmountType keyrange_genamount;
	int b0,b1,g0,g1,m0,m1;
	int i,j,m,n;
	char keyrange[16];
	char buffer[128];

	inst=data;
	next_inst=inst;
	next_inst++;
	b0=inst->bagNdx;

        bag=&sfinfo.inbags.bag[b0];
	g0=bag->GenNdx;
	m0=bag->ModNdx;

	b1=next_inst->bagNdx;
	
        bag=&sfinfo.inbags.bag[b1];
	g1=bag->GenNdx;
	m1=bag->ModNdx;

	g_print("Inst %d - %s  bag %d-%d\n",inst-&sfinfo.inst[0],inst->name,b0,b1-1);
	print_bags(1,b0,b1,g0,g1,m0,m1);

        dlg=gtk_dialog_new_with_buttons(inst->name,
					GTK_WINDOW(window),
					GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_STOCK_OK,
					GTK_RESPONSE_NONE,
					NULL);
        g_signal_connect_swapped (GTK_OBJECT (dlg), 
                             "response", 
                             G_CALLBACK (gtk_widget_destroy),
                             GTK_OBJECT (dlg));


        table = gtk_table_new(8, 1, TRUE);
        gtk_widget_show(table);
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dlg)->vbox), table);

	i=0;
        m=0;
        n=0;
        strcpy(keyrange,"");
	for (j=g0; j<g1; j++) {
	    gen=&sfinfo.ingens.gen[j];
 	    if ( gen->oper == SF_keyRange ) {
		keyrange_genamount.shAmount=gen->amount;
		if ( keyrange_genamount.ranges.byLo == keyrange_genamount.ranges.byHi )
		    strcpy(keyrange,midi_key[keyrange_genamount.ranges.byLo]);
		else
		    sprintf(keyrange,"%s - %s",midi_key[keyrange_genamount.ranges.byLo],midi_key[keyrange_genamount.ranges.byHi]); 
	    };
	    if ( gen->oper == SF_sampleId ) {
		if ( i < MAX_NO_OF_INST_SAMPLES ) {
		    wt_inst_sample_idx[i]=gen->amount;	
		    sample=&sfinfo.sample[gen->amount];
		    if ( sample->endsample > sample->startsample ) {
                        if ( strcmp(keyrange,"") == 0 )
                            strcpy(buffer,sample->name);
                        else
		            sprintf(buffer,"%s : %s",keyrange,sample->name);
                        if ( sample->endloop > sample->startloop ) {
		            wt_inst_sample_button[i]=gtk_toggle_button_new_with_label(buffer);
		            gtk_widget_show(wt_inst_sample_button[i]);
	                    gtk_signal_connect (GTK_OBJECT (wt_inst_sample_button[i]), "toggled",
		        		GTK_SIGNAL_FUNC (wt_inst_sample_button_toggled_cb), 
					&wt_inst_sample_idx[i]);	
		        }
		        else {
		            wt_inst_sample_button[i]=gtk_button_new_with_label(buffer);
		            gtk_widget_show(wt_inst_sample_button[i]);
	                    gtk_signal_connect (GTK_OBJECT (wt_inst_sample_button[i]), "clicked",
		        		GTK_SIGNAL_FUNC (wt_inst_sample_button_clicked_cb), 
					&wt_inst_sample_idx[i]);	
		        };
	                gtk_table_attach_defaults(GTK_TABLE(table),wt_inst_sample_button[i],n,n+1,m,m+1);
		        m++;
	                i++;
		        if ( m == 12 ) {
			    m=0;
			    n++;
		        };
		        if ( n == 0 )
		             gtk_table_resize(GTK_TABLE(table),m,1);
			else
			    if ( m == 0 )
		                gtk_table_resize(GTK_TABLE(table),12,n);
		            else
		                gtk_table_resize(GTK_TABLE(table),12,n+1);
                    };
	        };
	    };
	};
	gtk_widget_show_all(dlg);

}

void wt_sample_button_clicked_cb (GtkWidget *widget, gpointer data)
{
	SFSampleInfo *sample;
	int idx;
	sample=data;
	idx=sample-&sfinfo.sample[0];
	print_arl_sample(idx);
	wt_sample[idx]=play_arl_sample(idx,3,idx,widget);
}

void wt_sample_button_toggled_cb (GtkWidget *widget, gpointer data)
{
	SFSampleInfo *sample;
	int idx;
	sample=data;
	idx=sample-&sfinfo.sample[0];
	if ( gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON (widget))) {
		print_arl_sample(idx);
		wt_sample[idx]=play_arl_sample(idx,3,idx,widget);
	}
	else {
		stop_play_sample(&wt_wave[wt_sample[idx]]);
	};
}

void parse_arl_file(char *filename)
{
    	FILE *file;
	int i,j,no_of_drum_kits;
        if (( file = fopen(filename, "rb"))) {
	    strcpy(sfinfo.filename,filename);
	    vortex_parse_arl(&sfinfo,file);
	    fclose(file);
	    gtk_table_resize(GTK_TABLE(wt_preset_table),sfinfo.npresets-1,1);
	    no_of_drum_kits=0;
	    for (i = 0; i < (sfinfo.npresets-1); i++) {
	        wt_preset_button[i]=gtk_button_new_with_label(sfinfo.preset[i].name);
	        gtk_widget_show(wt_preset_button[i]);
		if ( sfinfo.preset[i].bank == 0 ) {
		    j=sfinfo.preset[i].preset;
		}
		else {
		    j=128+no_of_drum_kits;
                    no_of_drum_kits++;		
		};	     
   	        gtk_table_attach_defaults(GTK_TABLE(wt_preset_table),wt_preset_button[i],0,1,j,j+1);	
	        gtk_signal_connect (GTK_OBJECT (wt_preset_button[i]), "clicked",
		    GTK_SIGNAL_FUNC (wt_preset_button_clicked_cb), &sfinfo.preset[i]);
	    };

	    gtk_table_resize(GTK_TABLE(wt_inst_table),sfinfo.ninsts-1,1);
	    for (i = 0; i < (sfinfo.ninsts-1); i++) {
	        wt_inst_button[i]=gtk_button_new_with_label(sfinfo.inst[i].name);
	        gtk_widget_show(wt_inst_button[i]);
	   	gtk_table_attach_defaults(GTK_TABLE(wt_inst_table),wt_inst_button[i],0,1,i,i+1);	
	        gtk_signal_connect (GTK_OBJECT (wt_inst_button[i]), "clicked",
		    GTK_SIGNAL_FUNC (wt_inst_button_clicked_cb), &sfinfo.inst[i]);
	    };

	    gtk_table_resize(GTK_TABLE(wt_sample_table),sfinfo.nsamples,1);
	    for (i=0; i<(sfinfo.nsamples-1); i++) {
		if ( sfinfo.sample[i].endloop > sfinfo.sample[i].startloop ) {
                    wt_sample_button[i]=gtk_toggle_button_new_with_label(sfinfo.sample[i].name);
	            gtk_widget_show(wt_sample_button[i]);
	            gtk_signal_connect (GTK_OBJECT (wt_sample_button[i]), "toggled",
		        GTK_SIGNAL_FUNC (wt_sample_button_toggled_cb), &sfinfo.sample[i]);
		}
		else {
                    wt_sample_button[i]=gtk_button_new_with_label(sfinfo.sample[i].name);
	            gtk_widget_show(wt_sample_button[i]);
	            gtk_signal_connect (GTK_OBJECT (wt_sample_button[i]), "clicked",
		        GTK_SIGNAL_FUNC (wt_sample_button_clicked_cb), &sfinfo.sample[i]);
                };
	   	gtk_table_attach_defaults(GTK_TABLE(wt_sample_table),wt_sample_button[i],0,1,i,i+1);	
                if ( sfinfo.sample[i].startsample == sfinfo.sample[i].endsample ) {
		    gtk_widget_set_sensitive(wt_sample_button[i],FALSE);
		};
	    };
        };
}

int play_dls_sample(int cue_idx,int type,int idx,GtkWidget *widget)
{
	WT_WAVE *wt;
	CUE *cue;
	SAMPLE *sample;
	WLOOP  *wloop;
	int i;
	if ( cue_idx < dlsinfo.pooltable.cCues ) {
		cue=&dlsinfo.cue_list[cue_idx];
		i=0;
		while (( i<16 ) && (wt_wave[i].going))
			i++;
		if ( i < 16 ) {
		    wt=&wt_wave[i];
		    wt->idx=idx;

		    wt->channels=cue->fmt.channels;
		    wt->samples_per_sec=cue->fmt.samplespersec;
		    wt->bits_per_sample= cue->fmt.bitspersample;
		    if ( cue->fmt.formattag == WAVE_FORMAT_PCM ) {
                        if ( cue->fmt.bitspersample == 16 ) {
	                    wt->format=SND_PCM_FORMAT_S16_LE;
			    wt->data16=cue->data;
			    wt->end_sample=cue->size/2;
			}
			else {
	                    wt->format=SND_PCM_FORMAT_U8;
			    wt->data8=(unsigned char *)cue->data;
			    wt->end_sample=cue->size;
			};
                    };
		    wt->start_sample=0;
		    
		    wt->start_loop=0;
		    wt->end_loop=0;
		    wt->loop=FALSE;

		    if ( cue->sample != NULL ) {
			sample=cue->sample;
			if ( sample->wsmpl.cSampleLoops > 0 ) {
			    wt->loop=TRUE;
			    wloop=&sample->wloop_list[0];
		            wt->start_loop=wloop->ulStart;
		            wt->end_loop=wloop->ulStart+wloop->ulLength;
		        };
		    };
/*
		    if ( wt->start_loop == wt->end_loop ) 
		        wt->loop=4*wt->samples_per_sec/cue->size;
		    else
		    	wt->loop=4*wt->samples_per_sec/wloop->ulLength;
*/
		    wt->button_type=type;
//		    gtk_widget_set_sensitive(widget,FALSE);	
		    play_wave_sample(wt);
		    return i;
        	};
	};
	return -1;
}

void wt_dls_inst_sample_button_clicked_cb (GtkWidget *widget, gpointer data)
{
	REGION 	*region;
	INSTInfo *inst;
	int cue_idx,idx;
	region=data;
	inst=&dlsinfo.inst_list[wt_select_dls_inst];
	print_dls_region(inst,region);
	cue_idx=region->wavelink.ulTableIndex;
	idx=region-&inst->region_list[0];
	print_dls_cue(&dlsinfo.cue_list[cue_idx]);
	play_dls_sample(cue_idx,5,idx,widget);
}

void wt_dls_inst_sample_button_toggled_cb (GtkWidget *widget, gpointer data)
{
	REGION 	*region;
	INSTInfo *inst;
	int cue_idx,idx;
	region=data;
	inst=&dlsinfo.inst_list[wt_select_dls_inst];
	print_dls_region(inst,region);
	cue_idx=region->wavelink.ulTableIndex;
	idx=region-&inst->region_list[0];
	print_dls_cue(&dlsinfo.cue_list[cue_idx]);
	if ( gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(widget))) {
	    print_dls_cue(&dlsinfo.cue_list[cue_idx]);
	    wt_dls_sample[cue_idx]=play_dls_sample(cue_idx,5,idx,widget);
        }
        else {
	    stop_play_sample(&wt_wave[wt_dls_sample[cue_idx]]);
	};	
}

void wt_dls_inst_button_clicked_cb (GtkWidget *widget, gpointer data)
{
	GtkWidget *dlg;
	GtkWidget *table;
	INSTInfo *inst;
	REGION   *region;
	CUE      *cue;
	char buffer[128];
	int i,j,k;
	inst=data;
	wt_select_dls_inst=inst-&dlsinfo.inst_list[0];
	print_dls_inst(inst);

        dlg=gtk_dialog_new_with_buttons(inst->name,
					GTK_WINDOW(window),
					GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_STOCK_OK,
					GTK_RESPONSE_NONE,
					NULL);
        g_signal_connect_swapped (GTK_OBJECT (dlg), 
                             "response", 
                             G_CALLBACK (gtk_widget_destroy),
                             GTK_OBJECT (dlg));

        table=gtk_table_new(12,1,FALSE);
        gtk_widget_show(table);
        gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dlg)->vbox), table);	

	j=0;
	k=0;
 	region=&inst->region_list[0];
        for (i=0; i<inst->header.cRegions; i++) {
	    cue=&dlsinfo.cue_list[region->wavelink.ulTableIndex];
	    if ( region->range.RangeKey.usLow == region->range.RangeKey.usHigh ) 
                sprintf(buffer,"%s : %s",midi_key[region->range.RangeKey.usLow],cue->name);
            else
                sprintf(buffer,"%s - %s : %s",midi_key[region->range.RangeKey.usLow],
			midi_key[region->range.RangeKey.usHigh],cue->name);

	    if ( region->sample != NULL ) {

		    if ( region->sample->wsmpl.cSampleLoops > 0 ) {
			    wt_dls_inst_sample_button[i]=gtk_toggle_button_new_with_label(buffer);
			    gtk_widget_show(wt_dls_inst_sample_button[i]);
			    gtk_signal_connect (GTK_OBJECT (wt_dls_inst_sample_button[i]), "toggled",
		        		GTK_SIGNAL_FUNC (wt_dls_inst_sample_button_toggled_cb), 
					region);	
		    }
		    else {
			    wt_dls_inst_sample_button[i]=gtk_button_new_with_label(buffer);
			    gtk_widget_show(wt_dls_inst_sample_button[i]);
			    gtk_signal_connect (GTK_OBJECT (wt_dls_inst_sample_button[i]), "clicked",
		        		GTK_SIGNAL_FUNC (wt_dls_inst_sample_button_clicked_cb), 
					region);	
		    };
	            gtk_table_attach_defaults(GTK_TABLE(table),wt_dls_inst_sample_button[i],k,k+1,j,j+1);
		    j++;

		    if ( j == 12 ) {
			j=0;
			k++;
		    };
		    if ( k == 0 )
		        gtk_table_resize(GTK_TABLE(table),j,1);
		    else
		        if ( j == 0 )
		            gtk_table_resize(GTK_TABLE(table),12,k);
	                else
		            gtk_table_resize(GTK_TABLE(table),12,j+1);
	    };
	    region++;
	};    
	gtk_widget_show_all(dlg);
}

void wt_dls_sample_button_clicked_cb (GtkWidget *widget, gpointer data)
{
	CUE *cue;
	int idx;
	cue=data;
	idx=cue-&dlsinfo.cue_list[0];
	print_dls_cue(cue);
	play_dls_sample(idx,6,idx,widget);
}

void parse_dls_file(char *filename)
{
    	FILE *file;
	INSTInfo *inst;
	CUE      *cue;
	int i,j,k;
        if (( file = fopen(filename, "rb"))) {
	    vortex_parse_dls(&dlsinfo,file);
	    fclose(file);
	    g_print("name = %s\n",dlsinfo.name);
	    g_print("version = %.8x.%.8x\n",dlsinfo.version.dwVersionMS,
					dlsinfo.version.dwVersionLS);
	    g_print("cInstruments = %d\n",dlsinfo.header.cInstruments);	
	    g_print("Cues = %d\n",dlsinfo.pooltable.cCues);

	    gtk_table_resize(GTK_TABLE(wt_preset_table),dlsinfo.header.cInstruments,1);
	    gtk_table_resize(GTK_TABLE(wt_inst_table),dlsinfo.header.cInstruments,1);
	    inst=&dlsinfo.inst_list[0];
            k=128;
	    for (i = 0; i < dlsinfo.header.cInstruments; i++) {
		if (( inst->header.Locale.ulBank == 0 ) || ( inst->header.Locale.ulBank == 0x80000000 )) {
                    if ( inst->header.Locale.ulBank == 0 )
		        j=inst->header.Locale.ulInstrument;
                    else {
                        j=k;
                        k++;       
		        gtk_table_resize(GTK_TABLE(wt_preset_table),128+k,1);
		    };
 
 	            wt_dls_preset_button[j]=gtk_button_new_with_label(inst->name);
	            gtk_widget_show(wt_dls_preset_button[j]);
	   	    gtk_table_attach_defaults(GTK_TABLE(wt_preset_table),wt_dls_preset_button[j],0,1,j,j+1);	
	            gtk_signal_connect (GTK_OBJECT (wt_dls_preset_button[j]), "clicked",
		    GTK_SIGNAL_FUNC (wt_dls_inst_button_clicked_cb), inst);
		};

	        wt_dls_inst_button[i]=gtk_button_new_with_label(inst->name);
	        gtk_widget_show(wt_dls_inst_button[i]);
	   	gtk_table_attach_defaults(GTK_TABLE(wt_inst_table),wt_dls_inst_button[i],0,1,i,i+1);	
	        gtk_signal_connect (GTK_OBJECT (wt_dls_inst_button[i]), "clicked",
		    GTK_SIGNAL_FUNC (wt_dls_inst_button_clicked_cb), inst);
		inst++;
	    };

	    gtk_table_resize(GTK_TABLE(wt_sample_table),dlsinfo.pooltable.cCues,1);

	    cue=&dlsinfo.cue_list[0];
	    for (i=0; i<dlsinfo.pooltable.cCues; i++) {
                wt_dls_sample_button[i]=gtk_button_new_with_label(cue->name);
	        gtk_widget_show(wt_dls_sample_button[i]);
	   	gtk_table_attach_defaults(GTK_TABLE(wt_sample_table),wt_dls_sample_button[i],0,1,i,i+1);	
	        gtk_signal_connect (GTK_OBJECT (wt_dls_sample_button[i]), "clicked",
		        GTK_SIGNAL_FUNC (wt_dls_sample_button_clicked_cb), cue);
		
		cue++;
	    };
	};
}

void wt_filesel_ok_sel(GtkWidget *widget, GtkFileSelection *fs )
{
	int i;
	char filename[FILENAME_MAX+1];
	strcpy(filename,gtk_file_selection_get_filename(GTK_FILE_SELECTION(fs)));
	if (( strstr(filename,".arl") > 0 ) || ( strstr(filename,".dls") > 0 ) ) {
		if ( sfinfo.loaded ) {
	            for (i = 0; i < (sfinfo.npresets-1); i++) {
		        gtk_widget_destroy(wt_preset_button[i]);
                    };
	            for (i = 0; i < (sfinfo.ninsts-1); i++) {
		        gtk_widget_destroy(wt_inst_button[i]);
                    };
		    for (i = 0; i < (sfinfo.nsamples-1); i++) {
			gtk_widget_destroy(wt_sample_button[i]);
		    };
/*
		    for (i= 0; i<8; i++) {
			if ( sfinfo.info[i].desc != NULL )
			    gtk_widget_destroy(wt_info_label[i]);
		    };
*/
		    vortex_free_arl(&sfinfo);
		};
		if ( dlsinfo.loaded ) {
		    for (i=0; i< dlsinfo.header.cInstruments; i++) {
		        gtk_widget_destroy(wt_dls_preset_button[i]);
		    };
		    for (i=0; i< dlsinfo.header.cInstruments; i++) {
		        gtk_widget_destroy(wt_dls_inst_button[i]);
		    };
		    for (i=0; i<dlsinfo.pooltable.cCues; i++) {
			gtk_widget_destroy(wt_dls_sample_button[i]);
		    };
		    vortex_free_dls(&dlsinfo);
		};
		if ( is_arl_file(filename) ) {
		    parse_arl_file(filename);						
		};
		if ( is_dls_file(filename) ) {
		    parse_dls_file(filename);
		};		
		gtk_widget_set_sensitive(wt_midi_button,TRUE);
	        gtk_widget_destroy(GTK_WIDGET(fs));
	};
}

void wt_wavetable_button_clicked_cb (GtkWidget *widget, gpointer data)
{
	GtkWidget *fs;
        fs=gtk_file_selection_new("Wavetable");
        gtk_widget_show(fs);
        gtk_file_selection_hide_fileop_buttons(GTK_FILE_SELECTION(fs));
        gtk_file_selection_set_select_multiple(GTK_FILE_SELECTION(fs),FALSE);
//	gtk_file_selection_complete(GTK_FILE_SELECTION(fs),"*.arl");
     /* Connect the ok_button to file_ok_sel function */
        g_signal_connect (G_OBJECT (GTK_FILE_SELECTION (fs)->ok_button),
		      "clicked", G_CALLBACK (wt_filesel_ok_sel), (gpointer) fs);
    
        g_signal_connect_swapped (G_OBJECT (GTK_FILE_SELECTION (fs)->cancel_button),
	                      "clicked", G_CALLBACK (gtk_widget_destroy),
			      G_OBJECT (fs));	     			
}

#define printchunk() g_print("%c%c%c%c \n",chunk.id[0],chunk.id[1],chunk.id[2],chunk.id[3])

unsigned short readword(FILE *fd)
{
	uint16 temp;
        fread(&temp,2,1,fd);
	return(XCHG_SHORT(temp));
}

void readmidichunk(MIDIChunk *chunk,FILE *fd)
{
	uint32 temp;
	fread(chunk->id,4,1,fd);
	fread(&temp,4,1,fd);
	chunk->size=XCHG_LONG(temp);
}

static int num_params(unsigned char stat)
{
	if (stat >= MIDI_CMD_COMMON_SYSEX)
		return(-1);	/* SysEx or meta */
	if (stat < MIDI_CMD_PGM_CHANGE || stat >= MIDI_CMD_BENDER)
		return(2);
	else
		return(1);
}

ulong stol(unsigned char *str,int len)
{
	int i;
	long retval;

	retval = 0L;
	for (i = 0; i < len; i++)
	{
		retval <<= 8;
		retval |= str[i];
	}
	return(retval);
}

int midi_getc(TRACKInfo *track)
{
    unsigned char *p;
    if ( track->pos < track->size ) {
        p=track->data + track->pos ;
	track->pos++;
	return (*p);
    }
    else {
	return(MFE_EOF);
    };
}

static long read_varlen(TRACKInfo *track)
{
	long value;
	int c;

	if ((value = midi_getc(track)) < 0) 
		return(value);
	else {
		if (value & 0x80)
		{
			value &= 0x7f;
			do
			{
				if ((c = midi_getc(track)) < 0) 
					return(c);
				else
				        value = (value << 7) + (c & 0x7f);
			} while (c & 0x80);
		}
	}
	return(value);
}

int meta_keysig(int sharpflat,int minor)
{
	g_print("sharpflat =%d   minor=%d\n",sharpflat,minor);
	return 0;
}

int meta_seqnum(short seqnum)
{
	g_print("Seqnum = %d\n",seqnum);
	return 0;
}

int meta_smpte(int hour,int min,int sec,int frame,int fract)
{
	g_print("smpte hour=%d min=%d sec=%d frame=%d fract=%d\n",hour,min,sec,frame,fract);
	return 0;
}

int meta_timesig(int numer,int denom,int clocks,int qnotes)
{
	g_print("time sig  numer=%d  denom=%d clocks=%d quotes=%d\n",numer,denom,clocks,qnotes);
	return 0;
}

int meta_tempo(unsigned long tempo)
{
	g_print("tempo=%.8lx\n",tempo);	
	return 0;
}

int meta_text(int msg_type, unsigned char *data,long length)
{
	long i;
	switch (msg_type) {
        case META_TEXT:		
	    g_print("TEXT - ");
	    break;
	case META_COPYRIGHT:	
	    g_print("COPYRIGHT - ");
	    break;
	case META_SEQNAME:
	    g_print("SEQUENCER NAME - ");
	    break;
	case META_INSTNAME:
	    g_print("INSTRUMENT NAME - ");
	    break;
	case META_LYRIC:	
	    g_print("LYRIC - ");
	    break;	
        case META_MARKER:
	    g_print("MARKER - ");
	    break;
	case META_CUEPT:
	    g_print("CUE POINT - ");
	    break;
	case META_PROGRAM:
	    g_print("PROGRAM - ");
	    break;
	case META_DEVICE:
	    g_print("DEVICE - ");
	    break;
	default:
	    g_print("meta text - ");
	    break;
	};
	for (i=0; i<length; i++)
	    g_print("%c",data[i]);
	g_print("\n");
	return 0;
}

int midi_note_off(unsigned char channel, unsigned char note, unsigned char velocity)
{	
	int idx;
	g_print("Note off - channel=%.2x   note=%.2x %s   velocity=%.2x\n",channel,note,midi_key[note],velocity);
        idx=midiinfo.channel[channel];
	if ( sfinfo.loaded ) {
	    if ( sfinfo.midi[idx].note[note].wt >= 0 ) {
//		stop_play_sample(&wt_wave[sfinfo.midi[idx].note[note].wt]);
	    };
	};
	if ( dlsinfo.loaded ) {
	    if ( dlsinfo.midi[idx].note[note].wt >= 0 ) {
//		stop_play_sample(&wt_wave[dlsinfo.midi[idx].note[note].wt]);
	    };
	};
	return 0;
}

// There is only 16 adb which are not enough for playing midi.
// Need to use 64 WT instead of using ADB. 

int midi_note_on(unsigned char channel, unsigned char note, unsigned char velocity)
{
	SFInst *inst;
	SFPreset *preset;
	SFSampleInfo *sample;
	INSTInfo *dls_inst;
	int idx;
    if ( velocity == 0 ) 
        return(midi_note_off(channel,note,velocity));
    else {
	g_print("Note on - channel=%.2x   note=%.2x %s   velocity=%.2x\n",channel,note,midi_key[note],velocity);
        idx=midiinfo.channel[channel];
	if ( sfinfo.loaded ) {
            preset=&sfinfo.preset[sfinfo.midi[idx].preset];
	    inst=&sfinfo.inst[sfinfo.midi[idx].inst];
	    if ( sfinfo.midi[idx].note[note].sample >= 0 ) {
		sample=&sfinfo.sample[sfinfo.midi[idx].note[note].sample];
		g_print("ARL %s sample=%.3d %dHz rootkey=%.2x %s\n",preset->name,sfinfo.midi[idx].note[note].sample,sample->samplerate,sample->originalpitch,midi_key[sample->originalpitch]);
//	        sfinfo.midi[idx].note[note].wt=play_arl_sample(sfinfo.midi[idx].note[note].sample,3,sfinfo.midi[idx].note[note].sample,wt_sample_button[sfinfo.midi[idx].note[note].sample]);
	    };
	};
	if ( dlsinfo.loaded ) {
            dls_inst=&dlsinfo.inst_list[dlsinfo.midi[idx].inst];
	    if ( dlsinfo.midi[idx].note[note].cue >= 0 ) {
		g_print("DLS %s cue=%d\n",dls_inst->name,dlsinfo.midi[idx].note[note].cue);
//	        dlsinfo.midi[idx].note[note].wt=play_dls_sample(dlsinfo.midi[idx].note[note].cue,0,6,wt_dls_sample_button[dlsinfo.midi[idx].note[note].cue]);
	    };
	};
    };
    return 0;
}

int midi_note_pressure(unsigned char channel, unsigned char note, unsigned char velocity)
{
	g_print("Pressure - channel=%d  note=%d %s   velocity=%d\n",channel,note,midi_key[note],velocity);
	return 0;
}

int midi_control(unsigned char channel, unsigned char c, unsigned char value)
{
	g_print("controller c=%d  value=%d\n",c,value);
	return 0;
}

int midi_pgm_change(unsigned char channel, unsigned char inst)
{
	midiinfo.channel[channel]=inst;
	g_print("program change - channel=%d instrument=%d\n",channel,inst);
	return 0;
}

int midi_bender(unsigned char channel, unsigned char p1, unsigned char p2)
{
	g_print("pitchbend %.4x\n",(unsigned short)(p2)*256+p1);
	return 0;
}

int midi_channel_pressure(unsigned char channel, unsigned char p1)
{
	g_print("chanpressure %d\n",p1);
	return 0;
}

void vortex_free_midi(MIDIInfo *midi)
{
	TRACKInfo *track;
	int i;
        for (i=0; i<midi->ntrks; i++) {
	     track=&midi->track_list[i];
	     if ( track->data ) 
		free(track->data);
        };
	if ( midi->track_list )
		free(midi->track_list);	
	midi->loaded=FALSE;
}

void vortex_parse_midi(MIDIInfo *midi,FILE *fd)
{
	MIDIChunk chunk;
	TRACKInfo *track;
	int i;

        readmidichunk(&chunk,fd);   

	midi->format=readword(fd);
	midi->ntrks=readword(fd);
	midi->division=readword(fd);
	g_print("format = %d   Tracks=%d   Division=%d\n",midi->format,midi->ntrks,midi->division);

	midi->track_list=NEW(TRACKInfo,midi->ntrks);
	if ( midi->track_list != NULL ) {
	    for (i=0; i<midi->ntrks; i++) {	
                track=&midi->track_list[i];
                readmidichunk(&chunk,fd);   
                track->size=chunk.size;
		track->data=malloc(track->size);
		if ( track->data != NULL ) {
		    fread(track->data,track->size,1,fd);
		};
	    };
	};
	midi->loaded=TRUE;
}

void play_track(TRACKInfo *track)
{
	int error;
	int stat;			/* Current status */
	long delta_time;		/* Time since last event */
	static unsigned char params[2];		/* Event parameters */
	unsigned char channel;
	int cur_param;			/* Parameter being currently read */
	int i;
	int c;		/* Input character */

	char *data;		/* Where the data is kept */
	long length;		/* Length of message */
	short seqnum;
	uint32 tempo;
	int hour,min,sec,frame,fract;
	int numer,denom,clocks,qnotes;
	int sharpflat,minor;
	int msg_type;	/* Type of meta-event */
	long pos;

	error=MFE_OK;
        
        pos=track->pos;
	error =MFE_OK;
        if ((delta_time = read_varlen(track)) < 0)
            error=((int) delta_time);	
        else {
            while (( (track->currenttime+delta_time) <= midiinfo.currenttime ) && ( error == MFE_OK )) {
                track->currenttime += track->deltatime += delta_time;
 	        printf("%.8lx track %.2d ",track->currenttime,track-&midiinfo.track_list[0]);
	        if ((stat = midi_getc(track)) < 0 ) 
			error=MFE_EOF;	
	        else {
		    if (stat & 0x80) {
			track->run_stat = (unsigned char) stat;		
			cur_param = 0;		
		    } 
                    else {
			params[0] = (unsigned char) stat;	
			cur_param = 1;			
		    }

		    for (i = num_params(track->run_stat)-cur_param; i > 0; i--, cur_param++) {
			if ((c = midi_getc(track)) < 0 ) 
			    error=MFE_EOF;
			else 
			    params[cur_param] = (unsigned char) c;	
		    }

                    if ( error == MFE_OK ) {
                        channel=track->run_stat & 0x0F;
			switch(track->run_stat & 0xf0) {	
			case MIDI_CMD_NOTE_OFF:	
			    error = midi_note_off(channel,params[0],params[1]);
			    track->deltatime = 0L;
			    break;
		
			case MIDI_CMD_NOTE_ON:	
			    error = midi_note_on(channel,params[0],params[1]);
			    track->deltatime = 0L;
			    break;
		
			case MIDI_CMD_NOTE_PRESSURE:	
			    error = midi_note_pressure(channel,params[0],params[1]);
			    track->deltatime = 0L;
			    break;
		
			case MIDI_CMD_CONTROL:	
			    error = midi_control(channel,params[0],params[1]);
			    track->deltatime = 0L;
			    break;
		
			case MIDI_CMD_PGM_CHANGE:	
			    error = midi_pgm_change(channel,params[0]);
			    track->deltatime = 0L;
			    break;
		
			case MIDI_CMD_CHANNEL_PRESSURE:	
			    error = midi_channel_pressure(channel,params[0]);
			    track->deltatime = 0L;
			    break;
		
			case MIDI_CMD_BENDER:	
			    error = midi_bender(channel,params[0],params[1]);
			    track->deltatime = 0L;
			    break;

			case MIDI_CMD_COMMON_SYSEX:	
                            switch(track->run_stat) {
			    case MIDI_CMD_COMMON_SYSEX:
				do {
				    c=midi_getc(track);
				} while (( c == 0xF7 ) && ( c != MFE_EOF ));
                                break;
			    case 0xFF:
			        if ((msg_type = midi_getc(track)) < 0 ) 
				    error=MFE_EOF;
			        else {
			            if ((length = read_varlen(track)) < 0)
				        error=((int) length);
			            else {
			   	        data=track->data + track->pos;
				        track->pos+=length;

			  	        switch(msg_type) {
			  	        case META_SEQNUM:		
				    	    if (length == 2) {	
					        seqnum = ((unsigned short)(data[0]) << 8) | data[1];
					        error = meta_seqnum(seqnum);
					        track->deltatime = 0L;
					    }	
					    else  
					        error=MFE_ILLMLEN;
					    break;	
				        case META_TEXT:		
				        case META_COPYRIGHT:	
				        case META_SEQNAME:	
	                                case META_INSTNAME:
				        case META_LYRIC:	
				        case META_MARKER:	
			 	        case META_CUEPT:	
					    error = meta_text((int) msg_type,data,length);
					    track->deltatime = 0L;
					    break;				
				        case META_EOT:		
					    if (length != 0L) 
					        error=MFE_ILLMLEN;
					    
					    printf("End of Track\n");
					    midiinfo.playing_tracks--;
					    error=MFE_EOF;
					    break;
				        case META_TEMPO:	
					    if (length == 3) {	
					        tempo = stol(data,3);
					        error = meta_tempo(tempo);
					        track->deltatime = 0L;
					    }
					    else
					        error=MFE_ILLMLEN;
					    break;
				        case META_SMPTE:	
					    if (length == 5) {
					        hour  = (int) data[0];
					        min   = (int) data[1];
					        sec   = (int) data[2];
					        frame = (int) data[3];
					        fract = (int) data[4];
			                        error = meta_smpte(hour,min,sec,frame,fract);
					        track->deltatime = 0L;
					    }
					    else
					        error=MFE_ILLMLEN;
					    break;
				        case META_TIMESIG:	
					    if (length == 4) {
					        numer  = (int) data[0];
					        denom  = (int) data[1];
					        clocks = (int) data[2];
					        qnotes = (int) data[3];	 
       	        			        error = meta_timesig(numer,denom,clocks,qnotes);
					        track->deltatime = 0L;
					    }
					    else
					        error=MFE_ILLMLEN;
					    break;
				        case META_KEYSIG:	
					    if (length == 2) {
		 			        sharpflat = (int) data[0];
					        minor     = (int) data[1];
					        if (sharpflat > 7 ||   sharpflat < -7 ||  (minor != 0 && minor != 1)) {  
					   	    error=MFE_ILLMVAL;
			 		        }
					        else {
						    error = meta_keysig(sharpflat,minor);
						    track->deltatime = 0L;
					        } 
					    }
					    else 
					        error=MFE_ILLMLEN;
					    break;
				        case META_SEQSPEC:	
					    hexdump(data,length);
					    track->deltatime = 0L;
					    break;
				        default:
		                            error = MFE_UKMETA;
					    break;
				       };
				    };
				};
				break;
			    default:
				break;
			    };
			    break;
		
			default:
			    error=MFE_ILLSTAT;
			    break;
			};
                    };
		};	
                if ( track->pos < track->size ) {
                    pos=track->pos;
                    delta_time = read_varlen(track);
		};
	    };
            if ( track->pos < track->size )
		track->pos=pos;	   
            if ( (track->currenttime+delta_time) < midiinfo.nexttime )
                midiinfo.nexttime=track->currenttime+delta_time;
        };
}

void vortex_play_midi(MIDIInfo *midi)
{
	TRACKInfo *track;
	int i,j,k;
        for (i=0; i<midi->ntrks; i++) {	
            track=&midi->track_list[i];
	    track->pos=0;	
	    track->currenttime=0;
	    track->deltatime=0;
	    track->run_stat=0;
            if ( i == 0 )
		track->prev=midi->ntrks-1;
	    else
                track->prev= i-1;
	    track->next= ( i+1 ) % midi->ntrks;
        };
        midi->currenttime=0;
        midi->playing_tracks=midi->ntrks;
        k=0;
	while ( 0 < midi->playing_tracks ) {
            i=0; 
            j=midi->playing_tracks;
            midi->nexttime=midi->currenttime+10000;
	    while ( i < j ) {
	        track=&midi->track_list[k];
                play_track(track);
                if ( track->pos < track->size )
		    k=track->next;
		else {
		    (&midi->track_list[track->prev])->next=track->next;
                    (&midi->track_list[track->next])->prev=track->prev;
                    k=track->next;
                    track->prev=-1;
                    track->next=-1;                    
		};
	        i++;
            } 
//	    hwmix_usleep((midi->nexttime-midi->currenttime)*midi->division*10);
            midi->currenttime=midi->nexttime;
	};
}

void parse_midi_file(char *filename)
{
    	FILE *file;
        if (( file = fopen(filename, "rb"))) {
	    vortex_parse_midi(&midiinfo,file);
	    fclose(file);
	    gtk_widget_set_sensitive(wt_play_midi_button,TRUE);
	};
}

void wt_play_midi_button_clicked_cb (GtkWidget *widget, gpointer data)
{
	gtk_widget_set_sensitive(wt_midi_button,FALSE);
	gtk_widget_set_sensitive(wt_stop_midi_button,TRUE);
	vortex_play_midi(&midiinfo);
	gtk_widget_set_sensitive(wt_midi_button,TRUE);
	gtk_widget_set_sensitive(wt_stop_midi_button,FALSE);
}

void wt_stop_midi_button_clicked_cb (GtkWidget *widget, gpointer data)
{
	gtk_widget_set_sensitive(wt_play_midi_button,TRUE);
	gtk_widget_set_sensitive(wt_midi_button,TRUE);
}

void wt_midi_filesel_ok_sel(GtkWidget *widget, GtkFileSelection *fs )
{
	char filename[FILENAME_MAX+1];
	strcpy(filename,gtk_file_selection_get_filename(GTK_FILE_SELECTION(fs)));
	if ( strstr(filename,".mid") > 0 ) {
		if ( midiinfo.loaded )
		    vortex_free_midi(&midiinfo);
	        gtk_widget_destroy(GTK_WIDGET(fs));
		parse_midi_file(filename);
	}
}

void wt_midi_button_clicked_cb (GtkWidget *widget, gpointer data)
{
	GtkWidget *fs;
        fs=gtk_file_selection_new("MIDI");
        gtk_widget_show(fs);
        gtk_file_selection_hide_fileop_buttons(GTK_FILE_SELECTION(fs));
        gtk_file_selection_set_select_multiple(GTK_FILE_SELECTION(fs),FALSE);
	gtk_file_selection_complete(GTK_FILE_SELECTION(fs),"*.mid");
     /* Connect the ok_button to file_ok_sel function */
        g_signal_connect (G_OBJECT (GTK_FILE_SELECTION (fs)->ok_button),
		      "clicked", G_CALLBACK (wt_midi_filesel_ok_sel), (gpointer) fs);
    
    /* Connect the cancel_button to destroy the widget */
        g_signal_connect_swapped (G_OBJECT (GTK_FILE_SELECTION (fs)->cancel_button),
	                      "clicked", G_CALLBACK (gtk_widget_destroy),
			      G_OBJECT (fs));	     			
}

static void create_midi(GtkWidget *main, GtkWidget *notebook, int page)
{
	GtkWidget *label;
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkWidget *button;
	GtkWidget *frame;
	GtkWidget *table;
	GtkWidget *scrolled_window;
	GtkWidget *combo;
        GList *glist;
	int i;

	vbox = gtk_vbox_new(FALSE, 0);
	gtk_widget_show(vbox);
	gtk_container_add(GTK_CONTAINER(notebook), vbox);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 6);


        label = gtk_label_new("MIDI");
        gtk_widget_show(label);
	gtk_notebook_set_tab_label(GTK_NOTEBOOK(notebook), 
				   gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), page), 
				   label);
  
	table = gtk_table_new(10, 1, TRUE);
	gtk_widget_show(table);
	gtk_container_add(GTK_CONTAINER(vbox), table);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_widget_show(hbox);
   	gtk_table_attach_defaults(GTK_TABLE(table),hbox,0,1,0,5);	

        frame = gtk_frame_new("Preset");
        gtk_widget_show(frame);
	gtk_container_add(GTK_CONTAINER(hbox), frame);

        scrolled_window = gtk_scrolled_window_new (NULL, NULL);    
    	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),
                                    GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
        gtk_widget_show (scrolled_window);    
	gtk_container_add(GTK_CONTAINER(frame), scrolled_window);

        wt_preset_table = gtk_table_new (16, 1, FALSE);        
        gtk_widget_show (wt_preset_table);
	gtk_scrolled_window_add_with_viewport (
                   GTK_SCROLLED_WINDOW (scrolled_window), wt_preset_table);

        frame = gtk_frame_new("Instruments");
        gtk_widget_show(frame);
	gtk_container_add(GTK_CONTAINER(hbox), frame);

        scrolled_window = gtk_scrolled_window_new (NULL, NULL);    
    	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),
                                    GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
        gtk_widget_show (scrolled_window);    
	gtk_container_add(GTK_CONTAINER(frame), scrolled_window);
   
	wt_inst_table = gtk_table_new(16, 1, FALSE);
	gtk_widget_show(wt_inst_table);
	gtk_scrolled_window_add_with_viewport (
                   GTK_SCROLLED_WINDOW (scrolled_window), wt_inst_table);

        frame = gtk_frame_new("Samples");
        gtk_widget_show(frame);
	gtk_container_add(GTK_CONTAINER(hbox), frame);

        scrolled_window = gtk_scrolled_window_new (NULL, NULL);
       	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),
                                    GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
        gtk_widget_show (scrolled_window);    
	gtk_container_add(GTK_CONTAINER(frame), scrolled_window);
   
	wt_sample_table = gtk_table_new(16, 1, FALSE);
	gtk_widget_show(wt_sample_table);
	gtk_scrolled_window_add_with_viewport (
                   GTK_SCROLLED_WINDOW (scrolled_window), wt_sample_table);
/*
        frame = gtk_frame_new("Info");
        gtk_widget_show(frame);
	gtk_container_add(GTK_CONTAINER(hbox), frame);

        scrolled_window = gtk_scrolled_window_new (NULL, NULL);    
    	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),
                                    GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
        gtk_widget_show (scrolled_window);    
	gtk_container_add(GTK_CONTAINER(frame), scrolled_window);

        wt_info_table = gtk_table_new (16, 1, FALSE);        
        gtk_widget_show (wt_info_table);
	gtk_scrolled_window_add_with_viewport (
                   GTK_SCROLLED_WINDOW (scrolled_window), wt_info_table);
 
*/
  	wt_scope_drawing = gtk_drawing_area_new();
  	gtk_widget_show (wt_scope_drawing);
   	gtk_table_attach_defaults(GTK_TABLE(table),wt_scope_drawing,0,1,5,9);	

        gtk_widget_set_usize(wt_scope_drawing, 512, 48);
        gtk_signal_connect(GTK_OBJECT(wt_scope_drawing), 
	  	"expose_event",
                  	GTK_SIGNAL_FUNC(wt_scope_drawing_expose_event), NULL);
        gtk_signal_connect(GTK_OBJECT(wt_scope_drawing), 
		"configure_event",
                       	GTK_SIGNAL_FUNC(wt_scope_drawing_configure_event), NULL);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_widget_show(hbox);
   	gtk_table_attach_defaults(GTK_TABLE(table),hbox,0,1,9,10);	


        combo=gtk_combo_new();
	gtk_widget_show(combo);
	gtk_container_add(GTK_CONTAINER(hbox), combo);	

        glist = NULL;  
        for (i=0; i<4; i++) {
            glist = g_list_append (glist, scope_type_name[i]);
        };
        gtk_combo_set_popdown_strings (GTK_COMBO (combo), glist);
	g_list_free(glist);
        gtk_entry_set_editable(GTK_ENTRY(GTK_COMBO(combo)->entry),FALSE);
        gtk_signal_connect(GTK_OBJECT(GTK_COMBO(combo)->entry),
                         "changed",
                         GTK_SIGNAL_FUNC(wt_scope_type_changed_cb),
                         NULL);

	wt_play_midi_button = gtk_button_new_with_label("Parse MIDI");
        gtk_widget_show(wt_play_midi_button);
	gtk_container_add(GTK_CONTAINER(hbox), wt_play_midi_button);

        gtk_signal_connect (GTK_OBJECT (wt_play_midi_button), "clicked",
		    GTK_SIGNAL_FUNC (wt_play_midi_button_clicked_cb),NULL);
	gtk_widget_set_sensitive(wt_play_midi_button,FALSE);

	wt_stop_midi_button = gtk_button_new_with_label("Stop");
        gtk_widget_show(wt_stop_midi_button);
	gtk_container_add(GTK_CONTAINER(hbox), wt_stop_midi_button);

        gtk_signal_connect (GTK_OBJECT (wt_stop_midi_button), "clicked",
		    GTK_SIGNAL_FUNC (wt_stop_midi_button_clicked_cb),NULL);
	gtk_widget_set_sensitive(wt_stop_midi_button,FALSE);

	wt_midi_button = gtk_button_new_with_label("Select MIDI");
        gtk_widget_show(wt_midi_button);
	gtk_container_add(GTK_CONTAINER(hbox), wt_midi_button);

        gtk_signal_connect (GTK_OBJECT (wt_midi_button), "clicked",
		    GTK_SIGNAL_FUNC (wt_midi_button_clicked_cb),NULL);
	gtk_widget_set_sensitive(wt_midi_button,FALSE);

 
	button = gtk_button_new_with_label("Select Sound Bank File");
        gtk_widget_show(button);
	gtk_container_add(GTK_CONTAINER(hbox), button);

        gtk_signal_connect (GTK_OBJECT (button), "clicked",
		    GTK_SIGNAL_FUNC (wt_wavetable_button_clicked_cb),NULL);

        strcpy(sfinfo.filename,"gm.dls");
	if ( card_is_au8810 ) {
	    strcpy(sfinfo.filename,"au10gm.arl");
	};
	if ( card_is_au8820 ) {
	    strcpy(sfinfo.filename,"aurealgm");
	};
	if ( card_is_au8830 ) {	
	    strcpy(sfinfo.filename,"au30gm.arl");
	};
        if ( is_arl_file(sfinfo.filename) ) {
	    parse_arl_file(sfinfo.filename);
	    gtk_widget_set_sensitive(wt_midi_button,TRUE);		
        }
	else {
	    if ( is_dls_file(sfinfo.filename) ) {
		parse_dls_file(sfinfo.filename);
	        gtk_widget_set_sensitive(wt_midi_button,TRUE);				
	    };
	};
}

/******************************************************************************************/
void spdif_rate_set_toggle_button_cb (GtkWidget *widget, gpointer data)
{
	snd_ctl_elem_list_t *clist;
	int *rate,err,controls,cidx;
	rate=data;
	snd_ctl_elem_list_alloca(&clist);
	if ((err = snd_ctl_elem_list(ctl, clist)) < 0) {
		g_print("snd_ctl_elem_list : %s\n", snd_strerror(err));
		return;
	}
	if ((err = snd_ctl_elem_list_alloc_space(clist, snd_ctl_elem_list_get_count(clist))) < 0) {
		g_print("snd_ctl_elem_list_alloc_space : %s\n", snd_strerror(err));
		return;
	}
	if ((err = snd_ctl_elem_list(ctl, clist)) < 0) {
		g_print("snd_ctl_elem_list : %s\n", snd_strerror(err));
		return;
	}	
	controls = snd_ctl_elem_list_get_used(clist);
	for (cidx = 0; cidx < controls; cidx++) {
		if (!strcmp(snd_ctl_elem_list_get_name(clist, cidx), spdif_str))
		    break;
	}
	snd_ctl_elem_id_alloca(&id);
	snd_ctl_elem_list_get_id(clist, spdif_id, id);
	snd_ctl_elem_value_alloca(&spdif_out);
	snd_ctl_elem_value_set_id(spdif_out, id);
	if ((err = snd_ctl_elem_read(ctl, spdif_out)) < 0) {
	    g_print("snd_ctl_elem_read : %s\n", snd_strerror(err));
	    return;
	}
	snd_ctl_elem_value_get_iec958(spdif_out, &iec958);
	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget))) {
	   g_print("spdif %d\n",*rate);
	   if ( iec958.status[0] & IEC958_AES0_PROFESSIONAL) {
		iec958.status[0] & (~IEC958_AES0_PRO_FS);
		switch(*rate){
		case 32000:
			iec958.status[3] |= IEC958_AES0_PRO_FS_32000;
			break;
		case 44100:
			iec958.status[3] |= IEC958_AES0_PRO_FS_44100;
			break;
		case 48000:
			iec958.status[3] |= IEC958_AES0_PRO_FS_48000;
			break;		   
		};
	   }
	   else {
		iec958.status[3] & (~IEC958_AES3_CON_FS);
		switch(*rate){
		case 32000:
			iec958.status[3] |= IEC958_AES3_CON_FS_32000;
			break;
		case 44100:
			iec958.status[3] |= IEC958_AES3_CON_FS_44100;
			break;
		case 48000:
			iec958.status[3] |= IEC958_AES3_CON_FS_48000;
			break;		   
		};
	    };
	};	
	snd_ctl_elem_value_set_iec958(spdif_out, &iec958);
	if ((err = snd_ctl_elem_write(ctl, spdif_out)) < 0) {
		g_print("snd_ctl_elem_write : %s\n", snd_strerror(err));
		return;
	}
	if ((err = snd_ctl_elem_read(ctl, spdif_out)) < 0) {
		g_print("snd_ctl_elem_write : %s\n", snd_strerror(err));
		return;
	}
	snd_ctl_elem_value_get_iec958(spdif_out, &iec958);
};	

static void create_spdif(GtkWidget *main, GtkWidget *notebook, int page)
{
	GtkWidget *label;
	GtkWidget *vbox;
	GtkWidget *frame;
	GtkWidget *hbox;

	snd_ctl_elem_list_t *clist;
	unsigned int controls, cidx;
	char status[256];
	int i,err;

	vbox = gtk_vbox_new(FALSE, 0);
	gtk_widget_show(vbox);
	gtk_container_add(GTK_CONTAINER(notebook), vbox);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 6);

        label = gtk_label_new("SPDIF In/Out");
        gtk_widget_show(label);
	gtk_notebook_set_tab_label(GTK_NOTEBOOK(notebook), 
				   gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), page), 
				   label);

	snd_ctl_elem_list_alloca(&clist);
	if ((err = snd_ctl_elem_list(ctl, clist)) < 0) {
		g_print("snd_ctl_elem_list : %s\n", snd_strerror(err));
		return;
	}
	if ((err = snd_ctl_elem_list_alloc_space(clist, snd_ctl_elem_list_get_count(clist))) < 0) {
		g_print("snd_ctl_elem_list_alloc_space : %s\n", snd_strerror(err));
		return;
	}
	if ((err = snd_ctl_elem_list(ctl, clist)) < 0) {
		g_print("snd_ctl_elem_list : %s\n", snd_strerror(err));
		return;
	}
	controls = snd_ctl_elem_list_get_used(clist);
	for (cidx = 0; cidx < controls; cidx++) {
		if (!strcmp(snd_ctl_elem_list_get_name(clist, cidx), spdif_str))
		    break;
	}
	if (cidx < controls) {
		spdif_id=cidx;
//		g_print("%s\n",snd_ctl_elem_list_get_name(clist, cidx));
		snd_ctl_elem_id_alloca(&id);
		snd_ctl_elem_list_get_id(clist, cidx, id);
		snd_ctl_elem_value_alloca(&spdif_out);
		snd_ctl_elem_value_set_id(spdif_out, id);
		if ((err = snd_ctl_elem_read(ctl, spdif_out)) < 0) {
		    g_print("snd_ctl_elem_read : %s\n", snd_strerror(err));
		    return;
		}
	
		snd_ctl_elem_value_get_iec958(spdif_out, &iec958);
			
		frame= gtk_frame_new(snd_ctl_elem_id_get_name(id));
		gtk_widget_show(frame);
	        gtk_container_add(GTK_CONTAINER(vbox), frame);			

		sprintf(status,"AES0=0x%2x,AES1=0x%2x,AES2=0x%2x,AES3=0x%2x\n",
		       	iec958.status[0], iec958.status[1], 
			iec958.status[2], iec958.status[3]);
		label=gtk_label_new(status);
		gtk_widget_show(label);
		gtk_container_add(GTK_CONTAINER(frame), label);
						
          	frame= gtk_frame_new("Mode");
		gtk_widget_show(frame);
	        gtk_container_add(GTK_CONTAINER(vbox), frame);			
		if ( iec958.status[0] & IEC958_AES0_PROFESSIONAL) 
			label=gtk_label_new("professional");
		else
			label=gtk_label_new("consumer");
		gtk_widget_show(label);
		gtk_container_add(GTK_CONTAINER(frame), label);

		frame= gtk_frame_new("Data");
		gtk_widget_show(frame);
	        gtk_container_add(GTK_CONTAINER(vbox), frame);			
		if (!(iec958.status[0] & IEC958_AES0_NONAUDIO)) 
		    label=gtk_label_new("audio");
		else 
		    label=gtk_label_new("non-audio");
		gtk_widget_show(label);
		gtk_container_add(GTK_CONTAINER(frame), label);

	        frame= gtk_frame_new("Rate");
		gtk_widget_show(frame);
	        gtk_container_add(GTK_CONTAINER(vbox), frame);	

		hbox = gtk_hbox_new(FALSE, 0);
		gtk_widget_show(hbox);
                gtk_container_add(GTK_CONTAINER(frame), hbox);

	        spdif_rate_radio[0] = gtk_radio_button_new_with_label (NULL, "32000 Hz");
                spdif_rate_radio[1] = gtk_radio_button_new_with_label(gtk_radio_button_group (GTK_RADIO_BUTTON (spdif_rate_radio[0])), "44100 Hz");
                spdif_rate_radio[2] = gtk_radio_button_new_with_label(gtk_radio_button_group (GTK_RADIO_BUTTON (spdif_rate_radio[0])), "48000 Hz");
		for (i = 0; i < 3; i++) {
		        gtk_widget_show(spdif_rate_radio[i]);
                        gtk_box_pack_start(GTK_BOX(hbox), spdif_rate_radio[i], TRUE, TRUE, 2);
	    		gtk_signal_connect (GTK_OBJECT (spdif_rate_radio[i]), 							"toggled",
			                        GTK_SIGNAL_FUNC (spdif_rate_set_toggle_button_cb),
						&spdif_rate[i]);
		};
		if ( iec958.status[0] & IEC958_AES0_PROFESSIONAL) {
		    switch (iec958.status[0] & IEC958_AES0_PRO_FS) {
		    case IEC958_AES0_PRO_FS_44100:
		        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(spdif_rate_radio[1]),TRUE);
			break;
		    case IEC958_AES0_PRO_FS_48000:
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(spdif_rate_radio[2]),TRUE);
			break;
		    case IEC958_AES0_PRO_FS_32000:
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(spdif_rate_radio[0]),TRUE);
			break;
		    default:
			break;
		    }
                }
		else {
		    switch (iec958.status[3] & IEC958_AES3_CON_FS) {
		    case IEC958_AES3_CON_FS_44100:
		        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(spdif_rate_radio[1]),TRUE);
			break;
		    case IEC958_AES3_CON_FS_48000:
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(spdif_rate_radio[2]),TRUE);
			break;
		    case IEC958_AES3_CON_FS_32000:
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(spdif_rate_radio[0]),TRUE);
			break;
		    default:
			break;
		    }
		};

		frame= gtk_frame_new("Copyright");
		gtk_widget_show(frame);
	        gtk_container_add(GTK_CONTAINER(vbox), frame);			
    		if (iec958.status[0] & IEC958_AES0_CON_NOT_COPYRIGHT) 
		    label=gtk_label_new("permitted");
	        else 
		    label=gtk_label_new("protected");
	        gtk_widget_show(label);
		gtk_container_add(GTK_CONTAINER(frame), label);


		frame= gtk_frame_new("Emphasis");
		gtk_widget_show(frame);
	        gtk_container_add(GTK_CONTAINER(vbox), frame);			

		if ((iec958.status[0] & IEC958_AES0_CON_EMPHASIS) != IEC958_AES0_CON_EMPHASIS_5015)
		    label=gtk_label_new("none");
	        else 
		    label=gtk_label_new("50/15us");
		    
	        gtk_widget_show(label);
		gtk_container_add(GTK_CONTAINER(frame), label);

		frame= gtk_frame_new("Category");
		gtk_widget_show(frame);
	        gtk_container_add(GTK_CONTAINER(vbox), frame);	
		
		for (i = 0; i < ARRAY_SIZE(con_category); i++) {
		    if ((iec958.status[1] & IEC958_AES1_CON_CATEGORY) == 	con_category[i].val) {
			label=gtk_label_new(con_category[i].name);
			break;
		    }
		}
		if (i >= ARRAY_SIZE(con_category)) {
		    sprintf(status,"unknown 0x%x", iec958.status[1] & 	IEC958_AES1_CON_CATEGORY);
		    label=gtk_label_new(status);
		}
		gtk_widget_show(label);			
		gtk_container_add(GTK_CONTAINER(frame),label);

		frame= gtk_frame_new("Original");
		gtk_widget_show(frame);
	        gtk_container_add(GTK_CONTAINER(vbox), frame);			

		if (iec958.status[1] & IEC958_AES1_CON_ORIGINAL) 
		    label=gtk_label_new("original");
		else 
		    label=gtk_label_new("1st generation");
	        gtk_widget_show(label);
		gtk_container_add(GTK_CONTAINER(frame), label);

		frame= gtk_frame_new("Clock");
		gtk_widget_show(frame);
	        gtk_container_add(GTK_CONTAINER(vbox), frame);			

		switch (iec958.status[3] & IEC958_AES3_CON_CLOCK) {
		case IEC958_AES3_CON_CLOCK_1000PPM:
		    label=gtk_label_new("1000 ppm");
		    break;
		case IEC958_AES3_CON_CLOCK_50PPM:
		    label=gtk_label_new("50 ppm");
		    break;
		case IEC958_AES3_CON_CLOCK_VARIABLE:
		    label=gtk_label_new("variable pitch");
		    break;
		default:
		    label=gtk_label_new("unknown");
		    break;
		}
		gtk_widget_show(label);
		gtk_container_add(GTK_CONTAINER(frame), label);
	}
	else {
		label=gtk_label_new("SPDIF not working until au88x0 driver include PCM control for ICE958");
		gtk_widget_show(label);
  	        gtk_container_add(GTK_CONTAINER(vbox), label);
	};	
}

/******************************************************************************************/

static void generate_sine(signed short *samples, int channel, int count, double *_phase) {
	double phase = *_phase;
	double max_phase = 1.0 / freq;
	double step = 1.0 / (double)rate;
	double res;
	int    chn, ires;

	while (count-- > 0) {
		res = sin((phase * 2 * M_PI) / max_phase - M_PI) * 32767;
		ires = res;
		for(chn=0;chn<channels;chn++) {
			if (chn==channel)
				*samples++ = ires;
			else
				*samples++ = 0;
		}
		phase += step;
		if (phase >= max_phase)
			phase -= max_phase;
	}
	*_phase = phase;
}

static int set_hwparams(snd_pcm_t *handle, snd_pcm_hw_params_t *params, snd_pcm_access_t access) {
	unsigned int rrate;
	int          err, dir;
	snd_pcm_uframes_t     period_size_min;
	snd_pcm_uframes_t     period_size_max;
	snd_pcm_uframes_t     buffer_size_min;
	snd_pcm_uframes_t     buffer_size_max;
	snd_pcm_uframes_t     buffer_time_to_size;

  /* choose all parameters */
	if ((err = snd_pcm_hw_params_any(handle, params) < 0 )) {
    		g_print("Broken configuration for playback: no configurations available: %s\n", snd_strerror(err));
    		return err;
  	}

  /* set the interleaved read/write format */
  	if ((err = snd_pcm_hw_params_set_access(handle, params, access)) < 0 ) {
		g_print("Access type not available for playback: %s\n", snd_strerror(err));
    		return err;
  	}

  /* set the sample format */
  	if (( err = snd_pcm_hw_params_set_format(handle, params, format)) < 0 ) {
		g_print("Sample format not available for playback: %s\n", snd_strerror(err));
		return err;
  	}

  /* set the count of channels */
  	if (( err = snd_pcm_hw_params_set_channels(handle, params, channels)) < 0 ) {
		g_print("Channels count (%i) not available for playbacks: %s\n", channels, snd_strerror(err));
		return err;
	}

  /* set the stream rate */
  	rrate = rate;
  	if ((err = snd_pcm_hw_params_set_rate(handle, params, rate, 0)) < 0 ) {
		g_print("Rate %iHz not available for playback: %s\n", rate, snd_strerror(err));
		return err;
	}

  	if (rrate != rate) {
    		g_print("Rate doesn't match (requested %iHz, get %iHz, err %d)\n", rate, rrate, err);
    		return -EINVAL;
  	}

	g_print("Rate set to %iHz\n", rrate);
  /* set the buffer time */
	buffer_time_to_size = ( (snd_pcm_uframes_t)buffer_time * rate) / 1000000;
  	err = snd_pcm_hw_params_get_buffer_size_min(params, &buffer_size_min);
  	err = snd_pcm_hw_params_get_buffer_size_max(params, &buffer_size_max);
	dir=0;
	err = snd_pcm_hw_params_get_period_size_min(params, &period_size_min,&dir);
	dir=0;
	err = snd_pcm_hw_params_get_period_size_max(params, &period_size_max,&dir);

	buffer_size = buffer_time_to_size;
	if (buffer_size_max < buffer_size) 
		buffer_size = buffer_size_max;
	if (buffer_size_min > buffer_size) 
		buffer_size = buffer_size_min;
	period_size=buffer_size/8;
	buffer_size=period_size*8;

	dir=0;
	if ((err = snd_pcm_hw_params_set_period_size_near(handle, params, &period_size, &dir)) < 0 ) {
		g_print("Unable to set period size %lu for playback: %s\n", period_size, snd_strerror(err));
		return err;
	}
	dir=0;
	if (( err = snd_pcm_hw_params_get_period_size(params, &period_size, &dir)) < 0 ) {
		g_print("Unable to get period size for playback: %s\n", snd_strerror(err));
        };                                                                                                                     
	if ((err = snd_pcm_hw_params_set_buffer_size_near(handle, params, &buffer_size)) < 0 ) {
    		g_print("Unable to set buffer size %lu for playback: %s\n", buffer_size, snd_strerror(err));
		return err;
	}
	err = snd_pcm_hw_params_get_buffer_size(params, &buffer_size);
	g_print("period_size = %lu\n",period_size);
	g_print("buffer_size = %lu\n",buffer_size);
	if (2*period_size > buffer_size) {
		g_print("buffer to small, could not use\n");
		return err;
	}

  /* write the parameters to device */
  	if (( err = snd_pcm_hw_params(handle, params)) < 0 ) {
		g_print("Unable to set hw params for playback: %s\n", snd_strerror(err));
		return err;
	}
  	return 0;
}

static int set_swparams(snd_pcm_t *handle, snd_pcm_sw_params_t *swparams) {
	int err;

  /* get the current swparams */
  	if ((err = snd_pcm_sw_params_current(handle, swparams)) < 0 ) {
		g_print("Unable to determine current swparams for playback: %s\n", snd_strerror(err));
    		return err;
  	}

  /* start the transfer when a period is full */
	if (( err = snd_pcm_sw_params_set_start_threshold(handle, swparams, period_size)) < 0 ) {
		g_print("Unable to set start threshold mode for playback: %s\n", snd_strerror(err));
		return err;
	}

  /* allow the transfer when at least period_size samples can be processed */
	if (( err = snd_pcm_sw_params_set_avail_min(handle, swparams, period_size)) < 0 ) {
		g_print("Unable to set avail min for playback: %s\n", snd_strerror(err));
		return err;
	}

  /* align all transfers to 1 sample */
	if (( err = snd_pcm_sw_params_set_xfer_align(handle, swparams, 1)) < 0 ) {
		g_print("Unable to set transfer align for playback: %s\n", snd_strerror(err));
		return err;
	}

  /* write the parameters to the playback device */
	if ((err = snd_pcm_sw_params(handle, swparams)) < 0 ) {
		g_print("Unable to set sw params for playback: %s\n", snd_strerror(err));
		return err;
	}
	return 0;
}

/*
 *   Underrun and suspend recovery
 */

static int xrun_recovery(snd_pcm_t *handle, int err) {
	if (err == -EPIPE) {	/* under-run */
    		if (( err = snd_pcm_prepare(handle)) < 0 ) {
			g_print("Can't recovery from underrun, prepare failed: %s\n", snd_strerror(err));
		};
		return 0;
	} 
  	else {
		if (err == -ESTRPIPE) {
			while ((err = snd_pcm_resume(handle)) == -EAGAIN)
      				sleep(1);	/* wait until the suspend flag is released */

    			if (err < 0) {
      				err = snd_pcm_prepare(handle);
      				if (err < 0)
        				g_print("Can't recovery from suspend, prepare failed: %s\n", snd_strerror(err));
    			};
    			return 0;
  		}
  		return err;
	};
}

static int write_loop(snd_pcm_t *handle, int channel, int periods, signed short *samples) {
	double phase = 0;
	signed short *ptr;
	int    err, cptr, n;

	for(n = 0; n < periods; n++) {
		generate_sine(samples, channel, period_size, &phase);
		ptr = samples;
		cptr = period_size;
		while (cptr > 0) {
			err = snd_pcm_writei(handle, ptr, cptr);
			if (err == -EAGAIN)
        			continue;
      			if (err < 0) {
        			if (xrun_recovery(handle, err) < 0) {
          				g_print("Write error: %s\n", snd_strerror(err));
	  				return -1;
        			}
        			break;	/* skip one period */
      			}
      			ptr += (err * channels);
      			cptr -= err;
    		}
  	}
  	return 0;
}

int elem_value_get_integer(snd_hctl_elem_t *elem)
{
    int err;
    snd_ctl_elem_info_alloca(&info);
    snd_ctl_elem_value_alloca(&val);
    if ( (err= snd_hctl_elem_info(elem, info)) < 0) {
	g_print("snd_hctl_elem_info : %s\n",snd_strerror(err));
    }
    else {
	if ((err = snd_hctl_elem_read(elem,val)) < 0) {
	    g_print("snd_hctl_elem_read : %s\n",snd_strerror(err));
	}
	else {
	      return(snd_ctl_elem_value_get_integer(val,0));          
	};
    };
    return -1;    
}

int elem_value_get_leftright(snd_hctl_elem_t *elem) 
{
    int err,result,i;
    snd_ctl_elem_info_alloca(&info);
    snd_ctl_elem_value_alloca(&val);
    if ( (err= snd_hctl_elem_info(elem, info)) < 0) {
	g_print("snd_hctl_elem_info : %s\n",snd_strerror(err));
    }
    else {
	if ((err = snd_hctl_elem_read(elem,val)) < 0) {
	    g_print("snd_hctl_elem_read : %s\n",snd_strerror(err));
	}
	else {
	      result=0;
	      for (i=0; i<2; i++) {
		  result<<=8;
		  result |= snd_ctl_elem_value_get_integer(val,i);
	      }
	      return(result);          
	};
    };
    return -1;    
}

int elem_value_get_boolean(snd_hctl_elem_t *elem)
{
    int err;
    snd_ctl_elem_info_alloca(&info);
    snd_ctl_elem_value_alloca(&val);
    if ( (err= snd_hctl_elem_info(elem, info)) < 0) {
	g_print("snd_hctl_elem_info : %s\n",snd_strerror(err));
    }
    else {
	if ((err = snd_hctl_elem_read(elem,val)) < 0) {
	    g_print("snd_hctl_elem_read : %s\n",snd_strerror(err));
	}
	else {
	      return(snd_ctl_elem_value_get_boolean(val,0));          
	};
    };
    return FALSE;    
}

void play_sine_wave(int chn)
{
 	snd_pcm_t            *handle;
	snd_pcm_hw_params_t  *hwparams;
  	snd_pcm_sw_params_t  *swparams;
  	int                   err;
  	signed short         *samples;

//      	rate = rate < 8000 ? 8000 : rate;
//      	rate = rate > 48000 ? 48000 : rate;

//	freq = (int)(gtk_adjustment_get_value(GTK_ADJUSTMENT(speaker_freq_adj)));
     
      	buffer_time = buffer_time < 1000 ? 1000 : buffer_time;
      	buffer_time = buffer_time > 1000000 ? 1000000 : buffer_time;

      	period_time = period_time < 1000 ? 1000 : period_time;
      	period_time = period_time > 1000000 ? 1000000 : period_time;

/*
	err = snd_output_stdio_attach(&output, stdout, 0);
  	if (err < 0) {
		g_print("Output failed: %s\n", snd_strerror(err));
		return;
  	}
*/
	snd_pcm_hw_params_alloca(&hwparams);
  	snd_pcm_sw_params_alloca(&swparams);

  	g_print("Stream parameters are %iHz, %s, %i channels\n", rate, snd_pcm_format_name(format), channels);
  	g_print("Sine wave rate is %.4fHz chn=%d\n", freq,chn);

  	if ((err = snd_pcm_open(&handle, card, SND_PCM_STREAM_PLAYBACK, SND_PCM_NONBLOCK)) < 0) {
    		g_print("Playback open error: %s\n", snd_strerror(err));
    		exit(EXIT_FAILURE);
  	}

  	if ((err = set_hwparams(handle, hwparams, SND_PCM_ACCESS_RW_INTERLEAVED)) < 0) {
    		g_print("Setting of hwparams failed: %s\n", snd_strerror(err));
    		exit(EXIT_FAILURE);
  	}
  	if ((err = set_swparams(handle, swparams)) < 0) {
    		g_print("Setting of swparams failed: %s\n", snd_strerror(err));
    		exit(EXIT_FAILURE);
  	}

  	samples = malloc((period_size * channels * snd_pcm_format_width(format)) / 8);
  	if (samples == NULL) {
    		g_print("No enough memory\n");
    		exit(EXIT_FAILURE);
  	}

        err = write_loop(handle, chn, ((rate*3)/period_size), samples);

      	if (err < 0) {
        	g_print("Transfer failed: %s\n", snd_strerror(err));
		exit(EXIT_FAILURE);
    	};
  
  	free(samples);
  	err=snd_pcm_close(handle);
        if ( err < 0 ) {
            g_print("snd_pcm_close err:%s\n",snd_strerror(err));
	};
}	

/**************************************************************************************/

#define AWIDTH 128

static void scope_render_pcm (int idx) {
  	int i,j;
  	gint32 ry, ly, rry, rly, try = 0, tly = 0, trry=0, trly=0;
  	gint16 *pointer16;
	unsigned char *pointer8;
	int    width,height,h2;
        int channels,bits_per_sample;
	gint32    hh;

	channels=hwmix[idx].channels;
	if ( channels == 0 ) 
	    channels=1;
	bits_per_sample=snd_pcm_format_physical_width(hwmix[idx].format);

	width=scope_drawing[idx]->allocation.width;
	h2=scope_drawing[idx]->allocation.height/(2*channels);
	height=h2+h2;

 	GDK_THREADS_ENTER();

	if ( bits_per_sample == 16 ) {
		pointer16 = & scope_data[idx][0];
		hh=262144/height;
		if ( hh * height < 262144 )
			hh++;
	}
	else {
		pointer8 = (unsigned char *) & scope_data[idx][0];
		hh=1024/height;
		if ( hh * height < 1024 )
			hh++;
	};
  	for ( i = 0; i < AWIDTH; i++ ) {
	    ly = 0;
	    ry = 0;
  	    switch (channels) {
	    case 1:
	        if ( bits_per_sample == 16 ) {
		    for (j=0; j<4; j++) {
		        ly += *pointer16; 
		        pointer16++;
		    };
	        }
	        else {
		    for (j=0; j<4; j++) {
		        ly += *pointer8; 
		        pointer8++;
		    };
	        };
  		ly /= hh;
		break;
	    case 2:
	        if ( bits_per_sample == 16 ) {
		    for (j=0; j<4; j++) {
		        ly += *pointer16; 
		        pointer16++;
	      	        ry += *pointer16; 
		        pointer16++;
		    };		
	        }
	        else {
		    for (j=0; j<4; j++) {
		        ly += *pointer8; 
		        pointer8++;
		        ry+=*pointer8;
		        pointer8++;
		    };
                };
   		ly /= hh;
                ry /= hh;
		break;
	   case 4:
	        rry=0;
	        rly=0;
	        if ( bits_per_sample == 16 ) {
		    for (j=0; j<4; j++) {
		        ly += *pointer16; 
		        pointer16++;
	      	        ry += *pointer16; 
		        pointer16++;
		        rly += *pointer16; 
		        pointer16++;
	      	        rry += *pointer16; 
		        pointer16++;
		    };
		};
                ly /= hh;
    	        ry /= hh;
    		rly /= hh;
    	        rry /= hh;
                break;
	    };
	    if ( scope_type != 3 ) {

// FIXME - SND_PCM_FORMAT_U8 not display correctly

   	        ly += h2;
    	        ry += h2;
                rry+= h2;
		rly+= h2;
	    };

// FIXME - Wave is inverted in the scope.

    	    switch (scope_type) {
	    case 0:			// dot
		switch (channels) {
		case 4:
		    gdk_draw_point(scope_pixmap[idx], penWhite, i, height+height+height+rry); 
		    gdk_draw_point(scope_pixmap[idx], penWhite, i, height+height+rly); 
		case 2:
		    gdk_draw_point(scope_pixmap[idx], penWhite, i, height+ry);
		case 1:
		    gdk_draw_point(scope_pixmap[idx], penWhite, i, ly); 
		    break;
		};
		break;
	    case 1:			// line
		if (i == 0) { 
		    switch (channels) {
		    case 4:
			gdk_draw_point(scope_pixmap[idx], penWhite, 0, height+height+height+rry);
			trry = rry;		
		        gdk_draw_point(scope_pixmap[idx], penWhite, 0, height+height+rly);
		        trly = rly;
		    case 2:
			gdk_draw_point(scope_pixmap[idx], penWhite, 0, height+ry);
			try = ry;		
		    case 1:
		        gdk_draw_point(scope_pixmap[idx], penWhite, 0, ly);
		        tly = ly;
			break;
		    };
		}
		else {
		    switch (channels) {
		    case 4:
		        gdk_draw_line(scope_pixmap[idx], penWhite, 
				i - 1, height+height+height+trry, i, height+height+height+rry);
		        trry = rry;
		        gdk_draw_line(scope_pixmap[idx], penWhite, 
				i - 1, height+height+trly, i,    height+height+rly);
	  	        trly = rly;
		    case 2:
		        gdk_draw_line(scope_pixmap[idx], penWhite, 
				i - 1, height+try, i, height+ry);
		        try = ry;
		    case 1:
		        gdk_draw_line(scope_pixmap[idx], penWhite, i - 1, tly, i,    ly);
	  	        tly = ly;
			break;
                    };
      		};
		break;
	    case 2:			// filled
		switch (channels) {
		case 4:
		    if (rry != h2) 
		        gdk_draw_line(scope_pixmap[idx], penWhite, 
		i, height+height+height+((rry > 0) ? h2 : h2-1), i, height+height+height+rry);
		    if (rly != h2) 
		        gdk_draw_line(scope_pixmap[idx], penWhite, 
				i, height+height+((rly > 0) ? h2 : h2-1), i, height+height+rly);
		case 2:
		    if (ry != h2) 
		        gdk_draw_line(scope_pixmap[idx], penWhite, 
				i, height+((ry > 0) ? h2 : h2-1), i, height+ry);
		case 1:
		    if (ly != h2) 
		        gdk_draw_line(scope_pixmap[idx], penWhite, i, ((ly > 0) ? h2 : h2-1), i, ly);
		    break;
		};
		break;
	    case 3:			// filled 2
		switch (channels) {
		case 4:
		    rry /= 2;
		    gdk_draw_line(scope_pixmap[idx], penWhite, i, height+height+height+h2-1-rry, i, height+height+height+h2+rry);
		    rly /= 2;
		    gdk_draw_line(scope_pixmap[idx], penWhite, i, height+height+h2-1-rly, i, height+height+h2+rly);
		case 2:
		    ry /= 2;
		    gdk_draw_line(scope_pixmap[idx], penWhite, i, height+h2-1-ry, i, height+h2+ry);
		case 1:
		    ly /= 2;
		    gdk_draw_line(scope_pixmap[idx], penWhite, i, h2-1-ly, i, h2+ly);
		    break;
		};
		break;
	    };
	};
	gdk_window_clear(scope_drawing[idx]->window);
  	GDK_THREADS_LEAVE();
}

void redraw_scope(int idx)
{
       if ( GTK_WIDGET_VISIBLE(scope_drawing[idx]) ) {
		    scope_render_pcm(idx);
		    gdk_draw_rectangle(scope_pixmap[idx],
			   scope_drawing[idx]->style->black_gc,
			   TRUE,
			   0, 0,
			   scope_drawing[idx]->allocation.width,
			   scope_drawing[idx]->allocation.height);
	};	
}

void redraw_all_scope(void)
{
	int idx;
	for (idx=0; idx<no_of_scope; idx++) 
	    redraw_scope(idx);
}

static int get_scope(GtkWidget *widget)
{
	int idx;
	for (idx=0; idx<no_of_scope; idx++) {
		if ( widget == scope_drawing[idx] ) {
			return idx;
		};
	};	
	return 0;
}

gint scope_drawing_configure_event(GtkWidget *widget, GdkEventConfigure *event)
{
	int idx;
	idx=get_scope(widget);
	if (scope_pixmap[idx] != NULL)
		gdk_pixmap_unref(scope_pixmap[idx]);
 	scope_pixmap[idx] = gdk_pixmap_new(window->window, 
                          widget->allocation.width, widget->allocation.height,
			  gdk_visual_get_best_depth());
  	gdk_draw_rectangle(scope_pixmap[idx], penBlack, TRUE, 0, 0, 
			  widget->allocation.width, widget->allocation.height);
  	gdk_window_set_back_pixmap(scope_drawing[idx]->window, scope_pixmap[idx], 0);
	redraw_scope(idx); 
	return TRUE;
}

gint scope_drawing_expose_event(GtkWidget *widget, GdkEventExpose *event)
{
	int idx;
	idx=get_scope(widget);
	gdk_window_clear_area (widget->window,
                         event->area.x, event->area.y,
                         event->area.width, event->area.height);
  	gdk_gc_set_clip_rectangle (widget->style->fg_gc[widget->state],&event->area);

	redraw_scope(idx);
  	gdk_gc_set_clip_rectangle (widget->style->fg_gc[widget->state],NULL);
	return TRUE;
}

int alsa_playing(HWMIX_WAVE *p)
{
	if (!p->going || p->paused)
		return FALSE;
	return(snd_pcm_state(p->alsa_pcm) == SND_PCM_STATE_RUNNING);
}

static void xrun_recover(HWMIX_WAVE *p)
{
	int err;
	if (snd_pcm_state(p->alsa_pcm) == SND_PCM_STATE_XRUN) {
		if ((err = snd_pcm_prepare(p->alsa_pcm)) < 0)
			g_warning("snd_pcm_prepare() failed.");
	}
}

static snd_pcm_sframes_t alsa_get_avail(HWMIX_WAVE *p)
{
	snd_pcm_sframes_t ret;
	if ((ret = snd_pcm_avail_update(p->alsa_pcm)) == -EPIPE)
		xrun_recover(p);
	else {
	    if (ret < 0) {
		g_warning("snd_pcm_avail_update() failed: %s",snd_strerror(-ret));
		return 0;
	    }
	    else
		return ret;
        };
	if ((ret = snd_pcm_avail_update(p->alsa_pcm)) < 0) {
		g_warning("snd_pcm_avail_update() failed: %s",snd_strerror(-ret));
		return 0;
	}
	return ret;
}

int alsa_free(HWMIX_WAVE *p)
{
        int err;
	if (p->paused)
		return 0;
	else {
		if (p->force_start &&
		    snd_pcm_state(p->alsa_pcm) == SND_PCM_STATE_PREPARED) {
			if ((err = snd_pcm_start(p->alsa_pcm)) < 0)
				g_warning("snd_pcm_start() failed: %s", snd_strerror(err));
		}
		p->force_start = TRUE;
		return snd_pcm_frames_to_bytes(p->alsa_pcm, alsa_get_avail(p));
	}
}

void alsa_pause(HWMIX_WAVE *p,int pause_resume)
{
	if (pause_resume)
		p->paused = TRUE;
	if (p->alsa_can_pause)
		snd_pcm_pause(p->alsa_pcm, pause_resume);
	else if (pause_resume)
		snd_pcm_drop(p->alsa_pcm);
	if (!pause_resume)
		p->paused = FALSE;
}

void alsa_close(HWMIX_WAVE *p)
{
	int err, started;

	started = p->going;
	p->going = 0;

	if (p->alsa_pcm != NULL) {
		if (started) {
			if ((err = snd_pcm_drop(p->alsa_pcm)) < 0)
				g_warning("snd_pcm_drop() failed: %s",snd_strerror(err));
                };
		if ((err = snd_pcm_close(p->alsa_pcm)) < 0)
			g_warning("snd_pcm_close() failed: %s",snd_strerror(err));
		p->alsa_pcm = NULL;
	}
/*
	if (p->mmap) {
		g_free(p->buffer);
		p->buffer = NULL;

		g_free(p->areas);
		p->areas = NULL;
	}
*/
}

int alsa_get_written_time(HWMIX_WAVE *p)
{
	return (p->alsa_total_written * 1000 / p->alsa_bps);
}

static int alsa_setup(HWMIX_WAVE *p)
{
	int err;
	snd_pcm_t	    *alsa_pcm;
	snd_pcm_hw_params_t *hwparams;
	snd_pcm_sw_params_t *swparams;
	int bits_per_sample;
	snd_pcm_uframes_t alsa_buffer_size, alsa_period_size;
        unsigned int rrate;
        int dir;
	char device[32];


        strcpy(device,card);
	if ( p->v3d ) 
	    strcat(device,",2");

	err = snd_pcm_open(&p->alsa_pcm, device,
				SND_PCM_STREAM_PLAYBACK,
				SND_PCM_NONBLOCK);
	
	if ( err < 0 ) {
		g_warning("Failed to open pcm device (%s): %s",device, snd_strerror(err));
		p->alsa_pcm = NULL;
		return -1;
	};
	alsa_pcm=p->alsa_pcm;

	snd_pcm_hw_params_alloca(&p->hwparams);
        hwparams=p->hwparams;
 	err = snd_pcm_hw_params_any(alsa_pcm, hwparams);
	if ( err < 0 ) {
		g_warning("No configuration available for playback: %s", snd_strerror(err));
		return -1;
	};
/*
	if (p->mmap) {
	    err = snd_pcm_hw_params_set_access(alsa_pcm, hwparams,
						SND_PCM_ACCESS_MMAP_INTERLEAVED);
	    if ( err < 0 ) {
		g_message("Cannot set mmap mode: %s.", snd_strerror(err));
		p->mmap = 0;
	    };
	};
*/
//	if (!p->mmap ) {
	    err = snd_pcm_hw_params_set_access(alsa_pcm, hwparams,SND_PCM_ACCESS_RW_INTERLEAVED);
	    if ( err < 0 ) {
		g_warning("Cannot set direct write mode: %s",snd_strerror(err));
		return -1;
            };
//	}
 	err = snd_pcm_hw_params_set_format(alsa_pcm, hwparams, p->format);
        if ( err < 0 ) {
		g_warning("Sample format not available for playback: %s",snd_strerror(err));
		return -1;
	}

  	err=snd_pcm_hw_params_set_channels(alsa_pcm, hwparams, p->channels);
        if (err < 0) {
                g_print("Channels count (%i) not available for playbacks: %s\n", p->channels, snd_strerror(err));
                return -1;
        }
	dir=0;
        rrate=p->rate;
	err=snd_pcm_hw_params_set_rate_near(alsa_pcm, hwparams, &rrate, &dir);
	if ( err < 0 ) {
		g_warning("No usable samplerate available.");
		return -1;
	}
        if ( rrate != p->rate ) {
		g_warning("Rate doesn't match (requested %iHz, get %iHz)\n", p->rate, rrate);
                return -1;
        };

	err=snd_pcm_hw_params_get_period_size_max(hwparams,&alsa_period_size,&dir);
	if ( err < 0 ) {
		g_print("snd_pcm_hw_params_get_period_size_max failed : %s\n",snd_strerror(err));
		return -1;
	};

        err=snd_pcm_hw_params_set_period_size(p->alsa_pcm,hwparams,alsa_period_size,dir);
	if ( err < 0 ) {
		g_print("snd_pcm_hw_params_set_period_size failed : %s\n",snd_strerror(err));
		return -1;
	};

        alsa_buffer_size=alsa_period_size*4;
        err=snd_pcm_hw_params_set_buffer_size(p->alsa_pcm,hwparams,alsa_buffer_size);
	if ( err < 0 ) {
		g_print("snd_pcm_hw_params_set_buffer_size failed : %s\n",snd_strerror(err));
		return -1;
	};
	err=snd_pcm_hw_params(alsa_pcm, hwparams);
	if ( err < 0 ) {
		g_warning("snd_pcm_hw_params() failed: %s",snd_strerror(err));
		return -1;
	}

 	p->alsa_can_pause = snd_pcm_hw_params_can_pause(hwparams);

 	snd_pcm_sw_params_alloca(&p->swparams);
/*
        if ( err < 0 ) {
		g_warning("snd_pcm_sw_params_alloca() failed: %s",snd_strerror(err));
		return -1;
	};
*/
        swparams=p->swparams;
	snd_pcm_sw_params_current(alsa_pcm, swparams);
/*	
	if ( err < 0 ) {
		g_warning("snd_pcm_sw_params_current() failed: %s",snd_strerror(err));
		return -1;
	};
*/
	err = snd_pcm_sw_params_set_start_threshold(alsa_pcm,
                                 swparams, alsa_buffer_size - alsa_period_size);
        if ( err < 0 ) {
		g_warning("snd_pcm_sw_params_set_start_threshold() failed: %s", snd_strerror(err));
        };
	err=snd_pcm_sw_params(alsa_pcm, swparams);
	if ( err < 0 ) {
		g_warning("snd_pcm_sw_params() failed: %s", snd_strerror(err));
		return -1;
	}
	bits_per_sample = snd_pcm_format_physical_width(p->format);
	p->alsa_bps = (p->rate * bits_per_sample * p->channels) >> 3;
/*
	if (p->mmap) {
		p->buffer = g_malloc(alsa_period_size * bits_per_sample / 8 * p->channels);
		p->areas = g_malloc0(p->channels * sizeof(snd_pcm_channel_area_t));
		
   		for (chn = 0; chn < p->channels; chn++) {
			p->areas[chn].addr = p->buffer;
			p->areas[chn].first = chn * bits_per_sample;
			p->areas[chn].step = p->channels * bits_per_sample;
		}
	}
*/
	return 0;
}

void alsa_flush(HWMIX_WAVE *p,int time)
{
	p->alsa_total_written = (guint64) time * p->alsa_bps / 1000;
}

int alsa_get_output_time(HWMIX_WAVE *p)
{
	snd_pcm_sframes_t delay;
	ssize_t db = 0;

	if (!p->going)
		return 0;

	if (!snd_pcm_delay(p->alsa_pcm, &delay))
		db = snd_pcm_frames_to_bytes(p->alsa_pcm, delay);
	
	if (db < p->alsa_total_written)
		return ((p->alsa_total_written - db) * 1000 / p->alsa_bps);
	return 0;
}

static void alsa_mmap_audio(HWMIX_WAVE *p,char *data, int length)
{
	int cnt = 0, err;
	snd_pcm_uframes_t offset, frames, frame;
	const snd_pcm_channel_area_t *chan_areas = p->areas;
	int channel_offset = 0, channel;
	ssize_t sample_size, offset_bytes, step;

	alsa_get_avail(p);

	while (length > 0) {
		frames = snd_pcm_bytes_to_frames(p->alsa_pcm, length);
		if ((err = snd_pcm_mmap_begin(p->alsa_pcm, &chan_areas, &offset, &frames) < 0))
			g_warning("snd_pcm_mmap_begin() failed: %s", snd_strerror(err));

		cnt = snd_pcm_frames_to_bytes(p->alsa_pcm, frames);

		sample_size = snd_pcm_samples_to_bytes(p->alsa_pcm, 1);
		step = chan_areas[0].step / 8;
		offset_bytes = offset * step;

		for (frame = 0; frame < frames; frame++) {
			for (channel = 0; channel < p->channels; channel++) {
				char *ptr = chan_areas[channel].addr;
				memcpy(ptr + chan_areas[channel].first / 8 +
				       offset_bytes,
				       data + channel_offset, sample_size);
				channel_offset += sample_size;
			}
			offset_bytes += step;
		}
			
		err = snd_pcm_mmap_commit(p->alsa_pcm, offset, frames);
		if (err == -EPIPE)
			xrun_recover(p);
		else if (err < 0)
			g_warning("snd_pcm_mmap_commit() failed: %s", snd_strerror(err));
		else if (err != frames)
			g_warning("snd_pcm_mmap_commit returned %d, expected %d", err, (int)frames);	
		p->alsa_total_written += cnt;
		length -= cnt;
	}
}

static void alsa_write_audio(HWMIX_WAVE *p,gpointer data, int length)
{
	snd_pcm_sframes_t written_frames;

	while (length > 0) {
		int frames = snd_pcm_bytes_to_frames(p->alsa_pcm, length);
		written_frames = snd_pcm_writei(p->alsa_pcm, data, frames);	
		if (written_frames > 0) {
			int written = snd_pcm_frames_to_bytes(p->alsa_pcm,
							      written_frames);
			p->alsa_total_written += written;
			length -= written;
			data = (char*) data + written;
		}
		else if (written_frames == -EPIPE)
			xrun_recover(p);
		else {
			g_warning("write error: %s",snd_strerror(-written_frames));
			break;
		}
	}
}

void alsa_write(HWMIX_WAVE *p,gpointer data, int length)
{
	if (p->paused)
		return;
	p->force_start = FALSE;
	if (p->mmap)
		alsa_mmap_audio(p,data, length);
	else
		alsa_write_audio(p,data, length);
}

snd_pcm_format_t   wav_format_to_alsa_format(WaveFile *wf)
{
	switch (wf->bits_per_sample) {
	case 8:
	    return(SND_PCM_FORMAT_U8);
	case 16:
            return(SND_PCM_FORMAT_S16_LE);
	default:
	    return(SND_PCM_FORMAT_UNKNOWN);
	};		
}

int alsa_open(HWMIX_WAVE *p)
{
	
	p->mmap = 0;
	p->format=wav_format_to_alsa_format(p->wf);
	if ( p->format == SND_PCM_FORMAT_UNKNOWN )
	    return 0;
        p->rate=p->wf->samples_per_sec;
	p->channels=p->wf->channels;
	if (alsa_setup(p) < 0) {
		alsa_close(p);
		return 0;
	}
	p->alsa_total_written = 0;
	p->going = TRUE;
	p->paused = FALSE;
	p->force_start = FALSE;
	snd_pcm_prepare(p->alsa_pcm);
	return 1;
}

static void pause_play(HWMIX_WAVE *p,int pause_resume)
{
        if ( p->wf->going ) {
	    alsa_pause(p,pause_resume);
        };
}

static void stop_play(HWMIX_WAVE *p)
{
	if (p->wf && p->wf->going) {
		p->wf->going = 0;
		pthread_join(p->play_thread, NULL);
		alsa_close(p);
		g_free(p->wf);
		p->wf = NULL;
	}
}

void hwmix_vc_enable(int idx,int flag)
{
	int i;
	for (i=0; i<4; i++) {
	        gtk_widget_set_sensitive(hwmix_vc[idx].vc_hscale[i],flag);
	};
}

static void notify_end_of_play(HWMIX_WAVE *p)
{
        int i,j;
	switch (p->button_type) {
	case 1:
              gtk_widget_set_sensitive(hwmix_play_button[p->idx],TRUE);	
              gtk_widget_set_sensitive(hwmix_pause_button[p->idx],FALSE);
              gtk_widget_set_sensitive(hwmix_stop_button[p->idx],FALSE);
              gtk_widget_set_sensitive(hwmix_del_button[p->idx],TRUE);	
	      hwmix_vc_enable(p->idx,FALSE);
	      break;
	case 2:
              i=p->idx / 2;
              j=p->idx % 2;
	      gtk_widget_set_sensitive(speaker_button[i][j],TRUE);
	      break;
	case 3:
              gtk_widget_set_sensitive(v3d_play_button[p->idx],TRUE);	
              gtk_widget_set_sensitive(v3d_stop_button[p->idx],FALSE);	
	      break;
        };
}

static void *play_loop(void *arg)
{
	char *data;
	int bytes, blk_size, rate;
	int actual_read;
        int err;
	HWMIX_WAVE *p;
	WaveFile *wf;
	p=arg;
	wf=p->wf;       

	blk_size = 512 * (wf->bits_per_sample / 8) * wf->channels;
	rate = wf->samples_per_sec * wf->channels * (wf->bits_per_sample / 8);
	data = g_malloc(blk_size);
	while (wf->going) {
	    if (!wf->eof) {
		bytes = blk_size;
		if (wf->length - wf->position < bytes)
		    bytes = wf->length - wf->position;
		if (bytes > 0) {
		    actual_read = fread(data, 1, bytes, wf->file);
                  
		    if (actual_read == 0) {
		        wf->eof = TRUE;
//                        g_print("actual read=0\n");
			alsa_free(p);    
//                        g_print("alsa_free\n");
                        notify_end_of_play(p);
			if ( p->alsa_pcm != NULL ) {
		            if ((err = snd_pcm_close(p->alsa_pcm)) < 0)
			        g_warning("snd_pcm_close() failed: %s",snd_strerror(err));
		            p->alsa_pcm = NULL; 
                        };
//                        g_print("snd_pcm_close %s 1\n",p->filename);
                        fclose(wf->file);
                        p->wf->file=NULL;
			g_free(data);
                        pthread_detach(p->play_thread);
 		    }
		    else {
                        if ( p->button_type == 1 ) {
				memcpy(&scope_data[p->idx][0],data,actual_read);
				redraw_scope(p->idx);
			};
			while(alsa_free(p) < bytes && wf->going && wf->seek_to == -1)
			    hwmix_usleep(10000);
			if(wf->going && wf->seek_to == -1)
			    alsa_write_audio(p,data, bytes);
			wf->position += actual_read;
		    }
		}
		else {
		    wf->eof = TRUE;
//		    g_print("EOF %s\n",p->filename);
		    alsa_free(p);
                    if ( p->loop ) {
                        wf->eof=FALSE;
			wf->position = 0;
		        fseek(wf->file, wf->position + wf->data_offset, SEEK_SET);
                    }
                    else {
  		        hwmix_usleep(10000);
                        if ( p->button_type == 1 ) {
				memset(&scope_data[p->idx][0],0,4096);
				redraw_scope(p->idx);
			};
                        notify_end_of_play(p);
		        if ( p->alsa_pcm != NULL ) {
		            if ((err = snd_pcm_close(p->alsa_pcm)) < 0)
			        g_warning("snd_pcm_close() failed: %s",snd_strerror(err));
		            p->alsa_pcm = NULL;
                        };
//                        g_print("snd_pcm_close %s 2\n",p->filename);
                        g_free(data);
 	                fclose(wf->file);
                        p->wf->file=NULL;
                        pthread_detach(p->play_thread);
                    };
 		}
            }
	    else
		hwmix_usleep(10000);
	    if (wf->seek_to != -1) {
		wf->position = wf->seek_to * rate;
		fseek(wf->file, wf->position + wf->data_offset, SEEK_SET);
		alsa_flush(p,wf->seek_to * 1000);
		wf->seek_to = -1;
	    };
	};
        g_print("stop %s\n",p->filename);
	g_free(data);
	fclose(wf->file);
	pthread_exit(NULL);
}

static int is_wav_file(WaveFile *wf,char *filename)
{
	char magic[4];
	unsigned long len;
	char *ext;

    ext = strrchr(filename, '.');
    if (ext) {
	if (!strcasecmp(ext, ".wav")) {
	    if (( wf->file = fopen(filename, "rb"))) {
		fread(magic, 1, 4, wf->file);
		if (strncmp(magic, "RIFF", 4)==0) {
 		    read_le_long(wf->file, &len);
		    fread(magic, 1, 4, wf->file);
		    if (strncmp(magic, "WAVE", 4)==0) {
 		        for (;;) {
			    fread(magic, 1, 4, wf->file);
			    if (!read_le_long(wf->file, &len)) {
			   	    fclose(wf->file);
				    return FALSE;
			    }
			    if (!strncmp("fmt ", magic, 4))
				    break;
			    fseek(wf->file, len, SEEK_CUR);
		        }
		        if (len >= 16) {
		    	    read_le_short(wf->file, &wf->format_tag);
                            if ( wf->format_tag == WAVE_FORMAT_PCM ) {
		        	read_le_short(wf->file, &wf->channels);
		        	read_le_long(wf->file, &wf->samples_per_sec);
		        	read_le_long(wf->file, &wf->avg_bytes_per_sec);
		        	read_le_short(wf->file, &wf->block_align);
		        	read_le_short(wf->file, &wf->bits_per_sample);
	            	        fclose(wf->file);
		                return TRUE;
                            };
		        };
                    };
                };
		fclose(wf->file);
            };
	};
    };	
    return FALSE;
}

static void play_wav_file(HWMIX_WAVE *p)
{
	char magic[4];
	unsigned long len;
	int rate;
	WaveFile *wf;
	p->audio_error = FALSE;
	p->wf = g_malloc0(sizeof (WaveFile));
	if ((p->wf->file = fopen(p->filename, "rb"))) {
	        wf=p->wf;
		fread(magic, 1, 4, wf->file);
		if (strncmp(magic, "RIFF", 4)) {
			fclose(wf->file);
			g_free(p->wf);
			p->wf = NULL;
			return;
		}
		read_le_long(wf->file, &len);
		fread(magic, 1, 4, wf->file);
		if (strncmp(magic, "WAVE", 4)) {
			fclose(wf->file);
			g_free(p->wf);
			p->wf = NULL;
			return;
		}
		for (;;) {
			fread(magic, 1, 4, wf->file);
			if (!read_le_long(wf->file, &len)) {
				fclose(wf->file);
				g_free(p->wf);
				p->wf = NULL;
				return;
			}
			if (!strncmp("fmt ", magic, 4))
				break;
			fseek(wf->file, len, SEEK_CUR);
		}
		if (len < 16) {
			fclose(wf->file);
			g_free(p->wf);
			p->wf = NULL;
			return;
		}
		read_le_short(wf->file, &wf->format_tag);
                if ( wf->format_tag != WAVE_FORMAT_PCM ) {
		 	fclose(wf->file);
			g_free(p->wf);
			p->wf = NULL;
			return;
		}
		read_le_short(wf->file, &wf->channels);
		read_le_long(wf->file, &wf->samples_per_sec);
		read_le_long(wf->file, &wf->avg_bytes_per_sec);
		read_le_short(wf->file, &wf->block_align);
		read_le_short(wf->file, &wf->bits_per_sample);
		if (wf->bits_per_sample != 8 && wf->bits_per_sample != 16) {
			fclose(wf->file);
			g_free(p->wf);
			p->wf = NULL;
			return;
		}
		len -= 16;
		if (len)
			fseek(wf->file, len, SEEK_CUR);

		for (;;) {
			fread(magic, 4, 1, wf->file);

			if (!read_le_long(wf->file, &len))
			{
				fclose(wf->file);
				g_free(p->wf);
				p->wf = NULL;
				return;
			}
			if (!strncmp("data", magic, 4))
				break;
			fseek(wf->file, len, SEEK_CUR);
		}
		wf->data_offset = ftell(wf->file);
		wf->length = len;

		wf->position = 0;
		wf->going = 1;

 		if (alsa_open(p) == 0) {
			p->audio_error = TRUE;
			fclose(wf->file);
			g_free(p->wf);
			p->wf = NULL;
			return;
		}
		rate = wf->samples_per_sec * wf->channels * (wf->bits_per_sample / 8);

		wf->seek_to = -1;
/*
                g_print("Play %s  channels=%d  sample rate=%ld  byte rate=%ld  block align=%d   bits/sample=%d\n",
			p->filename,p->wf->channels,p->wf->samples_per_sec,
			p->wf->avg_bytes_per_sec,p->wf->block_align,p->wf->bits_per_sample);

*/
	        pthread_create(&p->play_thread, NULL, play_loop, p);
	}
}

/*
static int get_time(HWMIX_WAVE *p)
{
	if (p->audio_error)
		return -2;
	if (!p->wf)
		return -1;
	if (!p->wf->going || (p->wf->eof && !alsa_playing(p)))
		return -1;
	else 
		return alsa_get_output_time(p);
}

static void seek(HWMIX_WAVE *p,int time)
{
	p->wf->seek_to = time;

	p->wf->eof = FALSE;

	while (p->wf->seek_to != -1)
		hwmix_usleep(10000);
}
*/
static void capture_scope_render_pcm (void) {
  	int i,j;
  	gint32 ry, ly, try = 0, tly = 0;
  	gint16 *pointer16;
	unsigned char  *pointer8;
	int    width,height,h2;
	gint32    hh;


	GDK_THREADS_ENTER();

        h2=capture_scope_drawing->allocation.height/2;
        height=h2+h2;

        width=capture_scope_drawing->allocation.width;

        if ( capturing ) {
	    if ( capture_format == SND_PCM_FORMAT_S16_LE ) {
		pointer16=(gint16 *) &capture_data[0];
	        hh=capture_scale*65536/height;
	        if ( hh * height < (capture_scale*65536) )
		    hh++;
            }
	    else {
		pointer8=&capture_data[0];
		hh=capture_scale*256/height;
		if ( hh * height < (capture_scale*256) )
		    hh++;
	    };

 
  	    for ( i = 0; i < width; i++ ) {
	        ly = 0;
  	        if ( capture_channels == 1 ) {
		    if ( capture_format == SND_PCM_FORMAT_S16_LE ) {
		        for (j=0; j<capture_scale; j++) {
		            ly += *pointer16; 
		            pointer16++;
	                };
	            }
	            else {
		        for (j=0; j<capture_scale; j++) {
			    ly += *pointer8;
			    pointer8++;
		        };
		    };
  		    ly /= hh;
	        }
		else {
	            ry = 0;
		    if ( capture_format == SND_PCM_FORMAT_S16_LE ) {
	                for (j=0; j<capture_scale; j++) {
		            ly += *pointer16; 
		            pointer16++;
	      	            ry += *pointer16; 
		            pointer16++;
		        };
		    }
		    else {
	                for (j=0; j<capture_scale; j++) {
		            ly += *pointer8; 
		            pointer8++;
	      	            ry += *pointer8; 
		            pointer8++;
		        };
		    };		
   		    ly /= hh;
                    ry /= hh;
	        };
	        if ( capture_scope_type == 3 ) {
		    ly /= 2;
		    gdk_draw_line(capture_scope_pixmap, penWhite, i, h2-1-ly, i, h2+ly);
		    if ( capture_channels == 2) {
		        ry /= 2;
		        gdk_draw_line(capture_scope_pixmap, penRedLight, i, h2-1-ry, i, h2+ry);
	            };
		}
		else {
	  	    ly += h2;
    		    ry += h2;
                    switch (capture_scope_type) {
	            case 0:			// dot
	                gdk_draw_point(capture_scope_pixmap, penWhite, i, ly); 
		        if ( capture_channels == 2 ) 
		            gdk_draw_point(capture_scope_pixmap, penRedLight, i, ry);
		        break;
	            case 1:			// line
		        if (i == 0) { 
		            gdk_draw_point(capture_scope_pixmap, penWhite, 0, ly);
		            tly = ly;
		            if ( capture_channels == 2) {
			        gdk_draw_point(capture_scope_pixmap, penRedLight, 0, ry);
			        try = ry;		
		            };
		        }
		        else {
	                    gdk_draw_line(capture_scope_pixmap, penWhite, i - 1, tly, i,    ly); 
	  	            tly = ly;
		            if ( capture_channels == 2) {
		                gdk_draw_line(capture_scope_pixmap, penRedLight, i - 1, try, i, ry);  
		                try = ry;
                            };
      		        };
		        break;
	            case 2:			// filled
		        if (ly != h2) 
		            gdk_draw_line(capture_scope_pixmap, penWhite, i, ((ly > 0) ? h2 : h2-1), i, ly);		  
		        if (capture_channels == 2 ) {
		            if (ry != h2) 
		                gdk_draw_line(capture_scope_pixmap, penRedLight, i, ((ry > 0) ? h2 : h2-1), i, ry);  
		        };
		        break;
                    };
	        };
	    };
 	};
        gdk_window_clear(capture_scope_drawing->window);

  	GDK_THREADS_LEAVE();
}

void redraw_capture_scope(void)
{
       if ( GTK_WIDGET_VISIBLE(capture_scope_drawing) ) {
		    capture_scope_render_pcm();
		    gdk_draw_rectangle(capture_scope_pixmap,
			   capture_scope_drawing->style->black_gc,
			   TRUE,
			   0, 0,
			   capture_scope_drawing->allocation.width,
			   capture_scope_drawing->allocation.height);
	};	
}

gint capture_scope_drawing_configure_event(GtkWidget *widget, GdkEventConfigure *event)
{
	if (capture_scope_pixmap != NULL)
		gdk_pixmap_unref(capture_scope_pixmap);
 	capture_scope_pixmap = gdk_pixmap_new(window->window, 
                          widget->allocation.width, widget->allocation.height,
			  gdk_visual_get_best_depth());

/* PEN */

  	gdk_draw_rectangle(capture_scope_pixmap, penBlack, TRUE, 0, 0, 
			  widget->allocation.width, widget->allocation.height);
  	gdk_window_set_back_pixmap(capture_scope_drawing->window, capture_scope_pixmap, 0);
	redraw_capture_scope(); 
	return TRUE;
}

gint capture_scope_drawing_expose_event(GtkWidget *widget, GdkEventExpose *event)
{
	gdk_window_clear_area (widget->window,
                         event->area.x, event->area.y,
                         event->area.width, event->area.height);
  	gdk_gc_set_clip_rectangle (widget->style->fg_gc[widget->state],&event->area);

	redraw_capture_scope();
  	gdk_gc_set_clip_rectangle (widget->style->fg_gc[widget->state],NULL);
	return TRUE;
}

void capture_scope_type_changed_cb (GtkWidget *widget, gpointer data)
{
	char select_item[32];
	int idx;
	strcpy(select_item,gtk_entry_get_text (GTK_ENTRY(widget)));
	for (idx=0; idx<4; idx++) {
            if ( strcmp(scope_type_name[idx],select_item) == 0 ) {
		capture_scope_type=idx;
	        redraw_capture_scope();
		return;
            };
	};
}

// FIXME - Capture loop cannot capture sound

static void *capture_loop(void *arg)
{
	unsigned char *data;
	int  total_read,readed;
	snd_pcm_sframes_t read_frames;

        total_read=0;
        capturing=TRUE;	
	data=&capture_data[0];
	while (capturing) {
		do {
		    read_frames = snd_pcm_readi(capture_pcm, data, capture_period_size);
		} while ( read_frames == -EAGAIN );
	        g_print("read frames=%d  total read=%d\n",snd_pcm_frames_to_bytes(capture_pcm,read_frames),total_read);
		if (read_frames > 0) {
			readed = snd_pcm_frames_to_bytes(capture_pcm,read_frames);
			total_read += readed;
			data=(char *)data+readed;
                        if ( total_read  >= 4096 ) {
			    redraw_capture_scope();
			    data=&capture_data[0];			
			};
			snd_pcm_wait(capture_pcm,1000);
		}
		else {
                        if ( read_frames != -EPIPE ) {
                   	    g_warning("read error: %s",snd_strerror(-read_frames));
			    break;
                        };
		}
        };
        pthread_exit(NULL);
}

// FIXME - This function is intended to capture the output from ac97 mixer 
//         and display on a scope in order to study the effect of HRTF, ITD, ILD.

int start_capture(snd_pcm_format_t format,int channels,int rate)
{
	snd_pcm_hw_params_t *hwparams;
	snd_pcm_sw_params_t *swparams;
	snd_pcm_uframes_t boundary;
        int dir;
	int err;
	char capture_device[16];

	capture_format=format;
	capture_channels=channels;
	capture_rate=rate;
	sprintf(capture_device,"hw:%d,0",alsa_card_number);
	g_print("Capture device : %s\n",capture_device);
  	if ((err = snd_pcm_open(&capture_pcm, capture_device, SND_PCM_STREAM_CAPTURE, SND_PCM_NONBLOCK)) < 0) {
    		g_print("snd_pcm_open %s failed: %s\n", capture_device,snd_strerror(err));
		return err;
  	}
	snd_pcm_hw_params_malloc(&hwparams);
	if ((err = snd_pcm_hw_params_any(capture_pcm, hwparams) < 0 )) {
    		g_print("Broken configuration for capture: no configurations available: %s\n", snd_strerror(err));
    		return err;
  	}
  /* set the interleaved read/write format */
  	if ((err = snd_pcm_hw_params_set_access(capture_pcm, hwparams, SND_PCM_ACCESS_RW_INTERLEAVED)) < 0 ) {
		g_print("Access type not available for capture: %s\n", snd_strerror(err));
    		return err;
  	}
  /* set the sample format */
  	if (( err = snd_pcm_hw_params_set_format(capture_pcm, hwparams, capture_format)) < 0 ) {
		g_print("Sample format not available for capture: %s\n", snd_strerror(err));
		return err;
  	}
  /* set the count of channels */
  	if (( err = snd_pcm_hw_params_set_channels(capture_pcm, hwparams, capture_channels)) < 0 ) {
		g_print("Channels count (%i) not available for capture: %s\n", channels, snd_strerror(err));
		return err;
	}
	dir=1;
  	if ((err = snd_pcm_hw_params_set_rate_near(capture_pcm, hwparams, &capture_rate, &dir)) < 0 ) {
		g_print("Rate %iHz not available for capture: %s\n", rate, snd_strerror(err));
		return err;
	}
	if ( rate != capture_rate ) {
		g_print("set rate to %d instead of %d\n",capture_rate,rate);
	};
        if (( err =snd_pcm_hw_params_get_period_size_max(hwparams, &capture_period_size, &dir)) < 0 ) {
		g_print("Unable to get period size max for capture: %s\n", snd_strerror(err));
		return err;
	};
	capture_period_size=capture_period_size/16;
	capture_buffer_size=4*capture_period_size;
/*
	if ((err = snd_pcm_hw_params_set_period_size(capture_pcm, hwparams, period_size, dir)) < 0 ) {
		g_print("Unable to set period size %lu for capture: %s\n", period_size, snd_strerror(err));
		return err;
	}
	if ((err = snd_pcm_hw_params_set_buffer_size(capture_pcm, hwparams, buffer_size)) < 0 ) {
    		g_print("Unable to set buffer size %lu for capture: %s\n", buffer_size, snd_strerror(err));
		return err;
	}
*/
  /* write the parameters to device */
  	if (( err = snd_pcm_hw_params(capture_pcm, hwparams)) < 0 ) {
		g_print("Unable to set hw params for capture: %s\n", snd_strerror(err));
		return err;
	}
	snd_pcm_hw_params_free (hwparams);
  	snd_pcm_sw_params_malloc(&swparams);
	if ((err = snd_pcm_sw_params_malloc (&swparams)) < 0) {
		g_print ("cannot allocate software parameter structure (%s)\n",
                snd_strerror (err));
		return err;
	}
	if ((err = snd_pcm_sw_params_current (capture_pcm, swparams)) < 0) {
		g_print("cannot initialize software parameter structure (%s)\n",
			 snd_strerror (err));
		return err;
	}
	if ((err = snd_pcm_sw_params_get_boundary (swparams, &boundary)) < 0) {
		g_print ("cannot set access type (%s)\n",
			 snd_strerror (err));
		return err;
	}
	if ((err = snd_pcm_sw_params_set_start_threshold (capture_pcm, swparams, 128)) < 0) {
		g_print ( "cannot set access type (%s)\n",
			 snd_strerror (err));
		return err;
	}
	if ((err = snd_pcm_sw_params (capture_pcm, swparams)) < 0) {
		g_print ("cannot set sw-parameters (%s)\n",
			 snd_strerror (err));
		return err;
	} 
	snd_pcm_sw_params_free (swparams);
        if (( err = snd_pcm_prepare(capture_pcm)) < 0 ) {
		g_print("snd_pcm_prepare() failed: %s\n", snd_strerror(err));
		return err;
	}
        pthread_create(&capture_thread, NULL, capture_loop, NULL);
	return 0;
}

void stop_capture(void)
{
        int err;
        capturing=FALSE;
        pthread_join(capture_thread,NULL);
     	err=snd_pcm_close(capture_pcm);
	if ( err < 0 ) {
		g_print("snd pcm close failed: %s\n",snd_strerror(err));
	};
}

/***************************************************************************************/

void speaker_button_clicked_cb (GtkWidget *widget, gpointer data)
{
  	int     chn;
	int 	save_ac97_3d[2];
	int     save_ac97_3d_switch;
	int	save_sdac_master_switch;
	int     save_sdac_mixer_switch;
	int     save_stac9708_pcmoutpath;
	int     save_stac9708_4speakerstereo;
	int     i,j;
	char 	filename[FILENAME_MAX+1];
        WaveFile w;
   
	for (j=0; j<3; j++) {
	    for (i=0; i<2; i++) {
		if ( widget == speaker_button[j][i] ) {
//                    speaker_test.idx=j+j+i;
		    if ( gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(speaker_set_radio[0]))) {
			if ( j == 1 )
			    chn=i;
			else
			    return;
		    };
		    if ( gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(speaker_set_radio[1]))) {
			if  ( j == 0 )
			    chn=i;
			else
			    return;
		    };
		    if ( ac97_has_sdac ) {
			    if ( gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(speaker_set_radio[2]))) {
			        if ( j == 1 )
				    return; 
				else {
				    if ( j == 0 )
				        chn=i;
				    else
				        chn=i+2;
				};
			   };
		    };
                    speaker_wave[chn].idx=j+j+i;
		};
	    };
	};
	channels=2;
	if ( ac97_has_sdac ) {
		if ( gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(speaker_set_radio[2]))) {
			channels=4;
		};
	};

	if ( channels == 4 ) {

//      Sigmatel STAC9708 QUAD SPEAKER

//	save original value for restoring.

	    save_sdac_master_switch=elem_value_get_boolean(mix_mute_elem[sdac_master_idx]);
	    
            if ( sdac_mixer_idx > 0 )
	        save_sdac_mixer_switch=elem_value_get_boolean(mix_mute_elem[sdac_mixer_idx]);

	    save_stac9708_4speakerstereo=elem_value_get_integer(mix_enable_elem[stac9708_4speakerstereo_idx]);
	    save_stac9708_pcmoutpath=elem_value_get_integer(mix_enum_elem[stac9708_pcmoutpath_idx]);
	    save_ac97_3d_switch=elem_value_get_boolean(mix_enable_elem[ac97_3d_enable]);
	    for (i=0; i<no_of_ac97_3d; i++) 
		save_ac97_3d[i]=elem_value_get_integer(mix_3d_elem[i]);

//      Turn off 4-Speaker stereo

            gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mix_enable_button[stac9708_4speakerstereo_idx]),FALSE);		

//      Set PCM OUT Path to Post 3D

            gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mix_enum_radio[stac9708_pcmoutpath_idx+1]),TRUE);		

//    UnMute SDAC Master 

	    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mix_mute_button[sdac_master_idx]),TRUE);

//    Mute SDAC Mixer (Must be Muted to obtain 4 distinct channels )

            if ( sdac_mixer_idx > 0 )
	        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mix_mute_button[sdac_mixer_idx]),TRUE);

//    Turn off AC97 3D Control

            gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mix_enable_button[ac97_3d_enable]),FALSE);		
	    for (i=0; i<no_of_ac97_3d; i++ ) {
		gtk_adjustment_set_value(GTK_ADJUSTMENT(ac97_3d_adj[i]),0);
	    };
	};
	
	if ( channels == 4 ) 
            strcpy(filename,speaker_quad_wave[chn]);
	else
	    strcpy(filename,speaker_stereo_wave[chn]);

	if ( is_wav_file(&w,filename) ) {
	    gtk_widget_set_sensitive(widget,FALSE);
            strcpy(speaker_wave[chn].filename,filename);
            speaker_wave[chn].button_type=2;
	    speaker_wave[chn].v3d=FALSE;
            speaker_wave[chn].loop=FALSE;
 	    play_wav_file(&speaker_wave[chn]);
        }
	else {
 	    play_sine_wave(chn);	   
	};
	
	if ( channels == 4 ) {

//     Restore the previously saved value 

            gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mix_enable_button[stac9708_4speakerstereo_idx]),save_stac9708_4speakerstereo);		

            gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mix_enum_radio[stac9708_pcmoutpath_idx+save_stac9708_pcmoutpath]),TRUE);		

	    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mix_mute_button[sdac_master_idx]),save_sdac_master_switch);

            if ( sdac_mixer_idx > 0 )
	        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mix_mute_button[sdac_mixer_idx]),save_sdac_mixer_switch);

            gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mix_enable_button[ac97_3d_enable]),save_ac97_3d_switch);		
	    for (i=0; i<no_of_ac97_3d; i++ ) {
		gtk_adjustment_set_value(GTK_ADJUSTMENT(ac97_3d_adj[i]),save_ac97_3d[i]);
	    };
	};      
}

void speaker_set_toggle_button_cb (GtkWidget *widget, gpointer data)
{
	int idx;
	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(speaker_set_radio[0]))) {
	    for (idx=0; idx<2; idx++) {
		gtk_button_set_label(GTK_BUTTON(speaker_button[1][idx]),speaker_stereo_name[idx]);
		gtk_button_set_label(GTK_BUTTON(speaker_button[0][idx]),"  ");
		gtk_button_set_label(GTK_BUTTON(speaker_button[2][idx]),"  ");
                gtk_widget_set_sensitive(speaker_button[1][idx],TRUE);
                gtk_widget_set_sensitive(speaker_button[0][idx],FALSE);
                gtk_widget_set_sensitive(speaker_button[2][idx],FALSE);

	    };	  
	};	
	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(speaker_set_radio[1]))) {
	    for (idx=0; idx<2; idx++) {
	         gtk_button_set_label(GTK_BUTTON(speaker_button[0][idx]),speaker_quad_name[idx]);
	         gtk_button_set_label(GTK_BUTTON(speaker_button[1][idx])," ");
                 gtk_widget_set_sensitive(speaker_button[0][idx],TRUE);
                 gtk_widget_set_sensitive(speaker_button[1][idx],FALSE);
		 if ( ac97_has_sdac ) {
		     if ( elem_value_get_boolean(mix_enable_elem[stac9708_4speakerstereo_idx]) ) {
			 gtk_button_set_label(GTK_BUTTON(speaker_button[2][idx]),"Stereo copy");
	                 gtk_widget_set_sensitive(speaker_button[2][idx],TRUE);
		     }
		     else {
			 gtk_button_set_label(GTK_BUTTON(speaker_button[2][idx])," ");
	                 gtk_widget_set_sensitive(speaker_button[2][idx],FALSE);
		     };
		 }	
		 else {
		     gtk_button_set_label(GTK_BUTTON(speaker_button[2][idx])," ");
                     gtk_widget_set_sensitive(speaker_button[2][idx],FALSE);
                 };
	    };	  
	};	
	if ( ac97_has_sdac ) {
	    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(speaker_set_radio[2]))) {
		for (idx=0; idx<2; idx++) {
		    gtk_button_set_label(GTK_BUTTON(speaker_button[0][idx]),speaker_quad_name[idx]);
		    gtk_button_set_label(GTK_BUTTON(speaker_button[2][idx]),speaker_quad_name[idx+2]);
		    gtk_button_set_label(GTK_BUTTON(speaker_button[1][idx])," ");
                    gtk_widget_set_sensitive(speaker_button[0][idx],TRUE);
                    gtk_widget_set_sensitive(speaker_button[2][idx],TRUE);
                    gtk_widget_set_sensitive(speaker_button[1][idx],FALSE);
		};	  
	     };
	};
}

void set_speaker_table_row(GtkWidget *table,int row,char *s0,char *s1,char *s2, char *s3, char *s4, char *s5, char *s6)
{
     	GtkWidget *label;
    	
	label=gtk_label_new(s0);
    	gtk_widget_show(label);
    	gtk_table_attach_defaults(GTK_TABLE(table),label,0,1,row,row+1);
	label=gtk_label_new(s1);
    	gtk_widget_show(label);
    	gtk_table_attach_defaults(GTK_TABLE(table),label,1,2,row,row+1);
	label=gtk_label_new(s2);
    	gtk_widget_show(label);
    	gtk_table_attach_defaults(GTK_TABLE(table),label,2,3,row,row+1);
	label=gtk_label_new(s3);
    	gtk_widget_show(label);
    	gtk_table_attach_defaults(GTK_TABLE(table),label,3,4,row,row+1);
	label=gtk_label_new(s4);
    	gtk_widget_show(label);
    	gtk_table_attach_defaults(GTK_TABLE(table),label,4,5,row,row+1);
	label=gtk_label_new(s5);
    	gtk_widget_show(label);
    	gtk_table_attach_defaults(GTK_TABLE(table),label,5,6,row,row+1);
        if ( sdac_mixer_idx > 0 ) {
	    label=gtk_label_new(s6);
    	    gtk_widget_show(label);
    	    gtk_table_attach_defaults(GTK_TABLE(table),label,6,7,row,row+1);
        };
}

static void create_speaker(GtkWidget *main, GtkWidget *notebook, int page)
{
	GtkWidget *label;
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkWidget *frame;
	GtkWidget *table;
        int i,j;

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_widget_show(hbox);
	gtk_container_add(GTK_CONTAINER(notebook), hbox);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 6);

        label = gtk_label_new("Speaker");
        gtk_widget_show(label);
	gtk_notebook_set_tab_label(GTK_NOTEBOOK(notebook), 
				   gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), page), 
				   label);

	if ( ac97_has_sdac )
		no_of_speaker_set=3;
	else
		no_of_speaker_set=2;

        frame= gtk_frame_new("Speaker setup");
        gtk_widget_show(frame);
        gtk_container_add(GTK_CONTAINER(hbox), frame);

	vbox = gtk_vbox_new(FALSE, 0);
	gtk_widget_show(vbox);
	gtk_container_add(GTK_CONTAINER(frame), vbox);

	for (i=0; i<no_of_speaker_set; i++) {
            if ( i == 0 ) {
                speaker_set_radio[0] = gtk_radio_button_new_with_label (NULL, speaker_set_name[0]);
	    }
            else {
                speaker_set_radio[i] = gtk_radio_button_new_with_label(gtk_radio_button_group (GTK_RADIO_BUTTON (speaker_set_radio[0])), speaker_set_name[i]);
	    };
	    gtk_widget_show(speaker_set_radio[i]);
            gtk_box_pack_start(GTK_BOX(vbox), speaker_set_radio[i], TRUE, TRUE, 2);
	    gtk_signal_connect (GTK_OBJECT (speaker_set_radio[i]), 							"toggled",
			                        GTK_SIGNAL_FUNC (speaker_set_toggle_button_cb),
						NULL);
	};

	table=gtk_table_new(3,3,TRUE);
        gtk_widget_show(table);
	gtk_container_add(GTK_CONTAINER(vbox),table);
        for (j=0; j<3; j++) {
	    for (i=0; i<2; i++) {
		if ( j == 1 ) 
		    speaker_button[j][i]=gtk_button_new_with_label(speaker_stereo_name[i]);
		else
		    speaker_button[j][i]=gtk_button_new_with_label(" ");
		gtk_widget_show(speaker_button[j][i]);
		gtk_table_attach_defaults(GTK_TABLE(table),speaker_button[j][i],(i*2),(i*2)+1,j,j+1);
	        gtk_signal_connect (GTK_OBJECT (speaker_button[j][i]), "clicked",
			    GTK_SIGNAL_FUNC (speaker_button_clicked_cb),NULL);
	    };
	};
	if ( card_is_au8830 ) {
	    label = gtk_label_new("Please put out1lq.wav, out1rq.wav, out2lq.wav and out2rq.wav in the same directory to perform quad speaker test.");
            gtk_widget_show(label);
	    gtk_container_add(GTK_CONTAINER(vbox),label);
	};
/*
	if ( ac97_has_sdac ) {
		
		table=gtk_table_new(4,7,FALSE);
                gtk_widget_show(table);
	        gtk_container_add(GTK_CONTAINER(vbox),table);

                set_speaker_table_row(table,0," ",
                  "4-speaker\nstereo","AC97 3D","PCM\nOut Path","PCM",sdac_master,sdac_mixer);
		set_speaker_table_row(table,1,"2 channels\nto\n4 speakers",
                          "On"," "," ","Front & Rear"," "," ");
		set_speaker_table_row(table,2,"AC97 3D",
                          "Off","On","Pre 3D","Front","Rear","Mute");
		set_speaker_table_row(table,3,"Quad\nspeakers",
                          "Off"," ","Post 3D","Front","Rear","Mute");                 
	};
*/
}

/******************************************************************************************/
void free_resource(HWMIX_WAVE *p)
{
	int err;
	if ( p->alsa_pcm != NULL ) {
		if (p->going ) {
			if ((err = snd_pcm_drop(p->alsa_pcm)) < 0)
				g_warning("snd_pcm_drop() failed: %s",snd_strerror(err));
                };
		if (( err = snd_pcm_hw_free(p->alsa_pcm)) < 0 ) 
			g_warning("snd_pcm_hw_free() failed: %s",snd_strerror(err));
		if ((err = snd_pcm_close(p->alsa_pcm)) < 0)
			g_warning("snd_pcm_close() failed: %s",snd_strerror(err));
		p->alsa_pcm = NULL;
	};
}

void hwmix_vc_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
    int err,i,j,*p;
    int idx;
    snd_hctl_elem_t *elem;
    int m,count,type;
    char idname[32];

    p=data;
    idx=*p;    
    i = idx / 4;
    j = idx % 4;

    snd_ctl_elem_info_alloca(&info);
    snd_ctl_elem_value_alloca(&val);
    snd_ctl_elem_id_alloca(&id);

    m=0;
    for (elem = snd_hctl_first_elem(hctl); elem; elem = snd_hctl_elem_next(elem)) {
        if ( (err= snd_hctl_elem_info(elem, info)) < 0) {
	    g_print("snd_hctl_elem_info : %s\n",snd_strerror(err));
	}
	else {
            count =snd_ctl_elem_info_get_count(info);
	    type = snd_ctl_elem_info_get_type(info);
	    if ((err = snd_hctl_elem_read(elem, val)) < 0) {             
	        g_print("snd_hctl_elem_read : %s\n",snd_strerror(err));
	    }
	    else {
	        snd_hctl_elem_get_id(elem, id);
	        strcpy(idname,snd_ctl_elem_id_get_name(id));
                if ( strstr(idname,"HWMIX") > 0 ) {
	            switch (type) {
	            case SND_CTL_ELEM_TYPE_INTEGER:
			if ( i == m ) {	
    			    if ( (err= snd_hctl_elem_info(elem, info)) < 0) {
	                        g_print("snd_hctl_elem_info : %s\n",snd_strerror(err));
                            }
                            else {
	                        if ((err = snd_hctl_elem_read(elem,val)) < 0) {
	                            g_print("snd_hctl_elem_read : %s\n",snd_strerror(err));
	                        }
	                        else {
	                            if ( snd_ctl_elem_info_is_writable(info) ) {
	                                snd_ctl_elem_value_set_integer(val,j,adj->value);
	                                if ( (err = snd_hctl_elem_write(elem,val)) < 0 ) {
		                            g_print("snd_hctl_elem_write : %s\n",snd_strerror(err));
		                        };
	                            };
	                        };
                            };
			};	
			m++;
		  	break;
		    default:
			break;
                    };
                };
            };
        };
    };
}

void hwmix_dynamic_vc(int idx) 
{
	int i;
	if ( ! hwmix_vc[idx].created ) {
	    for (i=0; i<4; i++) {

                hwmix_vc[idx].vc_adj[i]=gtk_adjustment_new((gdouble)(8), 	
					(gdouble)(0), 
					(gdouble)(32767+1.0), 
        	    			(gdouble)(1.0),(gdouble)(1.0),(gdouble)(1.0));

		hwmix_vc[idx].vc_hscale[i]=gtk_hscale_new(GTK_ADJUSTMENT(hwmix_vc[idx].vc_adj[i]));
		gtk_widget_show(hwmix_vc[idx].vc_hscale[i]);
		gtk_container_add(GTK_CONTAINER(hwmix_vbox[idx]), hwmix_vc[idx].vc_hscale[i]);

	        gtk_scale_set_value_pos(GTK_SCALE(hwmix_vc[idx].vc_hscale[i]),GTK_POS_LEFT);
		gtk_scale_set_draw_value(GTK_SCALE(hwmix_vc[idx].vc_hscale[i]),1);
		gtk_scale_set_digits(GTK_SCALE(hwmix_vc[idx].vc_hscale[i]),0);

        	gtk_widget_set_usize(hwmix_vc[idx].vc_hscale[i], 102, 22);
 		gtk_signal_connect (GTK_OBJECT (hwmix_vc[idx].vc_adj[i]), 
					"value_changed",
					GTK_SIGNAL_FUNC (hwmix_vc_adj_value_changed_cb), 
					&hwmix_vc_idx[idx*4+i]);
            };
	};
	hwmix_vc[idx].created=TRUE;
}

void hwmix_play_button_clicked_cb (GtkWidget *widget, gpointer data)
{
     int idx,*p;
     p=data;
     idx=*p;
     hwmix[idx].v3d=FALSE;
     play_wav_file(&hwmix[idx]);
     gtk_widget_set_sensitive(hwmix_play_button[idx],FALSE);
     if ( hwmix[idx].alsa_can_pause ) {
         gtk_widget_set_sensitive(hwmix_pause_button[idx],TRUE);
     };
     gtk_widget_set_sensitive(hwmix_stop_button[idx],TRUE);     
     gtk_widget_set_sensitive(hwmix_del_button[idx],FALSE);     

     hwmix_dynamic_vc(idx);
     hwmix_vc_enable(idx,TRUE);
}

void hwmix_pause_button_toggled_cb (GtkWidget *widget, gpointer data)
{
     int idx,*p;
     p=data;
     idx=*p;
     if ( gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)) ) {
         pause_play(&hwmix[idx],1);
         gtk_widget_set_sensitive(hwmix_stop_button[idx],FALSE);     
     }
     else {
         pause_play(&hwmix[idx],0);          
         gtk_widget_set_sensitive(hwmix_stop_button[idx],TRUE);     
     };
}

void hwmix_stop_button_clicked_cb (GtkWidget *widget, gpointer data)
{
     int idx,*p;
     p=data;
     idx=*p;
     stop_play(&hwmix[idx]);
     gtk_widget_set_sensitive(hwmix_play_button[idx],TRUE);
     gtk_widget_set_sensitive(hwmix_pause_button[idx],FALSE);
     gtk_widget_set_sensitive(hwmix_stop_button[idx],FALSE);	
     gtk_widget_set_sensitive(hwmix_del_button[idx],TRUE);	
     hwmix_vc_enable(idx,FALSE);
}

void hwmix_loop_button_toggled_cb (GtkWidget *widget, gpointer data)
{
     int idx,*p;
     p=data;
     idx=*p;
     	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) 
		hwmix[idx].loop=TRUE;
	else 
    		hwmix[idx].loop=FALSE;	
}

void hwmix_display_total_channels(void)
{
	char buffer[10];
	sprintf(buffer,"Total: %d channels",hwmix_total_channels);
        gtk_label_set_text(GTK_LABEL(hwmix_total_chn_label),buffer);
}

void hwmix_del_button_clicked_cb (GtkWidget *widget, gpointer data)
{
     	int idx,*p;
     	p=data;
     	idx=*p;
        hwmix_total_channels-=hwmix[idx].channels;
        hwmix_display_total_channels();
	hwmix[idx].channels=0;
	strcpy(hwmix[idx].filename,"");
        gtk_label_set_text(GTK_LABEL(hwmix_label[idx])," ");
        gtk_label_set_text(GTK_LABEL(hwmix_rate_label[idx])," ");
        gtk_label_set_text(GTK_LABEL(hwmix_chn_label[idx])," ");
	gtk_widget_set_sensitive(hwmix_play_button[idx],FALSE);
        gtk_widget_set_sensitive(hwmix_pause_button[idx],FALSE);
        gtk_widget_set_sensitive(hwmix_stop_button[idx],FALSE);	
        gtk_widget_set_sensitive(hwmix_del_button[idx],FALSE);	
        gtk_widget_set_sensitive(hwmix_loop_button[idx],FALSE);	
	redraw_scope(idx);
}

enum {
    COL_FILENAME = 0,
    COL_FORMAT,
    COL_RATE,
    COL_CHANNEL
};

int hwmix_add_filename(GtkFileSelection *fs,WaveFile *w,char *filename)
{
	GtkWidget *message_dlg;
	GtkWidget *hbox;
    	int idx;
	char channel[8];
        char rate[10];
//	GtkTreeIter  iter;

        if ( ( hwmix_total_channels + w->channels ) >  16 ) {
            message_dlg=gtk_message_dialog_new (GTK_WINDOW(fs),
                                        GTK_DIALOG_DESTROY_WITH_PARENT,
                                        GTK_MESSAGE_WARNING,
                                        GTK_BUTTONS_OK,
                                        "Total channel cannot exceed 16");
            gtk_widget_show(message_dlg);       
            g_signal_connect_swapped (GTK_OBJECT (message_dlg), 
                             "response", 
            G_CALLBACK (gtk_widget_destroy), GTK_OBJECT (message_dlg));  
	    return 0;
        }
	else {
            hwmix_total_channels+=w->channels;
	    hwmix_display_total_channels();
	    
            idx=0;
            while (( idx < no_of_hwmixer ) && ( strcmp(hwmix[idx].filename,"") ))
	        idx++;
            if ( idx < 16 ) {	
		if ( idx >= no_of_hwmixer ) {

		    gtk_table_resize(GTK_TABLE(hwmix_table),no_of_hwmixer+1,4);

	            hbox=gtk_hbox_new(FALSE,0);
		    gtk_widget_show(hbox);
		    gtk_table_attach_defaults(GTK_TABLE(hwmix_table),hbox,0,1,idx,idx+1);
  	
		    hwmix_label[idx]=gtk_label_new(" ");
		    gtk_widget_show(hwmix_label[idx]); 
  		    gtk_container_add (GTK_CONTAINER (hbox), hwmix_label[idx]);

		    hwmix_rate_label[idx]=gtk_label_new(" ");
		    gtk_widget_show(hwmix_rate_label[idx]); 
  		    gtk_container_add (GTK_CONTAINER (hbox), hwmix_rate_label[idx]);

		    hwmix_chn_label[idx]=gtk_label_new(" ");
		    gtk_widget_show(hwmix_chn_label[idx]); 
  		    gtk_container_add (GTK_CONTAINER (hbox), hwmix_chn_label[idx]);

		    hwmix[idx].channels=w->channels;


	            hwmix_vbox[idx]=gtk_vbox_new(FALSE,0);
		    gtk_widget_show(hwmix_vbox[idx]);
		    gtk_table_attach_defaults(GTK_TABLE(hwmix_table),hwmix_vbox[idx],1,2,idx,idx+1);
  	scope_drawing[idx] = gtk_drawing_area_new();
  	gtk_widget_show (scope_drawing[idx]);
        gtk_table_attach_defaults(GTK_TABLE(hwmix_table),scope_drawing[idx],2,3,idx,idx+1);
        gtk_widget_set_usize(scope_drawing[idx], 128, 48);
        gtk_signal_connect(GTK_OBJECT(scope_drawing[idx]), 
	  	"expose_event",
                  	GTK_SIGNAL_FUNC(scope_drawing_expose_event), &scope_idx[idx]);
        gtk_signal_connect(GTK_OBJECT(scope_drawing[idx]), 
		"configure_event",
                       	GTK_SIGNAL_FUNC(scope_drawing_configure_event), &scope_idx[idx]);
//	g_print("set events\n");
//        gtk_widget_set_events(scope_drawing[idx], GDK_EXPOSURE_MASK);
//	g_print("Scope drawing ok\n");
		    	
        	    hbox=gtk_hbox_new(TRUE,0);
		    gtk_widget_show(hbox);
		    gtk_table_attach_defaults(GTK_TABLE(hwmix_table),hbox,3,4,idx,idx+1);

	          
		    hwmix_play_button[idx]=gtk_button_new_with_label("Play");
		    gtk_widget_show(hwmix_play_button[idx]);
  		    gtk_container_add (GTK_CONTAINER (hbox), hwmix_play_button[idx]);
		    gtk_signal_connect (GTK_OBJECT (hwmix_play_button[idx]), "clicked",
			    GTK_SIGNAL_FUNC (hwmix_play_button_clicked_cb), &hwmix_idx[idx]);
        	    gtk_widget_set_sensitive(hwmix_play_button[idx],FALSE);


		    hwmix_pause_button[idx]=gtk_toggle_button_new_with_label("Pause");
		    gtk_widget_show(hwmix_pause_button[idx]);
  		    gtk_container_add (GTK_CONTAINER (hbox), hwmix_pause_button[idx]);
		    gtk_signal_connect (GTK_OBJECT (hwmix_pause_button[idx]), "toggled",
			    GTK_SIGNAL_FUNC (hwmix_pause_button_toggled_cb), &hwmix_idx[idx]);

        	    gtk_widget_set_sensitive(hwmix_pause_button[idx],FALSE);
	
		    hwmix_stop_button[idx]=gtk_button_new_with_label("Stop");
		    gtk_widget_show(hwmix_stop_button[idx]);
  		    gtk_container_add (GTK_CONTAINER (hbox), hwmix_stop_button[idx]);
		    gtk_signal_connect (GTK_OBJECT (hwmix_stop_button[idx]), "clicked",
			    GTK_SIGNAL_FUNC (hwmix_stop_button_clicked_cb), &hwmix_idx[idx]);
        	    gtk_widget_set_sensitive(hwmix_stop_button[idx],FALSE);

		    hwmix_loop_button[idx]=gtk_toggle_button_new_with_label("Loop");
		    gtk_widget_show(hwmix_loop_button[idx]);
  		    gtk_container_add (GTK_CONTAINER (hbox), hwmix_loop_button[idx]);
		    gtk_signal_connect (GTK_OBJECT (hwmix_loop_button[idx]), "toggled",
			    GTK_SIGNAL_FUNC (hwmix_loop_button_toggled_cb), &hwmix_idx[idx]);
        	    gtk_widget_set_sensitive(hwmix_loop_button[idx],FALSE);

		    hwmix_del_button[idx]=gtk_button_new_with_label("Del");
		    gtk_widget_show(hwmix_del_button[idx]);
  		    gtk_container_add (GTK_CONTAINER (hbox), hwmix_del_button[idx]);
		    gtk_signal_connect (GTK_OBJECT (hwmix_del_button[idx]), "clicked",
			    GTK_SIGNAL_FUNC (hwmix_del_button_clicked_cb), &hwmix_idx[idx]);
        	    gtk_widget_set_sensitive(hwmix_del_button[idx],FALSE);

                    no_of_hwmixer++;
		    no_of_scope++;
		};

 	        strcpy(hwmix[idx].filename,filename);
                sprintf(rate,"%ld Hz",w->samples_per_sec);
	        switch (w->channels) {
	        case 1:
		    strcpy(channel,"Mono");
		    break;
	        case 2:
		    strcpy(channel,"Stereo");
		    break;
		case 4:
		    strcpy(channel,"Quad");
		    break;
		default:
		    strcpy(channel," ");
		    break;
		};	
                gtk_label_set_text(GTK_LABEL(hwmix_label[idx]),strrchr(filename, '/')+1);
		gtk_label_set_text(GTK_LABEL(hwmix_rate_label[idx]),rate);
                gtk_label_set_text(GTK_LABEL(hwmix_chn_label[idx]),channel);
		gtk_label_set_justify(GTK_LABEL(hwmix_label[idx]),GTK_JUSTIFY_LEFT);
                gtk_widget_set_sensitive(hwmix_play_button[idx],TRUE);
                gtk_widget_set_sensitive(hwmix_del_button[idx],TRUE);
                gtk_widget_set_sensitive(hwmix_loop_button[idx],TRUE);
/*
	        gtk_list_store_append (hwmix_list_store, &iter);
	        gtk_list_store_set(hwmix_list_store, &iter, 
			COL_FILENAME, filename, 
			COL_FORMAT, snd_pcm_format_name(wav_format_to_alsa_format(w)),
			COL_RATE, w->samples_per_sec,
			COL_CHANNEL, w->channels, -1);
*/
            };
            return 1;
      };
}
/*
static void onRowActivated (GtkTreeView        *view,
                  GtkTreePath        *path,
                  GtkTreeViewColumn  *col,
                  gpointer            data) 
{
    GtkTreeModel *model;
    GtkTreeIter   iter;
   
    g_print ("Row has been double-clicked. Removing row.\n");

    model = gtk_tree_view_get_model(view);

    if (!gtk_tree_model_get_iter(model, &iter, path))
      return; 

     gtk_list_store_remove(GTK_LIST_STORE(model), &iter);
}
*/

void hwmix_filesel_ok_sel(GtkWidget *widget, GtkFileSelection *fs )
{
	WaveFile w;
        int i;
	char filename[FILENAME_MAX+1];
        gchar **files;

        if ( gtk_file_selection_get_select_multiple(GTK_FILE_SELECTION(fs)) ) {
            files = gtk_file_selection_get_selections(GTK_FILE_SELECTION(fs));
	    if ( files ) {
	        for(i=0; (files[i] && ( hwmix_total_channels < 16)); i++) {
		    strcpy(filename,files[i]);
	            if ( is_wav_file(&w,filename) ) {
		        hwmix_add_filename(fs,&w,filename);
                    };
	        };
	    };
            g_strfreev(files);
            gtk_widget_destroy(GTK_WIDGET(fs));
        }
        else {
            strcpy(filename,gtk_file_selection_get_filename(GTK_FILE_SELECTION(fs)));
            if ( is_wav_file(&w,filename) ) {
   	        if ( hwmix_add_filename(fs,&w,filename) ) {
                    gtk_widget_destroy(GTK_WIDGET(fs));
                };
            };
        };
}

void hwmix_add_button_cb (GtkWidget *widget, gpointer data)
{
	GtkWidget *filesel;
	
        filesel=gtk_file_selection_new("Maximum 16 Mono/8 Stereo/4 Quad wave");
        gtk_widget_show(filesel);
        gtk_file_selection_hide_fileop_buttons(GTK_FILE_SELECTION(filesel));
        gtk_file_selection_set_select_multiple(GTK_FILE_SELECTION(filesel),TRUE);
	gtk_file_selection_complete(GTK_FILE_SELECTION(filesel),"*.wav");

//        gtk_dialog_add_button(GTK_DIALOG(filesel),"Add all",123);
//	gtk_dialog_set_response_sensitive(GTK_DIALOG(filesel),123,TRUE);

        g_signal_connect (G_OBJECT (GTK_FILE_SELECTION (filesel)->ok_button),
		      "clicked", G_CALLBACK (hwmix_filesel_ok_sel), (gpointer) filesel);
    
        g_signal_connect_swapped (G_OBJECT (GTK_FILE_SELECTION (filesel)->cancel_button),
	              "clicked", G_CALLBACK (gtk_widget_destroy), G_OBJECT (filesel));	  
        
//        g_signal_connect_swapped (GTK_OBJECT (filesel), 
//                             "response", 
//                             G_CALLBACK (hwmix_filesel_cb), GTK_OBJECT (filesel)); 
}

void hwmix_start_capture_button_clicked_cb (GtkWidget *widget, gpointer data)
{
        gtk_widget_set_sensitive(hwmix_start_capture_button,FALSE);
        start_capture(SND_PCM_FORMAT_S16_LE,2,48000);
        gtk_widget_set_sensitive(hwmix_stop_capture_button,TRUE);
}

void hwmix_stop_capture_button_clicked_cb (GtkWidget *widget, gpointer data)
{
        gtk_widget_set_sensitive(hwmix_stop_capture_button,FALSE);
        stop_capture();
        gtk_widget_set_sensitive(hwmix_start_capture_button,TRUE);
}

void scope_type_changed_cb (GtkWidget *widget, gpointer data)
{
	char select_item[32];
	int idx;
	strcpy(select_item,gtk_entry_get_text (GTK_ENTRY(widget)));
	for (idx=0; idx<4; idx++) {
            if ( strcmp(scope_type_name[idx],select_item) == 0 ) {
		scope_type=idx;
	        redraw_all_scope();
		return;
            };
	};
}

void hwmix_capture_button_clicked_cb(GtkWidget *widget, gpointer data)
{
	GtkWidget *dlg;
	GtkWidget *hbox;

       dlg=gtk_dialog_new_with_buttons("Capture",
					GTK_WINDOW(window),
					GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_STOCK_OK,
					GTK_RESPONSE_NONE,
					NULL);
        g_signal_connect_swapped (GTK_OBJECT (dlg), 
                             "response", 
                             G_CALLBACK (gtk_widget_destroy),
                             GTK_OBJECT (dlg));

  	capture_scope_drawing = gtk_drawing_area_new();
  	gtk_widget_show (capture_scope_drawing);

        gtk_widget_set_usize(capture_scope_drawing, 512, 192);
        gtk_signal_connect(GTK_OBJECT(capture_scope_drawing), 
	  	"expose_event",
                  	GTK_SIGNAL_FUNC(capture_scope_drawing_expose_event), NULL);
        gtk_signal_connect(GTK_OBJECT(capture_scope_drawing), 
		"configure_event",
                       	GTK_SIGNAL_FUNC(capture_scope_drawing_configure_event), NULL);

        gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dlg)->vbox), capture_scope_drawing);

	hbox=gtk_hbox_new(FALSE,0);
	gtk_widget_show(hbox);
        gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dlg)->vbox), hbox);

	hwmix_start_capture_button=gtk_button_new_with_label("Record");
        gtk_widget_show(hwmix_start_capture_button);
	gtk_container_add(GTK_CONTAINER(hbox), hwmix_start_capture_button);
        gtk_signal_connect (GTK_OBJECT (hwmix_start_capture_button), "clicked",
			    GTK_SIGNAL_FUNC (hwmix_start_capture_button_clicked_cb), NULL);

	hwmix_stop_capture_button=gtk_button_new_with_label("Stop");
        gtk_widget_show(hwmix_stop_capture_button);
	gtk_container_add(GTK_CONTAINER(hbox), hwmix_stop_capture_button);
        gtk_signal_connect (GTK_OBJECT (hwmix_stop_capture_button), "clicked",
			    GTK_SIGNAL_FUNC (hwmix_stop_capture_button_clicked_cb), NULL);
        gtk_widget_set_sensitive(hwmix_stop_capture_button,FALSE);
	gtk_widget_show_all(dlg);
}

static void create_hwmixer(GtkWidget *main, GtkWidget *notebook, int page)
{
	GtkWidget *label;
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkWidget *combo;
	GtkWidget *button;
        GList *glist;

//	GtkWidget *frame;
/*
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *column;
*/
	int i;

	vbox = gtk_vbox_new(FALSE, 0);
	gtk_widget_show(vbox);
	gtk_container_add(GTK_CONTAINER(notebook), vbox);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 6);

        label = gtk_label_new("Hardware Mixing");
        gtk_widget_show(label);
	gtk_notebook_set_tab_label(GTK_NOTEBOOK(notebook), 
				   gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), page), 
				   label);

        hwmix_total_channels=0;
	no_of_hwmixer=0;
/*
	hwmix_list_store = gtk_list_store_new (4, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_LONG, G_TYPE_INT);
	hwmix_tree = gtk_tree_view_new();	
	gtk_tree_view_set_model(GTK_TREE_VIEW(hwmix_tree), GTK_TREE_MODEL(hwmix_list_store));	

	renderer = gtk_cell_renderer_text_new();
        column = gtk_tree_view_column_new_with_attributes("Filename", renderer, "text",0, NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(hwmix_tree), column);

	gtk_container_add (GTK_CONTAINER (vbox), hwmix_tree);
        gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(hwmix_tree),FALSE);
	gtk_widget_set_usize(hwmix_tree, 300, 300);
	gtk_widget_show (hwmix_tree);
        g_signal_connect(hwmix_tree, "row-activated", G_CALLBACK(onRowActivated), NULL);
*/
	hwmix_table=gtk_table_new(no_of_hwmixer,4,FALSE);
        gtk_widget_show(hwmix_table);
	gtk_container_add(GTK_CONTAINER(vbox),hwmix_table);

        hbox=gtk_hbox_new(FALSE,0);
        gtk_widget_show(hbox);
	gtk_container_add(GTK_CONTAINER(vbox), hbox);

        hwmix_total_chn_label=gtk_label_new(" ");
        gtk_widget_show(hwmix_total_chn_label);
	gtk_container_add(GTK_CONTAINER(hbox), hwmix_total_chn_label);       

        combo=gtk_combo_new();
	gtk_widget_show(combo);
	gtk_container_add(GTK_CONTAINER(hbox), combo);	

        glist = NULL;  
        for (i=0; i<4; i++) {
            glist = g_list_append (glist, scope_type_name[i]);
        };
        gtk_combo_set_popdown_strings (GTK_COMBO (combo), glist);
	g_list_free(glist);
        gtk_entry_set_editable(GTK_ENTRY(GTK_COMBO(combo)->entry),FALSE);
        gtk_signal_connect(GTK_OBJECT(GTK_COMBO(combo)->entry),
                         "changed",
                         GTK_SIGNAL_FUNC(scope_type_changed_cb),
                         NULL);

	button=gtk_button_new_with_label("Capture");
	gtk_widget_show(button);
	gtk_container_add(GTK_CONTAINER(hbox), button);
        gtk_signal_connect (GTK_OBJECT (button), "clicked",
			    GTK_SIGNAL_FUNC (hwmix_capture_button_clicked_cb), NULL);
	gtk_widget_set_sensitive(button,FALSE);

        hwmix_add_button=gtk_button_new_with_label("Add Sound");
        gtk_widget_show(hwmix_add_button);
	gtk_container_add(GTK_CONTAINER(hbox), hwmix_add_button);
        gtk_signal_connect (GTK_OBJECT (hwmix_add_button), "clicked",
			    GTK_SIGNAL_FUNC (hwmix_add_button_cb), NULL);
}

/******************************************************************************************/

int line_intersect_x_plane(double x1,double y1,double x2,double y2,double px,double py0,double py1)
{
        double iy;
        if ( x1 != x2 ) {
	    	if ((( x1 <= px ) && ( px <= x2 ) ) || (( x2 <= px ) && ( px <= x1 ))) {
	        	iy=y1+(y2-y1)*(px-x1)/(x2-x1);
	       	    	if ( ( py0 <= iy ) && ( iy <= py1 ))
				return 1;            
		};
	}
	return 0;
}

int line_intersect_y_plane(double x1,double y1,double x2,double y2,double py,double px0,double px1)
{
	double ix;
        if ( y1 != y2 ) {
		if ((( y1 <= py ) && ( py <= y2 )) || (( y2 <= py ) && ( py <= y1 ))) {
			ix=x1+(x2-x1)*(py-y1)/(y2-y1);
			if ( ( px0 <= ix ) && ( ix <= px1 )) 
        	     		return 1;
		};
	};
	return 0;
}

int line_intersect_rect(double x1,double y1,double x2,double y2,V3D_RECT *rect)
{
    if (( reflect_rc1 != rect ) && ( reflect_rc2 != rect )) {
	if ( x1 != x2 ) {
		if (line_intersect_x_plane(x1,y1,x2,y2,rect->left,rect->bottom, rect->top))
	                return 1;
                if (line_intersect_x_plane(x1,y1,x2,y2,rect->right,rect->bottom,rect->top))
	                return 1;
	};
	if ( y1 != y2 ) {
	        if (line_intersect_y_plane(x1,y1,x2,y2,rect->bottom,rect->left,rect->right))
	                return 1;
                if (line_intersect_y_plane(x1,y1,x2,y2,rect->top,rect->left,rect->right))
	                return 1;
        };    
    };
    return 0;
}

#define fsign(v) (v < 0.0? -1.0:1.0)

int line_intersect_circle(double x1,double y1,double x2,double y2,double r)
{
	double d,dx,dy,dr2,disc;
	double ix1,iy1,ix2,iy2;
	dx=(x2-x1);
	dy=(y2-y1);
	dr2=dx*dx+dy*dy;
	d=(x1*y2-x2*y1);
	disc=r*r*dr2-d*d;
	if ( disc < 0.0 )
	    return 0;
	else {
            if ( disc == 0.0 ) {                      
//                if ( r == v3d_obstruction_radius ) 
//		    return 0;
                ix1=d*dy/dr2;
                iy1=-d*dx/dr2;
	        if ( x1 < x2 ) {                       
		    if (( x1 < ix1  ) && ( ix1 < x2 ))
		        return 1;
		}
		else {
		    if (( x2 < ix1  ) && ( ix1 < x1 ))
			return 1;
		};	
            }
	    else {
                ix1=(d*dy+fsign(dy)*dx*sqrt(disc))/dr2;
                iy1=(-d*dx+fabs(dy)*sqrt(disc))/dr2;

	        if ( x1 < x2 ) {
		    if (( x1 < ix1  ) && ( ix1 < x2 ))
		 	return 1;
                    ix2=(d*dy-fsign(dy)*dx*sqrt(disc))/dr2;
                    iy2=(-d*dx-fabs(dy)*sqrt(disc))/dr2;
		    if (( x1 < ix2  ) && ( ix2 < x2 ))
			return 1;
		}
		else {
		    if (( x2 < ix1  ) && ( ix1 < x1 ))
			return 1;
                    ix2=(d*dy-fsign(dy)*dx*sqrt(disc))/dr2;
                    iy2=(-d*dx-fabs(dy)*sqrt(disc))/dr2;
		    if (( x2 < ix2  ) && ( ix2 < x1 ))
			return 1;
		};	
	    };
 	    return 0;
	};
}

int direct_path(double x1, double y1, double x2, double y2)
{
	int i;
        for (i=0; i<v3d_no_of_obj; i++) 
            if ( line_intersect_rect(x1,y1,x2,y2,&v3d_obj[i]) )
		return 0;
	if ( v3d_reflection.flag == 2 ) 
            if ( ( sqrt(x1*x1+y1*y1) - 1.0 ) * ( sqrt(x2*x2+y2*y2) - 1.0 ) < 0 )
		return 0;
	if ( v3d_obstruction_circular ) 
		if ( line_intersect_circle(x1,y1,x2,y2,v3d_obstruction_radius) ) 
		     return 0;
        return 1;
}

void add_reflection_point(int i,double x, double y, double ix, double iy)
{
	int n;
        reflect_rc1=NULL;
	reflect_rc2=NULL;
	if ( direct_path(v3d_source[i].x,v3d_source[i].y,ix,iy) && 
	         direct_path(ix,iy,v3d_listener.x,v3d_listener.y)) {
            if ( v3d_1st_order_reflection ) {
                n=v3d_source[i].no_of_1st_reflect;
                v3d_source[i].reflection[n].x=x;
                v3d_source[i].reflection[n].y=y;
                v3d_source[i].reflection[n].ix=ix;
                v3d_source[i].reflection[n].iy=iy;                   
                v3d_source[i].no_of_1st_reflect++;
            };
            if ( v3d_wavetracing == 0 ) {
 	        gdk_draw_line(v3d_drawing->window,pen3DSource[i],
			      v3d_ox+v3d_m2*v3d_source[i].x,
	                      v3d_oy-v3d_m2*v3d_source[i].y,
			      v3d_ox+v3d_m2*ix,
	                      v3d_oy-v3d_m2*iy);
		gdk_draw_line(v3d_drawing->window,pen3DSource[i],
			      v3d_ox+v3d_m2*ix,
	                      v3d_oy-v3d_m2*iy,
			      v3d_ox+v3d_m2*v3d_listener.x,
	                      v3d_oy-v3d_m2*v3d_listener.y);
            };
        };
}

void reflect_by_outer_circle(int i,double r)
{
	double sp,lp,ix,iy,ip;
       
	sp=sqrt(v3d_source[i].x*v3d_source[i].x+v3d_source[i].y*v3d_source[i].y);
        lp=sqrt(v3d_listener.x*v3d_listener.x+v3d_listener.y*v3d_listener.y);
	if ((  r < sp  ) && ( r < lp )) {

// FIXME - This should be the reflection from outer circular surface with radius r
      
                ix=v3d_listener.x+(v3d_source[i].x-v3d_listener.x)*(1.0-lp)/(2.0-lp-sp);
		iy=v3d_listener.y+(v3d_source[i].y-v3d_listener.y)*(1.0-lp)/(2.0-lp-sp);

		ip=sqrt(ix*ix+iy*iy);
		ix=ix*v3d_obstruction_radius/ip;
		iy=iy*v3d_obstruction_radius/ip;
		
                reflect_rc1=NULL;
                reflect_rc2=NULL;
                if ( direct_path(ix,iy,v3d_source[i].x,v3d_source[i].y) ) {
		    if ( direct_path(ix,iy,v3d_listener.x,v3d_listener.y) ) {
//                       if ( v3d_wavetracing == 0 ) {
                             gdk_draw_line(v3d_drawing->window,pen3DSource[i],
					v3d_ox+v3d_m2*v3d_source[i].x,
	                                v3d_oy-v3d_m2*v3d_source[i].y,
					v3d_ox+v3d_m2*ix,
	                                v3d_oy-v3d_m2*iy);
	  		     gdk_draw_line(v3d_drawing->window,pen3DSource[i],
					v3d_ox+v3d_m2*ix,
	                                v3d_oy-v3d_m2*iy,
					v3d_ox+v3d_m2*v3d_listener.x,
	                                v3d_oy-v3d_m2*v3d_listener.y);
//                        };
		    };			
	        };
	};
}

double orientation(double x,double y)
{
    double r;
    r=sqrt(x*x+y*y);
    if ( x >= 0 ) 
        return(PI+asin(y/r));
    else 
        if ( y > 0 )
            return(PI+PI-asin(y/r));
        else
       	    return(-asin(y/r));
}

void reflect_by_inner_circle(int i,double r)
{
	double sp,lp,ix,iy,ip,dx,dy,dp;
        double so,lo;
       
	sp=sqrt(v3d_source[i].x*v3d_source[i].x+v3d_source[i].y*v3d_source[i].y);
        lp=sqrt(v3d_listener.x*v3d_listener.x+v3d_listener.y*v3d_listener.y);
        if (( sp <=r ) && ( lp <= r )) { 
            if (( lp == 0 ) || ( sp == 0 ))  {
                if (( lp == 0 ) && ( sp == 0 )) {
                   ix=r;                                                      
                   iy=0.0;
		   add_reflection_point(i,r+r,0.0,r,0.0);
	           add_reflection_point(i,-r-r,0.0,-r,0.0);
                }
		else {
 	            if ( sp == 0 ) {
		        ix=v3d_listener.x/lp;
		        iy=v3d_listener.y/lp;
		        add_reflection_point(i,2*ix-v3d_source[i].x,
                                               2*iy-v3d_source[i].y,ix,iy);
	                add_reflection_point(i,-2*ix-v3d_source[i].x,
                                               -2*iy-v3d_source[i].y,-ix,-iy);
                    }
		    else {
                        ix=v3d_source[i].x/sp;
	                iy=v3d_source[i].y/sp;
		        add_reflection_point(i,2*ix-v3d_source[i].x,
                                               2*iy-v3d_source[i].y,ix,iy);
	                add_reflection_point(i,-2*ix-v3d_source[i].x,
                                               -2*iy-v3d_source[i].y,-ix,-iy);
		    };
		};
   	    }
 	    else {
                if (( v3d_source[i].x == v3d_listener.x ) &&
                    ( v3d_source[i].y == v3d_listener.y )) {
                    ix=v3d_source[i].x/sp;
	            iy=v3d_source[i].y/sp;
		    add_reflection_point(i,2*ix-v3d_source[i].x,
                                           2*iy-v3d_source[i].y,ix,iy);
	            add_reflection_point(i,-2*ix-v3d_source[i].x,
                                           -2*iy-v3d_source[i].y,-ix,-iy);
                }
		else {
                    if ( lp == sp ) {                         
                        ix=(v3d_source[i].x+v3d_listener.x)/2;
			iy=(v3d_source[i].y+v3d_listener.y)/2;
                        ip=sqrt(ix*ix+iy*iy);
                        ix=ix/ip;
                        iy=iy/ip;
		        add_reflection_point(i,ix+ix-v3d_listener.x,iy+iy-v3d_listener.y,ix,iy);
			ix=-ix;
			iy=-iy;
		        add_reflection_point(i,ix+ix-v3d_listener.x,iy+iy-v3d_listener.y,ix,iy);
		    }
		    else {
                        if (( lp == r ) || ( sp == r )) {
                            lo=orientation(v3d_listener.x,v3d_listener.y);
                            so=orientation(v3d_source[i].x,v3d_source[i].y);
                            if (( lo == so ) || ( lo == ( so + PI ))) {
                                if ( lp == r ) {
				    ix=-v3d_listener.x;
                                    iy=-v3d_listener.y;
				    dx=(v3d_source[i].x-ix);
				    dy=(v3d_source[i].y-iy);
				    dp=sqrt(dx*dx+dy*dy);
			            add_reflection_point(i,
					v3d_listener.x+(ix-v3d_listener.x)*(2.0+dp)/2.0,
					v3d_listener.y+(iy-v3d_listener.y)*(2.0+dp)/2.0,ix,iy);
                                }
			        else {
				    ix=-v3d_source[i].x;
                                    iy=-v3d_source[i].y;
				    dx=(v3d_listener.x-ix);
				    dy=(v3d_listener.y-iy);
				    dp=sqrt(dx*dx+dy*dy);
			            add_reflection_point(i,
					v3d_source[i].x+(ix-v3d_source[i].x)*(2.0+dp)/2.0,
					v3d_source[i].y+(iy-v3d_source[i].y)*(2.0+dp)/2.0,ix,iy);
				};
			    };
                        }
			else {

// FIXME - This should be the reflection from inner circular wall with radius 1.0
          
                            ix=v3d_listener.x+(v3d_source[i].x-v3d_listener.x)*(1.0-lp)/(2.0-lp-sp);
      	                    iy=v3d_listener.y+(v3d_source[i].y-v3d_listener.y)*(1.0-lp)/(2.0-lp-sp);
                            ip=sqrt(ix*ix+iy*iy);
    	                    ix=ix/ip;
	                    iy=iy/ip;
			     
                            dx=(v3d_source[i].x-ix);
			    dy=(v3d_source[i].y-iy);
			    dp=sqrt(dx*dx+dy*dy);
 
			    dx=(ix-v3d_listener.x);
			    dy=(iy-v3d_listener.y);
			    ip=sqrt(dx*dx+dy*dy);

			    add_reflection_point(i,
				        v3d_listener.x+dx*(ip+dp)/ip,
					v3d_listener.y+dy*(ip+dp)/ip,ix,iy);
			    ix=-ix;
			    iy=-iy;
                            dx=(v3d_source[i].x-ix);
			    dy=(v3d_source[i].y-iy);
			    dp=sqrt(dx*dx+dy*dy);

			    dx=(ix-v3d_listener.x);
			    dy=(iy-v3d_listener.y);
			    ip=sqrt(dx*dx+dy*dy);

			    add_reflection_point(i,
					v3d_listener.x+dx*(ip+dp)/ip,
					v3d_listener.y+dy*(ip+dp)/ip,ix,iy);
			};
                    };
                };
            };
 	};
}
	
void reflect_x(int i, double x0,double bottom,double top)
{
	double iy;
	iy=v3d_source[i].y+(x0-v3d_source[i].x)*(v3d_listener.y-v3d_source[i].y)/(x0+x0-v3d_source[i].x-v3d_listener.x);
	if (( bottom <= iy ) && ( iy <= top )) 
	    if ( direct_path(v3d_source[i].x,v3d_source[i].y,x0,iy) ) 
		if ( direct_path(x0,iy,v3d_listener.x,v3d_listener.y) ) { 
                    add_reflection_point(i,x0+x0-v3d_source[i].x,v3d_source[i].y,x0,iy);
                    if ( ! v3d_wavetracing ) {
     			gdk_draw_line(v3d_drawing->window,pen3DSource[i],
				v3d_ox+v3d_m2*v3d_source[i].x,
                                v3d_oy-v3d_m2*v3d_source[i].y,
				v3d_ox+v3d_m2*x0,
                                v3d_oy-v3d_m2*iy);
     			gdk_draw_line(v3d_drawing->window,pen3DSource[i],
				v3d_ox+v3d_m2*x0,
                                v3d_oy-v3d_m2*iy,
				v3d_ox+v3d_m2*v3d_listener.x,
                                v3d_oy-v3d_m2*v3d_listener.y);
                     };
		};
}

void reflect_y(int i,double y0,double left,double right)
{
	double ix;
        ix=v3d_source[i].x+(y0-v3d_source[i].y)*(v3d_listener.x-v3d_source[i].x)/(y0+y0-v3d_source[i].y-v3d_listener.y);
	if ( ( left <= ix ) && ( ix <= right )) 
	    if ( direct_path(v3d_source[i].x,v3d_source[i].y,ix,y0) ) 
		if (  direct_path(ix,y0,v3d_listener.x,v3d_listener.y) ) {
                    add_reflection_point(i,v3d_source[i].x,y0+y0-v3d_source[i].y,ix,y0);
                    if ( v3d_wavetracing == 0 ) {
     			gdk_draw_line(v3d_drawing->window,pen3DSource[i],
				v3d_ox+v3d_m2*v3d_source[i].x,
                                v3d_oy-v3d_m2*v3d_source[i].y,
				v3d_ox+v3d_m2*ix,
                                v3d_oy-v3d_m2*y0);
     			gdk_draw_line(v3d_drawing->window,pen3DSource[i],
				v3d_ox+v3d_m2*ix,
                                v3d_oy-v3d_m2*y0,
				v3d_ox+v3d_m2*v3d_listener.x,
                                v3d_oy-v3d_m2*v3d_listener.y);
                     };
                };	
}

void reflect_rect(int i,V3D_RECT *rc)
{
	reflect_rc1=rc;
        reflect_rc2=NULL;
	if (( v3d_source[i].x < rc->left ) && ( v3d_listener.x  < rc->left ))
	    reflect_x(i,rc->left,rc->bottom,rc->top);
	if (( v3d_source[i].x > rc->right ) && ( v3d_listener.x  > rc->right ))
            reflect_x(i,rc->right,rc->bottom,rc->top);

        if (( v3d_source[i].y < rc->bottom ) && ( v3d_listener.y  < rc->bottom ))
            reflect_y(i,rc->bottom,rc->left,rc->right);
	if (( v3d_source[i].y > rc->top ) && ( v3d_listener.y  > rc->top ))
            reflect_y(i,rc->top,rc->left,rc->right);
}

void check_2nd_reflect_path(int i,double x0,double y0,double x1,double y1)
{
	if ( direct_path(v3d_source[i].x,v3d_source[i].y,x0,y0) ) 
	    if ( direct_path(x0,y0,x1,y1) ) 
	        if (  direct_path(x1,y1,v3d_listener.x,v3d_listener.y) ) 
                         if ( v3d_2nd_order_reflection ) {
  				gdk_draw_line(v3d_drawing->window,penGrey,
					v3d_ox+v3d_m2*v3d_source[i].x,
                                	v3d_oy-v3d_m2*v3d_source[i].y,
					v3d_ox+v3d_m2*x0,
                                	v3d_oy-v3d_m2*y0);
     			 	gdk_draw_line(v3d_drawing->window,penGrey,
					v3d_ox+v3d_m2*x0,
                                	v3d_oy-v3d_m2*y0,
					v3d_ox+v3d_m2*x1,
                                	v3d_oy-v3d_m2*y1);
     			 	gdk_draw_line(v3d_drawing->window,penGrey,
					v3d_ox+v3d_m2*x1,
                                	v3d_oy-v3d_m2*y1,
					v3d_ox+v3d_m2*v3d_listener.x,
                                	v3d_oy-v3d_m2*v3d_listener.y);
			};
}

void reflect_xy(int i,double x0,double bottom,double top,double y0,double left,double right)
{
     double rx,ry,ix,iy;
     rx=x0+x0-v3d_source[i].x;
     ry=y0+y0-v3d_source[i].y;
     iy=v3d_listener.y+(ry-v3d_listener.y)*(x0-v3d_listener.x)/(rx-v3d_listener.x);
     if (( bottom <= iy ) && ( iy <= top )) {
         ix=v3d_listener.x+(rx-v3d_listener.x)*(y0-v3d_listener.y)/(ry-v3d_listener.y);
         ix=(x0+x0-ix);
         if (( left <= ix ) && ( ix <= right )) 
	     check_2nd_reflect_path(i,ix,y0,x0,iy);
     };
     ix=v3d_listener.x+(rx-v3d_listener.x)*(y0-v3d_listener.y)/(ry-v3d_listener.y);
     if (( left <= ix ) && ( ix <= right )) {
         iy=v3d_listener.y+(ry-v3d_listener.y)*(x0-v3d_listener.x)/(rx-v3d_listener.x);
         iy=(y0+y0-iy);
         if (( bottom <= iy ) && ( iy <= top )) 
	     check_2nd_reflect_path(i,x0,iy,ix,y0);
     };
}

void  reflect_xx(int i,double x0,double bottom0,double top0,double x1,double bottom1,double top1)
{
     double rx,ry,iy0,iy1;
     rx=(x1+x1-(x0+x0-v3d_source[i].x));
     ry=v3d_source[i].y;
     iy1=v3d_listener.y+(ry-v3d_listener.y)*(x1-v3d_listener.x)/(rx-v3d_listener.x);
     if (( bottom1 <= iy1 ) && ( iy1 <= top1 )) {
         if ((( v3d_source[i].y < iy1 ) && ( iy1 < v3d_listener.y )) ||
             (( v3d_listener.y < iy1 ) && ( iy1 < v3d_source[i].y ))) {
             iy0=v3d_listener.y+(ry-v3d_listener.y)*(x1+x1-x0)/(rx-v3d_listener.x);
             if (( bottom0 <= iy0 ) && ( iy0 <= top0 )) 
	         if ((( v3d_source[i].y < iy0 ) && ( iy0 < v3d_listener.y )) ||
	             (( v3d_listener.y < iy0 ) && ( iy0 < v3d_source[i].y ))) 
	             if ( ! line_intersect_x_plane(v3d_source[i].x,v3d_source[i].y,x0,iy0,x1,bottom1,top1) ) 
                         if ( ! line_intersect_x_plane(x1,iy1,v3d_listener.x,v3d_listener.y,x0,bottom0,top0) ) 
                             check_2nd_reflect_path(i,x0,iy0,x1,iy1);
         };
     };
     rx=(x0+x0-(x1+x1-v3d_source[i].x));
     ry=v3d_source[i].y;
     iy1=v3d_listener.y+(ry-v3d_listener.y)*(x0-v3d_listener.x)/(rx-v3d_listener.x);
     if (( bottom0 <= iy1 ) && ( iy1 <= top0 )) {
         if ((( v3d_source[i].y < iy1 ) && ( iy1 < v3d_listener.y )) ||
             (( v3d_listener.y < iy1 ) && ( iy1 < v3d_source[i].y ))) {
             iy0=v3d_listener.y+(ry-v3d_listener.y)*(x0+x0-x1)/(rx-v3d_listener.x);
             if (( bottom1 <= iy0 ) && ( iy0 <= top1 )) 
	         if ((( v3d_source[i].y < iy0 ) && ( iy0 < v3d_listener.y )) ||
	             (( v3d_listener.y < iy0 ) && ( iy0 < v3d_source[i].y ))) 
                     if ( ! line_intersect_x_plane(v3d_source[i].x,v3d_source[i].y,x1,iy0,x0,bottom0,top0) ) 
                         if ( ! line_intersect_x_plane(x0,iy1,v3d_listener.x,v3d_listener.y,x1,bottom1,top1) ) 
                             check_2nd_reflect_path(i,x1,iy0,x0,iy1);
         };
     };     
}

void reflect_yy(int i,double y0,double left0,double right0,double y1,double left1,double right1)
{
     double rx,ry,ix0,ix1;
     ry=(y1+y1-(y0+y0-v3d_source[i].y));
     rx=v3d_source[i].x;
     ix1=v3d_listener.x+(rx-v3d_listener.x)*(y1-v3d_listener.y)/(ry-v3d_listener.y);
     if (( left1 <= ix1 ) && ( ix1 <= right1 )) {
         if ((( v3d_source[i].x < ix1 ) && ( ix1 < v3d_listener.x )) ||
             (( v3d_listener.x < ix1 ) && ( ix1 < v3d_source[i].x ))) {
             ix0=v3d_listener.x+(rx-v3d_listener.x)*(y1+y1-y0)/(ry-v3d_listener.y);
  	     if (( left0 <= ix0 ) && ( ix0 <= right0 )) 
                 if ((( v3d_source[i].x < ix0 ) && ( ix0 < v3d_listener.x )) ||
                     (( v3d_listener.x < ix0 ) && ( ix0 < v3d_source[i].x ))) 
                     if ( ! line_intersect_y_plane(v3d_source[i].x,v3d_source[i].y,ix0,y0,y1,left1,right1) ) 
		         if ( ! line_intersect_y_plane(ix1,y1,v3d_listener.x,v3d_listener.y,y0,left0,right0) ) 
	 	             check_2nd_reflect_path(i,ix0,y0,ix1,y1);
	 };
     };
     ry=(y0+y0-(y1+y1-v3d_source[i].y));
     rx=v3d_source[i].x;
     ix1=v3d_listener.x+(rx-v3d_listener.x)*(y0-v3d_listener.y)/(ry-v3d_listener.y);
     if (( left0 <= ix1 ) && ( ix1 <= right0 )) {
         if ((( v3d_source[i].x < ix1 ) && ( ix1 < v3d_listener.x )) ||
             (( v3d_listener.x < ix1 ) && ( ix1 < v3d_source[i].x ))) {
             ix0=v3d_listener.x+(rx-v3d_listener.x)*(y0+y0-y1)/(ry-v3d_listener.y);
  	     if (( left1 <= ix0 ) && ( ix0 <= right1 )) 
                 if ((( v3d_source[i].x < ix0 ) && ( ix0 < v3d_listener.x )) ||
                     (( v3d_listener.x < ix0 ) && ( ix0 < v3d_source[i].x ))) 
                     if ( ! line_intersect_y_plane(v3d_source[i].x,v3d_source[i].y,ix0,y1,y0,left0,right0) ) 
		         if ( ! line_intersect_y_plane(ix1,y0,v3d_listener.x,v3d_listener.y,y1,left1,right1) ) 
		             check_2nd_reflect_path(i,ix0,y1,ix1,y0);
	 };
     };
}

struct {
    int flag;
    double a1,a2;
} arc[21];
int no_of_arc;

void arc_minus(double a1,double a2)
{
     int i;
 //    g_print("minus %f - %f\n",180*a1/PI,180*a2/PI);
     for (i=0; i<no_of_arc; i++) {
          if ( arc[i].flag ) {
	      if (( a1 <= arc[i].a1 ) && ( arc[i].a2 <= a2 )) 
  		        arc[i].flag=0;
              else {
	          if (( arc[i].a1 < a1 ) && ( a2 < arc[i].a2 )) {
			arc[no_of_arc].flag=1;
			arc[no_of_arc].a1=a2;
	                arc[no_of_arc].a2=arc[i].a2;
	                arc[i].a2=a1;
	 		no_of_arc++;
	 		return;
		  }
		  else {
	              if (( arc[i].a1 == a1 ) && ( a2 < arc[i].a2 )) {
	                  arc[i].a1=a2;
			  return;
		      };
	              if (( arc[i].a2 == a2 ) && ( arc[i].a1 < a1)) {
			  arc[i].a2=a1;
			  return;
		      };
	              if (( arc[i].a1 < a2 ) && ( a2 < arc[i].a2 )) 
			  arc[i].a1=a2;
		      if (( a1 < arc[i].a2 ) && ( arc[i].a1 < a1 ) ) 
			  arc[i].a2=a1;
		  };
              };
	  };
     };
}

void arc_minus_angle_between(double o, double a)
{
    double a1,a2;
    a1=(o-a);
    a2=(o+a);

    if ( a1 < 0 ) {
        arc_minus(0.0,a2);
        arc_minus(PI+PI+a1,PI+PI);
    }
    else {
        if ( a2 <= (PI+PI) )
	    arc_minus(a1,a2);
        else {
            if ( (PI+PI) <= a1 ) 
	       arc_minus(a1-PI-PI,a2-PI-PI);
            else {
               arc_minus(a1,PI+PI);
               arc_minus(0,a2-PI-PI);
            };
        };                                                   
    };
}

void calc_draw_wave(int idx,double cx,double cy,double jj,int reflection_order)
{
	double left,right,bottom,top;
	double orientation;
	double cxcx_cycy;
	double jjjj;
	double a1,a2;
	double radius;
	int i,m;
        int no_blocking;
        
        jjjj=jj*jj;
	cxcx_cycy=cx*cx+cy*cy;
        no_blocking=1;

	i=0; 
	while (( i<v3d_no_of_obj ) && no_blocking ) {
            left=v3d_obj[i].left;
	    right=v3d_obj[i].right;
	    bottom=v3d_obj[i].bottom;
	    top=v3d_obj[i].top;	

	    if ( cx == left ) {
                if ( cy != top )
                    arc_minus(0.0,PI/2);
                    if ( cy != bottom )
                        arc_minus(3*PI/2,PI+PI);  
	    };			    
	    if ( cx == right ) {
                if ( cy != top )
                    arc_minus(PI/2,PI);
		    if ( cy != bottom )
                        arc_minus(PI,3*PI/2);
            };

	    if ((  reflection_order == 0 ) || 
                (( reflection_order == 1 ) && ( i != v3d_reflect_from_obj ))) {

	 	if ( cx < left ) {
                    if (( bottom <= cy ) && ( cy <= top )) {
		        if ( left <= ( cx + jj ) ) {
		            a2=atan((top-cy)/(left-cx));
	                    if ( jjjj < ((top-cy)*(top-cy)+(left-cx)*(left-cx)) ) 
                                a2=atan(sqrt(jjjj-(left-cx)*(left-cx))/(left-cx));
                            arc_minus(0.0,a2);
		            a1=PI+PI+atan((bottom-cy)/(left-cx));
	                    if ( jjjj < ((bottom-cy)*(bottom-cy)+(left-cx)*(left-cx) ) ) 
                                a1=PI+PI+atan(-sqrt(jjjj-(left-cx)*(left-cx))/(left-cx));
		            arc_minus(a1,PI+PI);
                        };
		    };
	        };
	        if ( right < cx ) {
                    if (( bottom <= cy ) && ( cy <= top )) {
		        if ( ( cx -jj ) <= right ) {
	    	            a1=PI+atan((top-cy)/(right-cx));
		            if ( jjjj < ( (top-cy)*(top-cy)+(right-cx)*(right-cx))) 
                                a1=PI+atan(sqrt(jjjj-(right-cx)*(right-cx))/(right-cx));
		   	    a2=PI+atan((bottom-cy)/(right-cx));
                            if ( jjjj < ( (bottom-cy)*(bottom-cy)+(right-cx)*(right-cx))) 
				a2=PI+atan(-sqrt(jjjj-(right-cx)*(right-cx))/(right-cx));
			    arc_minus(a1,a2);
                        };
		    };
		};
		if ( cy < bottom ) {
                    if (( left < cx ) && ( cx < right )) {
  		        if ( bottom <= ( cy + jj ) ) {
                            a1=atan((bottom-cy)/(right-cx));
		            a2=PI+atan((bottom-cy)/(left-cx));
                	    if ( jjjj < ((bottom-cy)*(bottom-cy)+(right-cx)*(right-cx))) 
                                a1=atan((bottom-cy)/sqrt(jjjj-(bottom-cy)*(bottom-cy)));
                            if ( jjjj < ((bottom-cy)*(bottom-cy)+(left-cx)*(left-cx)) ) 
 			        a2=PI+atan((cy-bottom)/sqrt(jjjj-(bottom-cy)*(bottom-cy)));
			    arc_minus(a1,a2);
                        };
		    };
	        };
		if ( top < cy ) {
	            if (( left < cx ) && ( cx < right )) {
			if ( ( cy -jj ) <= top ) {
                            a2=PI+PI+atan((top-cy)/(right-cx));
                            a1=PI+atan((top-cy)/(left-cx));
			    if ( jjjj < ((top-cy)*(top-cy)+(right-cx)*(right-cx)) ) 
                                a2=PI+PI+atan((top-cy)/sqrt(jjjj-(top-cy)*(top-cy)));
			    if ( jjjj < ((top-cy)*(top-cy)+(left-cx)*(left-cx)) ) 
                                a1=PI+atan((cy-top)/sqrt(jjjj-(top-cy)*(top-cy)));
			    arc_minus(a1,a2);
	                };   
		    };
		};
		if ( ( cx < left ) && ( top < cy ) ) {
		    if ( jjjj > ((left-cx)*(left-cx)+(top-cy)*(top-cy)) ) {
			a1=PI+PI+atan((bottom-cy)/(left-cx));
                        a2=PI+PI+atan((top-cy)/(right-cx));
                        if ( jjjj < ((right-cx)*(right-cx)+(top-cy)*(top-cy)) ) 
                            a2=PI+PI+atan((top-cy)/sqrt(jjjj-(top-cy)*(top-cy)));
                        if ( jjjj < ((left-cx)*(left-cx)+(bottom-cy)*(bottom-cy)) ) 
                            a1=PI+PI+atan(sqrt(jjjj-(left-cx)*(left-cx))/(cx-left)); 
			arc_minus(a1,a2);
		    };
		}; 
                if ( ( right < cx ) && ( top < cy ) ) {
		    if ( jjjj > ((right-cx)*(right-cx)+(top-cy)*(top-cy)) ) {
                        a1=PI+atan((top-cy)/(left-cx));
                        a2=PI+atan((bottom-cy)/(right-cx));
                        if ( jjjj < ((left-cx)*(left-cx)+(top-cy)*(top-cy)) ) 
                            a1=PI+atan((top-cy)/-sqrt(jjjj-(top-cy)*(top-cy)));
                        if ( jjjj < ((right-cx)*(right-cx)+(bottom-cy)*(bottom-cy)) ) 
                            a2=PI+atan(sqrt(jjjj-(right-cx)*(right-cx))/(cx-right)); 
			arc_minus(a1,a2);
		    };
		};
		if ( ( cx < left ) && ( cy < bottom )) {
		    if ( jjjj > ((left-cx)*(left-cx)+(bottom-cy)*(bottom-cy)) ) {
                        a1=atan((bottom-cy)/(right-cx));
                        a2=atan((top-cy)/(left-cx));
                        if ( jjjj < ((right-cx)*(right-cx)+(bottom-cy)*(bottom-cy)) ) 
		            a1=atan((bottom-cy)/sqrt(jjjj-(bottom-cy)*(bottom-cy))); 
                        if ( jjjj < ((left-cx)*(left-cx)+(top-cy)*(top-cy)) ) 
		            a2=atan(sqrt(jjjj-(left-cx)*(left-cx))/(left-cx));
		        arc_minus(a1,a2);
		    };
	        };
		if ( ( right < cx ) && ( cy < bottom )) {
		    if ( jjjj > ((right-cx)*(right-cx)+(bottom-cy)*(bottom-cy)) ) {
                        a1=PI+atan((top-cy)/(right-cx));
                        a2=PI+atan((bottom-cy)/(left-cx));
                        if ( jjjj < ((left-cx)*(left-cx)+(bottom-cy)*(bottom-cy)) ) 
                            a2=PI+atan((cy-bottom)/sqrt(jjjj-(bottom-cy)*(bottom-cy)));  
                        if ( jjjj < ((right-cx)*(right-cx)+(top-cy)*(top-cy)) ) 
                            a1=PI+atan(sqrt(jjjj-(right-cx)*(right-cx))/(right-cx));
			arc_minus(a1,a2);
		    };
		};
            };
            no_blocking=0;
	    for (m=0; m<no_of_arc; m++) {
	        if ( arc[m].flag ) 
	            no_blocking=1;
	    };
	    i++;
	};

        if ( v3d_obstruction_circular ) {
            if ( cx >= 0 ) 
                orientation=PI+asin(cy/sqrt(cxcx_cycy));
	    else {
                if ( cy > 0 )
	            orientation=PI+PI-asin(cy/sqrt(cxcx_cycy));
                else
	            orientation=-asin(cy/sqrt(cxcx_cycy));
	    };
            radius=v3d_obstruction_radius;
            if ( (cxcx_cycy) >= radius*radius ) {                           
                if ( (cxcx_cycy) == radius*radius ) 
                    arc_minus_angle_between(orientation,PI/2);
                else { 
                    if ( jj < ( sqrt(cxcx_cycy -radius*radius) ))  
                        arc_minus_angle_between(orientation,acos((cxcx_cycy+jjjj-radius*radius)/(2*jj*sqrt(cxcx_cycy)))); 
                    else 
                        arc_minus_angle_between(orientation,asin(radius/sqrt(cxcx_cycy)));
                };
            };
        };
	for (m=0; m<no_of_arc; m++) {
            if ( arc[m].flag ) {
   	        gdk_draw_arc(v3d_drawing->window,pen3DSource[idx],
        		      FALSE,
        		      v3d_ox+v3d_m2*cx-v3d_m2*jj, 
			      v3d_oy-v3d_m2*cy-v3d_m2*jj, 
			      v3d_m2*(jj+jj),v3d_m2*(jj+jj),
			      64*180*arc[m].a1/PI,64*180*(arc[m].a2-arc[m].a1)/PI);
	    };
	};
}

// Diffraction/Interfence not implemented

void wavetracing(int idx)
{
	double a1,a2;
        double ix,iy,ixix_iyiy;
	double cx,cy,cxcx_cycy;
	double dx,dy,dxdx_dydy;
        double orientation;
	double jj,jjjj;
	double limit;
	int i,j,k,r;	
	int no_blocking;

        k=(sqrt(8.0)/v3d_wavelength)+1;
        limit=k*v3d_wavelength;
        if ( v3d_source[idx].no_of_1st_reflect >  0 ) 
            k=(sqrt(32.0)/v3d_wavelength)+1;

        jj=v3d_wavelength;
	for (j=1; j<k; j++) {
            cx=v3d_source[idx].x;
            cy=v3d_source[idx].y;
            jjjj=jj*jj;

            no_blocking=1;
            i=0;
            while (( i < v3d_no_of_obj ) &&  no_blocking ) {
	        if ((v3d_obj[i].left < cx) && (cx < v3d_obj[i].right)) 
	            if ((v3d_obj[i].bottom < cy) && (cy < v3d_obj[i].top)) 
		        no_blocking=0;
	        i++;
	    };
            if ( no_blocking ) {
	        if ( v3d_obstruction_circular ) {
	  	    if (( fabs(cx) < v3d_obstruction_radius ) &&
                        ( fabs(cy) < v3d_obstruction_radius ))
		        no_blocking=0;
                };
            };
            if ( no_blocking ) {
	        arc[0].flag=1;
                arc[0].a1=0.0;
		arc[0].a2=PI+PI;
                no_of_arc=1;
	        if ( v3d_source[idx].cone_outer_angle < 360.0 ) {
                    arc_minus_angle_between(PI+v3d_source[idx].direction,(PI-PI*v3d_source[idx].cone_outer_angle/360.0));
                };
                cxcx_cycy=cx*cx+cy*cy;
                if ( jj < limit ) {
                    if ( v3d_reflection.flag == 1 ) {
                        if ( (sqrt(cxcx_cycy) + jj ) > 1.0 ) {
                            if ( cx > 0 ) {
                                a1=acos((1.0-cx)/jj);
                                arc_minus(0,a1);
                                arc_minus(PI+PI-a1,PI+PI);           
                            } 
                            else {
                                a1=acos((1.0+cx)/jj);
                                arc_minus(PI-a1,PI+a1);
                            };
                            if ( cy > 0 ) {
                                a2=acos((1.0-cy)/jj);
                                arc_minus(PI/2-a2,PI/2+a2);
                            }
                            else {
                                a2=acos((1.0+cy)/jj);
                                arc_minus(3*PI/2-a2,3*PI/2+a2);
                            };
                        };
                        if ( jj > (fabs(cx) + 1.0) ) {
                            if ( cx > 0 ) {
                                a1=acos((1.0+cx)/jj);
                                arc_minus(PI-a1,PI+a1);           
                            } 
                            else {
                                a1=acos((1.0-cx)/jj);
                                arc_minus(0,a1);
                                arc_minus(PI+PI-a1,PI+PI);           
                            };
                        };
                        if ( jj > (fabs(cy) + 1.0) ) {
                            if ( cy > 0 ) {
                                a2=acos((1.0+cy)/jj);
                                arc_minus(3*PI/2-a2,3*PI/2+a2);
                            }
                            else {
                                a2=acos((1.0-cy)/jj);
                                arc_minus(PI/2-a2,PI/2+a2);
                            };
                        };
 		    };
                    if ( v3d_reflection.flag == 2 ) {
                        if (( cx == 0.0 ) && ( cy == 0.0 )) {
                            if ( jj > 1.0 )
                                arc_minus(0,PI+PI);
                        }
                        else {
                            if ( v3d_source[idx].move == 1 ) {
                                if ( v3d_source[idx].x_range == 1.0 ) 
				    if ( v3d_source[idx].y_range == 1.0 ) 
					cxcx_cycy=1.0;
                            };
                            if ( cxcx_cycy > 1.0 ) {
                                if ( jj > (sqrt(cxcx_cycy) - 1.0) ) {
                        	    if ( cx >= 0 ) 
                                        orientation=PI+asin(cy/sqrt(cxcx_cycy));
			            else {
                                        if ( cy > 0 )
		                            orientation=PI+PI-asin(cy/sqrt(cxcx_cycy));
                                        else
		                     	    orientation=-asin(cy/sqrt(cxcx_cycy));
			     	    };        
                                    if ( jj < sqrt(cxcx_cycy-1.0) )  
                                        arc_minus_angle_between(orientation,acos((jjjj+cxcx_cycy-1.0)/(2*jj*sqrt(cxcx_cycy))));            
                                    else 
                                        arc_minus_angle_between(orientation,asin(1.0/sqrt(cxcx_cycy)));
				};                                    
                            }
                            else {
                                if ( jj > ( 1.0 + sqrt(cxcx_cycy)) ) {
                                    arc_minus(0,PI+PI);
                                }
  				else {
                                    if ( ( sqrt(cxcx_cycy) + jj ) >= 1.0 ) {
                                        if ( cx >= 0 ) {
                                            if ( cy >= 0 ) 
                                                orientation=asin(cy/sqrt(cxcx_cycy));
                                            else
                                                orientation=PI+PI+asin(cy/sqrt(cxcx_cycy));
	                                }
			                else 
	                                    orientation=PI-asin(cy/sqrt(cxcx_cycy));
                                        arc_minus_angle_between(orientation,(PI-acos((cxcx_cycy+jjjj-1.0)/(2*jj*sqrt(cxcx_cycy)))));
                                    };
                                };
                            };
                        };                            
	            };
		    calc_draw_wave(idx,cx,cy,jj,0);
	        };

// first order reflection

                for (r=0; r<v3d_source[idx].no_of_1st_reflect; r++) {		
		    dx=(v3d_source[idx].x-v3d_source[idx].reflection[r].ix);
		    dy=(v3d_source[idx].y-v3d_source[idx].reflection[r].iy);
		    if ( jj > sqrt(dx*dx+dy*dy) ) {
			dx=(v3d_listener.x-v3d_source[idx].reflection[r].x);
			dy=(v3d_listener.y-v3d_source[idx].reflection[r].y);
		        dxdx_dydy=dx*dx+dy*dy;			
			if ( jj < sqrt(dxdx_dydy) ) {
                            cx=v3d_source[idx].reflection[r].x;
                            cy=v3d_source[idx].reflection[r].y;

    	                    arc[0].flag=1;
        	            arc[0].a1=0.0;
                            arc[0].a2=PI+PI;
                            no_of_arc=1;
                   
              	            if ( dx >= 0 ) 
                                orientation=PI+asin(dy/sqrt(dxdx_dydy));
		            else {
                                if ( dy > 0 )
		                    orientation=PI+PI-asin(dy/sqrt(dxdx_dydy));
                                else
		                    orientation=-asin(dy/sqrt(dxdx_dydy));
		            };
                            arc_minus_angle_between(orientation,(31+sqrt(dxdx_dydy)/4.0)*PI/32);

                            ix=v3d_source[idx].reflection[r].ix;
                            iy=v3d_source[idx].reflection[r].iy;
                            v3d_reflect_from_obj=-1;          
                            i=0;
			    while (( i< v3d_no_of_obj ) && ( v3d_reflect_from_obj == -1 )) { 
			        if (( ix == v3d_obj[i].left ) || ( ix == v3d_obj[i].right )) 
                                    if (( v3d_obj[i].bottom < iy ) && ( iy <= v3d_obj[i].top )) 
				        v3d_reflect_from_obj=i;
			        if (( iy == v3d_obj[i].bottom ) || ( iy == v3d_obj[i].top )) 
                                    if (( v3d_obj[i].left < ix ) && ( ix <= v3d_obj[i].right )) 
				        v3d_reflect_from_obj=i;
			        i++;
			    };
			    if ( v3d_reflect_from_obj == -1 ) {
                                ixix_iyiy=ix*ix+iy*iy;
                                if ( v3d_reflection.flag == 2 )
                                    if ( ixix_iyiy == 1.0 )
				        v3d_reflect_from_obj = -2;
			        if ( v3d_reflect_from_obj == -1 ) 
                                    if ( v3d_reflection.flag == 1 ) 
                                        if (( fabs(ix) == 1.0 ) || ( fabs(iy) == 1.0 )) 
				            v3d_reflect_from_obj = -3;
			    }; 
                            calc_draw_wave(idx,cx,cy,jj,1);
                        };
                    };
                };
	    };
  	    jj+=v3d_wavelength;
        };
}

void reflect_by_circle_and_obj(int i,int j)
{
     int k,f;
     double vx,vy,vp,ix,iy,ip;
     double left,right,bottom,top;
     double lp,sx,sy,sp;
     double rx,ry;
            
     sp=sqrt(v3d_source[i].x*v3d_source[i].x+v3d_source[i].y*v3d_source[i].y);
     lp=sqrt(v3d_listener.x*v3d_listener.x+v3d_listener.y*v3d_listener.y);

     left=v3d_obj[j].left;
     right=v3d_obj[j].right;
     bottom=v3d_obj[j].bottom;
     top=v3d_obj[j].top;

    for (k=0; k<4; k++) {
		
        if (( v3d_source[i].x != 0 ) || ( v3d_source[i].y != 0.0 )) {
	    f=0;
            switch (k) {
	    case 0:
      	        if ( v3d_listener.x < left ) {
      		    vx=left*2-v3d_listener.x;
      		    vy=v3d_listener.y;
      		    rx=left;
      		    f=1;
      		};   
      		break;
       	    case 1:
      		if ( v3d_listener.x > right ) {
                    vx=right*2-v3d_listener.x;
                    vy=v3d_listener.y;
                    rx=right;
                    f=1;
                };      
                break;
            case 2:
                if ( v3d_listener.y < bottom ) {
                    vx=v3d_listener.x;
                    vy=bottom*2-v3d_listener.y;
                    ry=bottom;
                    f=1;
                };   
                break;
            case 3:
                if ( v3d_listener.y > top ) {
                    vx=v3d_listener.x;
                    vy=top*2-v3d_listener.y;
                    ry=top;
                    f=1;
                };
                break;
            };
            if ( f ) {
                vp=sqrt(vx*vx+vy*vy);
                ix=vx+(v3d_source[i].x-vx)*(1.0-vp)/(2.0-vp-sp);
                iy=vy+(v3d_source[i].y-vy)*(1.0-vp)/(2.0-vp-sp);
                ip=sqrt(ix*ix+iy*iy);
                ix=ix/ip;
                iy=iy/ip;

                if ( f ) {
                    reflect_rc1=NULL;
                    reflect_rc2=NULL;
                    if ( line_intersect_rect(ix,iy,vx,vy,&v3d_obj[j]) ) { 
                        if ( direct_path(v3d_source[i].x,v3d_source[i].y,ix,iy) ) {     
 	
	                    switch (k) {
	                    case 0:
	   		    case 1:
	                        ry=vy+(iy-vy)*(rx-vx)/(ix-vx);
	                        f = (( bottom <= ry ) && ( ry <= top ));
			        break;
			    case 2:
			    case 3:
	                        rx=vx+(ix-vx)*(ry-vy)/(iy-vy);
	                        f = (( left <= rx ) && ( rx <= right ));
                                break;
			    };
	                    if ( f ) {
	                        reflect_rc1=NULL;
	                        reflect_rc2=&v3d_obj[j];
	                        check_2nd_reflect_path(i,ix,iy,rx,ry);
                            };
	                };
	            };
	        };
		ix=-ix;
		iy=-iy;
                if ( f ) {	  
  	            reflect_rc1=NULL;
	            reflect_rc2=NULL;
	            if ( line_intersect_rect(ix,iy,vx,vy,&v3d_obj[j]) ) { 
                        if ( direct_path(v3d_source[i].x,v3d_source[i].y,ix,iy) ) {     
	                 	
	                    switch (k) {
	                    case 0:
			    case 1:
	                        ry=vy+(iy-vy)*(rx-vx)/(ix-vx);
	                        f = (( bottom <= ry ) && ( ry <= top ));
			        break;
			    case 2:
			    case 3:
	                        rx=vx+(ix-vx)*(ry-vy)/(iy-vy);
	                        f = (( left <= rx ) && ( rx <= right ));
			        break;
			    };
	                    if ( f ) {
	                        reflect_rc1=NULL;
	                        reflect_rc2=&v3d_obj[j];
	                        check_2nd_reflect_path(i,ix,iy,rx,ry);
                            };
	                }; 
                    };
		};
            };	
	};

        if (( v3d_listener.x != 0.0 ) || ( v3d_listener.y != 0.0 )) {

	    f=0;
            switch (k) {
	    case 0:
        	if ( v3d_source[i].x < left ) {
        	    sx=left*2-v3d_source[i].x;
        	    sy=v3d_source[i].y;
        	    rx=left;
        	    f=1;
        	};   
        	break;
            case 1:
        	if ( v3d_source[i].x > right ) {
        	    sx=right*2-v3d_source[i].x;
        	    sy=v3d_source[i].y;
        	    rx=right;
        	    f=1;
        	};      
        	break;
            case 2:
        	if ( v3d_source[i].y < bottom ) {
        	    sx=v3d_source[i].x;
        	    sy=bottom*2-v3d_source[i].y;
        	    ry=bottom;
        	    f=1;
        	};   
        	break;
            case 3:
        	if ( v3d_source[i].y > top ) {
        	    sx=v3d_source[i].x;
        	    sy=top*2-v3d_source[i].y;
        	    ry=top;
        	    f=1;
        	};
        	break;
            };
            if ( f ) {
        	vp=sqrt(vx*vx+vy*vy);
        	sp=(sx*sx+sy*sy);
                ix=v3d_listener.x+(sx-v3d_listener.x)*(1.0-lp)/(2.0-lp-sp);
        	iy=v3d_listener.y+(sx-v3d_listener.y)*(1.0-lp)/(2.0-lp-sp);
        	ip=sqrt(ix*ix+iy*iy);
        	ix=ix/ip;
        	iy=iy/ip;

                if ( f ) {		
        	    reflect_rc1=NULL;
        	    reflect_rc2=NULL;
        	    if ( line_intersect_rect(ix,iy,sx,sy,&v3d_obj[j]) ) { 
        	        if ( direct_path(v3d_listener.x,v3d_listener.y,ix,iy) ) {        
	
        	            switch (k) {
        	            case 0:
			    case 1:
        	                ry=sy+(iy-sy)*(rx-sx)/(ix-sx);
        	                f = (( bottom <= ry ) && ( ry <= top ));
			        break;
			    case 2:
			    case 3:
        	                rx=sx+(ix-sx)*(ry-sy)/(iy-sy);
        	                f = (( left <= rx ) && ( rx <= right ));
			        break;
			    };
        	            if ( f ) {
        	                reflect_rc1=NULL;
        	                reflect_rc2=&v3d_obj[j];
        	                check_2nd_reflect_path(i,rx,ry,ix,iy);
        	            };
                        };
        	    };
	        };
		ix=-ix;
		iy=-iy;
                if ( f ) {		
         	    reflect_rc1=NULL;
        	    reflect_rc2=NULL;
        	    if ( line_intersect_rect(ix,iy,sx,sy,&v3d_obj[j]) ) { 
        	        if ( direct_path(v3d_listener.x,v3d_listener.y,ix,iy) ) {
	 
        	            switch (k) {
        	            case 0:
			    case 1:
        	                ry=sy+(iy-sy)*(rx-sx)/(ix-sx);
        	                f = (( bottom <= ry ) && ( ry <= top ));
			        break;
			    case 2:
			    case 3:
        	                rx=sx+(ix-sx)*(ry-sy)/(iy-sy);
        	                f = (( left <= rx ) && ( rx <= right ));
			        break;
			    };
        	            if ( f ) {
        	                reflect_rc1=NULL;
        	                reflect_rc2=&v3d_obj[j];
        	                check_2nd_reflect_path(i,rx,ry,ix,iy);
                            };
        	        };
        	    };
	        };
	    };
	};
    };
}

void second_order_reflection(int i)
{
    int j,k;
    double sx,sy;
    double left,right,bottom,top;
    double k_left,k_right,k_bottom,k_top;

    sx=v3d_source[i].x;
    sy=v3d_source[i].y;
    for (j=0; j<v3d_no_of_obj; j++) {
        reflect_rc1=&v3d_obj[j];
        left=v3d_obj[j].left;
        right=v3d_obj[j].right;
        bottom=v3d_obj[j].bottom;
        top=v3d_obj[j].top;

        for (k=(j+1); k<v3d_no_of_obj; k++) {

            reflect_rc2=&v3d_obj[k];
            k_left=v3d_obj[k].left;
            k_right=v3d_obj[k].right;
            k_bottom=v3d_obj[k].bottom;
            k_top=v3d_obj[k].top;

            if ( left > k_right ) 
                if (( sx < left ) || ( v3d_listener.x  < left )) 
                    if (( sx > k_right ) || ( v3d_listener.x  > k_right )) 
                        reflect_xx(i,k_right,k_bottom,k_top,left,bottom,top);
            if ( right < k_left ) 
                if (( sx > right ) || ( v3d_listener.x  > right )) 
                    if (( sx < k_left ) || ( v3d_listener.x  < k_left )) 
                        reflect_xx(i,right,bottom,top,k_left,k_bottom,k_top);
            if ( bottom > k_top ) 
                if (( sy < bottom ) || ( v3d_listener.y  < bottom )) 
                    if (( sy > k_top ) || ( v3d_listener.y  > k_top )) 
                        reflect_yy(i,k_top,k_left,k_right,bottom,left,right);
            if ( top < k_bottom ) 
                if (( sy > top ) || ( v3d_listener.y  > top )) 
                    if (( sy < k_bottom ) || ( v3d_listener.y  < k_bottom )) 
                        reflect_yy(i,top,left,right,k_bottom,k_left,k_right);

            if (( v3d_listener.x < left ) && ( sx < left )) {
                if (( v3d_listener.y < k_bottom ) && ( sy < k_bottom )) 
                    reflect_xy(i,left,bottom,top,k_bottom,k_left,k_right);
                if (( v3d_listener.y > k_top ) && ( sy > k_top )) 
                    reflect_xy(i,left,bottom,top,k_top,k_left,k_right);
            };
            if (( v3d_listener.x > right ) && ( sx > right )) {
                if (( v3d_listener.y < k_bottom ) && ( sy < k_bottom )) 
                    reflect_xy(i,right,bottom,top,k_bottom,k_left,k_right);
                if (( v3d_listener.y > k_top ) && ( sy > k_top )) 
                    reflect_xy(i,right,bottom,top,k_top,k_left,k_right);
            };
            if (( v3d_listener.y < bottom ) && ( sy < bottom )) {
                if (( v3d_listener.x < k_left ) && ( sx < k_left )) 
                    reflect_xy(i,k_left,k_bottom,k_top,bottom,left,right);
                if (( v3d_listener.x > k_right ) && ( sx > k_right )) 
                    reflect_xy(i,k_right,k_bottom,k_top,bottom,left,right);
            };
            if (( v3d_listener.y > top ) && ( sy > top )) {
                if (( v3d_listener.x < k_left ) && ( sx < k_left )) 
                    reflect_xy(i,k_left,k_bottom,k_top,top,left,right);
                if (( v3d_listener.x > k_right ) && ( sx > k_right )) 
                    reflect_xy(i,k_right,k_bottom,k_top,top,left,right);
            };
        };
    };
    if ( v3d_reflection.flag == 1 ) {
	reflect_rc1=NULL;
        for (j=0; j<v3d_no_of_obj; j++) {
	    reflect_rc2=&v3d_obj[j];
            left=v3d_obj[j].left;
            right=v3d_obj[j].right;
            bottom=v3d_obj[j].bottom;
            top=v3d_obj[j].top;
            if (( sx < left ) || ( v3d_listener.x  < left )) 
	        reflect_xx(i,-1.0,-1.0,1.0,left, bottom, top);
            if (( sx < left ) && ( v3d_listener.x  < left )) 
 	        reflect_xy(i,-1.0,-1.0,1.0,left, bottom, top);
            if (( sx > right ) || ( v3d_listener.x  > right )) 
	        reflect_xx(i, right, bottom, top, 1.0,-1.0,1.0);
            if (( sx > right ) && ( v3d_listener.x  > right )) 
                reflect_xy(i, 1.0,-1.0,1.0, right, bottom, top);
            if (( sy < bottom ) || ( v3d_listener.y  < bottom )) 
		reflect_yy(i,-1.0,-1.0,1.0, bottom, left, right);
            if (( sy < bottom ) && ( v3d_listener.y  < bottom )) 
                reflect_xy(i,-1.0,-1.0,1.0, bottom, left, right);
            if (( sy > top ) || ( v3d_listener.y  > top )) 
 	        reflect_yy(i, top, left, right, 1.0,-1.0,1.0);
            if (( sy > top ) && ( v3d_listener.y  > top )) 
                reflect_xy(i, 1.0,-1.0,1.0, top, left, right);
        };
    };
    if ( v3d_reflection.flag == 2 ) {
        for (j=0; j<v3d_no_of_obj; j++) {
            reflect_by_circle_and_obj(i,j);       
	};    
    };
}

void calc_no_of_obj(void)
{
	int i;

        v3d_no_of_obj=0;      
	for (i=0; i<4; i++) {
            if ( v3d_occlusion[i].flag ) {
                memcpy(&v3d_obj[v3d_no_of_obj],&v3d_occlusion[i].rc,sizeof(V3D_RECT));
		v3d_no_of_obj++;
	    }
	    else {
		if ( v3d_exclusion[i].flag ) {
                    memcpy(&v3d_obj[v3d_no_of_obj],&v3d_exclusion[i].rc[0],sizeof(V3D_RECT));
       		    v3d_no_of_obj++;
                    memcpy(&v3d_obj[v3d_no_of_obj],&v3d_exclusion[i].rc[1],sizeof(V3D_RECT));
      		    v3d_no_of_obj++;
		}
	        else {
	            if ( v3d_obstruction[i].flag ) {
                        memcpy(&v3d_obj[v3d_no_of_obj],&v3d_obstruction[i].rc,sizeof(V3D_RECT));
    		        v3d_no_of_obj++;
                    };
		};
	    };
	};
}

void redraw_v3d(void)
{
	GtkWidget *widget;
	int m,ox,oy;
	int ww,hh,size;
	int i,j,r;
#ifdef HAVE_LIBOPENAL
	ALint error;
	ALint state;
	if ( no_of_3d_play > 0 ) {
            for (i=0; i<MAX_NO_OF_3D_SOURCE; i++) {
                if ( v3d_source[i].play ) {
		    alGetSourcei(source[i],AL_SOURCE_STATE,&state);
		    if ((error = alGetError()) == AL_NO_ERROR)	{
		        if ( state == AL_STOPPED ) {
			    g_print("Source %d stop..\n",i);
		            no_of_3d_play--; 
		    	    v3d_source[i].play=0;	
			    gtk_widget_set_sensitive(v3d_play_button[i],TRUE);
			    gtk_widget_set_sensitive(v3d_stop_button[i],FALSE);
			};
		    };
		};
	    };
	};
#endif
	widget=v3d_drawing;
	gdk_window_clear_area (widget->window,
                         0,0,widget->allocation.width,widget->allocation.height);

	ox=widget->allocation.width>>1;
	oy=widget->allocation.height>>1;
	m=MIN(widget->allocation.width,widget->allocation.height)-1-6;		
	v3d_ox=ox;	
	v3d_oy=oy;
	v3d_m2=m/2;

	if ( v3d_showpath ) {
	    for(i=0; i<MAX_NO_OF_3D_SOURCE; i++) {
		if ( v3d_source[i].move_type == 1 ) {
		    ww=m*v3d_source[i].x_range;
		    hh=m*v3d_source[i].y_range;
  		    gdk_draw_arc (widget->window,penBlack,
        		    FALSE,
        		    (widget->allocation.width-ww)>>1, 
			    (widget->allocation.height-hh)>>1, 
			    ww,hh,
			    0, 64*360);
		};
		if ( v3d_source[i].move_type == 2 ) {
                    gdk_draw_line(v3d_drawing->window,penBlack,
			    v3d_ox+v3d_m2*v3d_source[i].p_range*cos(v3d_source[i].o),
	                    v3d_oy-v3d_m2*v3d_source[i].p_range*sin(v3d_source[i].o),
			    v3d_ox-v3d_m2*v3d_source[i].p_range*cos(v3d_source[i].o),
	                    v3d_oy+v3d_m2*v3d_source[i].p_range*sin(v3d_source[i].o));
		};
	    };
	};

	for (i=0; i<4; i++) {
/*
	    if ( v3d_occlusion[i].flag ) {
		    gdk_draw_rectangle(widget->window,penBlack,TRUE,
			    ox+m*v3d_occlusion[i].rc.left/2,
			    oy-m*v3d_occlusion[i].rc.top/2,
			    m*(v3d_occlusion[i].rc.right-v3d_occlusion[i].rc.left)/2,
		  	    m*(v3d_occlusion[i].rc.top-v3d_occlusion[i].rc.bottom)/2);		
	    };	
*/
	    if ( v3d_obstruction[i].flag || v3d_occlusion[i].flag ) {
		    gdk_draw_rectangle(widget->window,penBlack,TRUE,
			    ox+m*v3d_obstruction[i].rc.left/2,
			    oy-m*v3d_obstruction[i].rc.top/2,
			 m*(v3d_obstruction[i].rc.right-v3d_obstruction[i].rc.left)/2,
		  	 m*(v3d_obstruction[i].rc.top-v3d_obstruction[i].rc.bottom)/2);		
	    };	
	    if ( v3d_exclusion[i].flag || v3d_occlusion[i].flag  ) {
                for (j=0; j<2; j++) {
		    gdk_draw_rectangle(widget->window,penBlack,TRUE,
			    ox+m*v3d_exclusion[i].rc[j].left/2,
			    oy-m*v3d_exclusion[i].rc[j].top/2,
			    m*(v3d_exclusion[i].rc[j].right-v3d_exclusion[i].rc[j].left)/2,
		  	    m*(v3d_exclusion[i].rc[j].top-v3d_exclusion[i].rc[j].bottom)/2);
		};
	    };	
	};
	if ( v3d_obstruction_circular ) {
                r=m*v3d_obstruction_radius;
	        gdk_draw_arc (widget->window,
                	penBlack,
                	TRUE,
                	((widget->allocation.width-r)>>1), 
                        ((widget->allocation.height-r)>>1), r,r,
	        	0, 64*360);
	};
        if  ( v3d_reflection.flag == 2 ) {
		for (j=m; j<(m+6); j+=2) {
  	        	gdk_draw_arc (widget->window,
                		penBlack,
                		FALSE,
               			(widget->allocation.width-j)>>1, 
				(widget->allocation.height-j)>>1, j,j,
	        		0, 64*360);
		};
	};	
	if ( v3d_reflection.flag == 1 ) {
	        gdk_draw_rectangle(widget->window, penBlack, TRUE,
        	        ((widget->allocation.width-m)>>1)-3,
			((widget->allocation.height-m)>>1)-3,3,m+7);
        	gdk_draw_rectangle(widget->window, penBlack, TRUE,
        	        ((widget->allocation.width+m)>>1)+1,
			((widget->allocation.height-m)>>1)-3,3,m+7);
	        gdk_draw_rectangle(widget->window, penBlack, TRUE,
        	        ((widget->allocation.width-m)>>1)-3,
			((widget->allocation.height-m)>>1)-3,m+7,3);
        	gdk_draw_rectangle(widget->window, penBlack, TRUE,
        	        ((widget->allocation.width-m)>>1)-3,
			((widget->allocation.height+m)>>1)+1,m+7,3);
	};
        if ( v3d_raytracing || v3d_2nd_order_reflection ) {
	    for (i=0; i<MAX_NO_OF_3D_SOURCE; i++) {
                v3d_1st_order_reflection=TRUE;
                v3d_source[i].no_of_1st_reflect=0;
		v3d_source[i].no_of_2nd_reflect=0;
	        if ( v3d_source[i].play || ( v3d_source_select && ( i == v3d_source_select_idx))) {
                    for(j=0; j<v3d_no_of_obj; j++) {
                        reflect_rect(i,&v3d_obj[j]);
                    };
                    if ( v3d_obstruction_circular ) {
			reflect_by_outer_circle(i,v3d_obstruction_radius);
		    };
		    if ( v3d_reflection.flag == 1 ) {

	                reflect_rc1=NULL;                    
			reflect_rc2=NULL;
        	        reflect_x(i, 1.0,-1.0,1.0);
			reflect_x(i,-1.0,-1.0,1.0);
			reflect_y(i, 1.0,-1.0,1.0);
			reflect_y(i,-1.0,-1.0,1.0);

			reflect_xx(i,-1.0,-1.0,1.0, 1.0,-1.0,1.0);
			reflect_yy(i,-1.0,-1.0,1.0, 1.0,-1.0,1.0);

			reflect_xy(i, 1.0,-1.0,1.0, 1.0,-1.0,1.0);
			reflect_xy(i, 1.0,-1.0,1.0,-1.0,-1.0,1.0);
			reflect_xy(i,-1.0,-1.0,1.0, 1.0,-1.0,1.0);
			reflect_xy(i,-1.0,-1.0,1.0,-1.0,-1.0,1.0);

                    }
		    else {
			if ( v3d_reflection.flag == 2 )
			    reflect_by_inner_circle(i,1.0);
		    };


                    v3d_1st_order_reflection=FALSE;
                    if ( v3d_2nd_order_reflection ) {
                        second_order_reflection(i);
		    };

                    reflect_rc1=NULL;
		    reflect_rc2=NULL;
		    if ( direct_path(v3d_source[i].x,v3d_source[i].y,v3d_listener.x,v3d_listener.y) ) { 
                        if ( v3d_wavetracing == 0 ) {
     			    gdk_draw_line(widget->window,pen3DSource[i],
				v3d_ox+v3d_m2*v3d_source[i].x,v3d_oy-v3d_m2*v3d_source[i].y,
				v3d_ox+v3d_m2*v3d_listener.x,v3d_oy-v3d_m2*v3d_listener.y);
			};
                    };
		};				
	    };	
        };
	if ( v3d_wavetracing > 0 ) {
            v3d_1st_order_reflection=0;
	    for (i=0; i<MAX_NO_OF_3D_SOURCE; i++) {
		if (( v3d_source[i].play ) || ( i == v3d_source_select_idx )) {
		    wavetracing(i);
		};
	    };
	};
        for (i=0; i<MAX_NO_OF_3D_SOURCE; i++) {
            if ( v3d_source[i].z <= v3d_listener.z ) {
	       size=5+((v3d_source[i].z+1)*4);
               gdk_draw_rectangle(widget->window,pen3DSource[i],TRUE,
       	            ox+m*v3d_source[i].x/2-((size-1)/2),
       	            oy-m*v3d_source[i].y/2-((size-1)/2),
       	            size,size);
             };
        };
	gdk_draw_line(widget->window,penGrey,
                ox+m*v3d_listener.x/2,
                oy-m*v3d_listener.y/2,
	        ox+m*v3d_listener.x/2+m*cos(v3d_listener.direction)*cos(v3d_listener.elevation)/8,
                oy-m*v3d_listener.y/2-m*sin(v3d_listener.direction)*cos(v3d_listener.elevation)/8);
        gdk_draw_rectangle(widget->window,penGrey,TRUE,
                (widget->allocation.width+m*v3d_listener.x)/2-4,
                (widget->allocation.height-m*v3d_listener.y)/2-4,
                9,9);
        for (i=0; i<MAX_NO_OF_3D_SOURCE; i++) {
            if ( v3d_source[i].z > v3d_listener.z ) {
	       size=5+((v3d_source[i].z+1)*4);
               gdk_draw_rectangle(widget->window,pen3DSource[i],TRUE,
       	            ox+m*v3d_source[i].x/2-((size-1)/2),
       	            oy-m*v3d_source[i].y/2-((size-1)/2),
       	            size,size);
             };
        };
}

void redraw_listener_pos(void)
{
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(v3d_listener_x_spin),v3d_listener.x);
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(v3d_listener_y_spin),v3d_listener.y);
}

void calc_source_pos(int i)
{
	char buffer[32];
	sprintf(buffer,"%.3f",v3d_source[i].x*v3d_scaleup);
	gtk_label_set_text(GTK_LABEL(v3d_x_pos_label[i]),buffer);
	sprintf(buffer,"%.3f",v3d_source[i].y*v3d_scaleup);
	gtk_label_set_text(GTK_LABEL(v3d_y_pos_label[i]),buffer);
	sprintf(buffer,"%.3f",v3d_source[i].z*v3d_scaleup);
	gtk_label_set_text(GTK_LABEL(v3d_z_pos_label[i]),buffer);
#ifdef HAVE_LIBOPENAL
	alSource3f(source[i], AL_POSITION, v3d_source[i].x*v3d_scaleup, v3d_source[i].y*v3d_scaleup, v3d_source[i].z*v3d_scaleup);
#endif
}

void update_source_spin(int i)
{
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(v3d_x_spin[i]),v3d_source[i].x);
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(v3d_y_spin[i]),v3d_source[i].y);
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(v3d_z_spin[i]),v3d_source[i].z);
}

void redraw_v3d_drawing(void)
{
        if ( GTK_WIDGET_VISIBLE(v3d_drawing) ) {
		    redraw_listener_pos();
		    redraw_v3d();
		    gdk_draw_rectangle(v3d_pixmap,
			   v3d_drawing->style->black_gc,
			   TRUE,
			   0, 0,
			   v3d_drawing->allocation.width,
			   v3d_drawing->allocation.height);
	};	
}

gint v3d_update_callback(gpointer data)
{
     int i;
#ifdef HAVE_LIBOPENAL
     ALint  error;
#endif
     for (i=0; i<MAX_NO_OF_3D_SOURCE; i++) {
        switch ( v3d_source[i].move ) {
	case 1:
	    v3d_source[i].angle+=v3d_source[i].speed;
	    v3d_source[i].x=cos(v3d_source[i].angle)*v3d_source[i].x_range;
	    v3d_source[i].y=sin(v3d_source[i].angle)*v3d_source[i].y_range;
#ifdef HAVE_LIBOPENAL
	    if ( v3d_source[i].speed >= 0 ) {
                v3d_source[i].direction=v3d_source[i].angle+PI/2;
                alSource3f(source[i], AL_VELOCITY, -sin(v3d_source[i].angle)*v3d_scaleup, cos(v3d_source[i].angle)*v3d_scaleup, 0.0);
	    }
            else {
                v3d_source[i].direction=v3d_source[i].angle-PI/2;
                alSource3f(source[i], AL_VELOCITY, sin(v3d_source[i].angle)*v3d_scaleup, -cos(v3d_source[i].angle)*v3d_scaleup, 0.0);
            }
	    if ((error = alGetError()) != AL_NO_ERROR)
		DisplayALError((ALbyte *) "alSource3f AL_VELOCITY : \n", error);
#endif
	    break;
	case 2:
	    if ( v3d_source[i].dir ) {
		v3d_source[i].p += v3d_source[i].increment;
		if ( v3d_source[i].p >= v3d_source[i].p_range ) {
		    v3d_source[i].p=v3d_source[i].p_range;
		    v3d_source[i].dir=0;
		};			
	    }
	    else {
		v3d_source[i].p -=v3d_source[i].increment;
		if ( v3d_source[i].p <= (-v3d_source[i].p_range) ) {
		    v3d_source[i].p =(-v3d_source[i].p_range);
		    v3d_source[i].dir=1;
		};
	    };
	    v3d_source[i].x=v3d_source[i].p*cos(v3d_source[i].o);
	    v3d_source[i].y=v3d_source[i].p*sin(v3d_source[i].o);
#ifdef HAVE_LIBOPENAL
            if ( v3d_source[i].dir > 0 ) {
		v3d_source[i].direction=v3d_source[i].o;
  	        alSource3f(source[i], AL_VELOCITY, v3d_source[i].increment*cos(v3d_source[i].o)*v3d_scaleup,v3d_source[i].increment*sin(v3d_source[i].o)*v3d_scaleup, 0.0);
            }
	    else {
		v3d_source[i].direction=v3d_source[i].o+PI;
   	        alSource3f(source[i], AL_VELOCITY, -v3d_source[i].increment*cos(v3d_source[i].o)*v3d_scaleup,-v3d_source[i].increment*sin(v3d_source[i].o)*v3d_scaleup, 0.0);
            }
	    if ((error = alGetError()) != AL_NO_ERROR)
		DisplayALError((ALbyte *) "alSource3f AL_VELOCITY : \n", error);
#endif
	    break;
	case 3:
	    v3d_source[i].oh_angle+=v3d_source[i].oh_speed;
	    v3d_source[i].x=cos(v3d_source[i].oh_angle)*cos(v3d_source[i].o);
	    v3d_source[i].y=cos(v3d_source[i].oh_angle)*sin(v3d_source[i].o);
            v3d_source[i].z=sin(v3d_source[i].oh_angle);
#ifdef HAVE_LIBOPENAL
	    if ( v3d_source[i].oh_speed >= 0 ) {
	        alSource3f(source[i], AL_VELOCITY, -v3d_source[i].z*cos(v3d_source[i].o)*v3d_scaleup, -v3d_source[i].z*sin(v3d_source[i].o)*v3d_scaleup, cos(v3d_source[i].oh_angle)*v3d_scaleup);
	    }
	    else {
	        alSource3f(source[i], AL_VELOCITY, v3d_source[i].z*cos(v3d_source[i].o)*v3d_scaleup, v3d_source[i].z*sin(v3d_source[i].o)*v3d_scaleup, -cos(v3d_source[i].oh_angle)*v3d_scaleup);
            }
	    if ((error = alGetError()) != AL_NO_ERROR)
		DisplayALError((ALbyte *) "alSource3f AL_VELOCITY : \n", error);

            if ( v3d_source[i].z < 0 ) 
 	        v3d_source[i].direction=v3d_source[i].o;
            else
 	        v3d_source[i].direction=PI+v3d_source[i].o;
#endif
 	    break;
        };
#ifdef HAVE_LIBOPENAL
	if ( v3d_source[i].move ) {
	    while ( v3d_source[i].direction >= (PI+PI) )
		v3d_source[i].direction-=(PI+PI);
	    while ( v3d_source[i].direction < 0.0 )
		v3d_source[i].direction+=(PI+PI);
            alSource3f(source[i], AL_DIRECTION, cos(v3d_source[i].direction), sin(v3d_source[i].direction), 0.0);
	    if ((error = alGetError()) != AL_NO_ERROR)
		    DisplayALError((ALbyte *) "alSource3f AL_DIRECTION : \n", error);
	};
#endif
        calc_source_pos(i);
    };
    redraw_v3d_drawing();
    return no_of_3d_move;
}

void v3d_w_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	int idx,*p;
	p=data;
	idx=*p;
	v3d_source[idx].x_range=adj->value;
	v3d_source[idx].x=cos(v3d_source[idx].angle)*v3d_source[idx].x_range;
	calc_source_pos(idx);
	redraw_v3d_drawing();
}

void v3d_h_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	int idx,*p;
	p=data;
	idx=*p;
	v3d_source[idx].y_range=adj->value;
	v3d_source[idx].y=sin(v3d_source[idx].angle)*v3d_source[idx].y_range;
	calc_source_pos(idx);
	redraw_v3d_drawing();
}

void v3d_x_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	int idx,*p;
	p=data;
	idx=*p;
	v3d_source[idx].x=adj->value;
        calc_source_pos(idx);
	redraw_v3d_drawing();
}

void v3d_y_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	int idx,*p;
	p=data;
	idx=*p;
	v3d_source[idx].y=adj->value;
        calc_source_pos(idx);
	redraw_v3d_drawing();
}

void v3d_z_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	int idx,*p;
	p=data;
	idx=*p;
	v3d_source[idx].z=adj->value;
        calc_source_pos(idx);
	redraw_v3d_drawing();
}

void v3d_speed_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	int idx,*p;
	p=data;
	idx=*p;
	v3d_source[idx].speed=adj->value;
	redraw_v3d_drawing();
}

void v3d_r_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	int idx,*p;
	p=data;
	idx=*p;
	v3d_source[idx].p_range=adj->value;
	redraw_v3d_drawing();
}

void v3d_o_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	int idx,*p;
	p=data;
	idx=*p;

	v3d_source[idx].o=adj->value;
        v3d_source[idx].x=v3d_source[idx].p*cos(v3d_source[idx].o);
        v3d_source[idx].y=v3d_source[idx].p*sin(v3d_source[idx].o);
	calc_source_pos(idx);
	redraw_v3d_drawing();
}

void v3d_v_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	int idx,*p;
	p=data;
	idx=*p;
	v3d_source[idx].increment=adj->value;
	redraw_v3d_drawing();
}

void v3d_oh_o_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	int idx,*p;
	p=data;
	idx=*p;
	v3d_source[idx].o=adj->value;
	v3d_source[idx].x=cos(v3d_source[idx].oh_angle)*cos(v3d_source[idx].o);
	v3d_source[idx].y=cos(v3d_source[idx].oh_angle)*sin(v3d_source[idx].o);
	calc_source_pos(idx);
	redraw_v3d_drawing();
}

void v3d_oh_speed_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	int idx,*p;
	p=data;
	idx=*p;
	v3d_source[idx].oh_speed=adj->value;
	redraw_v3d_drawing();
}

void v3d_cone_inner_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	int idx,*p;
	p=data;
	idx=*p;
	if ( adj->value <= v3d_source[idx].cone_outer_angle ) {
		v3d_source[idx].cone_inner_angle=adj->value;	
		gtk_spin_button_set_range(GTK_SPIN_BUTTON(v3d_cone_outer_spin[idx]),adj->value,360.0);		
	};
}

void v3d_cone_outer_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	int idx,*p;
	p=data;
	idx=*p;
	v3d_source[idx].cone_outer_angle=adj->value;	
	redraw_v3d_drawing();
}

#ifdef HAVE_LIBOPENAL
void update_openal_listener(void)
{
	ALint	error;
	alListener3f(AL_POSITION,v3d_listener.x*v3d_scaleup,v3d_listener.y*v3d_scaleup,v3d_listener.z*v3d_scaleup);
	if ((error = alGetError()) != AL_NO_ERROR )
		DisplayALError("alListener3f AL_POSITION : ", error);
}
#endif
void v3d_listener_x_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	v3d_listener.x=adj->value;
#ifdef HAVE_LIBOPENAL
	update_openal_listener();
#endif
	redraw_v3d_drawing();
}

void v3d_listener_y_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	v3d_listener.y=adj->value;
#ifdef HAVE_LIBOPENAL
	update_openal_listener();
#endif
	redraw_v3d_drawing();
}

void v3d_listener_z_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	v3d_listener.z=adj->value;
#ifdef HAVE_LIBOPENAL
	update_openal_listener();
#endif
	redraw_v3d_drawing();
}

void v3d_listener_direction_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
#ifdef HAVE_LIBOPENAL
	ALint error;
        ALfloat listenerOri[]={0.0,1.0,0.0, 0.0,0.0,1.0};
#endif
	v3d_listener.direction=adj->value;
#ifdef HAVE_LIBOPENAL
	listenerOri[0]=cos(v3d_listener.direction);
	listenerOri[1]=sin(v3d_listener.direction);
	alListenerfv(AL_ORIENTATION,listenerOri);
	if ((error = alGetError()) != AL_NO_ERROR)
		DisplayALError((ALbyte *) "alListenerfv ORIENTATION : ", error);
#endif
	redraw_v3d_drawing();
}

void v3d_wavelength_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	v3d_wavelength=adj->value;
	redraw_v3d_drawing();
}

/*
void v3d_listener_speed_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	double change;
	
	change=adj->value-v3d_listener.speed;
	v3d_listener.x += ( change * cos(v3d_listener.elevation) * cos(v3d_listener.direction));
	if ( v3d_listener.x < -1.0 )
		v3d_listener.x=-1.0;
	if ( v3d_listener.x > 1.0 )
		v3d_listener.x=1.0;

	v3d_listener.y += ( change * cos(v3d_listener.elevation) * sin(v3d_listener.direction));
	if ( v3d_listener.y < -1.0 )
		v3d_listener.y=-1.0;
	if ( v3d_listener.y > 1.0 )
		v3d_listener.y=1.0;

	v3d_listener.z += ( change * sin(v3d_listener.elevation));
	if ( v3d_listener.z < -1.0 )
		v3d_listener.z=-1.0;
	if ( v3d_listener.z > 1.0 )
		v3d_listener.z=1.0;
 
	v3d_listener.speed=adj->value;
	redraw_v3d_drawing();
}

void v3d_listener_elevation_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	v3d_listener.elevation=adj->value;
	redraw_v3d_drawing();
}
*/

void v3d_scaleup_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	int idx;
	v3d_scaleup=adj->value;
	for (idx=0; idx<MAX_NO_OF_3D_SOURCE; idx++)
	    calc_source_pos(idx);
	redraw_v3d_drawing();
}

void v3d_reflection_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	redraw_v3d_drawing();
}

void v3d_reverb_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	redraw_v3d_drawing();
}

void v3d_obstruction_radius_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	v3d_obstruction_radius=adj->value;
        if ( v3d_obstruction_radius > v3d_obstruction_min_radius ) 
		v3d_obstruction_circular=1;
	else
		v3d_obstruction_circular=0;
	redraw_v3d_drawing();
}

void setup_object(void)
{
     double outer,thick;
        outer=v3d_obstruction_size;
        thick=0.05;

	v3d_occlusion[0].rc.left=-outer;
	v3d_occlusion[0].rc.right=-outer+thick;
	v3d_occlusion[0].rc.bottom=-outer;
	v3d_occlusion[0].rc.top=outer;

	v3d_occlusion[1].rc.left=outer-thick;
	v3d_occlusion[1].rc.right=outer;
	v3d_occlusion[1].rc.bottom=-outer;
	v3d_occlusion[1].rc.top=outer;

	v3d_occlusion[2].rc.left=-outer;
	v3d_occlusion[2].rc.right=outer;
	v3d_occlusion[2].rc.bottom=outer-thick;
	v3d_occlusion[2].rc.top=outer;

	v3d_occlusion[3].rc.left=-outer;
	v3d_occlusion[3].rc.right=outer;
	v3d_occlusion[3].rc.bottom=-outer;
	v3d_occlusion[3].rc.top=-outer+thick;

	v3d_obstruction[0].rc.left=-outer;
	v3d_obstruction[0].rc.right=-outer+thick;
	v3d_obstruction[0].rc.bottom=-0.035;
	v3d_obstruction[0].rc.top=0.035;

	v3d_obstruction[1].rc.left=outer-thick;
	v3d_obstruction[1].rc.right=outer;
	v3d_obstruction[1].rc.bottom=-0.035;
	v3d_obstruction[1].rc.top=0.035;

	v3d_obstruction[2].rc.left=-0.035;
	v3d_obstruction[2].rc.right=0.035;
	v3d_obstruction[2].rc.bottom=outer-thick;
	v3d_obstruction[2].rc.top=outer;

	v3d_obstruction[3].rc.left=-0.035;
	v3d_obstruction[3].rc.right=0.035;
	v3d_obstruction[3].rc.bottom=-outer;
	v3d_obstruction[3].rc.top=-outer+thick;


	v3d_exclusion[0].rc[0].left=-outer;
	v3d_exclusion[0].rc[0].right=-outer+thick;
	v3d_exclusion[0].rc[0].bottom=-outer;
	v3d_exclusion[0].rc[0].top=-0.025;

	v3d_exclusion[0].rc[1].left=-outer;
	v3d_exclusion[0].rc[1].right=-outer+thick;
	v3d_exclusion[0].rc[1].bottom=0.025;
	v3d_exclusion[0].rc[1].top=outer;

	v3d_exclusion[1].rc[0].left=outer-thick;
	v3d_exclusion[1].rc[0].right=outer;
	v3d_exclusion[1].rc[0].bottom=-outer;
	v3d_exclusion[1].rc[0].top=-0.025;

	v3d_exclusion[1].rc[1].left=outer-thick;
	v3d_exclusion[1].rc[1].right=outer;
	v3d_exclusion[1].rc[1].bottom=0.025;
	v3d_exclusion[1].rc[1].top=outer;


	v3d_exclusion[2].rc[0].left=-outer;
	v3d_exclusion[2].rc[0].right=-0.025;
	v3d_exclusion[2].rc[0].bottom=outer-thick;
	v3d_exclusion[2].rc[0].top=outer;

	v3d_exclusion[2].rc[1].left=0.025;
	v3d_exclusion[2].rc[1].right=outer;
	v3d_exclusion[2].rc[1].bottom=outer-thick;
	v3d_exclusion[2].rc[1].top=outer;

	v3d_exclusion[3].rc[0].left=-outer;
	v3d_exclusion[3].rc[0].right=-0.025;
	v3d_exclusion[3].rc[0].bottom=-outer;
	v3d_exclusion[3].rc[0].top=-outer+thick;

	v3d_exclusion[3].rc[1].left=0.025;
	v3d_exclusion[3].rc[1].right=outer;
	v3d_exclusion[3].rc[1].bottom=-outer;
	v3d_exclusion[3].rc[1].top=-outer+thick;
        calc_no_of_obj();
}

void v3d_obstruction_size_adj_value_changed_cb (GtkAdjustment *adj, gpointer data)
{
	v3d_obstruction_size=adj->value;
	setup_object();
	redraw_v3d_drawing();
}

gint v3d_configure_event(GtkWidget *widget, GdkEventConfigure *event)
{
	if (v3d_pixmap != NULL)
		gdk_pixmap_unref(v3d_pixmap);
	v3d_pixmap = gdk_pixmap_new(widget->window,
				    widget->allocation.width,
				    widget->allocation.height,
				    -1);
	gdk_draw_rectangle(v3d_pixmap,
			   widget->style->black_gc,
			   TRUE,
			   0, 0,
			   widget->allocation.width,
			   widget->allocation.height);
	redraw_v3d(); 
	return TRUE;
}

gint v3d_expose_event(GtkWidget *widget, GdkEventExpose *event)
{
	gdk_window_clear_area (widget->window,
                         event->area.x, event->area.y,
                         event->area.width, event->area.height);
  	gdk_gc_set_clip_rectangle (widget->style->fg_gc[widget->state],&event->area);
	redraw_v3d_drawing();
  	gdk_gc_set_clip_rectangle (widget->style->fg_gc[widget->state],NULL);
	return TRUE;
}

void recalc_move_parameter(int idx,int i)
{
        switch(i) {
        case 0:
		update_source_spin(idx);
                break;
	case 1:
                v3d_source[idx].angle=v3d_source[idx].o;
		break;
        case 2:
                gtk_spin_button_set_value(GTK_SPIN_BUTTON(v3d_o_spin[idx]),v3d_source[idx].o);
		break;
	case 3:
                gtk_spin_button_set_value(GTK_SPIN_BUTTON(v3d_oh_o_spin[idx]),v3d_source[idx].o);
		break;
        };
}

gint v3d_button_press_event( GtkWidget *widget, GdkEventButton *event )
{
      int idx;
      int x1,y1;
      if (( event->button == 1 ) || ( event->button == 3 )) {
	    v3d_source_select=0;
	    if ( event->button == 1 ) {
	        v3d_listener_select=0;
                v3d_raytracing=1;
            };
	    if ( event->button == 3 ) 
		v3d_2nd_order_reflection=TRUE;
	    x1= v3d_ox+v3d_m2*v3d_listener.x;
	    if ( ( (x1-4) <= event->x ) && ( event->x <= (x1+4) )) { 
	        y1= v3d_oy-v3d_m2*v3d_listener.y;
		    if ( ( (y1-4) <= event->y ) && ( event->y <= (y1+4) )) { 
			v3d_listener_select=1;
                    };
	    };
	    if ( ! v3d_listener_select ) {
 	        for (idx=0; idx<MAX_NO_OF_3D_SOURCE; idx++) {
                    if ( v3d_source[idx].move == 0 ) {		
			x1=v3d_ox+v3d_m2*v3d_source[idx].x;
			if ( ( (x1-4) <= event->x ) && ( event->x <= (x1+4) )) { 
			    y1=v3d_oy-v3d_m2*v3d_source[idx].y;
			    if ( ( (y1-4) <= event->y ) && ( event->y <= (y1+4) )) { 
				v3d_source_select=1;
				v3d_source_select_idx=idx;
		                gtk_notebook_set_current_page(GTK_NOTEBOOK(v3d_nb),idx);
			    };
		        };
		    };
		};
	    };
	    redraw_v3d_drawing();
	};
        return TRUE;
}

gint v3d_button_release_event( GtkWidget *widget, GdkEventButton *event )
{
        if (( event->button == 1 ) || ( event->button == 3 )) {
            if ( ! gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(v3d_wavetracing_button)) ) 
 	        v3d_raytracing=0;
	    if (((v3d_ox - v3d_m2) <= event->x ) && ( event->x <= (v3d_ox + v3d_m2)) ) {
	        if (((v3d_oy - v3d_m2) <= event->y ) && ( event->y <= (v3d_oy + v3d_m2)) ) {
	            if ( v3d_listener_select ) {
                        v3d_listener.x= (event->x-v3d_ox)/v3d_m2;
			v3d_listener.y= (v3d_oy-event->y)/v3d_m2;
			redraw_listener_pos();
		    };
		    if ( v3d_source_select ) {
		        v3d_source[v3d_source_select_idx].x= (event->x-v3d_ox)/v3d_m2;
			v3d_source[v3d_source_select_idx].y= (v3d_oy-event->y)/v3d_m2;
                        calc_source_pos(v3d_source_select_idx);
                        update_source_spin(v3d_source_select_idx);
		    };
		};
	    };
	    v3d_source_select=0;
	    v3d_listener_select=0;
	    if ( event->button == 3 ) 
		v3d_2nd_order_reflection=FALSE;
	};	
        redraw_v3d_drawing();
        return TRUE;
}

gint v3d_motion_notify_event( GtkWidget *widget, GdkEventMotion *event )
{
 	if (((v3d_ox - v3d_m2) <= event->x ) && ( event->x <= (v3d_ox + v3d_m2)) ) {
  	    if (((v3d_oy - v3d_m2) <= event->y ) && ( event->y <= (v3d_oy + v3d_m2)) ) {
		if ( v3d_listener_select ) {
                    v3d_listener.x= (event->x-v3d_ox)/v3d_m2;
		    v3d_listener.y= (v3d_oy-event->y)/v3d_m2;    
		    redraw_listener_pos();
		};
		if ( v3d_source_select ) {
		    if ( v3d_source[v3d_source_select_idx].move == 0 ) {
	                v3d_source[v3d_source_select_idx].x= (event->x-v3d_ox)/v3d_m2;
		        v3d_source[v3d_source_select_idx].y= (v3d_oy-event->y)/v3d_m2;
		    };
		};
                redraw_v3d_drawing();
	    };  
	};
        return TRUE;
}

void v3d_move_toggle_button_cb (GtkWidget *widget, gpointer data)
{
	int i;
	int idx,*p;
	p=data;
	idx=*p;
     	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
		v3d_source[idx].move=v3d_source[idx].move_type;
                if ( v3d_source[idx].move ) {
			if ( no_of_3d_move == 0 ) {
		        	v3d_timer_tag=gtk_timeout_add(50, v3d_update_callback, NULL);
			};
			no_of_3d_move++;
		        for (i=0; i<MAX_NO_OF_MOVE_TYPE; i++) {
			    gtk_widget_set_sensitive(v3d_move_type_radio[idx][i],FALSE);
			};
		}
		else {
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widget),FALSE);
		};
    	} 
	else {
                if ( v3d_source[idx].move ) {
	    		v3d_source[idx].move=0;	
			no_of_3d_move--;
			if ( no_of_3d_move == 0 ) 
				gtk_timeout_remove(v3d_timer_tag);
		        for (i=0; i<MAX_NO_OF_MOVE_TYPE; i++) 
			    gtk_widget_set_sensitive(v3d_move_type_radio[idx][i],TRUE);
		};
    	}
}

void v3d_showpath_toggle_button_cb (GtkWidget *widget, gpointer data)
{
        if ( gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)) ) 
		v3d_showpath=1;
	else
		v3d_showpath=0;
	redraw_v3d_drawing();
}

void v3d_wavetracing_toggle_button_cb (GtkWidget *widget, gpointer data)
{
        if ( gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)) ) {
		v3d_raytracing=1;
		v3d_wavetracing=1;
        }
	else {
		v3d_raytracing=1;
		v3d_wavetracing=0;
        };
	redraw_v3d_drawing();
}

void v3d_move_type_toggle_button_cb (GtkWidget *widget, gpointer data)
{
        int i;
	int idx,*p;
	p=data;
	idx=*p;
        if ( v3d_source[idx].move == 0 ) {
	    for (i=0;i<MAX_NO_OF_MOVE_TYPE; i++) 
        	if ( gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(v3d_move_type_radio[idx][i])) ) {
		    v3d_source[idx].move_type=i;
		    gtk_widget_set_sensitive(v3d_move_button[idx],v3d_source[idx].move_type > 0);
                    recalc_move_parameter(idx,i);
		    gtk_notebook_set_current_page(GTK_NOTEBOOK(v3d_move_nb[idx]),i);
	        };
 	}
	else {
            gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(v3d_move_type_radio[idx][v3d_source[idx].move_type]),TRUE);
	};
	calc_source_pos(idx);
}

void v3d_loop_button_toggled_cb (GtkWidget *widget, gpointer data)
{
	int idx,*p;
	p=data;
	idx=*p;
     	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) 
		v3d_wave[idx].loop=TRUE;
	else 
    		v3d_wave[idx].loop=FALSE;	
#ifdef HAVE_LIBOPENAL
	alSourcei(source[idx], AL_LOOPING, v3d_wave[idx].loop );
#endif
}

void v3d_filesel_ok_sel(GtkWidget *widget, GtkFileSelection *fs )
{
#ifdef HAVE_LIBOPENAL
	ALint	error;
	ALsizei size,freq;
	ALenum	format;
	ALvoid	*data;
        ALsizei bits;
#endif
	WaveFile w;
	GtkWidget *message_dlg;
	char filename[FILENAME_MAX+1];
        strcpy(filename,gtk_file_selection_get_filename(GTK_FILE_SELECTION(fs)));
	if ( is_wav_file(&w,filename) ) {
            if ( w.channels == 1 ) {
	        strcpy(v3d_source[v3d_fs_idx].filename,filename);

#ifdef HAVE_LIBOPENAL
        	alutLoadWAV(filename, &data, &format, &size, & bits, &freq);
		if ((error = alGetError()) != AL_NO_ERROR) 
			DisplayALError((ALbyte *) "alutLoadWAVFile : ", error);
		alBufferData(g_Buffers[v3d_fs_idx],format,data,size,freq);
		if ((error = alGetError()) != AL_NO_ERROR) 
			DisplayALError((ALbyte *) "alBufferData : ", error);
	        alutUnloadWAV(format,data,size,freq);
		if ((error = alGetError()) != AL_NO_ERROR) 
			DisplayALError((ALbyte *) "alutUnloadWAV : ", error);
#endif
	        gtk_widget_set_sensitive(v3d_play_button[v3d_fs_idx],TRUE);
	        gtk_widget_set_sensitive(v3d_loop_button[v3d_fs_idx],TRUE);
	        gtk_widget_destroy(GTK_WIDGET(fs));
            }
	    else {
		message_dlg=gtk_message_dialog_new (GTK_WINDOW(fs),
                                        GTK_DIALOG_DESTROY_WITH_PARENT,
                                        GTK_MESSAGE_ERROR,
                                        GTK_BUTTONS_OK,
                                        "Only mono sound can be spatialized");
                gtk_widget_show(message_dlg);       
                g_signal_connect_swapped (GTK_OBJECT (message_dlg), 
                             "response", 
                             G_CALLBACK (gtk_widget_destroy), GTK_OBJECT (message_dlg)); 
	    };		
 	};
}

void v3d_wave_button_clicked_cb (GtkWidget *widget, gpointer data)
{
	GtkWidget *filesel;
	int idx,*p;
	p=data;
	idx=*p;
	v3d_fs_idx=idx;
        filesel=gtk_file_selection_new("Mono Sound");
        gtk_widget_show(filesel);
        gtk_file_selection_hide_fileop_buttons(GTK_FILE_SELECTION(filesel));
        gtk_file_selection_set_select_multiple(GTK_FILE_SELECTION(filesel),FALSE);
	gtk_file_selection_complete(GTK_FILE_SELECTION(filesel),"*.wav");

        g_signal_connect (G_OBJECT (GTK_FILE_SELECTION (filesel)->ok_button),
		      "clicked", G_CALLBACK (v3d_filesel_ok_sel), (gpointer) filesel);
    
        g_signal_connect_swapped (G_OBJECT (GTK_FILE_SELECTION (filesel)->cancel_button),
	              "clicked", G_CALLBACK (gtk_widget_destroy), G_OBJECT (filesel));	     
}

void v3d_play_button_clicked_cb (GtkWidget *widget, gpointer data)
{
	int idx,*p;
	p=data;
	idx=*p;
	v3d_source[idx].play=1;
	g_print("Source %d playing...\n",idx+1);
        strcpy(v3d_wave[idx].filename,v3d_source[idx].filename);
        v3d_wave[idx].v3d=TRUE; 
        v3d_wave[idx].loop=gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(v3d_loop_button[idx]));
	v3d_wave[idx].button_type=3;
        v3d_wave[idx].idx=idx;
	gtk_widget_set_sensitive(v3d_play_button[idx],FALSE);
	gtk_widget_set_sensitive(v3d_stop_button[idx],TRUE);
#ifdef HAVE_LIBOPENAL

	alSourcei(source[idx], AL_BUFFER, g_Buffers[idx]);
	alSource3f(source[idx], AL_POSITION, v3d_source[idx].x,v3d_source[idx].y,v3d_source[idx].z);	
	alSourcei(source[idx], AL_LOOPING, v3d_wave[idx].loop );
	alSourcePlay(source[idx]);
	g_print("openal source %d\n",idx);
#else
        play_wav_file(&v3d_wave[idx]);	
#endif	
        no_of_3d_play++;
	redraw_v3d_drawing();
}

void v3d_stop_button_clicked_cb (GtkWidget *widget, gpointer data)
{
	int idx,*p;
	p=data;
	idx=*p;
	g_print("Source %d stop..\n",idx+1);
        no_of_3d_play--; 
 	v3d_source[idx].play=0;	
	gtk_widget_set_sensitive(v3d_play_button[idx],TRUE);
	gtk_widget_set_sensitive(v3d_stop_button[idx],FALSE);
#ifdef HAVE_LIBOPENAL
	alSourceStop(source[idx]);
#else
        stop_play(&v3d_wave[idx]);
#endif
	redraw_v3d_drawing();
}

void v3d_occlusion_toggle_button_cb (GtkWidget *widget, gpointer data)
{
	int idx,*p;
	p=data;
	idx=*p;
     	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) 
		v3d_occlusion[idx].flag=1;
	else 
    		v3d_occlusion[idx].flag=0;	
        calc_no_of_obj();
	redraw_v3d_drawing();
}

void v3d_exclusion_toggle_button_cb (GtkWidget *widget, gpointer data)
{
	int idx,*p;
	p=data;
	idx=*p;
     	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) 
		v3d_exclusion[idx].flag=1;
	else 
    		v3d_exclusion[idx].flag=0;	
        calc_no_of_obj();
	redraw_v3d_drawing();
}

void v3d_obstruction_toggle_button_cb (GtkWidget *widget, gpointer data)
{
	int idx,*p;
	p=data;
	idx=*p;
     	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) 
		v3d_obstruction[idx].flag=1;
	else 
    		v3d_obstruction[idx].flag=0;	
        calc_no_of_obj();	
	redraw_v3d_drawing();
}

void v3d_reflection_type_changed_cb (GtkWidget *widget, gpointer data)
{
	char select_item[32];
	int idx;
	strcpy(select_item,gtk_entry_get_text (GTK_ENTRY(widget)));
	for (idx=0; idx<3; idx++) 
            if ( strcmp(v3d_reflection_type[idx],select_item) == 0 ) {
		v3d_reflection.flag=idx;
	        redraw_v3d_drawing();
		return;
            };
}

void v3d_reverb_toggle_button_cb (GtkWidget *widget, gpointer data)
{
     	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) 
		v3d_reverb=1;
	else 
    		v3d_reverb=0;	
	redraw_v3d_drawing();
}
#ifdef HAVE_LIBOPENAL
void init_alsa_3d(void)
{
	ALint	error;
        ALfloat listenerOri[]={0.0,1.0,0.0, 0.0,0.0,1.0};
	int i;

	Device = alcOpenDevice(NULL); 	
	Context=alcCreateContext(Device,NULL);
	alcMakeContextCurrent(Context);

	alGenBuffers(NUM_BUFFERS, g_Buffers);
	if ((error = alGetError()) != AL_NO_ERROR) 
		DisplayALError((ALbyte *) "alGenBuffers :", error);
	alGenSources(MAX_NO_OF_3D_SOURCE, source);
	if ((error = alGetError()) != AL_NO_ERROR) 
		DisplayALError((ALbyte *) "alGenSources :", error);

	for (i = 0; i < MAX_NO_OF_3D_SOURCE; i++) {
		alSource3f(source[i],AL_POSITION,v3d_source[i].x,v3d_source[i].y,v3d_source[i].z);
		if ((error = alGetError()) != AL_NO_ERROR) 
			DisplayALError((ALbyte *) "alSourcefv AL_POSITION : \n", error);
		alSource3f(source[i], AL_VELOCITY, 0.0, 0.0, 0.0);
		if ((error = alGetError()) != AL_NO_ERROR)
			DisplayALError((ALbyte *) "alSource3f AL_VELOCITY : \n", error);
		alSourcef(source[i],AL_PITCH,1.0f);
		if ((error = alGetError()) != AL_NO_ERROR)
			DisplayALError((ALbyte *) "alSourcef AL_PITCH : \n", error);
		alSourcef(source[i],AL_GAIN,1.0f);
		if ((error = alGetError()) != AL_NO_ERROR)
			DisplayALError((ALbyte *) "alSourcef AL_GAIN : \n", error);
		alSourcei(source[i],AL_LOOPING,AL_FALSE);
		if ((error = alGetError()) != AL_NO_ERROR) 
			DisplayALError((ALbyte *) "alSourcei AL_LOOPING true: \n", error);
	}

	alListener3f(AL_POSITION,v3d_listener.x,v3d_listener.y,v3d_listener.z);
	if ((error = alGetError()) != AL_NO_ERROR)	
		DisplayALError("alListenerfv POSITION : ", error);
	alListener3f(AL_VELOCITY,0.0,0.0,0.0);
	if ((error = alGetError()) != AL_NO_ERROR)
		DisplayALError((ALbyte *) "alListenerfv VELOCITY : ", error);
	alListenerfv(AL_ORIENTATION,listenerOri);
	if ((error = alGetError()) != AL_NO_ERROR)
		DisplayALError((ALbyte *) "alListenerfv ORIENTATION : ", error);
	alGetError();

        v3d_init=1;
}

void shutdown_alsa_3d(void)
{
	ALint	error;

	alGetError(); 
	alDeleteBuffers(NUM_BUFFERS, g_Buffers);
	if ((error = alGetError()) != AL_NO_ERROR) 
		DisplayALError("alDeleteBuffers : \n", error);
	alDeleteSources(MAX_NO_OF_3D_SOURCE, source); 
	error = alGetError();
	if ((error = alGetError()) != AL_NO_ERROR)
		DisplayALError("alDeleteSources : \n", error);

	Context=alcGetCurrentContext();
	Device=alcGetContextsDevice(Context);
	alcDestroyContext(Context);
	alcCloseDevice(Device);

	g_print("Openal shutdown\n");
}
#endif
void user_change_page(GtkNotebook *notebook,
                      GtkNotebookPage *page,
                      gint page_num,
                      gpointer data)
{
/*
    if ( page_num == v3d_page ) {
    };
    if ( gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook)) == v3d_page ) {
    };
*/
}

static void create_3D(GtkWidget *main, GtkWidget *notebook, int page)
{
	GtkWidget *label;
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkWidget *frame;
	GtkWidget *table;
	GtkWidget *nb;
	GtkWidget *main_hbox;
	GtkWidget *move_vbox;
	GtkWidget *move_nb;
	GtkWidget *combo;
        GList *glist;
	int i,j;
	char buffer[128];

	v3d_scaleup=1.0;
	v3d_showpath=0;
	v3d_wavelength=0.1;

	v3d_raytracing=0;
	v3d_wavetracing=0;
	for(i=0; i<4; i++) {
		v3d_occlusion[i].flag=0;
		v3d_exclusion[i].flag=0;
		v3d_obstruction[i].flag=0;
	};
	v3d_obstruction_min_radius=0.05;
	v3d_obstruction_radius=v3d_obstruction_min_radius;
	v3d_reflection.flag=0;

	v3d_reflection.rc.left=-1.0;
	v3d_reflection.rc.right=1.0;
	v3d_reflection.rc.bottom=-1.0;
	v3d_reflection.rc.top=1.0;

	v3d_obstruction_size=0.5;

        setup_object();

        no_of_3d_play=0;
	no_of_3d_move=0;
	for (i=0;i<MAX_NO_OF_3D_SOURCE; i++) {
	    v3d_source[i].idx=i;
	    v3d_source[i].move=0;
	    v3d_source[i].move_type=0;             
	    v3d_source[i].play=0;
            v3d_source[i].no_of_1st_reflect=0;
            v3d_source[i].no_of_2nd_reflect=0;

	    v3d_source[i].x=cos(PI*initial_o[i]/8.0);
	    v3d_source[i].y=sin(PI*initial_o[i]/8.0);
	    v3d_source[i].z=0.0;
 	    v3d_source[i].speed=0.025;
	    v3d_source[i].angle=PI*initial_o[i]/8.0;
	    v3d_source[i].x_range=1.0;
	    v3d_source[i].y_range=1.0;
 	    v3d_source[i].p=1.0;
	    v3d_source[i].dir=0;
	    v3d_source[i].increment=0.025;
	    v3d_source[i].p_range=1.0;            
	    v3d_source[i].o=PI*initial_o[i]/8.0;
	    v3d_source[i].oh_angle=0.0;
            v3d_source[i].oh_speed=0.025;

	    v3d_source[i].direction=PI+PI*initial_o[i]/8.0;
	    v3d_source[i].cone_inner_angle=360.0;
	    v3d_source[i].cone_outer_angle=360.0;

	    strcpy(v3d_source[i].filename,"");
	};

// Listener ( Stay at the origin but can move around by mouse )

	v3d_listener.x=0.0;
	v3d_listener.y=0.0;
	v3d_listener.z=0.0;
	v3d_listener.direction=PI/2;
	v3d_listener.elevation=0.0;

#ifdef HAVE_LIBOPENAL
	init_alsa_3d();
#endif
	main_hbox = gtk_hbox_new(FALSE, 0);
	gtk_widget_show(main_hbox);
	gtk_container_add(GTK_CONTAINER(notebook), main_hbox);
	gtk_container_set_border_width(GTK_CONTAINER(main_hbox), 6);

        label = gtk_label_new("3D Sound");
        gtk_widget_show(label);
	gtk_notebook_set_tab_label(GTK_NOTEBOOK(notebook), 
				   gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), page), 
				   label);

        vbox = gtk_vbox_new(FALSE, 0);
        gtk_widget_show(vbox);
        gtk_container_add(GTK_CONTAINER(main_hbox), vbox);

	v3d_drawing=gtk_drawing_area_new();
        gtk_widget_set_name(v3d_drawing, "V3D");
	gtk_widget_set_usize(v3d_drawing, 207, 207);
	gtk_widget_show(v3d_drawing);
	gtk_container_add(GTK_CONTAINER(vbox), v3d_drawing);

        gtk_signal_connect(GTK_OBJECT(v3d_drawing), 
		  	"expose_event",
		        GTK_SIGNAL_FUNC(v3d_expose_event), NULL);
	gtk_signal_connect(GTK_OBJECT(v3d_drawing), 		
			"configure_event",
			GTK_SIGNAL_FUNC(v3d_configure_event), NULL);

        g_signal_connect (G_OBJECT (v3d_drawing), "motion_notify_event",
		    G_CALLBACK (v3d_motion_notify_event), NULL);
        g_signal_connect (G_OBJECT (v3d_drawing), "button_press_event",
		    G_CALLBACK (v3d_button_press_event), NULL);
        g_signal_connect (G_OBJECT (v3d_drawing), "button_release_event",
		    G_CALLBACK (v3d_button_release_event), NULL);

	gtk_widget_set_events(v3d_drawing, GDK_EXPOSURE_MASK
 		         | GDK_LEAVE_NOTIFY_MASK
			 | GDK_BUTTON_PRESS_MASK
                         | GDK_BUTTON_RELEASE_MASK
			 | GDK_POINTER_MOTION_MASK);
#ifdef HAVE_LIBOPENAL
	strcpy(buffer,"OpenAL Version : ");
	strcat(buffer,alGetString(AL_VERSION));
	strcat(buffer,"    Renderer : ");
	strcat(buffer,alGetString(AL_RENDERER));
	label=gtk_label_new(buffer);
        gtk_widget_show(label);
        gtk_container_add(GTK_CONTAINER(vbox), label);
#endif
	frame=gtk_frame_new("Listener");
	gtk_widget_show(frame);
	gtk_container_add(GTK_CONTAINER(vbox), frame);

	table = gtk_table_new(2,3,FALSE);
	gtk_widget_show(table);
	gtk_container_add(GTK_CONTAINER(frame), table);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_widget_show(hbox);
        gtk_table_attach_defaults(GTK_TABLE(table),hbox,0,1,0,1);		

        label=gtk_label_new("X");
        gtk_widget_show(label);
        gtk_container_add(GTK_CONTAINER(hbox), label);

	v3d_listener_x_adj=gtk_adjustment_new((gdouble)(v3d_listener.x), 	
				    (gdouble)(-1.0), 
				    (gdouble)(1.0), 
        	    		    (gdouble)0.001,(gdouble)0.001,(gdouble)0.001);

        v3d_listener_x_spin=gtk_spin_button_new(GTK_ADJUSTMENT(v3d_listener_x_adj),0.01,3);
	gtk_widget_show(v3d_listener_x_spin);
	gtk_container_add(GTK_CONTAINER(hbox), v3d_listener_x_spin);
	gtk_signal_connect (GTK_OBJECT (v3d_listener_x_adj), 
				"value_changed",
				GTK_SIGNAL_FUNC (v3d_listener_x_adj_value_changed_cb), 
				NULL);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_widget_show(hbox);
        gtk_table_attach_defaults(GTK_TABLE(table),hbox,1,2,0,1);		

        label=gtk_label_new("Y");
        gtk_widget_show(label);
        gtk_container_add(GTK_CONTAINER(hbox), label);

	v3d_listener_y_adj=gtk_adjustment_new((gdouble)(v3d_listener.y), 	
				    (gdouble)(-1.0), 
				    (gdouble)(1.0), 
        	    		    (gdouble)0.001,(gdouble)0.001,(gdouble)0.001);

        v3d_listener_y_spin=gtk_spin_button_new(GTK_ADJUSTMENT(v3d_listener_y_adj),0.01,3);
	gtk_widget_show(v3d_listener_y_spin);
	gtk_container_add(GTK_CONTAINER(hbox), v3d_listener_y_spin);
	gtk_signal_connect (GTK_OBJECT (v3d_listener_y_adj), 
				"value_changed",
				GTK_SIGNAL_FUNC (v3d_listener_y_adj_value_changed_cb), 
				NULL);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_widget_show(hbox);
        gtk_table_attach_defaults(GTK_TABLE(table),hbox,2,3,0,1);		

        label=gtk_label_new("Z");
        gtk_widget_show(label);
        gtk_container_add(GTK_CONTAINER(hbox), label);

	v3d_listener_z_adj=gtk_adjustment_new((gdouble)(v3d_listener.z), 	
				    (gdouble)(-1.0), 
				    (gdouble)(1.0), 
        	    		    (gdouble)0.001,(gdouble)0.001,(gdouble)0.001);

        v3d_listener_z_spin=gtk_spin_button_new(GTK_ADJUSTMENT(v3d_listener_z_adj),0.01,3);
	gtk_widget_show(v3d_listener_z_spin);
	gtk_container_add(GTK_CONTAINER(hbox), v3d_listener_z_spin);
	gtk_signal_connect (GTK_OBJECT (v3d_listener_z_adj), 
				"value_changed",
				GTK_SIGNAL_FUNC (v3d_listener_z_adj_value_changed_cb), 
				NULL);

	label=gtk_label_new("Direction");
	gtk_widget_show(label);
        gtk_table_attach_defaults(GTK_TABLE(table),label,0,1,1,2);		
	
	v3d_listener_direction_adj=gtk_adjustment_new((gdouble)(v3d_listener.direction), 	
				    (gdouble)(-PI), 
				    (gdouble)(PI)+0.001, 
        	    		    (gdouble)0.001,(gdouble)0.001,(gdouble)0.001);

        v3d_listener_direction_hscale=gtk_hscale_new(GTK_ADJUSTMENT(v3d_listener_direction_adj));
	gtk_widget_show(v3d_listener_direction_hscale);
	gtk_scale_set_value_pos(GTK_SCALE(v3d_listener_direction_hscale),1);
	gtk_scale_set_draw_value(GTK_SCALE(v3d_listener_direction_hscale),0);
	gtk_scale_set_digits(GTK_SCALE(v3d_listener_direction_hscale),3);
        gtk_table_attach_defaults(GTK_TABLE(table),v3d_listener_direction_hscale,1,3,1,2);
	gtk_signal_connect (GTK_OBJECT (v3d_listener_direction_adj), 
				"value_changed",
				GTK_SIGNAL_FUNC (v3d_listener_direction_adj_value_changed_cb), 
				NULL);
/*

	frame=gtk_frame_new("Elevation");
	gtk_widget_show(frame);
        gtk_table_attach_defaults(GTK_TABLE(table),frame,1,2,1,2);		
	
	v3d_listener_elevation_adj=gtk_adjustment_new((gdouble)(v3d_listener.elevation), 	
				    (gdouble)(-PI), 
				    (gdouble)(PI)+0.001, 
        	    		    (gdouble)0.001,(gdouble)0.001,(gdouble)0.001);

        v3d_listener_elevation_hscale=gtk_hscale_new(GTK_ADJUSTMENT(v3d_listener_elevation_adj));

	        
	gtk_widget_show(v3d_listener_elevation_hscale);
	gtk_scale_set_value_pos(GTK_SCALE(v3d_listener_elevation_hscale),1);
        gtk_scale_set_draw_value(GTK_SCALE(v3d_listener_elevation_hscale),1);
	gtk_scale_set_digits(GTK_SCALE(v3d_listener_elevation_hscale),3);
	gtk_container_add(GTK_CONTAINER(frame), v3d_listener_elevation_hscale);

        gtk_signal_connect (GTK_OBJECT (v3d_listener_elevation_adj), 
				"value_changed",
				GTK_SIGNAL_FUNC (v3d_listener_elevation_adj_value_changed_cb), 
				NULL);

	v3d_listener_speed_adj=gtk_adjustment_new((gdouble)(0.0), 	
				    (gdouble)(-10000.0), 
				    (gdouble)( 10000.0), 
        	    		    (gdouble)0.01,(gdouble)0.01,(gdouble)0.01);

        v3d_listener_speed_spin=gtk_spin_button_new(GTK_ADJUSTMENT(v3d_listener_speed_adj),0.01,2);
        gtk_widget_show(v3d_listener_speed_spin);
        gtk_table_attach_defaults(GTK_TABLE(table),v3d_listener_speed_spin,2,3,1,2);		

        gtk_signal_connect (GTK_OBJECT (v3d_listener_speed_adj), 
				"value_changed",
				GTK_SIGNAL_FUNC (v3d_listener_speed_adj_value_changed_cb), 
				NULL);
*/
        vbox = gtk_vbox_new(FALSE, 0);
        gtk_widget_show(vbox);
        gtk_container_add(GTK_CONTAINER(main_hbox), vbox);

	frame=gtk_frame_new("Sources");
	gtk_widget_show(frame);
	gtk_container_add(GTK_CONTAINER(vbox), frame);

        nb = gtk_notebook_new();
	v3d_nb=nb;
        gtk_widget_show(nb);
	gtk_container_add(GTK_CONTAINER(frame), nb);

	for (i=0; i<MAX_NO_OF_3D_SOURCE; i++) {

            move_vbox=gtk_vbox_new(FALSE,0);
	    gtk_widget_show(move_vbox);
	    gtk_container_add(GTK_CONTAINER(nb), move_vbox);
	    gtk_container_set_border_width(GTK_CONTAINER(move_vbox), 6);

		hbox=gtk_hbox_new(FALSE,0);
		gtk_widget_show(hbox);
		gtk_container_add(GTK_CONTAINER(move_vbox), hbox);

		label=gtk_label_new("X");
		gtk_widget_show(label);
		gtk_container_add(GTK_CONTAINER(hbox), label);

		v3d_x_pos_label[i]=gtk_label_new("        ");
		gtk_widget_show(v3d_x_pos_label[i]);
		gtk_container_add(GTK_CONTAINER(hbox), v3d_x_pos_label[i]);

		label=gtk_label_new("Y");
		gtk_widget_show(label);
		gtk_container_add(GTK_CONTAINER(hbox), label);

		v3d_y_pos_label[i]=gtk_label_new("        ");
		gtk_widget_show(v3d_y_pos_label[i]);
		gtk_container_add(GTK_CONTAINER(hbox), v3d_y_pos_label[i]);

		label=gtk_label_new("Z");
		gtk_widget_show(label);
		gtk_container_add(GTK_CONTAINER(hbox), label);

		v3d_z_pos_label[i]=gtk_label_new("        ");
		gtk_widget_show(v3d_z_pos_label[i]);
		gtk_container_add(GTK_CONTAINER(hbox), v3d_z_pos_label[i]);
                calc_source_pos(i);
/*
		v3d_cone_inner_adj[i]=gtk_adjustment_new(
                                            (gdouble)(v3d_source[i].cone_inner_angle), 	
					    (gdouble)0.0, 
					    (gdouble)(v3d_source[i].cone_outer_angle), 
	        	    		    (gdouble)1.0,(gdouble)1.0,(gdouble)1.0);

		v3d_cone_inner_spin[i]=gtk_spin_button_new(GTK_ADJUSTMENT(v3d_cone_inner_adj[i]),1.0,0);
		gtk_widget_show(v3d_cone_inner_spin[i]);
		gtk_spin_button_set_update_policy (GTK_SPIN_BUTTON(v3d_cone_inner_spin[i]),GTK_UPDATE_IF_VALID);
		gtk_container_add(GTK_CONTAINER(hbox), v3d_cone_inner_spin[i]);

	        gtk_signal_connect (GTK_OBJECT (v3d_cone_inner_adj[i]), 
				"value_changed",
				GTK_SIGNAL_FUNC (v3d_cone_inner_adj_value_changed_cb), 
				&v3d_source[i].idx);
*/
		v3d_cone_outer_adj[i]=gtk_adjustment_new(
                                            (gdouble)(v3d_source[i].cone_outer_angle), 	
					    (gdouble)0.0, 
					    (gdouble)360.0, 
	        	    		    (gdouble)1.0,(gdouble)1.0,(gdouble)1.0);

		v3d_cone_outer_spin[i]=gtk_spin_button_new(GTK_ADJUSTMENT(v3d_cone_outer_adj[i]),1.0,0);
		gtk_widget_show(v3d_cone_outer_spin[i]);
		gtk_spin_button_set_update_policy (GTK_SPIN_BUTTON(v3d_cone_outer_spin[i]),GTK_UPDATE_IF_VALID);
		gtk_container_add(GTK_CONTAINER(hbox), v3d_cone_outer_spin[i]);

	        gtk_signal_connect (GTK_OBJECT (v3d_cone_outer_adj[i]), 
				"value_changed",
				GTK_SIGNAL_FUNC (v3d_cone_outer_adj_value_changed_cb), 
				&v3d_source[i].idx);
 
	    table=gtk_table_new(3,1,FALSE);
	    gtk_widget_show(table);
	    gtk_container_add(GTK_CONTAINER(move_vbox), table);

	    sprintf(buffer,"%d",i+1);
	    label = gtk_label_new(buffer);
            gtk_widget_show(label);
	    gtk_notebook_set_tab_label(GTK_NOTEBOOK(nb), 
				   gtk_notebook_get_nth_page(GTK_NOTEBOOK(nb), i), label);

	    hbox = gtk_hbox_new(FALSE, 0);
	    gtk_widget_show(hbox);
            gtk_table_attach_defaults(GTK_TABLE(table),hbox,0,1,2,3);	

 	    v3d_play_button[i]=gtk_button_new_with_label("Play");
	    gtk_widget_show(v3d_play_button[i]);
	    gtk_container_add(GTK_CONTAINER(hbox), v3d_play_button[i]);
       	    gtk_signal_connect (GTK_OBJECT (v3d_play_button[i]), "clicked",
			    GTK_SIGNAL_FUNC (v3d_play_button_clicked_cb),&v3d_source[i].idx);

            gtk_widget_set_sensitive(v3d_play_button[i],FALSE );

 	    v3d_stop_button[i]=gtk_button_new_with_label("Stop");
	    gtk_widget_show(v3d_stop_button[i]);
	    gtk_container_add(GTK_CONTAINER(hbox), v3d_stop_button[i]);
       	    gtk_signal_connect (GTK_OBJECT (v3d_stop_button[i]), "clicked",
			    GTK_SIGNAL_FUNC (v3d_stop_button_clicked_cb),&v3d_source[i].idx);
            gtk_widget_set_sensitive(v3d_stop_button[i],FALSE );

 	    v3d_loop_button[i]=gtk_toggle_button_new_with_label("Loop");
	    gtk_widget_show(v3d_loop_button[i]);
	    gtk_container_add(GTK_CONTAINER(hbox), v3d_loop_button[i]);
       	    gtk_signal_connect (GTK_OBJECT (v3d_loop_button[i]), "toggled",
			    GTK_SIGNAL_FUNC (v3d_loop_button_toggled_cb),&v3d_source[i].idx);

            gtk_widget_set_sensitive(v3d_loop_button[i],FALSE );

 	    v3d_wave_button[i]=gtk_button_new_with_label("Select Sound");
	    gtk_widget_show(v3d_wave_button[i]);
	    gtk_container_add(GTK_CONTAINER(hbox), v3d_wave_button[i]);
       	    gtk_signal_connect (GTK_OBJECT (v3d_wave_button[i]), "clicked",
			    GTK_SIGNAL_FUNC (v3d_wave_button_clicked_cb),&v3d_source[i].idx);

 	    hbox = gtk_hbox_new(FALSE, 0);
	    gtk_widget_show(hbox);
            gtk_table_attach_defaults(GTK_TABLE(table),hbox,0,1,0,1);

 	    v3d_move_button[i]=gtk_toggle_button_new_with_label("Move");
	    gtk_widget_show(v3d_move_button[i]);
	    gtk_container_add(GTK_CONTAINER(hbox), v3d_move_button[i]);
	    gtk_widget_set_sensitive(v3d_move_button[i],(v3d_source[i].move_type > 0 ));

       	    gtk_signal_connect (GTK_OBJECT (v3d_move_button[i]), "toggled",
			    GTK_SIGNAL_FUNC (v3d_move_toggle_button_cb),&v3d_source[i].idx);
	
            gtk_widget_set_sensitive(v3d_move_button[i],(v3d_source[i].move_type > 0) );

            for (j=0; j<MAX_NO_OF_MOVE_TYPE; j++) {
                if ( j == 0 ) {
                    v3d_move_type_radio[i][j] = gtk_radio_button_new_with_label (NULL, v3d_movement_type[0]);
	        }
                else {
                    v3d_move_type_radio[i][j] = gtk_radio_button_new_with_label(gtk_radio_button_group (GTK_RADIO_BUTTON (v3d_move_type_radio[i][0])), v3d_movement_type[j]);
	        };
                gtk_box_pack_start(GTK_BOX(hbox), v3d_move_type_radio[i][j], TRUE, TRUE, 2);
		gtk_widget_show(v3d_move_type_radio[i][j]);
  		gtk_signal_connect (GTK_OBJECT (v3d_move_type_radio[i][j]),
  	                         "toggled",
               		GTK_SIGNAL_FUNC (v3d_move_type_toggle_button_cb), &v3d_source[i].idx);
            };

            move_nb = gtk_notebook_new();
	    v3d_move_nb[i]=move_nb;
            gtk_widget_show(move_nb);
            gtk_table_attach_defaults(GTK_TABLE(table),move_nb,0,1,1,2);
	    gtk_notebook_set_show_tabs(GTK_NOTEBOOK(move_nb),FALSE);

	    move_vbox=gtk_vbox_new(FALSE,0);
	    gtk_widget_show(move_vbox);
	    gtk_container_add(GTK_CONTAINER(move_nb), move_vbox);	
	    gtk_container_set_border_width(GTK_CONTAINER(move_vbox), 6);
	            
	    label=gtk_label_new("Fixed");
            gtk_widget_show(label);
	    gtk_notebook_set_tab_label(GTK_NOTEBOOK(move_nb), 
				   gtk_notebook_get_nth_page(GTK_NOTEBOOK(move_nb), 0), label);

		hbox=gtk_hbox_new(FALSE,0);
		gtk_widget_show(hbox);
		gtk_container_add(GTK_CONTAINER(move_vbox), hbox);

		label=gtk_label_new("X");
		gtk_widget_show(label);
		gtk_container_add(GTK_CONTAINER(hbox), label);

		v3d_x_adj[i]=gtk_adjustment_new((gdouble)(v3d_source[i].x), 	
				    (gdouble)(-1.0), 
				    (gdouble)(1.0), 
        	    		    (gdouble)0.001,(gdouble)0.001,(gdouble)0.001);

		v3d_x_spin[i]=gtk_spin_button_new(GTK_ADJUSTMENT(v3d_x_adj[i]),0.001,3);
		gtk_widget_show(v3d_x_spin[i]);
		gtk_container_add(GTK_CONTAINER(hbox), v3d_x_spin[i]);
            	gtk_signal_connect (GTK_OBJECT (v3d_x_adj[i]), 
				"value_changed",
				GTK_SIGNAL_FUNC (v3d_x_adj_value_changed_cb), 
				&v3d_source[i].idx);

		label=gtk_label_new("Y");
		gtk_widget_show(label);
		gtk_container_add(GTK_CONTAINER(hbox), label);

		v3d_y_adj[i]=gtk_adjustment_new((gdouble)(v3d_source[i].y), 	
				    (gdouble)(-1.0), 
				    (gdouble)(1.0), 
        	    		    (gdouble)0.001,(gdouble)0.001,(gdouble)0.001);

		v3d_y_spin[i]=gtk_spin_button_new(GTK_ADJUSTMENT(v3d_y_adj[i]),0.001,3);
		gtk_widget_show(v3d_y_spin[i]);
		gtk_container_add(GTK_CONTAINER(hbox), v3d_y_spin[i]);
        	gtk_signal_connect (GTK_OBJECT (v3d_y_adj[i]), 
				"value_changed",
				GTK_SIGNAL_FUNC (v3d_y_adj_value_changed_cb), 
				&v3d_source[i].idx);

		label=gtk_label_new("Z");
		gtk_widget_show(label);
		gtk_container_add(GTK_CONTAINER(hbox), label);

		v3d_z_adj[i]=gtk_adjustment_new((gdouble)(v3d_source[i].z), 	
				    (gdouble)(-1.0), 
				    (gdouble)(1.0), 
        	    		    (gdouble)0.001,(gdouble)0.001,(gdouble)0.001);

		v3d_z_spin[i]=gtk_spin_button_new(GTK_ADJUSTMENT(v3d_z_adj[i]),0.001,3);
		gtk_widget_show(v3d_z_spin[i]);
		gtk_container_add(GTK_CONTAINER(hbox), v3d_z_spin[i]);
        	gtk_signal_connect (GTK_OBJECT (v3d_z_adj[i]), 
				"value_changed",
				GTK_SIGNAL_FUNC (v3d_z_adj_value_changed_cb), 
				&v3d_source[i].idx);

	    move_vbox=gtk_vbox_new(FALSE,0);
	    gtk_widget_show(move_vbox);
	    gtk_container_add(GTK_CONTAINER(move_nb), move_vbox);	
	    gtk_container_set_border_width(GTK_CONTAINER(move_vbox), 6);
            
	    label=gtk_label_new("Circular Movement");
            gtk_widget_show(label);
	    gtk_notebook_set_tab_label(GTK_NOTEBOOK(move_nb), 
				   gtk_notebook_get_nth_page(GTK_NOTEBOOK(move_nb), 1), label);

            table=gtk_table_new(3,2,FALSE);
	    gtk_widget_show(table);
	    gtk_container_add(GTK_CONTAINER(move_vbox), table);	

		label=gtk_label_new("X range");
		gtk_widget_show(label);
	        gtk_table_attach_defaults(GTK_TABLE(table),label,0,1,0,1);	
		
		v3d_w_adj[i]=gtk_adjustment_new((gdouble)(v3d_source[i].x_range), 	
				    (gdouble)(0.0), 
				    (gdouble)(+1.0)+0.01, 
        	    		    (gdouble)0.01,(gdouble)0.01,(gdouble)0.01);

        	v3d_w_hscale[i]=gtk_hscale_new(GTK_ADJUSTMENT(v3d_w_adj[i]));
        	gtk_widget_show(v3d_w_hscale[i]);
		gtk_scale_set_value_pos(GTK_SCALE(v3d_w_hscale[i]),1);
        	gtk_scale_set_draw_value(GTK_SCALE(v3d_w_hscale[i]),1);
		gtk_scale_set_digits(GTK_SCALE(v3d_w_hscale[i]),2);
                gtk_widget_set_usize(v3d_w_hscale[i], 102, 22);
	        gtk_table_attach_defaults(GTK_TABLE(table),v3d_w_hscale[i],1,2,0,1);	
        	gtk_signal_connect (GTK_OBJECT (v3d_w_adj[i]), 
				"value_changed",
				GTK_SIGNAL_FUNC (v3d_w_adj_value_changed_cb), 
				&v3d_source[i].idx);

		label=gtk_label_new("Y range");
		gtk_widget_show(label);
	        gtk_table_attach_defaults(GTK_TABLE(table),label,0,1,1,2);	

		v3d_h_adj[i]=gtk_adjustment_new((gdouble)(v3d_source[i].y_range), 	
				    (gdouble)(0.0), 
				    (gdouble)(+1.0)+0.01, 
        	    		    (gdouble)0.01,(gdouble)0.01,(gdouble)0.01);

        	v3d_h_hscale[i]=gtk_hscale_new(GTK_ADJUSTMENT(v3d_h_adj[i]));
        	gtk_widget_show(v3d_h_hscale[i]);
		gtk_scale_set_value_pos(GTK_SCALE(v3d_h_hscale[i]),1);
        	gtk_scale_set_draw_value(GTK_SCALE(v3d_h_hscale[i]),1);
		gtk_scale_set_digits(GTK_SCALE(v3d_h_hscale[i]),2);
                gtk_widget_set_usize(v3d_h_hscale[i], 102, 22);
	        gtk_table_attach_defaults(GTK_TABLE(table),v3d_h_hscale[i],1,2,1,2);	

        	gtk_signal_connect (GTK_OBJECT (v3d_h_adj[i]), 
				"value_changed",
				GTK_SIGNAL_FUNC (v3d_h_adj_value_changed_cb), 
				&v3d_source[i].idx);

		label=gtk_label_new("Speed");
		gtk_widget_show(label);
	        gtk_table_attach_defaults(GTK_TABLE(table),label,0,1,2,3);	

		v3d_speed_adj[i]=gtk_adjustment_new((gdouble)(v3d_source[i].speed), 	
				    (gdouble)(-0.1), 
				    (gdouble)(0.1), 
        	    		    (gdouble)0.001,(gdouble)0.001,(gdouble)0.001);
		v3d_speed_spin[i]=gtk_spin_button_new(GTK_ADJUSTMENT(v3d_speed_adj[i]),0.001,3);
		gtk_widget_show(v3d_speed_spin[i]);
	        gtk_table_attach_defaults(GTK_TABLE(table),v3d_speed_spin[i],1,2,2,3);	
		gtk_signal_connect (GTK_OBJECT (v3d_speed_adj[i]), 
				"value_changed",
				GTK_SIGNAL_FUNC (v3d_speed_adj_value_changed_cb), 
				&v3d_source[i].idx);
	   
	    move_vbox=gtk_vbox_new(FALSE,0);
	    gtk_widget_show(move_vbox);
	    gtk_container_add(GTK_CONTAINER(move_nb), move_vbox);	
	    gtk_container_set_border_width(GTK_CONTAINER(move_vbox), 6);
	                
	    label=gtk_label_new("Linear Movement");
            gtk_widget_show(label);
	    gtk_notebook_set_tab_label(GTK_NOTEBOOK(move_nb), 
				   gtk_notebook_get_nth_page(GTK_NOTEBOOK(move_nb), 2), label);

            table=gtk_table_new(3,2,FALSE);
	    gtk_widget_show(table);
	    gtk_container_add(GTK_CONTAINER(move_vbox), table);	

		label=gtk_label_new("P range");
		gtk_widget_show(label);
		gtk_label_set_justify(GTK_LABEL(label),GTK_JUSTIFY_LEFT);
	        gtk_table_attach_defaults(GTK_TABLE(table),label,0,1,0,1);	
	
		v3d_r_adj[i]=gtk_adjustment_new((gdouble)(v3d_source[i].p_range), 	
				    (gdouble)(0.0), 
				    (gdouble)(+1.0)+0.01, 
        	    		    (gdouble)0.01,(gdouble)0.01,(gdouble)0.01);

	        v3d_r_hscale[i]=gtk_hscale_new(GTK_ADJUSTMENT(v3d_r_adj[i]));
	        gtk_widget_show(v3d_r_hscale[i]);
		gtk_scale_set_value_pos(GTK_SCALE(v3d_r_hscale[i]),1);
	        gtk_scale_set_draw_value(GTK_SCALE(v3d_r_hscale[i]),1);
		gtk_scale_set_digits(GTK_SCALE(v3d_r_hscale[i]),2);
                gtk_widget_set_usize(v3d_r_hscale[i], 102, 22);

	        gtk_table_attach_defaults(GTK_TABLE(table),v3d_r_hscale[i],1,2,0,1);	
	
	        gtk_signal_connect (GTK_OBJECT (v3d_r_adj[i]), 
					"value_changed",
					GTK_SIGNAL_FUNC (v3d_r_adj_value_changed_cb), 
					&v3d_source[i].idx);

		label=gtk_label_new("Orienation");
		gtk_widget_show(label);
	        gtk_table_attach_defaults(GTK_TABLE(table),label,0,1,1,2);	
				
		v3d_o_adj[i]=gtk_adjustment_new((gdouble)(v3d_source[i].o), 	
					    (gdouble)(-PI), 
					    (gdouble)(+PI), 
	        	    		    (gdouble)0.001,(gdouble)0.001,(gdouble)0.001);

		v3d_o_spin[i]=gtk_spin_button_new(GTK_ADJUSTMENT(v3d_o_adj[i]),0.001,3);
		gtk_widget_show(v3d_o_spin[i]);
	        gtk_table_attach_defaults(GTK_TABLE(table),v3d_o_spin[i],1,2,1,2);	

	        gtk_signal_connect (GTK_OBJECT (v3d_o_adj[i]), 
				"value_changed",
				GTK_SIGNAL_FUNC (v3d_o_adj_value_changed_cb), 
				&v3d_source[i].idx);

		label=gtk_label_new("Speed");
		gtk_widget_show(label);
	        gtk_table_attach_defaults(GTK_TABLE(table),label,0,1,2,3);	
	
		v3d_v_adj[i]=gtk_adjustment_new((gdouble)(v3d_source[i].increment), 	
				    (gdouble)(0.0), 
				    (gdouble)(+0.1)+0.001, 
        	    		    (gdouble)0.001,(gdouble)0.001,(gdouble)0.001);

	        v3d_v_hscale[i]=gtk_hscale_new(GTK_ADJUSTMENT(v3d_v_adj[i]));
	        gtk_widget_show(v3d_v_hscale[i]);
		gtk_scale_set_value_pos(GTK_SCALE(v3d_v_hscale[i]),1);
	        gtk_scale_set_draw_value(GTK_SCALE(v3d_v_hscale[i]),1);
		gtk_scale_set_digits(GTK_SCALE(v3d_v_hscale[i]),3);
                gtk_widget_set_usize(v3d_v_hscale[i], 102, 22);
	        gtk_table_attach_defaults(GTK_TABLE(table),v3d_v_hscale[i],1,2,2,3);	


	        gtk_signal_connect (GTK_OBJECT (v3d_v_adj[i]), 
				"value_changed",
				GTK_SIGNAL_FUNC (v3d_v_adj_value_changed_cb), 
				&v3d_source[i].idx);

	    move_vbox=gtk_vbox_new(FALSE,0);
	    gtk_widget_show(move_vbox);
	    gtk_container_add(GTK_CONTAINER(move_nb), move_vbox);	
	    gtk_container_set_border_width(GTK_CONTAINER(move_vbox), 6);
           
	    label=gtk_label_new("Overhead Movement");
            gtk_widget_show(label);
	    gtk_notebook_set_tab_label(GTK_NOTEBOOK(move_nb), 
				   gtk_notebook_get_nth_page(GTK_NOTEBOOK(move_nb), 3), label);
	
            table=gtk_table_new(2,2,FALSE);
	    gtk_widget_show(table);
	    gtk_container_add(GTK_CONTAINER(move_vbox), table);	

		label=gtk_label_new("Orienation");
		gtk_widget_show(label);
	        gtk_table_attach_defaults(GTK_TABLE(table),label,0,1,0,1);
				
		v3d_oh_o_adj[i]=gtk_adjustment_new((gdouble)(v3d_source[i].o), 	
					    (gdouble)(-PI), 
					    (gdouble)(+PI), 
	        	    		    (gdouble)0.001,(gdouble)0.001,(gdouble)0.001);

		v3d_oh_o_spin[i]=gtk_spin_button_new(GTK_ADJUSTMENT(v3d_oh_o_adj[i]),0.001,3);
		gtk_widget_show(v3d_oh_o_spin[i]);
	        gtk_table_attach_defaults(GTK_TABLE(table),v3d_oh_o_spin[i],1,2,0,1);	

	        gtk_signal_connect (GTK_OBJECT (v3d_oh_o_adj[i]), 
				"value_changed",
				GTK_SIGNAL_FUNC (v3d_oh_o_adj_value_changed_cb), 
				&v3d_source[i].idx);

		label=gtk_label_new("Speed");
		gtk_widget_show(label);
	        gtk_table_attach_defaults(GTK_TABLE(table),label,0,1,1,2);	

		v3d_oh_speed_adj[i]=gtk_adjustment_new((gdouble)(v3d_source[i].oh_speed), 	
				    (gdouble)(-0.1), 
				    (gdouble)(0.1), 
        	    		    (gdouble)0.001,(gdouble)0.001,(gdouble)0.001);
		v3d_oh_speed_spin[i]=gtk_spin_button_new(GTK_ADJUSTMENT(v3d_oh_speed_adj[i]),0.001,3);
		gtk_widget_show(v3d_oh_speed_spin[i]);
	        gtk_table_attach_defaults(GTK_TABLE(table),v3d_oh_speed_spin[i],1,2,1,2);	
		gtk_signal_connect (GTK_OBJECT (v3d_oh_speed_adj[i]), 
				"value_changed",
				GTK_SIGNAL_FUNC (v3d_oh_speed_adj_value_changed_cb), 
				&v3d_source[i].idx);
	};

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_widget_show(hbox);
	gtk_container_add(GTK_CONTAINER(vbox), hbox);

	frame=gtk_frame_new("Scale");
	gtk_widget_show(frame);
	gtk_container_add(GTK_CONTAINER(hbox), frame);

	v3d_scaleup_adj=gtk_adjustment_new((gdouble)(v3d_scaleup), 	
				    (gdouble)(1.0), 
				    (gdouble)(100.0), 
        	    		    (gdouble)0.1,(gdouble)0.1,(gdouble)0.1);

        v3d_scaleup_spin=gtk_spin_button_new(GTK_ADJUSTMENT(v3d_scaleup_adj),0.1,1);
        gtk_widget_show(v3d_scaleup_spin);
	gtk_container_add(GTK_CONTAINER(frame), v3d_scaleup_spin);

        gtk_signal_connect (GTK_OBJECT (v3d_scaleup_adj), 
				"value_changed",
				GTK_SIGNAL_FUNC (v3d_scaleup_adj_value_changed_cb), 
				NULL);

	v3d_showpath_button=gtk_toggle_button_new_with_label("Path");
	gtk_widget_show(v3d_showpath_button);
	gtk_container_add(GTK_CONTAINER(hbox), v3d_showpath_button);
        gtk_signal_connect (GTK_OBJECT (v3d_showpath_button), "toggled",
			    GTK_SIGNAL_FUNC (v3d_showpath_toggle_button_cb),NULL);

	
	v3d_wavetracing_button=gtk_toggle_button_new_with_label("Tracing");
	gtk_widget_show(v3d_wavetracing_button);
   	gtk_signal_connect (GTK_OBJECT (v3d_wavetracing_button), "toggled",
			    GTK_SIGNAL_FUNC (v3d_wavetracing_toggle_button_cb),NULL);
	gtk_container_add(GTK_CONTAINER(hbox), v3d_wavetracing_button);

	v3d_wavelength_adj=gtk_adjustment_new((gdouble)(v3d_wavelength), 	
				    (gdouble)(0.01), 
				    (gdouble)(1.0)+0.01, 
        	    		    (gdouble)0.01,(gdouble)0.01,(gdouble)0.01);

        v3d_wavelength_hscale=gtk_hscale_new(GTK_ADJUSTMENT(v3d_wavelength_adj));
	gtk_widget_show(v3d_wavelength_hscale);
	gtk_scale_set_value_pos(GTK_SCALE(v3d_wavelength_hscale),1);
	gtk_scale_set_draw_value(GTK_SCALE(v3d_wavelength_hscale),1);
	gtk_scale_set_digits(GTK_SCALE(v3d_wavelength_hscale),2);
	gtk_container_add(GTK_CONTAINER(hbox), v3d_wavelength_hscale);
        gtk_widget_set_usize(v3d_wavelength_hscale, 102,22);

	gtk_signal_connect (GTK_OBJECT (v3d_wavelength_adj), 
				"value_changed",
				GTK_SIGNAL_FUNC (v3d_wavelength_adj_value_changed_cb), 
				NULL);

	frame=gtk_frame_new("Reverb");
	gtk_widget_show(frame);
	gtk_container_add(GTK_CONTAINER(vbox), frame);

	v3d_reverb_adj=gtk_adjustment_new((gdouble)(0), 	
				    (gdouble)(0.0), 
				    (gdouble)(1.0)+0.1, 
        	    		    (gdouble)0.1,(gdouble)0.1,(gdouble)0.1);
        v3d_reverb_hscale=gtk_hscale_new(GTK_ADJUSTMENT(v3d_reverb_adj));
        gtk_widget_show(v3d_reverb_hscale);
	gtk_scale_set_value_pos(GTK_SCALE(v3d_reverb_hscale),1);
        gtk_scale_set_draw_value(GTK_SCALE(v3d_reverb_hscale),0);
//	gtk_scale_set_digits(GTK_SCALE(v3d_scaleup_hscale),1);
	gtk_container_add(GTK_CONTAINER(frame), v3d_reverb_hscale);

        gtk_signal_connect (GTK_OBJECT (v3d_reverb_adj), 
				"value_changed",
				GTK_SIGNAL_FUNC (v3d_reverb_adj_value_changed_cb), 
				NULL);	

	frame=gtk_frame_new("Reflection");
	gtk_widget_show(frame);
	gtk_container_add(GTK_CONTAINER(vbox), frame);
	hbox = gtk_hbox_new(FALSE, 0);
	gtk_widget_show(hbox);
	gtk_container_add(GTK_CONTAINER(frame), hbox);	

        combo=gtk_combo_new();
	gtk_widget_show(combo);
	gtk_container_add(GTK_CONTAINER(hbox), combo);	

        glist = NULL;  
        for (i=0; i<3; i++) {
            glist = g_list_append (glist, v3d_reflection_type[i]);
        };
        gtk_combo_set_popdown_strings (GTK_COMBO (combo), glist);
	g_list_free(glist);
//        gtk_combo_set_use_arrows_always(GTK_COMBO(combo),TRUE);	
        gtk_entry_set_editable(GTK_ENTRY(GTK_COMBO(combo)->entry),FALSE);
        gtk_signal_connect(GTK_OBJECT(GTK_COMBO(combo)->entry),
                         "changed",
                         GTK_SIGNAL_FUNC(v3d_reflection_type_changed_cb),
                         NULL);

	label=gtk_label_new("Delay");
	gtk_widget_show(label);
	gtk_container_add(GTK_CONTAINER(hbox), label);

	v3d_reflection_adj=gtk_adjustment_new((gdouble)(0), 	
				    (gdouble)(0.0), 
				    (gdouble)(1.0)+0.1, 
        	    		    (gdouble)0.1,(gdouble)0.1,(gdouble)0.1);

        v3d_reflection_hscale=gtk_hscale_new(GTK_ADJUSTMENT(v3d_reflection_adj));
        gtk_widget_show(v3d_reflection_hscale);
	gtk_scale_set_value_pos(GTK_SCALE(v3d_reflection_hscale),1);
        gtk_scale_set_draw_value(GTK_SCALE(v3d_reflection_hscale),0);
        gtk_widget_set_usize(v3d_reflection_hscale, 152, 22);

	gtk_container_add(GTK_CONTAINER(hbox), v3d_reflection_hscale);

        gtk_signal_connect (GTK_OBJECT (v3d_reflection_adj), 
				"value_changed",
				GTK_SIGNAL_FUNC (v3d_reflection_adj_value_changed_cb), 
				NULL);

        nb = gtk_notebook_new();
        gtk_widget_show(nb);
	gtk_container_add(GTK_CONTAINER(vbox), nb);

	for (j=0; j<3; j++) {
 	    hbox = gtk_hbox_new(FALSE, 0);
	    gtk_widget_show(hbox);
	    gtk_container_add(GTK_CONTAINER(nb), hbox);	
	    gtk_container_set_border_width(GTK_CONTAINER(hbox), 6);
            
	    label=gtk_label_new(v3d_oeo[j]);
            gtk_widget_show(label);
	    gtk_notebook_set_tab_label(GTK_NOTEBOOK(nb), 
				   gtk_notebook_get_nth_page(GTK_NOTEBOOK(nb), j), label);

	    for (i=0; i<4; i++) {
		switch(j) {
		case 0:
 	  	    v3d_occlusion_button[i]=gtk_toggle_button_new_with_label(v3d_oeo_label[i]);
	            gtk_widget_show(v3d_occlusion_button[i]);
	            gtk_container_add(GTK_CONTAINER(hbox), v3d_occlusion_button[i]);
	            gtk_signal_connect (GTK_OBJECT (v3d_occlusion_button[i]), "toggled",
			    GTK_SIGNAL_FUNC (v3d_occlusion_toggle_button_cb),&v3d_source[i].idx);
		    break;
		case 1:
	            v3d_exclusion_button[i]=gtk_toggle_button_new_with_label(v3d_oeo_label[i]);
	            gtk_widget_show(v3d_exclusion_button[i]);
	            gtk_container_add(GTK_CONTAINER(hbox), v3d_exclusion_button[i]);
	            gtk_signal_connect (GTK_OBJECT (v3d_exclusion_button[i]), "toggled",
			    GTK_SIGNAL_FUNC (v3d_exclusion_toggle_button_cb),&v3d_source[i].idx);
                    break;
		case 2:
 	            v3d_obstruction_button[i]=gtk_toggle_button_new_with_label(v3d_oeo_label[i]);
	            gtk_widget_show(v3d_obstruction_button[i]);
	            gtk_container_add(GTK_CONTAINER(hbox), v3d_obstruction_button[i]);
	            gtk_signal_connect (GTK_OBJECT (v3d_obstruction_button[i]), "toggled",
			    GTK_SIGNAL_FUNC (v3d_obstruction_toggle_button_cb),&v3d_source[i].idx);
                    break;
		};
	    };
	};

 	hbox = gtk_hbox_new(FALSE, 0);
	gtk_widget_show(hbox);
	gtk_container_add(GTK_CONTAINER(nb), hbox);	
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 6);
            
	label=gtk_label_new("Circular");
        gtk_widget_show(label);
	gtk_notebook_set_tab_label(GTK_NOTEBOOK(nb), 
				   gtk_notebook_get_nth_page(GTK_NOTEBOOK(nb), 3), label);

	v3d_obstruction_radius_adj=gtk_adjustment_new((gdouble)(v3d_obstruction_radius), 	
				    (gdouble)(v3d_obstruction_min_radius), 
				    (gdouble)(1.0)+0.01, 
        	    		    (gdouble)0.01,(gdouble)0.01,(gdouble)0.01);

        v3d_obstruction_radius_hscale=gtk_hscale_new(GTK_ADJUSTMENT(v3d_obstruction_radius_adj));
	gtk_widget_show(v3d_obstruction_radius_hscale);
        gtk_widget_set_usize(v3d_obstruction_radius_hscale, 102, 22);

	gtk_scale_set_value_pos(GTK_SCALE(v3d_obstruction_radius_hscale),1);
	gtk_scale_set_draw_value(GTK_SCALE(v3d_obstruction_radius_hscale),1);
	gtk_scale_set_digits(GTK_SCALE(v3d_obstruction_radius_hscale),2);
	gtk_container_add(GTK_CONTAINER(hbox), v3d_obstruction_radius_hscale);

	gtk_signal_connect (GTK_OBJECT (v3d_obstruction_radius_adj), 
				"value_changed",
				GTK_SIGNAL_FUNC (v3d_obstruction_radius_adj_value_changed_cb), 
				NULL);

 	hbox = gtk_hbox_new(FALSE, 0);
	gtk_widget_show(hbox);
	gtk_container_add(GTK_CONTAINER(nb), hbox);	
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 6);

	label=gtk_label_new("Size");
        gtk_widget_show(label);
	gtk_notebook_set_tab_label(GTK_NOTEBOOK(nb), 
				   gtk_notebook_get_nth_page(GTK_NOTEBOOK(nb), 4), label);

	v3d_obstruction_size_adj=gtk_adjustment_new((gdouble)(v3d_obstruction_size), 	
				    (gdouble)(0.2), 
				    (gdouble)(1.0)+0.01, 
        	    		    (gdouble)0.01,(gdouble)0.01,(gdouble)0.01);

        v3d_obstruction_size_hscale=gtk_hscale_new(GTK_ADJUSTMENT(v3d_obstruction_size_adj));
	gtk_widget_show(v3d_obstruction_size_hscale);
        gtk_widget_set_usize(v3d_obstruction_size_hscale, 102, 22);

	gtk_scale_set_value_pos(GTK_SCALE(v3d_obstruction_size_hscale),1);
	gtk_scale_set_draw_value(GTK_SCALE(v3d_obstruction_size_hscale),1);
	gtk_scale_set_digits(GTK_SCALE(v3d_obstruction_size_hscale),2);
	gtk_container_add(GTK_CONTAINER(hbox), v3d_obstruction_size_hscale);

	gtk_signal_connect (GTK_OBJECT (v3d_obstruction_size_adj), 
				"value_changed",
				GTK_SIGNAL_FUNC (v3d_obstruction_size_adj_value_changed_cb), 
				NULL);
}

/******************************************************************************************/

static void create_about(GtkWidget *main, GtkWidget *notebook, int page)
{
	GtkWidget *label;
	GtkWidget *vbox;
	char buffer[1024];

	vbox = gtk_vbox_new(FALSE, 0);
	gtk_widget_show(vbox);
	gtk_container_add(GTK_CONTAINER(notebook), vbox);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 6);

        label = gtk_label_new("About");
        gtk_widget_show(label);
	gtk_notebook_set_tab_label(GTK_NOTEBOOK(notebook), 
				   gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), page), 
				   label);
	strcpy(buffer,"Vortex Control Panel ");
	strcat(buffer,VC_VERSION);
	strcat(buffer,"\n");
	strcat(buffer,"Copyright(c) 2004-2005 by Claudia");
	strcat(buffer,"\n");
	strcat(buffer,"\n");
	strcat(buffer,"\n");
	strcat(buffer,snd_ctl_card_info_get_longname(hw_info));
	strcat(buffer,"\n");
	strcat(buffer,snd_ctl_card_info_get_mixername(hw_info));
	label=gtk_label_new(buffer);
        gtk_widget_show(label);
        gtk_container_add(GTK_CONTAINER(vbox), label);
}

int main(int argc, char **argv)
{
        GtkWidget *notebook;
        char title[128];
	int i, c, err;
	int page;
	
	snd_ctl_card_info_alloca(&hw_info);

        gtk_init(&argc, &argv);

	alsa_card_number=0;
	strcpy(card,"hw:0");
	while ((c = getopt(argc, argv, "c:")) != -1) {
		switch (c) {
		case 'c':
			i = atoi(optarg);
			if (i < 0 || i >= MAX_CARD_NUMBERS) {
				fprintf(stderr, "Invalid card number %d\n", i);
				exit(1);
			}
			alsa_card_number=i;
			sprintf(card, "hw:%d", alsa_card_number);
			break;
		default:
			fprintf(stderr, "usage: vortexcontrol [-c card#]\n");
			fprintf(stderr, "\t-c, --card\tAlsa card number to control\n");
			exit(1);
			break;
		}
	}
	if ((err = snd_ctl_open(&ctl, card, 0)) < 0) {
		fprintf(stderr, "snd_ctl_open: %s\n", snd_strerror(err));
//		exit(EXIT_FAILURE);
	}
	if ((err = snd_ctl_card_info(ctl, hw_info)) < 0) {
		fprintf(stderr, "snd_ctl_card_info: %s\n", snd_strerror(err));
		exit(EXIT_FAILURE);
	}
	strcpy(driver,snd_ctl_card_info_get_name(hw_info));

	if (!( card_is_au8810 || card_is_au8820 || card_is_au8830 )) {
		fprintf(stderr, "No Aureal Vortex cards found, sorry.\n", driver);
//		exit(EXIT_FAILURE);
	}
	if ((err = snd_hctl_open(&hctl, card, 0)) < 0) {
		fprintf(stderr,"snd_hctl_open %s : %s", card, snd_strerror(err));
		exit(EXIT_FAILURE);
	}
	if ((err = snd_hctl_load(hctl)) < 0) {
		fprintf(stderr,"snd_htcl_load : %s\n", snd_strerror(err));
		exit(EXIT_FAILURE);
	}

	eq_lock=1;
	eq_ungang=1;

        for (i=0; i<4; i++) {
            	memset(&speaker_wave[i],0,sizeof(HWMIX_WAVE));
            	speaker_wave[i].button_type=2;
        };
	for (i=0; i<16; i++) {
        	memset(&hwmix[i],0,sizeof(HWMIX_WAVE));
                hwmix[i].button_type=1;
                hwmix[i].loop=FALSE;
                hwmix[i].idx=i;
	       	memset(&v3d_wave[i],0,sizeof(HWMIX_WAVE));
                v3d_wave[i].button_type=3;
                v3d_wave[i].idx=i;
		memset(&wt_wave[i],0,sizeof(WT_WAVE));
	};

        sprintf(title, "Vortex Control Panel %s", VC_VERSION);
        window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
        gtk_window_set_title(GTK_WINDOW(window), title);
        gtk_signal_connect(GTK_OBJECT (window), "delete_event", 
			(GtkSignalFunc) gtk_main_quit, NULL);
        signal(SIGINT, (void *)gtk_main_quit);

        notebook = gtk_notebook_new();
        gtk_widget_show(notebook);
	gtk_container_add(GTK_CONTAINER(window), notebook);

	memset(&dlsinfo,0,sizeof(DLSInfo));
	memset(&sfinfo,0,sizeof(SFInfo));

        v3d_page=-1;
        v3d_init=0;
	page=0;    

	if ( card_is_au8810 || card_is_au8830 ) 
		create_equalizer(window,notebook,page++);
	create_mixer(window, notebook, page++);
	if ( card_is_au8820 || card_is_au8830 )
		create_midi(window, notebook, page++);
	if ( iec958_present )                           
		create_spdif(window, notebook, page++);
	create_speaker(window, notebook, page++);
	create_hwmixer(window, notebook, page++);
	if ( card_is_au8810 || card_is_au8830 ) {
                v3d_page=page;
		create_3D(window, notebook, page++);
                gtk_signal_connect (GTK_OBJECT (notebook), "switch-page",
                            (GtkSignalFunc) user_change_page, NULL);
        };
 	create_about(window, notebook, page++);

	gtk_widget_show(window);

	if ( card_is_au8810 || card_is_au8830 ) {
	        gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(eqlock_button),eq_lock);
                gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(eqpower_button),eq_power);
                gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(eqgang_button),1-eq_ungang);
        };

	gtk_main();

        for (i=0; i<4; i++) 
	    free_resource(&speaker_wave[i]);
	for (i=0; i<16; i++) {
		free_resource(&hwmix[i]);
                free_resource(&v3d_wave[i]);
	};
	if ( card_is_au8820 || card_is_au8830 ) {
	    if ( sfinfo.loaded )
		vortex_free_arl(&sfinfo);
	    if ( dlsinfo.loaded )
		vortex_free_dls(&dlsinfo);
	    if ( midiinfo.loaded )	
		vortex_free_midi(&midiinfo);
	};
#ifdef HAVE_LIBOPENAL        
	if ( v3d_init )
	    shutdown_alsa_3d();
#endif
	snd_hctl_close(hctl);
	snd_ctl_close(ctl);
	return EXIT_SUCCESS;
}
